{*******************************************************************************
   Unit
      sButtons.pas
   Description:
      TsSpeedButton and TBitBtn with GlyphList suppot, 3d captions and hilight
         support.
   Versions:
      '99
   Autor(s):
      Partial Copyright (c) 1995,98 Inprise Corporation
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   History:
      '99         -  August 1999
                     .Final fixups
                     .Added label notification messages to TsBibBtn
      '99 beta 2  -  June 1999
                     .Styles property for TBitBtn
      '99 beta 1	- 	June 1999
                     .Color property added
                     .Fixed flicker on MouseIn/MouseOut in TBitBtn
      1.0*        -  Feb - march 1999
      			      Initial release
   Comments:
      These components are part of sEditTools.

*   This implementation is an extend of the standart Delphi SpeedButton and BitBtn.
   Since these components have very pure developer interface I had no other choice, but
   recreate the components from the scratch (copying the source from the Buttons.pas
   VCL source) and make the modifications I need.
*******************************************************************************}

unit sButtons;

{$S-,W-,R-}
{$C PRELOAD}

interface

uses Windows, Messages, Classes, Controls, Forms, Graphics,
   sGraphics, sGlyphsList, sConsts, stdctrls, sCommonTypes, sDefaults;

type
   TsSpeedButton = class(TGraphicControl)
   private
      FAllowAllUp: Boolean;
      FCancel: Boolean;
      FDefault: Boolean;
      FDown: Boolean;
      FDragging: Boolean;
      FFlat: Boolean;
      FFont3D: TFont3D;
      FGlyphList: TsGlyphList;
      FGlyphId: Integer;
      FGroupIndex: Integer;
      FGlyph: Pointer;
      FHilightStyle: TTextHilightStyle;
      FHilight3d: THilight3d;
      FKind: TBitBtnKind;
      FLayout: TButtonLayout;
      FMargin: Integer;
      FModalResult: TModalResult;
      FModifiedGlyph: Boolean;
      FMouseInControl: Boolean;
      FSpacing: Integer;
      FTransparent: Boolean;

      procedure Font3dChanged(Sender: Tobject);
      procedure GlyphChanged(Sender: TObject);
      function GlyphStored: Boolean;
      function IsCustom: Boolean;
      procedure UpdateExclusive;

      function GetGlyph: TBitmap;
      function GetKind: TBitBtnKind;
      function GetNumGlyphs: TNumGlyphs;

      procedure SetAllowAllUp(Value: Boolean);
      procedure SetDown(Value: Boolean);
      procedure SetFlat(Value: Boolean);
      procedure SetFont3d(Value: TFont3d);
      procedure SetGlyph(Value: TBitmap);
      procedure SetGlyphList(Value: TsGlyphList);
      procedure SetGlyphId(Value: Integer);
      procedure SetGroupIndex(Value: Integer);
      procedure SetHilight3d(Value: THilight3d);
      procedure SetHilightStyle(Value: TTextHilightStyle);
      procedure SetKind(Value: TBitBtnKind);
      procedure SetLayout(Value: TButtonLayout);
      procedure SetNumGlyphs(Value: TNumGlyphs);
      procedure SetSpacing(Value: Integer);
      procedure SetTransparent(Value: Boolean);
      procedure SetMargin(Value: Integer);
      procedure UpdateTracking;

      procedure WMLButtonDblClk(var Message: TWMLButtonDown); message WM_LBUTTONDBLCLK;
      procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
      procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
      procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
      procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
      procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
      procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
      procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure STMGlyphIdChanged(var Message: TMessage); message STM_GLYPHIDCHANGED;
  protected
      FState: TButtonState;
      procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
      function GetPalette: HPALETTE; override;
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure Paint; override;
      property MouseInControl: Boolean read FMouseInControl;
  public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Click; override;
  published
      property Action;
      property AllowAllUp: Boolean read FAllowAllUp write SetAllowAllUp default False;
      property Anchors;
      property BiDiMode;
      property Cancel: Boolean read FCancel write FCancel stored IsCustom;
      property Color default clBtnFace;
      property Constraints;
      property Default: Boolean read FDefault write FDefault stored IsCustom;
      property Down: Boolean read FDown write SetDown default False;
      property Caption;
      property Enabled;
      property Flat: Boolean read FFlat write SetFlat default DEF_SpeedBtn_Flat;
      property Font;
      property Font3D: TFont3D read FFont3D write SetFont3D;
      property Glyph: TBitmap read GetGlyph write SetGlyph stored GlyphStored;
      property GlyphList: TsGlyphList read FGlyphList write SetGlyphList;
      property GlyphListId: Integer read FGlyphId write SetGlyphId default -1;
      property GroupIndex: Integer read FGroupIndex write SetGroupIndex default 0;
      property HilightStyle: TTextHilightStyle read FHilightStyle write SetHilightStyle default DEF_SpeedBtn_Hilight;
      property Hilight3d: THilight3d read FHilight3d write SetHilight3d;
      property Kind: TBitBtnKind read GetKind write SetKind default bkCustom;
      property Layout: TButtonLayout read FLayout write SetLayout default blGlyphLeft;
      property Margin: Integer read FMargin write SetMargin default -1;
      property ModalResult: TModalResult read FModalResult write FModalResult stored IsCustom;
      property NumGlyphs: TNumGlyphs read GetNumGlyphs write SetNumGlyphs default 1;
      property ParentFont;
      property ParentShowHint;
      property ParentBiDiMode;
      property PopupMenu;
      property ShowHint;
      property Spacing: Integer read FSpacing write SetSpacing default 4;
      property Transparent: Boolean read FTransparent write SetTransparent default DEF_SpeedBtn_Tramsparent;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
   end;

   TsBitBtn = class(TButton)
   private
      FCanvas: TCanvas;
      FBtnStyle: TsBitBtnStyle;
      FFocused: Boolean;
      FFont3D: TFont3D;
      FGlyph: Pointer;
      FGlyphList: TsGlyphList;
      FGlyphId: Integer;
      FHilightStyle: TTextHilightStyle;
      FHilight3d: THilight3d;
      FKind: TBitBtnKind;
      FLayout: TButtonLayout;
      FMargin: Integer;
      FModifiedGlyph: Boolean;
      FMouseInControl: Boolean;
      FSpacing: Integer;
      FState: TButtonState;
      FStyle: TButtonStyle;

      procedure DrawItem(const DrawItemStruct: TDrawItemStruct);
      procedure InvalidateCaption;

      procedure GlyphChanged(Sender: TObject);
      procedure Font3dChanged(Sender: Tobject);
      function IsCustom: Boolean;
      function IsCustomCaption: Boolean;
      function GlyphStored: Boolean;

      function GetGlyph: TBitmap;
      function GetKind: TBitBtnKind;
      function GetNumGlyphs: TNumGlyphs;

      procedure SetBtnStyle(Value: TsBitBtnStyle);
      procedure SetFont3d(Value: TFont3d);
      procedure SetGlyph(Value: TBitmap);
      procedure SetGlyphList(Value: TsGlyphList);
      procedure SetGlyphId(Value: Integer);
      procedure SetHilight3d(Value: THilight3d);
      procedure SetHilightStyle(Value: TTextHilightStyle);
      procedure SetKind(Value: TBitBtnKind);
      procedure SetLayout(Value: TButtonLayout);
      procedure SetMargin(Value: Integer);
      procedure SetNumGlyphs(Value: TNumGlyphs);
      procedure SetSpacing(Value: Integer);
      procedure SetStyle(Value: TButtonStyle);

      procedure CNMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
      procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
      procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
      procedure STMGlyphIdChanged(var Message: TMessage); message STM_GLYPHIDCHANGED;
  protected
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
      procedure CreateHandle; override;
      procedure CreateParams(var Params: TCreateParams); override;
      function GetPalette: HPALETTE; override;
      procedure SetButtonStyle(ADefault: Boolean); override;
  public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Click; override;
  published
      property Action;
      property Anchors;
      property BiDiMode;
      property Cancel stored IsCustom;
      property Caption stored IsCustomCaption;
      property Constraints;
      property Color default clBtnFace;
      property Default stored IsCustom;
      property Enabled;
      property BtnStyle: TsBitBtnStyle read FBtnStyle write SetBtnStyle default DEF_BitBtn_Style;
      property Font3D: TFont3D read FFont3D write SetFont3D;
      property Glyph: TBitmap read GetGlyph write SetGlyph stored GlyphStored;
      property GlyphList: TsGlyphList read FGlyphList write SetGlyphList;
      property GlyphListId: Integer read FGlyphId write SetGlyphId default -1;
      property HilightStyle: TTextHilightStyle read FHilightStyle write SetHilightStyle default DEF_BitBtn_Hilight;
      property Hilight3d: THilight3d read FHilight3d write SetHilight3d;

      property Kind: TBitBtnKind read GetKind write SetKind default bkCustom;
      property Layout: TButtonLayout read FLayout write SetLayout default blGlyphLeft;
      property Margin: Integer read FMargin write SetMargin default -1;
      property ModalResult stored IsCustom;
      property NumGlyphs: TNumGlyphs read GetNumGlyphs write SetNumGlyphs stored IsCustom default 1;
      property ParentShowHint;
      property ParentBiDiMode;
      property ShowHint;
      property Style: TButtonStyle read FStyle write SetStyle default bsAutoDetect;
      property Spacing: Integer read FSpacing write SetSpacing default 4;
      property TabOrder;
      property TabStop;
      property Visible;
      property OnEnter;
      property OnExit;
  end;

implementation
{.$R *.DCR}

uses SysUtils, ActnList, CommCtrl, ImgList, StdUtils;

{ TBitBtn data }
var
   BitBtnResNames: array[TBitBtnKind] of PChar = (
      nil, 'BBOK', 'BBCANCEL', 'BBHELP', 'BBYES', 'BBNO', 'BBCLOSE',
      'BBABORT', 'BBRETRY', 'BBIGNORE', 'BBALL');
   BitBtnCaptions: array[TBitBtnKind] of Pointer = (
      nil, @SOKButton, @SCancelButton, @SHelpButton, @SYesButton, @SNoButton,
      @SCloseButton, @SAbortButton, @SRetryButton, @SIgnoreButton, @SAllButton);
   BitBtnModalResults: array[TBitBtnKind] of TModalResult = (
      0, mrOk, mrCancel, 0, mrYes, mrNo, 0, mrAbort, mrRetry, mrIgnore, mrAll);

var
   BitBtnGlyphs: array[TBitBtnKind] of TBitmap;

function GetBitBtnGlyph(Kind: TBitBtnKind): TBitmap;
begin
   if BitBtnGlyphs[Kind] = nil then begin
      BitBtnGlyphs[Kind] := TBitmap.Create;
      BitBtnGlyphs[Kind].LoadFromResourceName(HInstance, BitBtnResNames[Kind]);
   end;
   Result := BitBtnGlyphs[Kind];
end;

type
   TButtonGlyph = class
   private
      FOriginal: TBitmap;
      FGlyphList: TGlyphList;
      FIndexs: array[TButtonState] of Integer;
      FTransparentColor: TColor;
      FNumGlyphs: TNumGlyphs;
      FPaint3D: TFont3D;
      FOnChange: TNotifyEvent;
      procedure GlyphChanged(Sender: TObject);
      procedure SetGlyph(Value: TBitmap);
      procedure SetNumGlyphs(Value: TNumGlyphs);
      procedure Invalidate;
      function CreateButtonGlyph(State: TButtonState): Integer;
      procedure DrawButtonGlyph(Canvas: TCanvas; const GlyphPos: TPoint; State: TButtonState;
         Transparent: Boolean);
      procedure DrawButtonText(Canvas: TCanvas; const Caption: string;
         TextBounds: TRect; State: TButtonState; BiDiFlags: LongInt; Font3D, Hilight3d: TFont3D;
         hilightStyle: TTextHilightStyle; Hilighted: TsHilighted);
      procedure CalcButtonLayout(Canvas: TCanvas; const Client: TRect;
         const Offset: TPoint; const Caption: string; Layout: TButtonLayout;
         Margin, Spacing: Integer; var GlyphPos: TPoint; var TextBounds: TRect;
         BiDiFlags: Longint);
   public
      constructor Create;
      destructor Destroy; override;
      function Draw(Canvas: TCanvas; const Client: TRect; const Offset: TPoint;
         const Caption: string; Layout: TButtonLayout; Margin, Spacing: Integer;
         State: TButtonState; Transparent: Boolean; BiDiFlags: Longint;
         Font3D, Hilight3d: TFont3D; hilightStyle: TTextHilightStyle; Hilighted: TsHilighted): TRect;
      property Glyph: TBitmap read FOriginal write SetGlyph;
      property NumGlyphs: TNumGlyphs read FNumGlyphs write SetNumGlyphs;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
   end;

var
   GlyphCache: TGlyphCache = nil;
   ButtonCount: Integer = 0;

constructor TButtonGlyph.Create;
var
   I: TButtonState;
begin
   inherited Create;
   FOriginal := TBitmap.Create;
   FOriginal.OnChange := GlyphChanged;
   FPaint3D := TFont3D.Create(nil);
   FTransparentColor := clOlive;
   FNumGlyphs := 1;
   for I := Low(I) to High(I) do
      FIndexs[I] := -1;
   if GlyphCache = nil then
      GlyphCache := TGlyphCache.Create;
end;

destructor TButtonGlyph.Destroy;
begin
   FOriginal.Free;
   Invalidate;
   if Assigned(GlyphCache) and GlyphCache.Empty then begin
      GlyphCache.Free;
      GlyphCache := nil;
   end;
   FPaint3D.Free;
   inherited Destroy;
end;

procedure TButtonGlyph.Invalidate;
var
   I: TButtonState;
begin
   for I := Low(I) to High(I) do begin
      if FIndexs[I] <> -1 then
         FGlyphList.Delete(FIndexs[I]);
      FIndexs[I] := -1;
   end;
   GlyphCache.ReturnList(FGlyphList);
   FGlyphList := nil;
end;

procedure TButtonGlyph.GlyphChanged(Sender: TObject);
begin
   if Sender = FOriginal then begin
      FTransparentColor := FOriginal.TransparentColor;
      Invalidate;
      if Assigned(FOnChange) then
         FOnChange(Self);
   end;
end;

procedure TButtonGlyph.SetGlyph(Value: TBitmap);
var
   Glyphs: Integer;
begin
   Invalidate;
   FOriginal.Assign(Value);
   if (Value <> nil) and (Value.Height > 0) then begin
      FTransparentColor := Value.TransparentColor;
      if Value.Width mod Value.Height = 0 then begin
         Glyphs := Value.Width div Value.Height;
         if Glyphs > 4 then
            Glyphs := 1;
         SetNumGlyphs(Glyphs);
      end;
   end;
end;

procedure TButtonGlyph.SetNumGlyphs(Value: TNumGlyphs);
begin
   if (Value <> FNumGlyphs) and (Value > 0) then begin
      Invalidate;
      FNumGlyphs := Value;
      GlyphChanged(Glyph);
   end;
end;

function TButtonGlyph.CreateButtonGlyph(State: TButtonState): Integer;
const
   ROP_DSPDxax = $00E20746;
var
   TmpImage, DDB, MonoBmp: TBitmap;
   IWidth, IHeight: Integer;
   IRect, ORect: TRect;
   I: TButtonState;
   DestDC: HDC;
begin
   if (State = bsDown) and (NumGlyphs < 3) then
      State := bsUp;
   Result := FIndexs[State];
   if Result <> -1 then
      Exit;
   if (FOriginal.Width or FOriginal.Height) = 0 then
      Exit;
   IWidth := FOriginal.Width div FNumGlyphs;
   IHeight := FOriginal.Height;
   if FGlyphList = nil then begin
      if GlyphCache = nil then
         GlyphCache := TGlyphCache.Create;
      FGlyphList := GlyphCache.GetList(IWidth, IHeight);
   end;
   TmpImage := TBitmap.Create;
   try
      TmpImage.Width := IWidth;
      TmpImage.Height := IHeight;
      IRect := Rect(0, 0, IWidth, IHeight);
      TmpImage.Canvas.Brush.Color := clBtnFace;
      TmpImage.Palette := CopyPalette(FOriginal.Palette);
      I := State;
      if Ord(I) >= NumGlyphs then
         I := bsUp;
      ORect := Rect(Ord(I) * IWidth, 0, (Ord(I) + 1) * IWidth, IHeight);
      case State of
         bsUp, bsDown, bsExclusive: begin
            TmpImage.Canvas.CopyRect(IRect, FOriginal.Canvas, ORect);
            if FOriginal.TransparentMode = tmFixed then
               FIndexs[State] := FGlyphList.AddMasked(TmpImage, FTransparentColor)
            else
               FIndexs[State] := FGlyphList.AddMasked(TmpImage, clDefault);
         end;
         bsDisabled: begin
            MonoBmp := nil;
            DDB := nil;
            try
               MonoBmp := TBitmap.Create;
               DDB := TBitmap.Create;
               DDB.Assign(FOriginal);
               DDB.HandleType := bmDDB;
               if NumGlyphs > 1 then with TmpImage.Canvas do begin
                  { Change white & gray to clBtnHighlight and clBtnShadow }
                  CopyRect(IRect, DDB.Canvas, ORect);
                  MonoBmp.Monochrome := True;
                  MonoBmp.Width := IWidth;
                  MonoBmp.Height := IHeight;

                  { Convert white to clBtnHighlight }
                  DDB.Canvas.Brush.Color := clWhite;
                  MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
                  Brush.Color := clBtnHighlight;
                  DestDC := Handle;
                  SetTextColor(DestDC, clBlack);
                  SetBkColor(DestDC, clWhite);
                  BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);

                  { Convert gray to clBtnShadow }
                  DDB.Canvas.Brush.Color := clGray;
                  MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
                  Brush.Color := clBtnShadow;
                  DestDC := Handle;
                  SetTextColor(DestDC, clBlack);
                  SetBkColor(DestDC, clWhite);
                  BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);

                  { Convert transparent color to clBtnFace }
                  DDB.Canvas.Brush.Color := ColorToRGB(FTransparentColor);
                  MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
                  Brush.Color := clBtnFace;
                  DestDC := Handle;
                  SetTextColor(DestDC, clBlack);
                  SetBkColor(DestDC, clWhite);
                  BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
               end else begin
                  { Create a disabled version }
                  with MonoBmp do begin
                     Assign(FOriginal);
                     HandleType := bmDDB;
                     Canvas.Brush.Color := clBlack;
                     Width := IWidth;
                     if Monochrome then begin
                        Canvas.Font.Color := clWhite;
                        Monochrome := False;
                        Canvas.Brush.Color := clWhite;
                     end;
                     Monochrome := True;
                  end;
                  with TmpImage.Canvas do begin
                     Brush.Color := clBtnFace;
                     FillRect(IRect);
                     Brush.Color := clBtnHighlight;
                     SetTextColor(Handle, clBlack);
                     SetBkColor(Handle, clWhite);
                     BitBlt(Handle, 1, 1, IWidth, IHeight,
                        MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
                     Brush.Color := clBtnShadow;
                     SetTextColor(Handle, clBlack);
                     SetBkColor(Handle, clWhite);
                     BitBlt(Handle, 0, 0, IWidth, IHeight,
                        MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
                  end;
               end;
            finally
               DDB.Free;
               MonoBmp.Free;
            end;
            FIndexs[State] := FGlyphList.AddMasked(TmpImage, clDefault);
         end;
      end;
   finally
      TmpImage.Free;
   end;
   Result := FIndexs[State];
   FOriginal.Dormant;
end;

procedure TButtonGlyph.DrawButtonGlyph(Canvas: TCanvas; const GlyphPos: TPoint;
   State: TButtonState; Transparent: Boolean);
var
   Index: Integer;
begin
   if FOriginal = nil then
      Exit;
   if (FOriginal.Width = 0) or (FOriginal.Height = 0) then
      Exit;
   Index := CreateButtonGlyph(State);
   with GlyphPos do
      //if Transparent or (State = bsExclusive) then
         ImageList_DrawEx(FGlyphList.Handle, Index, Canvas.Handle, X, Y, 0, 0,
            clNone, clNone, ILD_Transparent)
      {else
         ImageList_DrawEx(FGlyphList.Handle, Index, Canvas.Handle, X, Y, 0, 0,
            ColorToRGB(clBtnFace), clNone, ILD_Normal);}
end;

procedure TButtonGlyph.DrawButtonText(Canvas: TCanvas; const Caption: string;
      TextBounds: TRect; State: TButtonState; BiDiFlags: LongInt; Font3D, Hilight3d: TFont3D;
      hilightStyle: TTextHilightStyle; Hilighted: TsHilighted);
begin
   FPaint3D.Assign(Font3d);
   if State = bsDisabled then begin
      if NewStyleControls then begin
         FPaint3D.Position := spRightBottom;
         FPaint3D.Spacing := 1;
         FPaint3D.Active := TRUE;
      end;
      Canvas.Font.Color := clGrayText;
      FPaint3D.Color := clBtnHighlight;
      hilightStyle := hsNone;
   end;
   Canvas.Brush.Style := bsClear;
   PaintText( Canvas, PChar(Caption), Length(Caption), TextBounds,
      DT_CENTER or DT_VCENTER or BiDiFlags, FPaint3D, Hilight3d, hilightStyle, Hilighted);
end;

procedure TButtonGlyph.CalcButtonLayout(Canvas: TCanvas; const Client: TRect;
   const Offset: TPoint; const Caption: string; Layout: TButtonLayout; Margin,
   Spacing: Integer; var GlyphPos: TPoint; var TextBounds: TRect;
   BiDiFlags: LongInt);
var
   TextPos: TPoint;
   ClientSize, GlyphSize, TextSize: TPoint;
   TotalSize: TPoint;
begin
   if (BiDiFlags and DT_RIGHT) = DT_RIGHT then
      if Layout = blGlyphLeft then
         Layout := blGlyphRight
      else if Layout = blGlyphRight then
         Layout := blGlyphLeft;
   { calculate the item sizes }
   ClientSize := Point(Client.Right - Client.Left, Client.Bottom - Client.Top);

   if FOriginal <> nil then
      GlyphSize := Point(FOriginal.Width div FNumGlyphs, FOriginal.Height)
   else
      GlyphSize := Point(0, 0);

   if Length(Caption) > 0 then begin
      TextBounds := Rect(0, 0, Client.Right - Client.Left, 0);
      DrawText(Canvas.Handle, PChar(Caption), Length(Caption), TextBounds, DT_CALCRECT or BiDiFlags);
      TextSize := Point(TextBounds.Right - TextBounds.Left, TextBounds.Bottom - TextBounds.Top);
   end else begin
      TextBounds := Rect(0, 0, 0, 0);
      TextSize := Point(0,0);
   end;

   { If the layout has the glyph on the right or the left, then both the
    text and the glyph are centered vertically.  If the glyph is on the top
    or the bottom, then both the text and the glyph are centered horizontally.}
   if Layout in [blGlyphLeft, blGlyphRight] then begin
      GlyphPos.Y := (ClientSize.Y - GlyphSize.Y + 1) div 2;
      TextPos.Y := (ClientSize.Y - TextSize.Y + 1) div 2;
   end else begin
      GlyphPos.X := (ClientSize.X - GlyphSize.X + 1) div 2;
      TextPos.X := (ClientSize.X - TextSize.X + 1) div 2;
   end;

   { if there is no text or no bitmap, then Spacing is irrelevant }
   if (TextSize.X = 0) or (GlyphSize.X = 0) then
      Spacing := 0;

   { adjust Margin and Spacing }
   if Margin = -1 then begin
      if Spacing = -1 then begin
         TotalSize := Point(GlyphSize.X + TextSize.X, GlyphSize.Y + TextSize.Y);
         if Layout in [blGlyphLeft, blGlyphRight] then
            Margin := (ClientSize.X - TotalSize.X) div 3
         else
            Margin := (ClientSize.Y - TotalSize.Y) div 3;
         Spacing := Margin;
      end else begin
         TotalSize := Point(GlyphSize.X + Spacing + TextSize.X, GlyphSize.Y +
            Spacing + TextSize.Y);
         if Layout in [blGlyphLeft, blGlyphRight] then
            Margin := (ClientSize.X - TotalSize.X + 1) div 2
         else
            Margin := (ClientSize.Y - TotalSize.Y + 1) div 2;
      end;
   end else begin
      if Spacing = -1 then begin
         TotalSize := Point(ClientSize.X - (Margin + GlyphSize.X), ClientSize.Y -
            (Margin + GlyphSize.Y));
         if Layout in [blGlyphLeft, blGlyphRight] then
            Spacing := (TotalSize.X - TextSize.X) div 2
         else
            Spacing := (TotalSize.Y - TextSize.Y) div 2;
      end;
   end;

   case Layout of
      blGlyphLeft: begin
         GlyphPos.X := Margin;
         TextPos.X := GlyphPos.X + GlyphSize.X + Spacing;
      end;
      blGlyphRight: begin
         GlyphPos.X := ClientSize.X - Margin - GlyphSize.X;
         TextPos.X := GlyphPos.X - Spacing - TextSize.X;
      end;
      blGlyphTop: begin
         GlyphPos.Y := Margin;
         TextPos.Y := GlyphPos.Y + GlyphSize.Y + Spacing;
      end;
      blGlyphBottom: begin
         GlyphPos.Y := ClientSize.Y - Margin - GlyphSize.Y;
         TextPos.Y := GlyphPos.Y - Spacing - TextSize.Y;
      end;
   end;

   { fixup the result variables }
   with GlyphPos do begin
      Inc(X, Client.Left + Offset.X);
      Inc(Y, Client.Top + Offset.Y);
   end;
   OffsetRect(TextBounds, TextPos.X + Client.Left + Offset.X,
      TextPos.Y + Client.Top + Offset.X);
end;

function TButtonGlyph.Draw(Canvas: TCanvas; const Client: TRect; const Offset: TPoint;
         const Caption: string; Layout: TButtonLayout; Margin, Spacing: Integer;
         State: TButtonState; Transparent: Boolean; BiDiFlags: Longint;
         Font3D, Hilight3d: TFont3D; hilightStyle: TTextHilightStyle; Hilighted: TsHilighted): TRect;
var
   GlyphPos: TPoint;
begin
   CalcButtonLayout(Canvas, Client, Offset, Caption, Layout, Margin, Spacing,
      GlyphPos, Result, BiDiFlags);
   DrawButtonGlyph(Canvas, GlyphPos, State, Transparent);
   DrawButtonText(Canvas, Caption, Result, State, BiDiFlags,
      Font3D, Hilight3d, HilightStyle, Hilighted);
end;

{ TsSpeedButton }
constructor TsSpeedButton.Create(AOwner: TComponent);
begin
   FGlyph := TButtonGlyph.Create;
   TButtonGlyph(FGlyph).OnChange := GlyphChanged;

   inherited Create(AOwner);

   SetBounds(0, 0, 23, 22);
   ControlStyle := [csCaptureMouse, csDoubleClicks];
   ParentFont := True;
   Color := clBtnFace;

   FFlat := DEF_SpeedBtn_Flat;
   FFont3d := TFont3d.Create(Font3dChanged);
   FGlyphList := nil;
   FGlyphId := -1;
   FHilight3d := THilight3d.Create(Font3dChanged);
   FHilightStyle := DEF_SpeedBtn_Hilight;
   FLayout := blGlyphLeft;
   FMargin := -1;
   FModifiedGlyph := False;
   FSpacing := 4;
   Transparent := DEF_SpeedBtn_Tramsparent;
   Inc(ButtonCount);
end;

destructor TsSpeedButton.Destroy;
begin
   Dec(ButtonCount);
   inherited Destroy;
   FFont3D.Free;
   FHilight3D.Free;
   TButtonGlyph(FGlyph).Free;
end;

procedure TsSpeedButton.Paint;
const
   DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
   FillStyles: array[Boolean] of Integer = (BF_MIDDLE, 0);
var
   PaintRect: TRect;
   DrawFlags: Integer;
   Offset: TPoint;
   hilighted: TsHilighted;
begin
   if not Enabled then begin
      FState := bsDisabled;
      FDragging := False;
   end else if FState = bsDisabled then
      if FDown and (GroupIndex <> 0) then
         FState := bsExclusive
      else
         FState := bsUp;

   Canvas.Font := Self.Font;
   PaintRect := Rect(0, 0, Width, Height);
   if not FFlat then begin
      DrawFlags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
      if FState in [bsDown, bsExclusive] then
         DrawFlags := DrawFlags or DFCS_PUSHED;
      DrawFrameControl(Canvas.Handle, PaintRect, DFC_BUTTON, DrawFlags);
      Canvas.Brush.Color := Color;
      Canvas.FillRect(PaintRect);
   end else begin

      if (FState in [bsDown, bsExclusive]) or
         (FMouseInControl and (FState <> bsDisabled)) or
         (csDesigning in ComponentState) then
         DrawEdge(Canvas.Handle, PaintRect, DownStyles[FState in [bsDown, bsExclusive]],
            FillStyles[Transparent] or BF_RECT)
      else if not Transparent then begin
         Canvas.Brush.Color := Color;
         Canvas.FillRect(PaintRect);
      end;

      InflateRect(PaintRect, -1, -1);
      if not Transparent then begin
         Canvas.Brush.Color := Color;
         Canvas.FillRect(PaintRect);
      end;
   end;

   if FState in [bsDown, bsExclusive] then begin
      if (FState = bsExclusive) and (not FFlat or not FMouseInControl) then begin
         Canvas.Brush.Bitmap := AllocPatternBitmap(clBtnFace, clBtnHighlight);
         Canvas.FillRect(PaintRect);
      end;
      Offset.X := 1;
      Offset.Y := 1;
   end else begin
      Offset.X := 0;
      Offset.Y := 0;
   end;

   hilighted := hNone;
   if Enabled and not (csDesigning in ComponentState) then begin
      if FState = bsDown then
         hilighted := hHilight
      else if FMouseInControl then
         hilighted := hMouseOver;
   end;
   TButtonGlyph(FGlyph).Draw(Canvas, PaintRect, Offset, Caption, FLayout, FMargin,
      FSpacing, FState, Transparent, DrawTextBiDiModeFlags(0),
      FFont3D, FHilight3d, FHilightStyle, hilighted);
end;

procedure TsSpeedButton.UpdateTracking;
var
   P: TPoint;
begin
   if FFlat or (FHilightStyle <> hsNone) or
      (FHilight3D.Active and not FHilight3D.Equal(FFont3D)) then begin
      if Enabled then begin
         GetCursorPos(P);
         FMouseInControl := not (FindDragTarget(P, True) = Self);
         if FMouseInControl then
            Perform(CM_MOUSELEAVE, 0, 0)
         else
            Perform(CM_MOUSEENTER, 0, 0);
      end;
   end;
end;

procedure TsSpeedButton.Loaded;
var
   State: TButtonState;
begin
   inherited Loaded;
   if Enabled then
      State := bsUp
   else
      State := bsDisabled;
   if FGlyphList <> nil then
      Glyph := TBitMap(FGlyphList[FGlyphId]);
   TButtonGlyph(FGlyph).CreateButtonGlyph(State);
end;

procedure TsSpeedButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   inherited MouseDown(Button, Shift, X, Y);
   if (Button = mbLeft) and Enabled then begin
      if not FDown then begin
         FState := bsDown;
         Invalidate;
      end;
      FDragging := True;
   end;
end;

procedure TsSpeedButton.MouseMove(Shift: TShiftState; X, Y: Integer);
var
   NewState: TButtonState;
begin
   inherited MouseMove(Shift, X, Y);
   if FDragging then begin
      if not FDown then
         NewState := bsUp
      else
         NewState := bsExclusive;
      if (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight) then
         if FDown then
            NewState := bsExclusive else NewState := bsDown;
      if NewState <> FState then begin
         FState := NewState;
         Invalidate;
      end;
   end else if not FMouseInControl then
      UpdateTracking;
end;
    
procedure TsSpeedButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
   DoClick: Boolean;
begin
   inherited MouseUp(Button, Shift, X, Y);
   if FDragging then begin
      FDragging := False;
      DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
      if FGroupIndex = 0 then begin
         { Redraw face in-case mouse is captured }
         FState := bsUp;
         FMouseInControl := False;
         if DoClick and not (FState in [bsExclusive, bsDown]) then
            Invalidate;
      end else if DoClick then begin
         SetDown(not FDown);
         if FDown then
            Repaint;
      end else begin
         if FDown then
            FState := bsExclusive;
         Repaint;
      end;
      if DoClick then
         Click;
      UpdateTracking;
   end;
end;
    
procedure TsSpeedButton.Click;
var
   Form: TCustomForm;
begin
   Form := GetParentForm(Self);
   if Form <> nil then
      Form.ModalResult := ModalResult;
   inherited Click;
end;


procedure TsSpeedButton.Notification(AComponent: TComponent; Operation: TOperation);
begin
	inherited;
   if (AComponent = FGlyphList) and (Operation = opRemove) then begin
   	FGlyphList := nil;
      Glyph := nil;
   end;
end;

function TsSpeedButton.GetPalette: HPALETTE;
begin
   Result := Glyph.Palette;
end;

function TsSpeedButton.GetGlyph: TBitmap;
begin
	Result := nil;
   if (FGlyphList <> nil) and (FGlyphId > -1) then
   	Result := TBitMap(FGlyphList[FGlyphId]);
   if Result = nil then
      Result := TButtonGlyph(FGlyph).Glyph;
end;

procedure TsSpeedButton.SetGlyph(Value: TBitmap);
begin
   TButtonGlyph(FGlyph).Glyph := Value;
   Invalidate;
   FModifiedGlyph := TRUE;
end;

function TsSpeedButton.GlyphStored: Boolean;
begin
   Result := (Kind = bkCustom) and ((GlyphList = nil) or (GlyphListId = -1));
end;

function TsSpeedButton.GetNumGlyphs: TNumGlyphs;
begin
   Result := TButtonGlyph(FGlyph).NumGlyphs;
end;

procedure TsSpeedButton.SetNumGlyphs(Value: TNumGlyphs);
begin
   if Value < 0 then
      Value := 1
   else if Value > 4 then
      Value := 4;
   if Value <> TButtonGlyph(FGlyph).NumGlyphs then begin
      TButtonGlyph(FGlyph).NumGlyphs := Value;
      Invalidate;
   end;
end;

procedure TsSpeedButton.GlyphChanged(Sender: TObject);
begin
   Invalidate;
end;

procedure TsSpeedButton.UpdateExclusive;
var
   Msg: TMessage;
begin
   if (FGroupIndex <> 0) and (Parent <> nil) then begin
      Msg.Msg := CM_BUTTONPRESSED;
      Msg.WParam := FGroupIndex;
      Msg.LParam := Longint(Self);
      Msg.Result := 0;
      Parent.Broadcast(Msg);
   end;
end;

procedure TsSpeedButton.SetDown(Value: Boolean);
begin
   if FGroupIndex = 0 then
      Value := False;
   if Value <> FDown then begin
      if FDown and (not FAllowAllUp) then
         Exit;
      FDown := Value;
      if Value then begin
         if FState = bsUp then
            Invalidate;
         FState := bsExclusive
      end else begin
         FState := bsUp;
         Repaint;
      end;
      if Value then
         UpdateExclusive;
   end;
end;

procedure TsSpeedButton.SetFlat(Value: Boolean);
begin
   if Value <> FFlat then begin
      FFlat := Value;
      Invalidate;
   end;
end;
    
procedure TsSpeedButton.SetGroupIndex(Value: Integer);
begin
   if FGroupIndex <> Value then begin
      FGroupIndex := Value;
      UpdateExclusive;
   end;
end;

procedure TsSpeedButton.SetLayout(Value: TButtonLayout);
begin
   if FLayout <> Value then begin
      FLayout := Value;
      Invalidate;
   end;
end;

procedure TsSpeedButton.SetMargin(Value: Integer);
begin
   if (Value <> FMargin) and (Value >= -1) then begin
      FMargin := Value;
      Invalidate;
   end;
end;

procedure TsSpeedButton.SetSpacing(Value: Integer);
begin
   if Value <> FSpacing then begin
      FSpacing := Value;
      Invalidate;
   end;
end;

procedure TsSpeedButton.SetTransparent(Value: Boolean);
begin
   if Value <> FTransparent then begin
      FTransparent := Value;
      if Value then
         ControlStyle := ControlStyle - [csOpaque]
      else
         ControlStyle := ControlStyle + [csOpaque];
      Invalidate;
   end;
end;

procedure TsSpeedButton.SetAllowAllUp(Value: Boolean);
begin
   if FAllowAllUp <> Value then begin
      FAllowAllUp := Value;
      UpdateExclusive;
   end;
end;

procedure TsSpeedButton.WMLButtonDblClk(var Message: TWMLButtonDown);
begin
   inherited;
   if FDown then
      DblClick;
end;

procedure TsSpeedButton.CMEnabledChanged(var Message: TMessage);
const
   NewState: array[Boolean] of TButtonState = (bsDisabled, bsUp);
begin
   TButtonGlyph(FGlyph).CreateButtonGlyph(NewState[Enabled]);
   UpdateTracking;
   Repaint;
end;

procedure TsSpeedButton.CMButtonPressed(var Message: TMessage);
var
   Sender: TsSpeedButton;
begin
   if Message.WParam = FGroupIndex then begin
      Sender := TsSpeedButton(Message.LParam);
      if Sender <> Self then begin
         if Sender.Down and FDown then begin
            FDown := False;
            FState := bsUp;
            Invalidate;
         end;
         FAllowAllUp := Sender.AllowAllUp;
      end;
   end;
end;

procedure TsSpeedButton.STMGlyphIdChanged(var Message: TMessage);
begin
	if Message.WParam = FGlyphId then begin
      if Message.lParam = Message.wParam then
         Glyph := TBitMap(FGlyphList[FGlyphId])
      else
			GlyphListId := Message.LParam;
      FModifiedGlyph := FALSE;
   end;
end;

procedure TsSpeedButton.CNCommand(var Message: TWMCommand);
begin
   if Message.NotifyCode = BN_CLICKED then
      Click;
end;

procedure TsSpeedButton.CMDialogKey(var Message: TCMDialogKey);
begin
   with Message do
      if Enabled and Visible and (((CharCode = VK_RETURN) and Default) or
         ((CharCode = VK_ESCAPE) and FCancel)) and Parent.Showing and
         (KeyDataToShiftState(Message.KeyData) = []) then begin
         Click;
         Result := 1;
      end else
         inherited;
end;

procedure TsSpeedButton.CMDialogChar(var Message: TCMDialogChar);
begin
   with Message do
      if IsAccel(CharCode, Caption) and Enabled and Visible and
         (Parent <> nil) and Parent.Showing then begin
         Click;
         Result := 1;
      end else
         inherited;
end;

procedure TsSpeedButton.CMFontChanged(var Message: TMessage);
begin
   Invalidate;
end;

procedure TsSpeedButton.CMTextChanged(var Message: TMessage);
begin
   Invalidate;
end;

procedure TsSpeedButton.CMSysColorChange(var Message: TMessage);
begin
   with TButtonGlyph(FGlyph) do begin
      Invalidate;
      CreateButtonGlyph(FState);
   end;
end;

procedure TsSpeedButton.CMMouseEnter(var Message: TMessage);
begin
   inherited;
   { Don't draw a border if DragMode <> dmAutomatic since this button is meant to
    be used as a dock client. }
   if (FFlat or (FHilightStyle <> hsNone) or
      (FHilight3D.Active and not FHilight3D.Equal(FFont3D))) and
      not FMouseInControl and Enabled and (DragMode <> dmAutomatic)
      and (GetCapture = 0) then begin
      FMouseInControl := True;
      Repaint;
   end;
end;

procedure TsSpeedButton.CMMouseLeave(var Message: TMessage);
begin
   inherited;
   if (FFlat or (FHilightStyle <> hsNone) or
      (FHilight3D.Active and not FHilight3D.Equal(FFont3D))) and
      FMouseInControl and Enabled and not FDragging then begin
      FMouseInControl := False;
      Invalidate;
   end;
end;

procedure TsSpeedButton.ActionChange(Sender: TObject; CheckDefaults: Boolean);
   procedure CopyImage(ImageList: TCustomImageList; Index: Integer);
   begin
      with Glyph do begin
         Width := ImageList.Width;
         Height := ImageList.Height;
         Canvas.Brush.Color := clFuchsia;//! for lack of a better color
         Canvas.FillRect(Rect(0,0, Width, Height));
         ImageList.Draw(Canvas, 0, 0, Index);
      end;
   end;

begin
   inherited ActionChange(Sender, CheckDefaults);
   if Sender is TCustomAction then with TCustomAction(Sender) do begin
      { Copy image from action's imagelist }
      if (Glyph.Empty) and (ActionList <> nil) and (ActionList.Images <> nil) and
         (ImageIndex >= 0) and (ImageIndex < ActionList.Images.Count) then
         CopyImage(ActionList.Images, ImageIndex);
   end;
end;


procedure TsSpeedButton.SetGlyphList(Value: TsGlyphList);
begin
	if FGlyphList <> Value then begin
      if FGlyphList <> nil then
			FGlyphList.ChangeNotification(self, FALSE);
   	FGlyphList := Value;
      if FGlyphList <> nil then begin
			FGlyphList.ChangeNotification(self, TRUE);
         Glyph := TBitMap(FGlyphList[FGlyphId]);
      end else
         Glyph := nil;
   end;
end;

procedure TsSpeedButton.SetGlyphId(Value: Integer);
begin
	if FGlyphId <> Value then begin
		if csLoading in ComponentState then
   		FGlyphId := Value
		else begin
   		if FGlyphList = nil then
   			Exit;
      	if (Value > -1) then begin
            if FGlyphList[Value] = nil then
               Raise Exception.CreateFmt(SErrorGlyphNotFound, [Value]);
            if not FGlyphList[Value].InheritsFrom(TBitMap) then
      		   Raise Exception.Create(SErrorNeedsBmp);
         end;
   		FGlyphId := Value;
         Glyph := TBitMap(FGlyphList[FGlyphId]);
      end;
   end;
end;

procedure TsSpeedButton.SetKind(Value: TBitBtnKind);
begin
   if Value <> FKind then begin
      if Value <> bkCustom then begin
         Default := Value in [bkOK, bkYes];
         Cancel := Value in [bkCancel, bkNo];
         ModalResult := BitBtnModalResults[Value];
         Glyph := GetBitBtnGlyph(Value);
         NumGlyphs := 2;
         FModifiedGlyph := FALSE;
      end;
      FKind := Value;
      Invalidate;
   end;
end;

function TsSpeedButton.GetKind: TBitBtnKind;
begin
   if FKind <> bkCustom then
      if ((FKind in [bkOK, bkYes]) xor Default) or
         ((FKind in [bkCancel, bkNo]) xor Cancel) or
         (ModalResult <> BitBtnModalResults[FKind]) or
         FModifiedGlyph then
         FKind := bkCustom;
   Result := FKind;
end;

function TsSpeedButton.IsCustom: Boolean;
begin
   Result := Kind = bkCustom;
end;

procedure TsSpeedButton.Font3dChanged(Sender: Tobject);
begin
   Invalidate;
end;

procedure TsSpeedButton.SetFont3d(Value: TFont3d);
begin
   if FFont3d <> Value then
      FFont3d.Assign(Value);
end;

procedure TsSpeedButton.SetHilight3d(Value: THilight3d);
begin
   if FHilight3d <> Value then
      FHilight3d.Assign( Value);
end;

procedure TsSpeedButton.SetHilightStyle(Value: TTextHilightStyle);
begin
   if Value <> FHilightStyle then begin
      FHilightStyle := Value;
      if MouseInControl then
         Invalidate;
   end;
end;

{ TsBitBtn }

constructor TsBitBtn.Create(AOwner: TComponent);
begin
   FGlyph := TButtonGlyph.Create;
   TButtonGlyph(FGlyph).OnChange := GlyphChanged;
   inherited Create(AOwner);
   FCanvas := TCanvas.Create;
   FStyle := bsAutoDetect;
   FKind := bkCustom;
   FLayout := blGlyphLeft;
   FSpacing := 4;
   FMargin := -1;
   ControlStyle := ControlStyle + [csReflector];

   FGlyphList := nil;
   FGlyphId := -1;
   FModifiedGlyph := FALSE;

   FFont3d := TFont3d.Create(Font3dChanged);
   FHilight3d := THilight3d.Create(Font3dChanged);
   FHilightStyle := DEF_BitBtn_Hilight;

   FBtnStyle := DEF_BitBtn_Style;
end;

destructor TsBitBtn.Destroy;
begin
   inherited Destroy;
   FFont3D.Free;
   FHilight3D.Free;
   TButtonGlyph(FGlyph).Free;
   FCanvas.Free;
end;

procedure TsBitBtn.CreateHandle;
//var
//   State: TButtonState;
begin
   if Enabled then
      FState := bsUp
   else
      FState := bsDisabled;
   inherited CreateHandle;
   TButtonGlyph(FGlyph).CreateButtonGlyph(FState);
end;

procedure TsBitBtn.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
   with Params do Style := Style or BS_OWNERDRAW;
end;

procedure TsBitBtn.Loaded;
begin
   inherited;
   if FGlyphList <> nil then
      Glyph := TBitMap(FGlyphList[FGlyphId])
end;

procedure TsBitBtn.Notification(AComponent: TComponent; Operation: TOperation);
begin
	inherited;
   if (AComponent = FGlyphList) and (Operation = opRemove) then begin
   	FGlyphList := nil;
      Glyph := nil;
   end;
end;

procedure TsBitBtn.SetButtonStyle(ADefault: Boolean);
begin
   if ADefault <> FFocused then begin
      FFocused := ADefault;
      Refresh;
   end;
end;

procedure TsBitBtn.Click;
var
   Form: TCustomForm;
   Control: TWinControl;
begin
   case FKind of
      bkClose: begin
         Form := GetParentForm(Self);
         if Form <> nil then
            Form.Close
         else
            inherited Click;
      end;
      bkHelp: begin
         Control := Self;
         while (Control <> nil) and (Control.HelpContext = 0) do
            Control := Control.Parent;
         if Control <> nil then
            Application.HelpContext(Control.HelpContext)
         else
            inherited Click;
      end;
      else
         inherited Click;
   end;
end;

procedure TsBitBtn.CNMeasureItem(var Message: TWMMeasureItem);
begin
   with Message.MeasureItemStruct^ do begin
      itemWidth := Width;
      itemHeight := Height;
   end;
end;

procedure TsBitBtn.CNDrawItem(var Message: TWMDrawItem);
begin
   DrawItem(Message.DrawItemStruct^);
end;

procedure TsBitBtn.InvalidateCaption;
var
   hilighted: TsHilighted;
   R: TRect;
   GlyphPos: TPoint;
begin
   R := ClientRect;
   hilighted := hNone;
   if Enabled then begin
      if Focused then
         hilighted := hHilight
      else if FMouseInControl then
         hilighted := hMouseOver;
   end;

   FCanvas.Handle := GetDC(Handle);

   TButtonGlyph(FGlyph).CalcButtonLayout(FCanvas, R, Point(0,0), Caption, FLayout, FMargin,
      FSpacing, GlyphPos, R, DrawTextBiDiModeFlags(0));

   FCanvas.Brush.Color := Color;
   FCanvas.FillRect(R);

   TButtonGlyph(FGlyph).DrawButtonText(FCanvas, Caption, R, FState, DrawTextBiDiModeFlags(0),
      FFont3D, FHilight3d, FHilightStyle, Hilighted);
   ReleaseDC(Handle, FCanvas.Handle);
   FCanvas.Handle := 0;
end;

procedure TsBitBtn.DrawItem(const DrawItemStruct: TDrawItemStruct);
const
   DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
var
   IsDown, IsDefault: Boolean;
   R: TRect;
   Flags: Longint;
   hilighted: TsHilighted;
begin
   FCanvas.Handle := DrawItemStruct.hDC;
   R := ClientRect;

   with DrawItemStruct do begin
      IsDown := itemState and ODS_SELECTED <> 0;
      IsDefault := itemState and ODS_FOCUS <> 0;

      if not Enabled then
         FState := bsDisabled
      else if IsDown then
         FState := bsDown
      else
         FState := bsUp;
   end;

   if FBtnStyle in [btsFlat, btsFramed] then begin
      FCanvas.Brush.Color := Color;
      FCanvas.FillRect(R);

      if (FState = bsDown) or (FMouseInControl and (FState <> bsDisabled)) or
         ((FBtnStyle = btsFlat) and (csDesigning in ComponentState)) then
         DrawFlatEDGE(FCanvas, R, IsDown)
      else if FBtnStyle = btsFramed then begin
         FCanvas.Pen.Color := clWindowFrame;
         FCanvas.Pen.Width := 1;
         FCanvas.Brush.Style := bsClear;
         FCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
      end;
   end else begin
      Flags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
      if IsDown then
         Flags := Flags or DFCS_PUSHED;
      if DrawItemStruct.itemState and ODS_DISABLED <> 0 then
         Flags := Flags or DFCS_INACTIVE;

     { DrawFrameControl doesn't allow for drawing a button as the
      default button, so it must be done here. }
      if FFocused or IsDefault then begin
         FCanvas.Pen.Color := clWindowFrame;
         FCanvas.Pen.Width := 1;
         FCanvas.Brush.Style := bsClear;
         FCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);

         { DrawFrameControl must draw within this border }
         InflateRect(R, -1, -1);
      end;
      { DrawFrameControl does not draw a pressed button correctly }
      if IsDown then begin
         FCanvas.Pen.Color := clBtnShadow;
         FCanvas.Pen.Width := 1;
         FCanvas.Brush.Color := Color;//clBtnFace;
         FCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
         InflateRect(R, -1, -1);
      end else begin
         DrawFrameControl(DrawItemStruct.hDC, R, DFC_BUTTON, Flags);
         //R := ClientRect;
         //InflateRect(R, -2, -2);
         FCanvas.Brush.Color := Color;
         FCanvas.FillRect(R);
      end;
   end;

   if FFocused then begin
      R := ClientRect;
      InflateRect(R, -1, -1);
   end;

   FCanvas.Font := Self.Font;
   if IsDown then
      OffsetRect(R, 1, 1);

   hilighted := hNone;
   if Enabled then begin
      if Focused then
         hilighted := hHilight
      else if FMouseInControl then
         hilighted := hMouseOver;
   end;


   TButtonGlyph(FGlyph).Draw(FCanvas, R, Point(0,0), Caption, FLayout, FMargin,
      FSpacing, FState, False, DrawTextBiDiModeFlags(0), FFont3D, FHilight3d,
      FHilightStyle, hilighted);

   if FFocused and IsDefault then begin
      R := ClientRect;
      InflateRect(R, -3, -3);
      FCanvas.Pen.Color := clWindowFrame;
     	FCanvas.Brush.Color := Color;//clBtnFace;
      DrawFocusRect(FCanvas.Handle, R);
   end;

   FCanvas.Handle := 0;
end;

procedure TsBitBtn.CMFontChanged(var Message: TMessage);
begin
   inherited;
   Invalidate;
end;

procedure TsBitBtn.CMEnabledChanged(var Message: TMessage);
begin
   inherited;
   Invalidate;
   Notify( self, STM_ENABLEDCHANGENOTIFY, Integer(Enabled));   
end;

procedure TsBitBtn.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
   Perform(WM_LBUTTONDOWN, Message.Keys, Longint(Message.Pos));
end;

procedure TsBitBtn.CMMouseEnter(var Message: TMessage);
begin
   inherited;
   if not FMouseInControl and Enabled and (DragMode <> dmAutomatic) and (GetCapture = 0) then begin
      FMouseInControl := TRUE;
      if FBtnStyle <> btsDefault then
         Repaint
      else
         InvalidateCaption;
      Notify( self, STM_MOUSEENTERNOTIFY, 0);
   end;
end;

procedure TsBitBtn.CMMouseLeave(var Message: TMessage);
begin
   inherited;
   if FMouseInControl and Enabled then begin
      FMouseInControl := False;
      if FBtnStyle <> btsDefault then
         Repaint
      else
		   InvalidateCaption;
      Notify( self, STM_MOUSELEAVENOTIFY, 0);
   end;
end;

procedure TsBitBtn.STMGlyphIdChanged(var Message: TMessage);
begin
	if Message.WParam = FGlyphId then begin
      if Message.lParam = Message.wParam then
         Glyph := TBitMap(FGlyphList[FGlyphId])
      else
			GlyphListId := Message.LParam;
      FModifiedGlyph := FALSE;
   end;
end;

function TsBitBtn.GetPalette: HPALETTE;
begin
   Result := Glyph.Palette;
end;

procedure TsBitBtn.SetGlyph(Value: TBitmap);
begin
   TButtonGlyph(FGlyph).Glyph := Value as TBitmap;
   FModifiedGlyph := True;
   Invalidate;
end;

function TsBitBtn.GetGlyph: TBitmap;
begin
	Result := nil;
   if (FGlyphList <> nil) and (FGlyphId > -1) then
   	Result := TBitMap(FGlyphList[FGlyphId]);
   if Result = nil then
      Result := TButtonGlyph(FGlyph).Glyph;
end;

procedure TsBitBtn.GlyphChanged(Sender: TObject);
begin
   Invalidate;
end;

function TsBitBtn.IsCustom: Boolean;
begin
   Result := Kind = bkCustom;
end;

procedure TsBitBtn.SetStyle(Value: TButtonStyle);
begin
   if Value <> FStyle then begin
      FStyle := Value;
      Invalidate;
   end;
end;

procedure TsBitBtn.SetKind(Value: TBitBtnKind);
begin
   if Value <> FKind then begin
      if Value <> bkCustom then begin
         Default := Value in [bkOK, bkYes];
         Cancel := Value in [bkCancel, bkNo];

         if ((csLoading in ComponentState) and (Caption = '')) or
            (not (csLoading in ComponentState)) then begin
            if BitBtnCaptions[Value] <> nil then
               Caption := LoadResString(BitBtnCaptions[Value]);
         end;

         ModalResult := BitBtnModalResults[Value];
         TButtonGlyph(FGlyph).Glyph := GetBitBtnGlyph(Value);
         NumGlyphs := 2;
         FModifiedGlyph := False;
      end;
      FKind := Value;
      Invalidate;
   end;
end;

function TsBitBtn.IsCustomCaption: Boolean;
begin
   Result := AnsiCompareStr(Caption, LoadResString(BitBtnCaptions[FKind])) <> 0;
end;

function TsBitBtn.GetKind: TBitBtnKind;
begin
   if FKind <> bkCustom then
      if ((FKind in [bkOK, bkYes]) xor Default) or
         ((FKind in [bkCancel, bkNo]) xor Cancel) or
         (ModalResult <> BitBtnModalResults[FKind]) or
         FModifiedGlyph then
         FKind := bkCustom;
   Result := FKind;
end;

procedure TsBitBtn.SetLayout(Value: TButtonLayout);
begin
   if FLayout <> Value then begin
      FLayout := Value;
      Invalidate;
   end;
end;

function TsBitBtn.GetNumGlyphs: TNumGlyphs;
begin
   Result := TButtonGlyph(FGlyph).NumGlyphs;
end;

procedure TsBitBtn.SetNumGlyphs(Value: TNumGlyphs);
begin
   if Value < 0 then
      Value := 1
   else if Value > 4 then
      Value := 4;
   if Value <> TButtonGlyph(FGlyph).NumGlyphs then begin
      TButtonGlyph(FGlyph).NumGlyphs := Value;
      Invalidate;
   end;
end;

procedure TsBitBtn.SetSpacing(Value: Integer);
begin
   if FSpacing <> Value then begin
      FSpacing := Value;
      Invalidate;
   end;
end;

procedure TsBitBtn.SetMargin(Value: Integer);
begin
   if (Value <> FMargin) and (Value >= - 1) then begin
      FMargin := Value;
      Invalidate;
   end;
end;

procedure TsBitBtn.SetGlyphList(Value: TsGlyphList);
begin
	if FGlyphList <> Value then begin
      if FGlyphList <> nil then
			FGlyphList.ChangeNotification(self, FALSE);
   	FGlyphList := Value;
      if FGlyphList <> nil then begin
			FGlyphList.ChangeNotification(self, TRUE);
         Glyph := TBitMap(FGlyphList[FGlyphId]);
      end else
         Glyph := nil;
   end;
end;

procedure TsBitBtn.SetGlyphId(Value: Integer);
begin
	if FGlyphId <> Value then begin
		if csLoading in ComponentState then
   		FGlyphId := Value
		else begin
   		if FGlyphList = nil then
   			Exit;
      	if (Value > -1) then begin
            if FGlyphList[Value] = nil then
               Raise Exception.CreateFmt(SErrorGlyphNotFound, [Value]);
            if not FGlyphList[Value].InheritsFrom(TBitMap) then
      		   Raise Exception.Create(SErrorNeedsBmp);
         end;
   		FGlyphId := Value;
         Glyph := TBitMap(FGlyphList[FGlyphId]);
      end;
   end;
end;

function TsBitBtn.GlyphStored: Boolean;
begin
   Result := (Kind = bkCustom) and ((GlyphList = nil) or (GlyphListId = -1));
end;

procedure TsBitBtn.Font3dChanged(Sender: Tobject);
begin
   Invalidate;
end;

procedure TsBitBtn.SetBtnStyle(Value: TsBitBtnStyle);
begin
   if FBtnStyle <> Value then begin
      FBtnStyle := Value;
      Invalidate;
   end;
end;

procedure TsBitBtn.SetFont3d(Value: TFont3d);
begin
   if FFont3d <> Value then
      FFont3d.Assign(Value);
end;

procedure TsBitBtn.SetHilight3d(Value: THilight3d);
begin
   if FHilight3d <> Value then
      FHilight3d.Assign( Value);
end;

procedure TsBitBtn.SetHilightStyle(Value: TTextHilightStyle);
begin
   if Value <> FHilightStyle then begin
      FHilightStyle := Value;
      if FMouseInControl then
         Invalidate;
   end;
end;

procedure TsBitBtn.ActionChange(Sender: TObject; CheckDefaults: Boolean);
   procedure CopyImage(ImageList: TCustomImageList; Index: Integer);
   begin
      with Glyph do begin
         Width := ImageList.Width;
         Height := ImageList.Height;
         Canvas.Brush.Color := clFuchsia;//! for lack of a better color
         Canvas.FillRect(Rect(0,0, Width, Height));
         ImageList.Draw(Canvas, 0, 0, Index);
      end;
   end;

begin
   inherited ActionChange(Sender, CheckDefaults);
   if Sender is TCustomAction then with TCustomAction(Sender) do begin
      { Copy image from action's imagelist }
      if (Glyph.Empty) and (ActionList <> nil) and (ActionList.Images <> nil) and
         (ImageIndex >= 0) and (ImageIndex < ActionList.Images.Count) then
         CopyImage(ActionList.Images, ImageIndex);
   end;
end;

procedure DestroyLocals; far;
var
   I: TBitBtnKind;
begin
   for I := Low(TBitBtnKind) to High(TBitBtnKind) do
      BitBtnGlyphs[I].Free;
end;

initialization
   FillChar(BitBtnGlyphs, SizeOf(BitBtnGlyphs), 0);
   
finalization
   DestroyLocals;
end.
