{*******************************************************************************
   Unit
      sCombos.pas
   Description:
      Flat style combo boxes.
   Versions:
      '99
	sHistory:
      '99    - Aug. 1999
               .Final fixups
               .Image draw for path combo.
					.Delphi's clipboard hotkeys handling is fixed.
               .Fixed problem with loosing of selection when combo was sitted on TToolBar97
                  which was dragged.
      '99 beta 2- July 1999
               .Fixed problem with ListHilight.Font assignment
               .Fixed the problem with scroll bar under WinNT.
               .Fixed problem with the scroll bar when ListPopupAnimation = anNone
               .Fixed handling of pictures when LookupStyle = lsDropDown
               .Added new option in FontCombo foNoFixedPitchOnly. To make it event more
                  tunable I added new event OnAddFont where you can examine the font before it will
                  be added to the combo and deside if you want it or not  
               .Fixed 2 bugs with lsDropDown lookup style:
                  - Before it did not keep selected index
                  - Before it improparly handled not existed items 
               .Fixed problem with system icon in TsPathComboBox
               .Bug with picture repaint on dropping down list is fixed.
               .Animation tune up properties.
               .Property BorderStyle is added. Ctl3d and Flat properties are redundant
                  now and kept for back compatibility. Simulation of frame border
                  (like in TEdit when Ctl3D = FALSE) and no border.
               .Several small fixes.
		2.3('99 beta 1)-April-June 1999
               .Changed the disbaled style drawings policy
               .Valid chars property and validation events
               .Different selection styles
               .Very serious recources leak is fixed
					.DropDown and scroll animation
               .Fixed bug with the disabled font handling.
               .Changes in the item height to be not smaller then
               	height of the images.
      2.2a  -  Feb-March 1999
               .Added picture support for the drop down list
               .Added picture support to the edit area when style = csDropDown
               .First implementation of lookup - new property LookupStyle
               .Item and combo height calculations are changed to comply with
                  the new GetFontHeight.
               .Fixed "5 pixels" problem - before Text in the combo box was
                  cutted at Position about 5 pixels (FSysBtnWidth - FButtonWidth, in win98
                  depends of the system settings) on the Left of the button when style
                  not in [csSimple, csDropDown].
               .Changed the Style policy (thanks to RX). Now it is possible to draw
                  anything in the list even if Style <> csOwnerDrawXXX. This also
                  fixes the problem with list border under Windows '98.
                  Side effect - animation is disabled under Win '98.
      2.1h  -  23/12/1998
               .Stupid bug of mine (ReleaseDC(DC, Handle)!!) reported by S.Tkachenko
               is fixed...
      2.1g  -  21/12/1998
               .Fixup of the drop down list size.
               .New property DefaultHeight. When it is set to FALSE, the combo box sets
               the same height as the Office '97 one.
               .(INTERNAL CHANGE) Centralized the ItemHeight calculation.
               .Published DropDownCount for TsColor, TsFont combo boxes.
      2.1f  -  17/12/1998
               .Fixup of the serious problem (reported by Vladislav Necheporenko) -
               combobox was loosing the ListBox handle on recreation.
               .Border of the drop down list now looks exactly like in the Office '97
               "combos"
               .Problem with selections for DropDownList and OwnerDrawXXX styles of
               combo in Delphi 4 fixed.
      2.1e  -  3/12/1998
               .Default color for TsXXComboBox's drop down list changed to clWindow
      2.1d  -  30/11/1998
               .Changes suggested by Thomas von Stetten.
               They marked with TVS in the comment.
      2.1c  -  27/11/1998
               .Sending notifications about MouseEnter, MouseLeave events.
      2.0b  -  25/11/1998
               .Change enabled notification message implementation
               .Moving strings to sConsts
      2.0a  -  20/11/1998
               .Wait cursor for TPathCombo when filling up the list
      2.0*	- 	End of September
      			.Initial release
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      These components are part of sEditTools.

*     I did not track the versions before, so let's consider it as 2.0
**    TsColorComboBox and TsFontComboBox are rewrites of TColorComboBox and
		TFontComboBox taken from RX package.
*******************************************************************************}
unit sCombos;

interface

{$B-,S-,W-,R-}
{$C PRELOAD}
{$I S.Inc}

{
   To do:
      1. csOwnerDrawVariable style support
         a.	- OnMeasureItem is not called by TComboBox. Fix it.
         b. - Combo Height
         c. - List Height
         d. - Check what kind of changes have to be done in scrolling
}

uses Windows, Messages, Classes, sCommonTypes, Forms, Controls, Graphics,
   StdCtrls, stdUtils, sGraphics, ImgList, sDefaults;

type
   TsComboState = set of (msMouseInControl, msButtonPainted, msDropped,
   	msLock, msItemLock, msIndexLock, msTopIndexLock);
   TsTopBtmState = (tbsNone, tbsTop, tbsBottom);

   TsCustomComboBox = class(TCustomComboBox)
   private
      FAnimating: Boolean;
      FAniUpDuration: Boolean;
      FBorderStyle: TsBorderStyle;
      FButtonWidth: Integer;
      FChildHandle: HWND;
      FCtrlState: TsComboState;
      FDefaultHeight: Boolean;
      FDefListProc: Pointer;
      FDisabledFont: TFont;
      FDisabledStyle: TsDisabledStyle;
      FDosKey: Boolean;
      FEditOffset: Integer;
      FFlat: Boolean;
      FKey: Word;
      FLastY: Integer;
      FListBorderStyle: TsListBorderStyle;
      FListBrush: THandle;
      FListColor: TColor;
      FListCursor: TCursor;
      FListHilight: TsHilight;
      FListHandle: HWND;
      FListInstance: Pointer;
      FListPopupAnimation: TsListAnimationType;
      FListScrollAnimation: TsScrollAnimationStyle;
      FLookupStyle: TsLookupStyle;
      FSavedFont: TFont;
      FScrolling: Boolean;
      FScrollCount: Integer;
      FScrollCounter: Integer;
      FSolidBorder: Boolean;
      FSysBtnWidth: Integer;
      FTopBtn: TsTopBtmState;
      FValidChars: string;
      FOnChange: TNotifyEvent;
      FOnLookup: TNotifyEvent;
      FOnValidate: TValidateEvent;
      FOnValidateError: TNotifyEvent;

      function GetCtl3D: Boolean;

      procedure SetBorderStyle(const Value: TsBorderStyle);
      procedure SetCtl3D(const Value: Boolean);
      procedure SetDefaultHeight(const Value: Boolean);
      procedure SetDisabledStyle(const Value: TsDisabledStyle);
      procedure SetDisabledFont(const Value: TFont);
      procedure SetFlat(const Value: Boolean);
      procedure SetListBorderStyle(const Value: TsListBorderStyle);
      procedure SetListColor(const Value: TColor);
      procedure SetListHilight(const Value: TsHilight);
      procedure SetListWidth(const Value: Integer);

      function GetSolidBorder: Boolean;
      procedure SetSolidBorder;
      function GetButtonRect: TRect;
      function GetButtonWidth: Integer;

      function ValidCharsStored: Boolean;
      procedure InvalidateSelection;
      procedure RefreshScrollBar;
      procedure PaintButton;
      procedure PaintButtonBorder;
      procedure RedrawBorders;
      procedure SetupEnabled;
      function GetListHeight: Integer;
      function GetListWidth: Integer;
      procedure AniNCPaint(tDC: HDC; var R: TRect);
      procedure PaintItemTo(DC: HDC; R: TRect; Index: Integer; Selected: Boolean);
      procedure ListWndProc(var Message: TMessage);
      procedure WMChar(var Message: TMessage); message WM_CHAR;
      procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
      procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
      procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
      procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
      procedure CNFontChanged(var Message: TMessage); message CM_FONTCHANGED;
   protected
      procedure WndProc(var Message: TMessage); override;
      procedure ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer); override;
      procedure CreateWnd; override;
      procedure CreateParams(var Params: TCreateParams); override;
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
      procedure KeyPress(var Key: Char); override;
      procedure Change; override;
      procedure DoExit; override;

      procedure DoLookup; virtual;
      procedure DoChange; virtual;
      procedure ResetItemHeight; dynamic;
      function IsValidChar(const Key: Char): Boolean; virtual;
      procedure ValidateError; virtual;
      function Validate(var Pos: Integer): Boolean; virtual;

      function GetDefaultItemHeight: Integer; dynamic;
      function InvalidateEditImage: TRect; dynamic;
      function ItemImageAssigned(index: Integer): Boolean; dynamic;

      procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
      procedure DrawImage( index: Integer; rect: TRect); dynamic;

      {$IFDEF V1_COMP}      {Back compatibility}
      procedure ReadDefaultDraw(Reader: TReader);
      procedure WriteNothing(Writer: TWriter);
      procedure DefineProperties(Filer: TFiler); override;
      {$ENDIF}

      property BorderStyle: TsBorderStyle read FBorderStyle write SetBorderStyle default DEF_Combo_BorderStyle;
      property Ctl3D: Boolean read GetCtl3D write SetCtl3D stored FALSE;
      property DefaultHeight: Boolean read FDefaultHeight write SetDefaultHeight default DEF_Combo_DefaultHeight;
      property DisabledFont: TFont read FDisabledFont write SetDisabledFont;
      property DisabledStyle: TsDisabledStyle read FdisabledStyle write SetDisabledStyle default DEF_Combo_DisabledStyle;
      property DosKey: Boolean read FDosKey write FDosKey default DEF_Combo_DosKey;
      property Flat: Boolean read FFlat write SetFlat stored FALSE;//default DEF_Combo_Flat;
      property ListBorderStyle: TsListBorderStyle read FListBorderStyle write SetListBorderStyle default DEF_Combo_ListBorderStyle;
      property ListColor: TColor read FListColor write SetListColor default DEF_Combo_ListColor;
      property ListCursor: TCursor read FListCursor write FListCursor default DEF_Combo_ListCursor;
      property ListHilight: TsHilight read FListHilight write SetListHilight;
      property ListPopupAnimation: TsListAnimationType read FListPopupAnimation write FListPopupAnimation default DEF_Combo_ListPopupAnimation;
      property ListScrollAnimation: TsScrollAnimationStyle read FListScrollAnimation write FListScrollAnimation default DEF_Combo_ListScrollAnimation;
      property ListWidth: Integer read GetListWidth write SetListWidth default 0;
      property LookupStyle: TsLookupStyle read FLookupStyle write FLookupStyle default DEF_Combo_LookupStyle;
      property SolidBorder: Boolean read FSolidBorder;
      property ValidChars: String read FValidChars write FValidChars stored ValidCharsStored;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
      property OnLookup: TNotifyEvent read FOnLookup write FOnLookup;
      property OnValidate: TValidateEvent read FOnValidate write FOnValidate;
      property OnValidateError: TNotifyEvent read FOnValidateError write FOnValidateError;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure ScrollTopIndex(newIndex: Integer; animate: Boolean = TRUE);
      procedure ScrollToIndex(newIndex: Integer; animate: Boolean = TRUE; setTopIndex: Boolean = TRUE; keepView: Boolean = TRUE);
   end;

   TGetImageIdProc = procedure(Sender: TObject; ItemIndex: Integer; var imageIndex: Integer) of object;

   TsComboBox = class(TsCustomComboBox)
   private
      FImageIndexes: TsImageIdList;
      FImages: TCustomImageList;
      FImageChangeLink: TChangeLink;
      FSaveIndexes: TsImageIdList;
      FOnGetImageId: TGetImageIdProc;
      function GetImageId(Index: Integer): Integer;
      function GetItems: TStrings;
      procedure SetItems(Value: TStrings);
      procedure ImageListChange(Sender: TObject);
      procedure SetImageIndexes(const Value: TsImageIdList);
      procedure SetImages(const Value: TCustomImageList);
      procedure CBResetContent(var Message: TMessage); message CB_RESETCONTENT;
      procedure CBDeleteString(var Message: TMessage); message CB_DELETESTRING;
   protected
      procedure CreateWnd; override;
      procedure DestroyWnd; override;
      function ItemImageAssigned(index: Integer): Boolean; override;
      procedure DrawImage( index: Integer; rect: TRect); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
   published
      property Style; {Must be published before Items}
      property BorderStyle;
      property DosKey;
      property Font;
      property Color;
      property Ctl3D;
      property DisabledStyle;
      property DisabledFont;
      property ItemHeight;
      property DefaultHeight;
      property DragMode;
      property DragCursor;
      property DropDownCount;
      property Flat;
      property Images: TCustomImageList read FImages write SetImages;
      property Items: TStrings read GetItems write SetItems;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property LookupStyle;
      property MaxLength;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Sorted;
      property TabOrder;
      property TabStop;
      property Text;
      property Visible;
      property ValidChars;
      property ItemIndex default -1;
      property Enabled; // mast be after style and disabled style
      property ImageIndexes: TsImageIdList read FImageIndexes write SetImageIndexes;
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnDrawItem;
      property OnDropDown;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnGetImageId: TGetImageIdProc read FOnGetImageId write FOnGetImageId;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnLookup;
      property OnMeasureItem;
      property OnStartDrag;
      property OnValidate;
      property OnValidateError;
      {$IFDEF VER120}
      property Anchors;
      //property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

   { TColorComboBox }
   TsColorComboBox = class(TsCustomComboBox)
   private
      FColorValue: TColor;
      FDisplayNames: Boolean;
      FFullScheme: Boolean;
      procedure SetColorValue(NewValue: TColor);
      procedure SetDisplayNames(Value: Boolean);
      procedure SetFullScheme(Value: Boolean);
      procedure DrawColor(Index: Integer; Rect: TRect; small: Boolean);
   protected
      procedure CreateWnd; override;
      procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
      procedure DrawImage( index: Integer; rect: TRect); override;
      function GetDefaultItemHeight: Integer; override;
      function ItemImageAssigned(index: Integer): Boolean; override;
      procedure DoChange; override;
      procedure DoLookup; override;
      procedure BuildList; virtual;
   public
      constructor Create(AOwner: TComponent); override;
      property Text;
   published
      property Style default csDropDownList;
      property BorderStyle;
      property ColorValue: TColor read FColorValue write SetColorValue default DEF_ColorCombo_ColorValue;
      property DefaultHeight;
      property DisabledFont;
      property DisabledStyle;
      property DisplayNames: Boolean read FDisplayNames write SetDisplayNames default DEF_ColorCombo_DisplayNames;
      property DosKey;
      property DropDownCount;
      property Color;
      property Ctl3D;
      property DragMode;
      property DragCursor;
      property Flat;
      property Font;
      property FullScheme: Boolean read FFullScheme write SetFullScheme default FALSE;
      property ImeMode;
      property ImeName;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property LookupStyle;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property TabOrder;
      property TabStop;
      property Visible;
      property Enabled; // mast be after style and disabled style
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnDropDown;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnStartDrag;
      property OnValidate;
      property OnValidateError;
      {$IFDEF VER120}
      property Anchors;
      //property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

   { TsFontComboBox }
   TFontDevice = (fdScreen, fdPrinter, fdBoth);
   TFontListOption = (foAnsiOnly, foTrueTypeOnly, foFixedPitchOnly, foNoOEMFonts, foOEMFontsOnly, foScalableOnly, foNoFixedPitchOnly);
   TFontListOptions = set of TFontListOption;

   TsFontComboBox = class;

   TFontValidateEvent = procedure(Sender: TsFontComboBox; var Accept: Boolean; var LogFont: TLogFont;
      var TextMetric: TTextMetric; FontType: Integer) of object;


   TsFontComboBox = class(TsCustomComboBox)
   private
      FDevice: TFontDevice;
      FLastBitmap: TBitMap;
      FPreviewFont: Boolean;
      FUpdate: Boolean;
      FOptions: TFontListOptions;
      FOnAddFont: TFontValidateEvent;
      procedure SetFontName(const NewFontName: string);
      function GetFontName: string;
      function GetFontIndex(aFontName: String): Integer;
      function GetTrueTypeOnly: Boolean;
      procedure SetDevice(Value: TFontDevice);
      procedure SetOptions(Value: TFontListOptions);
      procedure SetTrueTypeOnly(Value: Boolean);
      procedure SetPreviewFont(Value: Boolean);
      procedure Reset;
      function GetBmp(index: Integer): TBitMap;
      procedure WMFontChange(var Message: TMessage); message WM_FONTCHANGE;
      procedure WMPaint(var Message: TMessage); message WM_PAINT;
   protected
      procedure BuildList; virtual;
      procedure DoChange; override;
      procedure CreateWnd; override;
      function GetDefaultItemHeight: Integer; override;
      function ItemImageAssigned(index: Integer): Boolean; override;
      function InvalidateEditImage: TRect; override;
      procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
      procedure DrawImage( index: Integer; rect: TRect); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      property Text;
   published
      property Style default csDropDownList;
      property BorderStyle;
      property Color;
      property Ctl3D;
      property DefaultHeight;
      property DisabledStyle;
      property DisabledFont;
      property Device: TFontDevice read FDevice write SetDevice default fdScreen;
      property DosKey;
      property DragMode;
      property DragCursor;
      property DropDownCount;
      property Flat;
      property Font;
      property FontName: string read GetFontName write SetFontName;
      property Options: TFontListOptions read FOptions write SetOptions default [];
      property ImeMode;
      property ImeName;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property LookupStyle;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property PreviewFont: Boolean read FPreviewFont write SetPreviewFont;
      property ShowHint;
      property TabOrder;
      property TabStop;
      property TrueTypeOnly: Boolean read GetTrueTypeOnly write SetTrueTypeOnly stored False; { obsolete, use Options instead }
      property Visible;
      property Enabled; // mast be after style and disabled style
      property OnAddFont: TFontValidateEvent read FOnAddFont write FOnAddFont;
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnDropDown;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnStartDrag;
      property OnValidate;
      property OnValidateError;
      {$IFDEF VER120}
      property Anchors;
      //property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

   TsFilesListOption = (flArchive, flDirectory, flDrives, flExclusive,
      flHidden, flReadOnly, flReadWrite, flSystem);
   TsFilesListOptions = set of TsFilesListOption;

   TsPathComboBox = class(TsCustomComboBox)
   private
      FFileMask: String;
      FIconHeight: Integer;
      FImageList: THandle;
      FLastImageIndex: Integer;
      FOptions: TsFilesListOptions;
      FPath: String;
      function FileMaskStored: Boolean;
      procedure SetFileMask(Value: String);
      procedure SetOptions(Value: TsFilesListOptions);
      procedure SetPath(Value: String);
   protected
      procedure CreateWnd; override;
      function GetDefaultItemHeight: Integer; override;
      function InvalidateEditImage: TRect; override;
      function ItemImageAssigned(index: Integer): Boolean; override;
      procedure DoChange; override;
      procedure DrawImage( index: Integer; rect: TRect); override;
      procedure WMPaint(var Message: TMessage); message WM_PAINT;
      procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
      procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure BeginUpdate;
      procedure EndUpdate;
      procedure FillList;
      property Items;
      property Text;
   published
      property Style default csDropDownList;
      property BorderStyle;
      property Color;
      property Ctl3D;
      property DefaultHeight;
      property DisabledFont;
      property DisabledStyle;
      property DosKey;
      property DragMode;
      property DragCursor;
      property DropDownCount;
      property FileMask: String read FFileMask write SetFileMask stored FileMaskStored;
      property Flat;
      property Font;
      property Path: String read FPath write SetPath;
      property ImeMode;
      property ImeName;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property LookupStyle;
      property Options: TsFilesListOptions read FOptions write SetOptions default [flReadWrite];
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Sorted;
      property TabOrder;
      property TabStop;
      property Visible;
      property Enabled; 
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnDropDown;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMeasureItem;
      property OnStartDrag;
      property OnValidate;
      property OnValidateError;
      {$IFDEF VER120}
      property Anchors;
      //property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;


implementation
{$R *.RES}
{.$R *.DCR}

uses sFlat, SysUtils, Printers, sFileUtils, ShellApi, Commctrl, Consts, sConsts;

{ TsCustomComboBox}
constructor TsCustomComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   BorderStyle := DEF_Edit_BorderStyle;
   ControlStyle := ControlStyle - [csFixedHeight];
   TControlCanvas(Canvas).Control := self;
   FButtonWidth := 11;
   FDefaultHeight := DEF_Combo_DefaultHeight;
   FDefListProc := nil;
   FDisabledFont := TFont.Create;
   FDisabledFont.Assign(Font);
   FDisabledStyle := DEF_Combo_DisabledStyle;
   FDosKey := DEF_Combo_DosKey;
   FSavedFont := TFont.Create;
	FSavedFont.Assign(Font);
   FEditOffset := 0;
   FListHandle := 0;
   FListInstance := MakeObjectInstance(ListWndProc);
   FListBorderStyle := DEF_Combo_ListBorderStyle;
   FListBrush := 0;
   FListColor := DEF_Combo_ListColor;
   FListCursor := DEF_Combo_ListCursor;
   FListHilight := TsHilight.Create;
   FListHilight.Font.AssignToFont(Font);

   FListPopupAnimation := DEF_Combo_ListPopupAnimation;
   FListScrollAnimation := DEF_Combo_ListScrollAnimation;
   FLookupStyle := DEF_Combo_LookupStyle;
   FSysBtnWidth := GetSystemMetrics(SM_CXVSCROLL);
end;

destructor TsCustomComboBox.Destroy;
begin
   FDisabledFont.Free;
   FSavedFont.Free;
   FLisTHilight.Free;
   if FListBrush <> 0 then
      DeleteObject(FListBrush);
   inherited;
   FreeObjectInstance(FListInstance);
end;

procedure TsCustomComboBox.CreateParams(var Params: TCreateParams);
const
  ComboBoxStyles: array[TComboBoxStyle] of DWORD =
    (CBS_DROPDOWN, CBS_SIMPLE, CBS_DROPDOWNLIST, CBS_OWNERDRAWFIXED, CBS_OWNERDRAWVARIABLE);
begin
   inherited CreateParams(Params);
   if not (Style in [csOwnerDrawFixed, csOwnerDrawVariable]) then
      Params.Style := (Params.Style and not CBS_DROPDOWNLIST) or CBS_OWNERDRAWFIXED or
         ComboBoxStyles[Style];
end;

procedure TsCustomComboBox.CreateWnd;
var
   dlStyle: LongInt;
begin
   ResetItemHeight;
   inherited;

   if FChildHandle <> EditHandle then
      FListHandle := FChildHandle;

   FDefListProc := Pointer(GetWindowLong(FListHandle, GWL_WNDPROC));
   SetWindowLong(FListHandle, GWL_WNDPROC, Longint(FListInstance));

   dlStyle := GetWindowLong(FListHandle, GWL_STYLE);
   if FListBorderStyle = lbsEdge then
      dlStyle := dlStyle and not WS_BORDER
   else
      dlStyle := dlStyle or WS_BORDER;

   SetWindowLong(FListHandle, GWL_STYLE, dlStyle);

   dlStyle := GetWindowLong(FListHandle, GWL_EXSTYLE);
   if FListBorderStyle = lbsEdge then
      dlStyle := dlStyle or WS_EX_CLIENTEDGE
   else
      dlStyle := dlStyle and not WS_EX_CLIENTEDGE;
   SetWindowLong(FListHandle, GWL_EXSTYLE, dlStyle);

   SetSolidBorder;
   //SetupEnabled;
end;

type
   TWinControlCrack = class(TWinControl)
      property Color;
   end;

procedure TsCustomComboBox.WndProc(var Message: TMessage);
const
   isVisible: array[Boolean] of Integer = (SWP_HIDEWINDOW, SWP_SHOWWINDOW);
var
   fh: Integer;
begin
   fh := 0; // to avoid compiler warning
   if (Message.Msg = WM_PARENTNOTIFY) then case LoWord(Message.wParam) of
      WM_CREATE:
         if FDefListProc <> nil then begin
            SetWindowLong(FListHandle, GWL_WNDPROC, Longint(FDefListProc));
            FDefListProc := nil;
            FChildHandle := Message.lParam;
         end else begin
            if FChildHandle = 0 then
               FChildHandle := Message.lParam
            else
               FListHandle := Message.lParam;
         end;
   end else if Message.Msg = WM_WINDOWPOSCHANGING then begin
      fh := GetFontHeight(Font);

      if FDefaultHeight then
          Inc( fh, 2);
      if not (Style in [csOwnerDrawFixed, csOwnerDrawVariable]) then
         TWMWindowPosMsg(Message).WindowPos^.cy := fh + 6
      else if Style = csOwnerDrawFixed then
         TWMWindowPosMsg(Message).WindowPos^.cy := ItemHeight+6;

      if Style in [csDropDown, csSimple] then begin
         if not Enabled and (DisabledStyle = dsEditTools) then
            SetWindowPos(EditHandle, 0, 0, 0, 0, 0, SWP_NOMOVE + SWP_NOSIZE + SWP_NOACTIVATE + SWP_HIDEWINDOW)
         else begin
            if FDefaultHeight or not FFlat then
               fh := ClientHeight - 6;
            SetWindowPos( EditHandle, 0,
                  3 + FEditOffset, 3,
                  ClientWidth - GetButtonWidth - 7 - FEditOffset, fh,
                  SWP_NOZORDER + SWP_NOACTIVATE + SWP_NOREDRAW);
         end;
      end;
   end else if (Message.Msg = CB_SHOWDROPDOWN) and (Message.wParam = 0) and
      (FLookupStyle = lsDropDown) and (Style in [csSimple, csDropDown]) then
      fh := ItemIndex;

   if (Message.Msg <> WM_NCPAINT) or {not FFlat}(FBorderStyle = bsDefault) then
	   inherited;

   case Message.Msg of
      CB_SHOWDROPDOWN: if (FLookupStyle = lsDropDown) and (Style in [csSimple, csDropDown]) and
         (Message.wParam = 0) and (fh <> ItemIndex) then begin
         ItemIndex := fh;
         InvalidateEditImage;
      end;
      WM_CTLCOLORLISTBOX: if FListColor <> clWindow then begin
         SetBkColor( Message.wParam, ColorToRGB(FListColor));
         Message.Result := FListBrush;
      end;
      CB_SETCURSEL: begin
         if FFlat and not (csLoading in ComponentState) then begin
         	if not (Style in [csSimple, csDropDown]) then
	            InvalidateSelection;
            RedrawBorders;
         end;
         if Message.wParam <> ItemIndex then
            DoChange;
      end;
      WM_NCPAINT: if FBorderStyle <> bsDefault then begin
         if not FScrolling then begin
            if (Style <> csSimple) and FFlat then
               PaintButton;
            RedrawBorders;
         end;
      end;
      WM_SETFOCUS, WM_KILLFOCUS: if not (csDesigning in ComponentState) then begin
         SetSolidBorder;
         if FFlat and not (Style in [csSimple, csDropDown]) and not (csLoading in ComponentState) then
            InvalidateSelection;
      end;
      CM_MOUSEENTER: if not (msMouseInControl in FCtrlState) and Enabled then begin
         Include(FCtrlState, msMouseInControl);
         SetSolidBorder;
         Notify( self, STM_MOUSEENTERNOTIFY, 0);
      end;
      CM_MOUSELEAVE: if (msMouseInControl in FCtrlState) and Enabled then begin
         Exclude(FCtrlState, msMouseInControl);
         SetSolidBorder;
         Notify( self, STM_MOUSELEAVENOTIFY, 0);
      end;
      CM_ENABLEDCHANGED: begin
         SetupEnabled;
         Notify( self, STM_ENABLEDCHANGENOTIFY, Ord(Enabled));
         Invalidate;
      end;
   end;
end;

function TsCustomComboBox.GetListHeight: Integer;
begin
	Result := ItemHeight * Min(DropDownCount, Items.Count) + 2;
   if FListBorderStyle = lbsEdge then
   	Inc(Result, 2);
   if Result <= 4  then
   	Result := 20;
end;

function TsCustomComboBox.GetListWidth: Integer;
begin
	Result := Perform(CB_GETDROPPEDWIDTH, 0, 0);
end;

procedure TsCustomComboBox.AniNCPaint(tDC: HDC; var R: TRect);
var
   hBrush: THandle;
   Brush: TBrush;
   sR: TRect;
   flag: Integer;

   SI: TScrollInfo;
   cy: Integer;

   sWidth: Integer;
   sArrowHeight: Integer;
   sDisabled: Boolean;
begin
   if FListBorderStyle = lbsSmallEdge then
   	InflateRect(R, 1, 1);
   if FListBorderStyle <> lbsDefault then
   	DrawEdge( tDC, R, EDGE_RAISED, BF_RECT or BF_ADJUST)
   else begin
   	hBrush := CreateSolidBrush(GetSysColor(COLOR_WINDOWFRAME));
      FrameRect( tDC, R, hBrush);
      DeleteObject(hBrush);
      InflateRect(R, -1, -1);
   end;

   FillRect( tDC, R, THandle(ColorToRGB(FLIstColor + 1)));

   sDisabled := GetWindowLong(Handle, GWL_STYLE) and CBS_DISABLENOSCROLL > 0;
   if sDisabled or (Items.Count > DropDownCount) then begin

      sWidth := GetSystemMetrics( SM_CXVSCROLL); // Width of arrow bitmap on vertical scroll bar
      sArrowHeight := GetSystemMetrics( SM_CYVSCROLL); // Height of arrow bitmap on vertical scroll bar

      if HeightOf(R) - 2 * sArrowHeight < 0 then
         sArrowHeight := HeightOf(R) div 2;

   	sR := R;
      sR.Left := sR.Right - sWidth;

      Brush := TBrush.Create;
      try
         Brush.Bitmap := AllocPatternBitmap(clBtnFace, clBtnHighlight);
         FillRect( tDC, sR, Brush.Handle);
      finally
         Brush.Free;
      end;

      sR.Bottom := sR.Top + sArrowHeight;
      flag := 0;
      if sDisabled then
         flag := flag or DFCS_INACTIVE;
      DrawFrameControl( tDC, sR, DFC_SCROLL, DFCS_SCROLLUP or flag);

      sR.Bottom := R.Bottom;
      sR.Top := sR.Bottom - sArrowHeight;
      DrawFrameControl( tDC, sR, DFC_SCROLL, DFCS_SCROLLDOWN or flag);

      if not sDisabled then begin
	      SI.cbSize:= SizeOf(SI);
   	   SI.fMask:= SIF_RANGE or SIF_POS or SIF_PAGE;
      	GetScrollInfo( FListHandle, SB_VERT, SI);

	      sR.Top := R.Top + sArrowHeight;
         cy := Trunc(0.5+(HeightOf(R) - 2 * sArrowHeight) * ((Integer(SI.nPage))/(SI.nMax+1)));
         if cy < sArrowHeight div 2 then
            cy := sArrowHeight div 2;
         sR.Top := sR.Top + Trunc(0.5+(((HeightOf(R) - 2 * sArrowHeight) - cy) /
               ((SI.nMax + 1) - Integer(SI.nPage)) * SI.nPos));


      	sR.Bottom := sR.Top + cy;

         if HeightOf(sR) < HeightOf(R) - 2 * sArrowHeight then begin
            DrawEdge( tDC, sR, EDGE_RAISED, BF_RECT or BF_ADJUST);
            FillRect( tDC, sR, THandle(COLOR_SCROLLBAR+1));
         end;
      end;

      // For tests
      //BitBlt( TForm(Parent).Canvas.Handle, 19, 19, WidthOf(R)+2, heightOf(R)+2, tDC, 0, 0, SRCCOPY);

      R.Right := R.Right - sWidth - 1;
      R.Bottom := R.Bottom - 1;

      if FListBorderStyle = lbsEdge then begin
         Dec(R.Right);
         Dec(R.Bottom);
      end;
   end else begin
   	R.Right := R.Right - 2;
      R.Bottom := R.Bottom - 2
   end;
end;

procedure TsCustomComboBox.PaintItemTo(DC: HDC; R: TRect; Index: Integer; Selected: Boolean);
const
   Actions: array[Boolean] of Word = (ODA_DRAWENTIRE, ODA_SELECT);
   States: array[Boolean] of Word = (ODS_DEFAULT, ODS_SELECTED);
var
   DrawItemStruct: TDrawItemStruct;
begin
   with DrawItemStruct do begin
      CtlType := ODT_COMBOBOX;
      CtlID := Handle;
      itemID := Index;
      itemAction := Actions[selected];
      itemState := States[selected];
      if selected and not (Style in [csSimple, csDropDown]) then
         itemState := itemState or ODS_FOCUS;
      hwndItem := Handle;
      hDC := DC;
      rcItem := R;
   end;
   Perform( CN_DRAWITEM, DrawItemStruct.CtlID, LongInt(@DrawItemStruct));
end;

procedure TsCustomComboBox.RefreshScrollBar;
var
   SI: TScrollInfo;
begin
   SI.cbSize := sizeof(SI);
   SI.fMask := SIF_ALL;
   GetScrollInfo(FListHandle, SB_VERT, SI);
   SetScrollInfo(FListHandle, SB_VERT, SI, TRUE);
end;

procedure TsCustomComboBox.ScrollTopIndex(newIndex: Integer; animate: Boolean = TRUE);
var
   TopIndex: Integer;
   step, steps: Integer;
   Scroller: TsListScroller;
begin
   if msDropped in FCtrlState then begin
      topIndex := Perform( CB_GETTOPINDEX, 0, 0);
      if (newIndex = topIndex) then
         Exit;
      if animate then begin
         step := 0;
         steps := 0;

         if not ((ItemIndex = -1) and (newIndex = 0)) then begin
            steps := topIndex - newIndex;
            if newIndex = -1 then
               Dec(steps);
            if topIndex + dropDownCount - steps >= Items.Count then
               Inc(steps, dropdowncount {- topIndex}  -1)
            else if NewIndex + DropDownCount >= Items.Count then
               Inc(steps, DropDownCount - 2);
         end;

         if steps <> 0 then begin
            step := 1;
            if steps < 0 then begin
               steps := -steps;
               step := -1;
            end;
         end;

         if step <> 0 then begin
            FScrolling := True;
            Scroller := TsListScroller.Create(FListHandle, ItemHeight, DropDownCount);
            Scroller.PaintItemProc := PaintItemTo;
            Scroller.WaitStep := steps * ListScrollWaitSteps;
            try
               CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 0, 0);
               while steps > 0 do begin
                  Scroller.Scroll( TopIndex, step, -1, FALSE);
                  Dec(steps);
               end;
            finally
         	   FScrolling := False;
               if (step < 0) and (topIndex + dropDownCount - steps >= Items.Count) then
                  CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, newIndex - dropdowncount+1, 0)
               else if newIndex < Items.Count - DropDownCount then
                  CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, newIndex, 0);
               Include(FCtrlState, msItemLock); // to awoid flicker..
               CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
               if Win32Platform = VER_PLATFORM_WIN32_NT then
                  RefreshScrollBar;
               Exclude(FCtrlState, msItemLock);
               Scroller.Free;
            end;
         end;
      end;
   end else
      CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, newIndex, 0);
end;

procedure TsCustomComboBox.ScrollToIndex(newIndex: Integer; animate: Boolean = TRUE; setTopIndex: Boolean = TRUE; keepView: Boolean = TRUE);
var
   TopIndex, Index: Integer;
   step, steps: Integer;
   Scroller: TsListScroller;

   procedure DrawItemInList( anIndex: Integer; selected: Boolean);
   var
      DC: HDC;
      R: TRect;
   begin
      DC := GetDC(FListHandle);
      Windows.GetClientRect(FListHandle, R);
      R.Top := (anIndex - topIndex) * ItemHeight;
      R.Bottom := R.Top + ItemHeight;
      PaintItemTo(DC, R, anIndex, selected);
      ReleaseDC( FListHandle, DC);
   end;

begin
   Index := ItemIndex;

   if (newIndex = Index) and not setTopIndex then
      Exit;

   if msDropped in FCtrlState then begin
      if Animate then begin
         step := 0;
         steps := 0;
         topIndex := Perform( CB_GETTOPINDEX, 0, 0);

         if Index - topIndex in [0..DropDownCount-1] then
            DrawItemInList( Index, FALSE); // "unselecting" item

         if newIndex - topIndex in [0..DropDownCount-1] then begin
            DrawItemInList( newIndex, TRUE); // "selecting" item
            if keepView then begin
               CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, newIndex, 0);
               Include(FCtrlState, msTopIndexLock);
               Exit;
            end;
         end;

         if newIndex < Index then
            setTopIndex := TRUE;

         if not ((Index = -1) and (newIndex = 0)) then begin
            if not setTopIndex then begin
               steps := index - newIndex;
               if Index = -1 then
                  Inc(steps);
            end else begin
               steps := topIndex - newIndex;
               if newIndex = -1 then
                  Dec(steps);
               if topIndex + dropDownCount - steps >= Items.Count then
                  Inc(steps, dropdowncount {- topIndex}  -1)
               else if NewIndex + DropDownCount >= Items.Count then
                  Inc(steps, DropDownCount - 2);
            end;
         end;

         if steps <> 0 then begin
            step := 1;
            if steps < 0 then begin
               steps := -steps;
               step := -1;
            end;
         end;

         if step <> 0 then begin
            Scroller := TsListScroller.Create(FListHandle, ItemHeight, DropDownCount);
            Scroller.PaintItemProc := PaintItemTo;
            Scroller.WaitStep := steps * ListScrollWaitSteps;

            try
               CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 0, 0);

               if step > 0 then
                  Index := topIndex
               else
                  Index := topIndex + DropDownCount - 1;

               while steps > 0 do begin
                  Dec(Index, step);
                  FScrolling := Index <> newIndex;
                  Scroller.Scroll( TopIndex, step, -1, Index = newIndex);
                  FScrolling := True;
                  Dec(steps);
               end;
            finally
         	   FScrolling := False;
               if setTopIndex then begin
                  if (step < 0) and (topIndex + dropDownCount - steps >= Items.Count) then
                  	CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, newIndex - dropdowncount+1, 0)
                  else if newIndex < Items.Count - DropDownCount then
                     CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, newIndex, 0);
               end else
                  CallWindowProc( FDefListProc, FListHandle, LB_SETTOPINDEX, topIndex - steps, 0);
               CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, newIndex, 0);
               Include(FCtrlState, msItemLock); // to awoid flicker..
               CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
               if Win32Platform = VER_PLATFORM_WIN32_NT then
                  RefreshScrollBar;
               Exclude(FCtrlState, msItemLock);
               Scroller.Free;
            end;
         end else {if newIndex < 1 then}
            CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, newIndex, 0);
      end else
         CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, newIndex, 0);
   end else
      CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, newIndex, 0);
end;

procedure TsCustomComboBox.ListWndProc(var Message: TMessage);
   procedure CallDefaultProc;
   begin
      with Message do
         Result := CallWindowProc(FDefListProc, FListHandle, Msg, WParam, LParam);
   end;
   procedure PaintListFrame;
   var
      DC: HDC;
      R: TRect;
   begin
      GetWindowRect(FListHandle, R);
      OffsetRect (R, -R.Left, -R.Top);
      DC := GetWindowDC(FListHandle);
      DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_ADJUST);
      ReleaseDC(FListHandle, DC);
   end;
   procedure CheckCursor;
   var
   	P: TPoint;
      R: TRect;
   begin
   	GetCursorPos(P);
      GetWindowRect(FListHandle, R);
      if (DropDownCount < Items.Count) or
      	(GetWindowLong(Handle, GWL_STYLE) and CBS_DISABLENOSCROLL > 0) then
         R.Right := R.Right - GetSystemMetrics(SM_CXVSCROLL) - 2;
      if PtInRect( R, P) then
      	Windows.SetCursor(Screen.Cursors[FListCursor])
      else
      	Windows.SetCursor(Screen.Cursors[Cursor]);
   end;

   function GetItemRect(anIndex, topIndex: Integer ): TRect;
   begin
      Windows.GetClientRect(FListHandle, Result);
      Result.Top := (anIndex - topIndex) * ItemHeight;
      Result.Bottom := Result.Top + ItemHeight;
   end;

   procedure DrawItemInList( anIndex, topIndex: Integer; selected: Boolean);
   var
      DC: HDC;
      R: TRect;
   begin
      DC := GetDC(FListHandle);
      R := GetItemRect(anIndex, topIndex);
      PaintItemTo(DC, R, anIndex, selected);
      ReleaseDC( FListHandle, DC);
   end;

   procedure DoScrollWindow;
   var
      step: Integer;
      Index, topIndex: Integer;
      h: Integer;
      scroller: TsListScroller;
   begin
      if FScrolling then
         Exit;
      topIndex := Perform(CB_GETTOPINDEX, 0, 0);
      step := 0;
      h := GetListHeight;

      if ((FLastY >=  h-4) or (FLastY < 0)) then begin
      	Inc(FScrollCount);
         if FTopBtn = tbsNone then
         	CallWindowProc( FDefListProc, FListHandle, LB_SETCURSEL, -1, 0);
      end else begin
      	FScrollCount := 0;
         FScrollCounter := 0;
         FTopBtn := tbsNone;
         CallDefaultProc;
         Exit;
      end;

      if (FLastY >= h-4) and
      	(topIndex  + DropDownCount < Items.Count) then
         step := -1
      else if (FLastY < 0) and (topIndex > 0) then
      	step := 1;

      if step = 0 then begin
      	if FTopBtn = tbsNone then begin
         	if FLastY >= ( h-4) then
            	FTopBtn := TbsBottom
            else
            	FTopBtn := tbsTop;
            CallDefaultProc;
         end;
      end else begin
      	Index := ItemIndex;
         Scroller := TsListScroller.Create(FListHandle, ItemHeight, DropDownCount);
         FScrolling := True;
         try
         	Scroller.PaintItemProc := PaintItemTo;
            Scroller.WaitTime := ListScrollWaitTime;
            if FScrollCount <> 0 then
            	Scroller.WaitStep := FScrollCount * ListScrollWaitSteps;
            Scroller.ScrollCounter := FScrollCounter;
            Scroller.Scroll( topIndex, step, index, FALSE);
            FScrollCounter := Scroller.ScrollCounter;
         finally
         	Scroller.Free;
            FScrolling := FALSE;
         end;
         CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 0, 0);
         CallDefaultProc;
      end;
   end;

   procedure HandleScrollBar;
   var
      Index, TopIndex, step, steps: Integer;
      SI: TScrollInfo;
      Scroller: TsListScroller;
   begin
      step := 0;
      steps := 1;
      TopIndex := Perform( CB_GETTOPINDEX, 0, 0);

      SI.cbSize:= SizeOf(SI);
      SI.fMask:= SIF_RANGE or SIF_POS or SIF_PAGE;
      GetScrollInfo( FListHandle, SB_VERT, SI);

      case TWMVScroll(Message).ScrollCode of
         SB_TOP: begin
            steps := topIndex;
            step := 1;
         end;
         SB_BOTTOM: begin
            steps := Items.Count - (topIndex + DropDownCount);
            step := -1;
         end;
         SB_LINEUP: if topIndex > 0 then
            step := 1;
         SB_LINEDOWN: if topIndex + DropDownCount < Items.Count then
            step := -1;
         SB_PAGEUP: begin
            steps := SI.nPage;
            if topIndex - steps < 0 then
               steps := topIndex;
            step := 1;//steps;
            TWMVScroll(Message).ScrollCode := SB_LINEUP;
         end;
         SB_PAGEDOWN: begin
            steps := SI.nPage;
            if topIndex + DropDownCount + steps >= Items.Count then
               steps := Items.Count - (topIndex + DropDownCount);
            step := -1;
            TWMVScroll(Message).ScrollCode := SB_LINEDOWN;
         end;
         SB_THUMBPOSITION, SB_THUMBTRACK: begin
            steps := SI.nPos - TWMVScroll(Message).Pos;
            if steps <> 0 then begin
	            step := 1;
   	         if steps < 0 then begin
      	         steps := -steps;
         	      step := -1;
            	end;
            end;
         end;
      end;
	   if step <> 0 then begin
         Index := ItemIndex;
         //FScrolling := TRUE;
         Scroller := TsListScroller.Create(FListHandle, ItemHeight, DropDownCount);
         try
            Scroller.PaintItemProc := PaintItemTo;
            Scroller.WaitStep := steps * ListScrollWaitSteps;
         	CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 0, 0);
            while steps > 0 do begin
               CallDefaultProc;
               Scroller.Scroll( TopIndex, step, index, FALSE);
               Dec(steps);
            end;
         finally
         	CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
            if Win32Platform = VER_PLATFORM_WIN32_NT then
               RefreshScrollBar;
            Scroller.Free;
         end;
         //FScrolling := FALSE;
      end else
         CallDefaultProc;
   end;

   procedure HandleScrollKeys(aMessage: TWMKey);
   var
      TopIndex, Index: Integer;
      step, steps, ii: Integer;
      Scroller: TsListScroller;
   begin
      TopIndex := Perform( CB_GETTOPINDEX, 0, 0);
      Index := CallWindowProc( FDefListProc, FListHandle, LB_GETCURSEL, 0, 0);
      step := 0;
      steps := 0;

      if Index = -1 then
         Inc(Index);
      {if Index >= 0 then }case aMessage.CharCode of
         VK_DOWN: begin
            if Index = Items.Count - 1 then
               Exit;
         	if (Index - TopIndex + 1 >= DropDownCount) and
                  (topIndex + DropDownCount < Items.Count) then begin
            	steps := 1;
            	step := -1;
         	end;
         end;
         VK_UP: begin
            if Index = 0 then
               Exit;
         	if (Index - TopIndex <= 0) and (TopIndex > 0) then begin
            	steps := 1;
            	step := 1;
            end;
         end;
         VK_NEXT, VK_END: begin
            if Index = Items.Count - 1 then
               Exit;
            if aMessage.CharCode = VK_NEXT then
               steps := DropDownCount-1
            else
               steps := Items.Count - Index;
            if Index + steps > TopIndex + DropDownCount then begin
	            if index + steps >= Items.Count then
   	            steps := Items.Count - (topIndex + DropDownCount);
               if steps <> 0 then
	      	      step := -1;
            end;
         end;
         VK_PRIOR, VK_HOME: begin
            if Index = 0 then
               Exit;
            if aMessage.CharCode = VK_PRIOR then
               steps := DropDownCount-1
            else
               steps := Index;
            if Index - steps < topIndex then begin
	            if topIndex - steps <= 0 then
   	            steps := topIndex;
               if steps <> 0 then
	      	      step := 1;
            end;
         end;
         else begin
            CallDefaultProc;
            Exit;
         end;
      end;
      FTopBtn := tbsNone;
	   if droppeddown and (step <> 0) then begin
      	if Index - topIndex in [0..DropDownCount-1] then
   	      DrawItemInList( Index, topIndex, FALSE); // "unselecting" item
         Scroller := TsListScroller.Create(FListHandle, ItemHeight, DropDownCount);
         try
            Scroller.PaintItemProc := PaintItemTo;
            Scroller.WaitStep := steps * ListScrollWaitSteps;
            Scroller.WaitTime := ListScrollWaitTime;
            for ii := 1 to steps do begin
               FScrolling := not (aMessage.CharCode in [VK_DOWN, VK_UP]) and (ii <> steps);
               Scroller.Scroll( TopIndex, step, -1, (aMessage.CharCode in [VK_DOWN, VK_UP]) or (ii = steps));
               FScrolling := TRUE;
            end;
         finally
            Scroller.Free;
            FScrolling := False;
         end;
         if Style in [csSimple, csDropDown] then begin
            CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 0, 0);
            CallDefaultProc;
            CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
            if Win32Platform = VER_PLATFORM_WIN32_NT then
               RefreshScrollBar;
         end else begin
         	Perform( WM_SETREDRAW, 0, 0);
         	CallDefaultProc;
            Include(FCtrlState, msItemLock);
	         CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
   	      Perform( WM_SETREDRAW, 1, 0);
            if Win32Platform = VER_PLATFORM_WIN32_NT then
               RefreshScrollBar;
            Exclude(FCtrlState, msItemLock);
         	InvalidateSelection;
         end;
      end else begin
         if Style in [csSimple, csDropDown] then
            CallDefaultProc
         else begin
	         Perform( WM_SETREDRAW, 0, 0);
   	      CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, 1, 0);
      	 	CallDefaultProc;
	         Include(FCtrlState, msItemLock);
   	      Perform( WM_SETREDRAW, 1, 0);
      	   Exclude(FCtrlState, msItemLock);
	         if CallWindowProc( FDefListProc, FListHandle, LB_GETCURSEL, 0, 0) <> Index then
		         InvalidateSelection;
         end;
      end;
	end;

begin
   case Message.Msg of
      WM_MOUSEMOVE:
         if saAuto in FListScrollAnimation then begin
      	   CheckCursor;
            if Message.wParam and MK_LBUTTON	> 0 then begin
               FLastY := TWMMouseMove(Message).YPos;
               DoScrollWindow;
            end else
	            CallDefaultProc;
            CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, Integer(TRUE), 0);
         end else
            CallDefaultProc;
      $118: // WM_SYSTIMER
         if saAuto in FListScrollAnimation then begin
            DoScrollWindow;
            CallWindowProc( FDefListProc, FListHandle, WM_SETREDRAW, Integer(TRUE), 0);
         end else
            CallDefaultProc;
      WM_LBUTTONUP: begin
         if Flat and not (Style in [csSimple, csDropDown]) then
           Perform( WM_SETREDRAW, Integer(FALSE), 0);
         CallDefaultProc;
         if Flat and not (Style in [csSimple, csDropDown]) then
            Perform( WM_SETREDRAW, Integer(TRUE), 0);
      end;
      WM_KEYDOWN:
         if saKey in FListScrollAnimation then
            HandleScrollKeys(TWMKey(message))
         else
            CallDefaultProc;
      WM_VSCROLL:
         if saScrollBar in FListScrollAnimation then
            HandleScrollBar
         else
            CallDefaultProc;
      WM_NCPAINT: if not FAnimating or (FListPopupAnimation = anNone) then begin
         if (FListPopupAnimation = anNone) or (FListBorderStyle <> lbsDefault) then
            CallDefaultProc;
         PaintListFrame;
      end;
      WM_WINDOWPOSCHANGING: with TWMWindowPosMsg(Message).WindowPos^ do begin
         cy := GetListHeight;
         y := ClientToScreen(Point(0, 0)).Y;
         FAniUpDuration := Y + Height + cy > Screen.Height;
         if FAniUpDuration then
         	y := y - cy
         else
         	y := y + Height;
      end;
      WM_SHOWWINDOW: begin
         if Boolean(Message.wParam) and not FAnimating then begin
				AnimateDropDown( FListHandle, @FAnimating,
               Rect( 0, 0, GetListWidth, GetListHeight), AniNCPaint, nil,
               FListPopupAnimation, TRUE, FAniUpDuration);
         end;
         CallDefaultProc;
      end;
      WM_ERASEBKGND:
         if FAnimating then
            Message.Result := 1;
      LB_SETTOPINDEX: begin
         if not (msTopIndexLock in FCtrlState) then
            CallDefaultProc;
         Exclude(FCtrlState, msTopIndexLock)
      end;
      CB_SETTOPINDEX: begin
         if (FLookupStyle <> lsDropDown) or not (msIndexLock in FCtrlState) then
            CallDefaultProc;
         Exclude( FCtrlState, msIndexLock);
      end;
      LB_SETCURSEL: begin
         if not (msIndexLock in FCtrlState) then begin
            CallDefaultProc;
         end;
         Exclude( FCtrlState, msIndexLock);
      end;
      LB_GETCURSEL:
         CallDefaultProc;
      else
         CallDefaultProc;
   end;
end;

procedure TsCustomComboBox.ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer);
begin
   inherited;
   if (ComboWnd = EditHandle) then case Message.Msg of
      WM_SETFOCUS, WM_KILLFOCUS:
         SetSolidBorder;
   end;
end;

procedure TsCustomComboBox.CNCommand(var Message: TWMCommand);
var
   R: TRect;
//   index: Integer;
begin
   case Message.NotifyCode of
      CBN_DROPDOWN:
         FCtrlState := FCtrlState + [msDropped, msIndexLock];
      CBN_CLOSEUP:
         FCtrlState := FCtrlState - [msDropped, msIndexLock];
   end;
   inherited;
   case Message.NotifyCode of
(*
   	CBN_DROPDOWN:{, CBN_SELCHANGE, CBN_SELENDOK:} if FFlat then begin
      	if not (Style in [csSimple, csDropDown]) then
         	InvalidateSelection
      	else if not FDefaultHeight then
         	RedrawBorders;
   	end;
*)
   	CBN_CLOSEUP: begin
	      if FFlat then begin
   	      R := GetButtonRect;
      	   Exclude(FCtrlState, msButtonPainted);
         	InvalidateRect( Handle, @R, FALSE);
	      end;
	   end;
   	CBN_SELCHANGE: begin
	      DoChange;
      end;
   end;
end;

procedure TsCustomComboBox.WMChar(var Message: TMessage);
var
	Index: Integer;
begin
   FKey := Message.wParam;
   if not (Style in [csSimple, csDropDown]) then begin
      Index := LocateInList(Items, char(Message.wParam), ItemIndex+1);
      if Index = -1 then
         Index := LocateInList(Items, char(Message.wParam));
      if Index = -1 then
         inherited
      else begin
         ScrollToIndex(Index, saKey in FListScrollAnimation, FALSE);
         InvalidateSelection;
      end;
   end else {if FLookupStyle <> lsDropDown then }begin
      Include(FCtrlState, msItemLock);
	   inherited;
      Exclude(FCtrlState, msItemLock);
   end;
end;

procedure TsCustomComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  FKey := Key;
end;

procedure TsCustomComboBox.KeyPress(var Key: Char);
var
   aForm: TCustomForm;
begin
   if (Key = #13) and FDosKey then begin
      aForm := GetParentForm(Self);
      if aForm <> nil then
         SendMessage( aForm.Handle, WM_NEXTDLGCTL, 0, 0);
      Key := #0;
   end else if not IsValidChar(Key) then begin
      ValidateError;
   	Key := #0
   end;
	inherited;
end;

procedure TsCustomComboBox.WMPaint(var Message: TWMPaint);
var
   R: TRect;
   DC: HDC;
   PS: TPaintStruct;
   x: Integer;
begin
   if not Enabled and (FDisabledStyle = dsOffice97) then
      Brush.Color := TWinControlCrack(Parent).Color
   else
      Brush.Color := Color;

   if not FFlat then begin
      inherited;
      if (Style in [csSimple, csDropDown]) and (Enabled or (DisabledStyle <> dsEditTools)) then
         InvalidateEditImage;
      if FBorderStyle <> bsDefault then
         RedrawBorders;
      Exit;
   end;

   if Style = csSimple then
      inherited
   else begin
      DC := BeginPaint(Handle, PS);
      try
         if Enabled or (FDisabledStyle <> dsEditTools) then begin
            R := PS.rcPaint;
            if R.Right > Width - FButtonWidth - 2 then
               R.Right := Width - FButtonWidth - 2;
            FillRect(DC, R, Brush.Handle);
            if RectInRect(GetButtonRect, PS.rcPaint) and not (msButtonPainted in FCtrlState) then
               PaintButton;
            x := 0;
            if not FFLat or (Style in [csSimple, csDropDown]) then
               x := ClientWidth - FSysBtnWidth - 2;
            ExcludeClipRect( DC, x, 0, ClientWidth, ClientHeight);
            if (Style = csDropDown) and DroppedDown then begin
               R := ClientRect;
               InflateRect(R, -2, -2);
               R.Right := Width - FButtonWidth - 3;
               Canvas.Brush.Color := Color;
               Canvas.FrameRect(R);
            end;
         end else begin
            Canvas.Brush.Color := Color;
			   Canvas.FillRect(ClientRect);
         end;
      finally
         EndPaint(Handle, PS);
         //Canvas.Handle := 0;
      end;
   end;
   InvalidateSelection;
   RedrawBorders;
   Message.Result := 0;
end;

procedure TsCustomComboBox.WMEraseBkgnd(var Message: TMessage);
{var
   aBrush: HBrush;
   R: TRect;}
begin
{   if not DoubleBuffered or (Message.wParam = Message.lParam) then begin
      if not Enabled and (FdisabledStyle = dsOffice97) then
         aBrush := CreateSolidBrush( ColorToRGB(TWinControlCrack(Parent).Color))
      else
         aBrush := CreateSolidBrush( ColorToRGB(Color));
      FillRect( TWMEraseBkgnd(Message).DC, ClientRect, aBrush);
      DeleteObject(aBrush);
   end;}
   Message.Result := 1;
end;

procedure TsCustomComboBox.CNFontChanged(var Message: TMessage);
    function EqualFonts(const f1, f2: TFont): Boolean;
    begin
    	Result :=(f1.Charset = f2.Charset) and
    				(f1.Height = f2.Height) and
    				(f1.Name = f2.Name) and
               (f1.Pitch = f2.Pitch) and
    				(f1.Size = f2.Size){ and
    				(f1.Style = f2.Style)};
    end;
begin
	if EqualFonts(Font, FListHilight.Font) then
		FListHilight.Font.AssignToFont(Font);
   RecreateWnd;
end;

function TsCustomComboBox.ItemImageAssigned(index: Integer): Boolean;
begin
   Result := FALSE;
end;

function TsCustomComboBox.InvalidateEditImage: TRect;
var
   index: Integer;
   Rgn: HRgn;
begin
   if FEditOffset = 0 then
      Exit;
	index := ItemIndex;
   if (Style in [csSimple, csDropDown]) and
   	not (FLookupStyle in [lsDropDown, lsAutoComplete]) and
   	(CompareText(Text, Items[index]) <> 0) then
   	index := -1;
   Result := Rect( 3, 3, 3 + FEditOffset, Height-3);
   Canvas.Brush.Color := Color;
   if not Enabled and (FDisabledStyle = dsOffice97) then
      Canvas.Brush.Color := TWinControlCrack(Parent).Color;
   if ItemImageAssigned(index) then begin
      Rgn := CreateRectRgn(3, 3, 3 + FEditOffset, Height-3);
      SelectClipRgn(Canvas.Handle, Rgn);
      DrawImage( index, Result);
      SelectClipRgn(Canvas.Handle, 0);
      DeleteObject(Rgn);
   end else
   	DrawImage( index, Result);
end;

procedure TsCustomComboBox.CNDrawItem(var Message: TWMDrawItem);
const
   EdgeD: array[Boolean] of Smallint = (1, 2);
   EdgeS: array[Boolean] of Integer = (EDGE_SUNKEN, EDGE_RAISED);
var
   State: TOwnerDrawState;
   isEdit: Boolean;
   R, imR: TRect;

begin
   if msItemLock in FCtrlState then
      Exit;
   with Message.DrawItemStruct^ do begin
      State := TOwnerDrawState(WordRec(LongRec(itemState).Lo).Lo);
      if FScrolling then
         Exclude(State, odSelected);

      Canvas.Handle := hDC;
      Canvas.Font := Font;
      Canvas.Brush := Brush;

      isEdit := ItemState and ODS_COMBOBOXEDIT > 0;

      // 1st. Call DrawItem method. Called when drawing the list item OR
      //		when drawing the edit box and style in [csDropDown,
      //		csOwnerDrawXXX].
      if not IsEdit then begin
         if odSelected in State then begin
            Canvas.Font.Assign(FListHilight.Font);
            Canvas.Brush.Color := FListHilight.Color;
            if FListHilight.Style = lhsUnderline then
               Canvas.Font.Style := Canvas.Font.Style + [fsUnderline];
         end else begin
            Canvas.Brush.Color := FListColor;
            Canvas.Font.Assign(Font);
         end;
      end else if not Enabled and (DisabledStyle = dsEditTools) then
         Canvas.Font.Assign(FDisabledFont)
      else
         Canvas.Font.Assign(Font);

      if not isEdit or (Style in [csDropDownList, csOwnerDrawFixed, csOwnerDrawVariable]) or
         (not Enabled and (DisabledStyle = dsEditTools)) then begin
         if isEdit then
            R := Rect( 4, 3, ClientWidth - GetButtonWidth - 4, Height-3)
         else
            R := rcItem;

         R.Left := R.Left + FEditOffset;
         if not IsEdit and (FListHilight.Style = lhsOffice97) and ItemImageAssigned(ItemId) then
            Inc( R.Left);

         DrawItem(itemID, R, State);
      end;

      // 2nd. If FEditOffset > 0 then drawing image
      if FEditOffset > 0 then begin
         if isEdit then
            imR := InvalidateEditImage
         else begin
            imR := rcItem;
            imR.Right := imR.Left + FEditOffset;
            if (FListHilight.Style = lhsOffice97) and ItemImageAssigned(ItemId) then begin
               if odSelected in State then
	               Canvas.Brush.Color := clBtnFace//FListColor;
               else
	               Canvas.Brush.Color := FListColor;
            end;
            DrawImage( ItemId, imR);
         end;
      end;
      if not IsEdit then begin
         if FListHilight.Style = lhsOffice97 then
            Canvas.Brush.Color := FListColor;
      	if FEditOffset > 0 then
         	Canvas.FillRect(Rect(imR.Right, R.Top, R.Left, R.Bottom));
      	if odSelected in State then case FListHilight.Style of
         	lhsFrame, lhsSunkedFrame: begin
               if FEditOffset > 0 then
	               UnionRect( R, R, imR);
               DrawFlatEDGE(Canvas, R, FListHilight.Style = lhsSunkedFrame);
            end;
            lhsEDGE, lhsSunkedEDGE: begin
               if FEditOffset > 0 then
	               UnionRect( R, R, imR);
               DrawEDGE(Canvas.Handle, R,
               	EdgeS[(FListHilight.Style = lhsEdge)], BF_RECT);
            end;
            lhsOffice97: if ItemImageAssigned(ItemId) then
               DrawFlatEDGE(Canvas, imR, FALSE);
         end;
      end;

      if (not FScrolling or (FTopBtn <> tbsNone)) and (odFocused in State)
      	and ((FListHilight.Style in [lhsDefault, lhsUnderline]) or IsEdit) then begin
            {if IsEdit then
               Canvas.Brush.Color := Color
            //else if odSelected in State then
            //   Canvas.Brush.Color := FListHilight.Color
            else
               Canvas.Brush.Color := FListHilight.Color;//FListColor;}
         if isEdit then
            rcItem := Rect( 4, 3, ClientWidth - GetButtonWidth - 4, Height-3)
         else
            Canvas.Font.Color := FListColor;
         DrawFocusRect(Canvas.Handle, rcItem);
      end;
   end;
   Canvas.Handle := 0;
end;

procedure TsCustomComboBox.Change;
begin
   inherited;
   if ([csDesigning, csLoading] * ComponentState = []) and (Style in [csSimple, csDropDown]) then
      DoLookup;
end;

procedure TsCustomComboBox.DoExit;
var
   Pos: Integer;
begin
  	if not Validate(pos) then begin
      if not (csDesigning in ComponentState) then begin
         //Include(FEditState, msReEnter);
         if Pos = -1 then
            SelectAll
         else
            SetCursor(Pos);
         SetFocus;
      end;
      ValidateError;
  	end;
   inherited;
end;

procedure TsCustomComboBox.DoLookup;
var
   tIndex, lIndex: Integer;
   S: string;
   p1, p2: Integer;
begin
   if (FLookupStyle = lsNone) or
      (Style in [csDropDownList, csOwnerDrawFixed, csOwnerDrawVariable]) or
      not (char(FKey) in [#8, char(VK_DELETE), ' ', '0'..'9', 'a'..'z', 'A'..'Z', #96..#105]) then
      Exit;
   lIndex := ItemIndex;
   if Assigned( FOnLookup) then
      FOnLookup(self)
   else case FLookupStyle of
      lsMatch: begin
      	Perform( CB_GETEDITSEL, Integer(@p1), Integer(@p2));
         S := Text;

         tIndex := Items.IndexOf(S);
         lIndex := LocateInList(Items, S);

         if ItemIndex <> tIndex then begin
            ScrollToIndex(tIndex, saLookup in FListScrollAnimation);
            Include( FCtrlState, msIndexLock);
            InvalidateEditImage;
         end else begin
            ScrollTopIndex(lIndex, saLookup in FListScrollAnimation);
            Include( FCtrlState, msIndexLock);
         end;
         if ItemIndex = -1 then begin
	         Text := S;
   	      Perform(CB_SETEDITSEL, 0, MAKELONG(p1, p2));
      	   SendMessage( EditHandle, EM_SCROLLCARET, 0,0);
         end else
            Perform(CB_SETEDITSEL, 0, MAKELONG(0, Length(S)));
      end;
      lsAutoComplete: begin
         Perform( CB_GETEDITSEL, Integer(@p1), Integer(@p2));
         S := Text;

         tIndex := LocateInList(Items, S);
         if not (FKey in [8, VK_DELETE]) and (tIndex <> -1) then
            S := Items[tIndex];

         Text := S;
         if tIndex = -1 then
            SendMessage( EditHandle, EM_SETSEL, p1, p2)
         else
            SendMessage( EditHandle, EM_SETSEL, p1, Length(Text));
         SendMessage( EditHandle, EM_SCROLLCARET, 0,0);
         ScrollToIndex(tIndex, saLookup in FListScrollAnimation);
         Include( FCtrlState, msIndexLock);
      end;
      lsDropDown: begin
         if not (FKey in [0, VK_ESCAPE, VK_RETURN]) then begin
            SendMessage( EditHandle, WM_SETREDRAW, Integer(FALSE), 0);
            Perform( CB_GETEDITSEL, Integer(@p1), Integer(@p2));
            S := Text;
            if not DroppedDown and not (FKey in [8, VK_DELETE]) then
               DroppedDown := TRUE;

            tIndex := LocateInList(Items, Text);
            if tIndex > -1 then begin
               if tIndex <> SendMessage( FListHandle, LB_GETCURSEL, 0, 0) then
                  ScrollToIndex(tIndex, saLookup in FListScrollAnimation);
               Include(FCtrlState, msIndexLock);
            end else
               ItemIndex := -1;
            Include(FCtrlState, msTopIndexLock);

            Text := S;
            SendMessage( EditHandle, WM_SETREDRAW, 1, 0);
            Perform(CB_SETEDITSEL, 0, MAKELONG(p1, p2));
            SendMessage( EditHandle, EM_SCROLLCARET, 0,0);
            if Win32Platform = VER_PLATFORM_WIN32_NT then
               RefreshScrollBar;
         end;
      end;
   end;
   if lIndex <> ItemIndex then
	   DoChange
   else if ItemIndex = -1 then
      InvalidateEditImage;
   FKey := 0;
end;

procedure TsCustomComboBox.DoChange;
begin
   if ([csLoading] * ComponentState = []) then begin
      if (Style in [csSimple, csDropDown])then
         InvalidateEditImage;
      if Assigned(FOnChange) then
	      FOnChange(Self);
   end;
end;

function TsCustomComboBox.IsValidChar(const Key: Char): Boolean;
begin
   Result := TRUE;
   if FValidChars > '' then
      Result := (Pos(key, FValidChars) > 0) or (Key in [#8,#13,#27]);
   if Result then
      Result := not (Key in [#0, #9, #10]);
end;

function TsCustomComboBox.Validate(var Pos: Integer): Boolean;
begin
   Result := TRUE;
   if Assigned(FOnValidate) then
      FOnValidate(self, Result, Pos);
end;

procedure TsCustomComboBox.ValidateError;
begin
   if Assigned(FOnValidateError) then
      FOnValidateError(self)
   else begin
	   MessageBeep( MB_ICONHAND);
      Abort;
   end;
end;

function TsCustomComboBox.ValidCharsStored: Boolean;
begin
   Result := FValidChars > '';
end;

{$IFDEF V1_COMP}
procedure TsCustomComboBox.ReadDefaultDraw(Reader: TReader);
begin
   if Reader.ReadBoolean then
      DisabledStyle := dsDefault
   else
      DisabledStyle := dsOffice97;
end;

procedure TsCustomComboBox.WriteNothing(Writer: TWriter);
begin
end;

procedure TsCustomComboBox.DefineProperties(Filer: TFiler);
begin
   inherited;
   Filer.DefineProperty( 'DefaultDraw', ReadDefaultDraw, WriteNothing, FALSE);
end;
{$ENDIF V1_COMP}

procedure TsCustomComboBox.DrawImage( index: Integer; rect: TRect);
begin
   Canvas.FillRect(Rect);
end;

procedure TsCustomComboBox.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
   S: string;
begin
   Canvas.FillRect(Rect);
   if Index > -1 then begin
      S := Items[Index];
      DrawText( Canvas.Handle, PChar(S), Length(S),
         	Rect, DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
   end;
end;

procedure TsCustomComboBox.SetFlat(const Value: Boolean);
begin
   if Value <> FFlat then begin
      if Value then
         FBorderStyle := bsFlat
      else
         FBorderStyle := bsDefault;
      FFlat := Value;
      //Invalidate;
      RecreateWnd;
   end;
end;

procedure TsCustomComboBox.SetCtl3D(const Value: Boolean);
begin
   if Ctl3D <> Value then begin
      if Value then
         BorderStyle := bsDefault
      else
         BorderStyle := _bsSingle;
   end;
end;

function TsCustomComboBox.GetCtl3D: Boolean;
begin
   Result := FBorderStyle in [bsDefault, bsFlat];
end;

procedure TsCustomComboBox.SetBorderStyle(const Value: TsBorderStyle);
begin
  if FBorderStyle <> Value then begin
      FBorderStyle := Value;
      FFlat := FBorderStyle = bsFlat;
      Invalidate;
  end;
end;

procedure TsCustomComboBox.SetListBorderStyle(const Value: TsListBorderStyle);
var
   dlStyle: LongInt;
begin
   if Value <> FListBorderStyle then begin
      FListBorderStyle := Value;
      if FListHandle <> 0 then begin
         dlStyle := GetWindowLong(FListHandle, GWL_STYLE);
         if FListBorderStyle = lbsEdge then
            dlStyle := dlStyle and not WS_BORDER
         else
            dlStyle := dlStyle or WS_BORDER;
         SetWindowLong(FListHandle, GWL_STYLE, dlStyle);

         dlStyle := GetWindowLong(FListHandle, GWL_EXSTYLE);
         if FListBorderStyle = lbsEdge then
            dlStyle := dlStyle or WS_EX_CLIENTEDGE
         else
            dlStyle := dlStyle and not WS_EX_CLIENTEDGE;
         SetWindowLong(FListHandle, GWL_EXSTYLE, dlStyle);
      end;
   end;
end;

procedure TsCustomComboBox.SetListColor(const Value: TColor);
begin
   if FListColor <> Value then begin
      FListColor := Value;
      if FListColor = clWindow then begin
         if FListBrush <> 0 then
            DeleteObject(FListBrush);
         FListBrush := 0;
      end else
         FListBrush := CreateSolidBrush(ColorToRGB(FLIstColor));
   end;
end;

procedure TsCustomComboBox.SetListHilight(const Value: TsHilight);
begin
	FListHilight.Assign(Value);
end;

procedure TsCustomComboBox.SetListWidth(const Value: Integer);
begin
	Perform( CB_SETDROPPEDWIDTH, Value,0);
end;


procedure TsCustomComboBox.SetDefaultHeight(const Value: Boolean);
begin
   if Value <> FDefaultHeight then begin
      FDefaultHeight := Value;
      RecreateWnd;
   end;
end;

procedure TsCustomComboBox.SetDisabledStyle(const Value: TsDisabledStyle);
begin
   if FDisabledStyle <> Value then begin
      FDisabledStyle := Value;
      if not Enabled then begin
         SetupEnabled;
         Invalidate;
      end;
   end;
end;

procedure TsCustomComboBox.SetDisabledFont(const Value: TFont);
begin
   FDisabledFont.Assign(Value);
   if not Enabled and (DisabledStyle = dsEditTools) then
      Invalidate;
end;

procedure TsCustomComboBox.ResetItemHeight;
begin
   if DefaultHeight then
      ItemHeight := GetDefaultItemHeight
   else
      ItemHeight := GetFontHeight(Font);
end;

function TsCustomComboBox.GetDefaultItemHeight: Integer;
begin
   Result := GetFontHeight(Font) + 2;
end;

function TsCustomComboBox.GetButtonWidth: Integer;
begin
   if Style = csSimple then
      Result := 0
   else begin
      if FFlat then
         Result := FButtonWidth
      else
         Result := FSysBtnWidth;
   end;
end;

procedure TsCustomComboBox.SetupEnabled;
const
   isVisible: array[Boolean] of Integer = (SWP_HIDEWINDOW, SWP_SHOWWINDOW);
begin
   if ([csLoading,csReading] * ComponentState = []) then begin
      if Style in [csSimple, csDropDown] then
         SetWindowPos(EditHandle, 0, 0, 0, 0, 0, SWP_NOMOVE + SWP_NOSIZE + SWP_NOACTIVATE +
            isVisible[Enabled or (DisabledStyle <> dsEditTools)])
   end;
end;

procedure TsCustomComboBox.InvalidateSelection;
const
   Actions: array[Boolean] of Word = (ODA_DRAWENTIRE,ODA_SELECT);
   States: array[Boolean] of Word = (ODS_DEFAULT, ODS_SELECTED);
var
   DrawItemStruct: TDrawItemStruct;
   IsFocused: Boolean;
   h: Integer;
   Index: Integer;
begin
   if (Style in [csSimple, csDropDown]) and (Enabled or (DisabledStyle <> dsEditTools)) then begin
      InvalidateEditImage;
      Exit;
   end;

   Index := ItemIndex;
   IsFocused := (GetFocus = Handle) and not (msDropped in FCtrlState);
   with DrawItemStruct do begin
      CtlType := ODT_COMBOBOX;
      CtlID := Handle;
      itemID := Index;
      itemAction := Actions[IsFocused];
      itemState := ODS_COMBOBOXEDIT or States[IsFocused];
      hwndItem := Handle;
      hDC := Canvas.Handle;
      if Style in [csOwnerDrawFixed, csOwnerDrawVariable] then
         h := ItemHeight
      else begin
         h := GetFontHeight(Font);
         if FDefaultHeight then
            Inc(h, 2);
      end;
      rcItem := Rect( 3, 3, ClientWidth - GetButtonWidth - 4, h + 3);
   end;
   Canvas.Brush.Color := Color;

   Perform( WM_DRAWITEM, DrawItemStruct.CtlID, LongInt(@DrawItemStruct));
   if IsFocused then
      Canvas.DrawFocusRect( Rect( 3, 3, Width - GetButtonWidth - 4, h+3));
end;

function TsCustomComboBox.GetButtonRect: TRect;
begin
   GetWindowRect(Handle, Result);
   OffsetRect(Result, -Result.Left, -Result.Top);
   Inc(Result.Left, ClientWidth - FButtonWidth);
   InflateRect(Result, 0, -1);
   OffsetRect(Result, -2, 0);
end;

procedure TsCustomComboBox.PaintButton;
var
   R: TRect;
   state: Integer;
begin
   R := GetButtonRect;
   InflateRect(R, 1, 0);
   state := DFCS_SCROLLDOWN or DFCS_FLAT;
   if not Enabled then
      state := state or DFCS_INACTIVE;
   if (msDropped in FCtrlState) then begin
      state := state or DFCS_PUSHED;
      Include(FCtrlState, msButtonPainted);
   end;
   DrawFrameControl(Canvas.Handle, R, DFC_SCROLL, state);
   PaintButtonBorder;
   ExcludeClipRect(Canvas.Handle, ClientWidth - FSysBtnWidth, 0, ClientWidth, ClientHeight);
end;

function TsCustomComboBox.GetSolidBorder: Boolean;
var
   focus: HWnd;
begin
   focus := GetFocus;
   Result := (csDesigning in ComponentState) or
      ((msDropped in FCtrlState) or
      ((focus <> 0) and ((focus = Handle) or (focus = EditHandle))) or
      ((msMouseInControl in FCtrlState) and IsFlatApplied));
end;

procedure TsCustomComboBox.SetSolidBorder;
var
   sb: Boolean;
begin
   sb := GetSolidBorder;
   if sb <> FSolidBorder then begin
      FSolidBorder := sb;
      RedrawBorders;
   end;
end;

procedure TsCustomComboBox.PaintButtonBorder;
const
   flags: array[Boolean] of Integer = ( EDGE_RAISED, EDGE_SUNKEN);
   Colors: array[Boolean] of TColor = (clBtnShadow, clWindow);
var
   R: TRect;
begin
   R := GetButtonRect;
   InflateRect(R, 0, -1);
   if FFlat then with Canvas do begin
      if SolidBorder then begin
         Brush.Color := Colors[msDropped in FCtrlState];
         FrameRect(R);
         Pen.Color := Colors[not (msDropped in FCtrlState)];
         Polyline([Point(R.Right-2, R.Top), Point(R.Left, R.Top),
               Point(R.Left, R.Bottom-1)]);
         Pen.Color := clWindow;
         PolyLine([Point(R.Left - 2, R.Top), Point(R.Left - 2, R.Bottom)]);
         Pen.Color := clBtnFace;
         Polyline([Point(R.Left-1, R.Top), Point(R.Left-1, R.Bottom)]);
      end else begin
         Brush.Color := clWindow;
         FrameRect(R);
         InflateRect(R, 0, -1);
         Pen.Color := self.Brush.Color;
         Polyline([Point(R.Left-2, R.Top), Point(R.Left-2, R.Bottom)]);
         Polyline([Point(R.Left-1, R.Top), Point(R.Left-1, R.Bottom)]);
      end;
   end;
end;

procedure TsCustomComboBox.RedrawBorders;
begin
   case FBorderStyle of
      bsFlat: begin
         InternalRedrawBorder( Handle, SolidBorder);
         if (Style <> csSimple) and (Enabled or (FdisabledStyle <> dsEditTools)) then
            PaintButtonBorder;
      end;
      _bsSingle:
         InternalRedrawSingleBorder(Handle, Color);
      _bsNone:
         InternalEriseBorder(Handle, Color);
   end;
end;

{TsComboBox}
constructor TsComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FSaveIndexes := nil;
   FImageChangeLink := TChangeLink.Create;
   FImageChangeLink.OnChange := ImageListChange;
   FImageIndexes := TsImageIdList.Create;
end;

destructor TsComboBox.Destroy;
begin
   FImageChangeLink.Free;
   FImageIndexes.Free;
   FSaveIndexes.Free;
   inherited;
end;

procedure TsComboBox.CreateWnd;
begin
   inherited;
   if FSaveIndexes <> nil then begin
      FImageIndexes.Assign(FSaveIndexes);
      FSaveIndexes.Free;
      FSaveIndexes := nil;
   end;
//   SendMessage( Handle, CB_SETITEMHEIGHT, 0,
//   	MAKELONG( MaxInteger([TrueTypeBMP.Height + 2, DeviceBMP.Height + 2, GetFontHeight(font) + 2]), 0));
end;

procedure TsComboBox.DestroyWnd;
begin
   if FImageIndexes.Count > 0 then begin
      FSaveIndexes := TsImageIdList.Create;
		FSaveIndexes.Assign(FImageIndexes);
   end;
   inherited DestroyWnd;
end;

procedure TsComboBox.SetItems(Value: TStrings);
begin
   Include( FCtrlState, msLock);
   inherited Items := Value;
   Exclude( FCtrlState, msLock);
end;

function TsComboBox.GetItems: TStrings;
begin
   Result := inherited Items;
end;

procedure TsComboBox.CBResetContent(var Message: TMessage);
begin
   if Items.Count = 0 then
      Exit;
   if not (msLock in FCtrlState) then
      FImageIndexes.Clear;
   inherited;
end;

procedure TsComboBox.CBDeleteString(var Message: TMessage);
begin
   FImageIndexes[Message.wParam] := -1;
   inherited;
end;

function TsComboBox.ItemImageAssigned(index: Integer): Boolean;
begin
   Result := GetImageId(Index) > -1;
end;

function TsComboBox.GetImageId(Index: Integer): Integer;
begin
   Result := FImageIndexes[index];
   if Assigned(FOnGetImageId) then
      FOnGetImageId( self, Index, Result);
end;

procedure TsComboBox.DrawImage( index: Integer; rect: TRect);
var
   imageIndex: Integer;
begin
   inherited;
   if FImages <> nil then begin
      imageIndex := GetImageId(Index);
      if imageIndex > -1 then
         ImageList_Draw( FImages.Handle, imageIndex, Canvas.Handle,
         Rect.Left + 1, (Rect.Top + Rect.Bottom - FImages.Height) div 2+1,
         ILD_TRANSPARENT);
   end;
end;

procedure TsComboBox.SetImageIndexes(const Value: TsImageIdList);
begin
   FImageIndexes.Assign(Value);
   if not (csLoading in ComponentState) then
      Invalidate;
end;

procedure TsComboBox.SetImages(const Value: TCustomImageList);
begin
   if FImages <> nil then
      FImages.UnRegisterChanges(FImageChangeLink);
   FImages := Value;
   if FImages <> nil then begin
      FImages.RegisterChanges(FImageChangeLink);
      FImages.FreeNotification(Self);
   	FEditOffset := Images.Width + 2;
   	if (ItemHeight < Images.Height + 2) then
      	SendMessage( Handle, CB_SETITEMHEIGHT, 0, MAKELONG( Images.Height + 2,0));
   end;
   if not (csLoading in ComponentState) then
      Invalidate;
end;

procedure TsComboBox.ImageListChange(Sender: TObject);
begin
   if Sender = Images then begin
      FEditOffset := Images.Width + 2;
   	if (ItemHeight < Images.Height + 2) then
      	SendMessage( Handle, CB_SETITEMHEIGHT, 0, MAKELONG( Images.Height + 2,0));
      Refresh;
   end;
end;


{ TsColorComboBox }
const
   ColorsInList = 16;
   ColorValues: array[1..ColorsInList] of TColor = (
      clBlack, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clGray,
      clSilver, clRed, clLime, clYellow, clBlue, clFuchsia, clAqua, clWhite);

   SysColorsInList = 25;
   SysColorValues: array[1..SysColorsInList] of TColor = (
      clScrollBar, clBackground, clActiveCaption, clInactiveCaption, clMenu,
      clWindow, clWindowFrame, clMenuText, clWindowText, clCaptionText,
      clActiveBorder, clInactiveBorder, clAppWorkSpace, clHighlight,
      clHighlightText, clBtnFace, clBtnShadow, clGrayText, clBtnText,
      clInactiveCaptionText, clBtnHighlight, cl3DDkShadow, cl3DLight,
      clInfoText, clInfoBk);

const
   ColorWidth = 22;

constructor TsColorComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   Style := csDropDownList;
   FColorValue := DEF_ColorCombo_ColorValue; { make default color selected }
   FDisplayNames := DEF_ColorCombo_DisplayNames;
   FEditOffset := ColorWidth + 6;
end;

procedure TsColorComboBox.CreateWnd;
begin
   inherited CreateWnd;
   BuildList;
   SetColorValue(FColorValue);
end;

procedure TsColorComboBox.BuildList;
var
   ii: Integer;
   ColorName: string;
begin
   Clear;
   for ii := 1 to ColorsInList do begin
      ColorName := Copy(ColorToString(ColorValues[ii]), 3, 40);
      Items.AddObject(ColorName, TObject(ColorValues[ii]));
   end;
   if FFullScheme then for ii := 1 to SysColorsInList do begin
      ColorName := Copy(ColorToString(SysColorValues[ii]), 3, 40);
      Items.AddObject(ColorName, TObject(SysColorValues[ii]));
   end;
end;

procedure TsColorComboBox.SetDisplayNames(Value: Boolean);
begin
   if DisplayNames <> Value then begin
      FDisplayNames := Value;
      if not FDisplayNames then begin
         FEditOffset := 0;
         Style := csDropDownList;
      end else
         FEditOffset := ColorWidth;
      SetColorValue(FColorValue);
   end;
end;

procedure TsColorComboBox.SetColorValue(NewValue: TColor);
var
   Item: Integer;
   CurrentColor: TColor;
begin
   if not FFullScheme then
      NewValue := TColor(ColorToRGB(NewValue));
   if (ItemIndex < 0) or (NewValue <> FColorValue) then begin
      for Item := 0 to Pred(Items.Count) do begin
         CurrentColor := TColor(Items.Objects[Item]);
         if CurrentColor = NewValue then begin
            FColorValue := NewValue;
            if ItemIndex <> Item then
               ItemIndex := Item;
            Break;
         end;
      end;
   end;
   InvalidateEditImage;
end;

procedure TsColorComboBox.SetFullScheme(Value: Boolean);
begin
   if FFullScheme <> Value then begin
      FFullScheme := Value;
      BuildList;
      SetColorValue(FColorValue);
   end;
end;

procedure TsColorComboBox.DrawColor(Index: Integer; Rect: TRect; small: Boolean);
var
   Safer: TColor;
begin
   Inc(Rect.Top, 2);
   Inc(Rect.Left, 2);
   Dec(Rect.Bottom, 2);

   if small then begin
      Rect.Right := Rect.Left + ColorWidth;
      offsetRect(Rect, 1, 0);
   end else
      Dec(Rect.Right, 3);

   with Canvas do begin
      Safer := Brush.Color;
      Pen.Color := clBlack;
      Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
      if index > -1 then
         Brush.Color := TColor(Items.Objects[Index])
      else
         Brush.Color := FColorValue;
      try
         InflateRect(Rect, -1, -1);
         FillRect(Rect);
      finally
         Brush.Color := Safer;
      end;
   end;
end;

procedure TsColorComboBox.DrawImage( index: Integer; rect: TRect);
begin
   Canvas.FillRect(Rect);
   DrawColor(index, Rect, TRUE);
end;

procedure TsColorComboBox.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
   if FDisplayNames then begin
      inherited;
   end else begin
      Canvas.FillRect(Rect);
      DrawColor(Index, Rect, FALSE);
   end;
end;

function TsColorComboBox.GetDefaultItemHeight: Integer;
begin
   Result := Max(GetFontHeight(Font) + 2, 9);
end;

function TsColorComboBox.ItemImageAssigned(index: Integer): Boolean;
begin
   Result := TRUE;
end;

procedure TsColorComboBox.DoChange;
begin
   FColorValue := TColor(Items.Objects[ItemIndex]);
   inherited;
end;

procedure TsColorComboBox.DoLookup;
var
   AColor: TColor;
   ii: integer;
begin
   inherited;
   if (Style in [csDropDownList, csOwnerDrawFixed, csOwnerDrawVariable]) then
      Exit;
   if {Items.IndexOf(inherited Text) < 0}itemIndex < 0 then begin
      Val(inherited Text, AColor, ii);
      if ii = 0 then begin
         FColorValue := AColor;
         InvalidateEditImage;
         inherited DoChange;
      end;
   end;
end;

{ TsFontComboBox }
const
   TrueTypeBMP: TBitMap = nil;
   DeviceBMP: TBitMap = nil;

function CreateBitmap(ResName: PChar): TBitmap;
begin
   Result := TBitmap.Create;
   Result.Handle := LoadBitmap(HInstance, ResName);
   if Result.Handle = 0 then begin
      Result.Free;
      Result := nil;
   end;
   if Result = nil then
      Raise Exception.CreateFmt( SErrorResourceNotFound, [ResName]);
end;

function EnumFontsProc(var LogFont: TLogFont; var TextMetric: TTextMetric;
   FontType: Integer; Data: Pointer): Integer; export; stdcall;
var
   Box: TsFontComboBox;
   Accept: Boolean;
   function IsValidFont: Boolean;
   begin
      Result := True;
      if foAnsiOnly in Box.Options then
         Result := Result and (LogFont.lfCharSet = ANSI_CHARSET);
      if foTrueTypeOnly in Box.Options then
         Result := Result and (FontType and TRUETYPE_FONTTYPE = TRUETYPE_FONTTYPE);
      if foFixedPitchOnly in Box.Options then
         Result := Result and (LogFont.lfPitchAndFamily and FIXED_PITCH = FIXED_PITCH);
      if foOEMFontsOnly in Box.Options then
         Result := Result and (LogFont.lfCharSet = OEM_CHARSET);
      if foNoOEMFonts in Box.Options then
         Result := Result and (LogFont.lfCharSet <> OEM_CHARSET);
      if foScalableOnly in Box.Options then
         Result := Result and (FontType and RASTER_FONTTYPE = 0);
      if foNoFixedPitchOnly in Box.Options then
         Result := Result and (LogFont.lfPitchAndFamily and FIXED_PITCH <> FIXED_PITCH);
   end;
begin
   Box := TsFontComboBox(Data);
   Accept := (Box.Items.IndexOf(StrPas(LogFont.lfFaceName)) < 0) and IsValidFont;
   if Assigned( Box.FOnAddFont) then
      Box.FOnAddFont(Box, Accept, LogFont, TextMetric, FontType);
   if Accept then
      Box.Items.AddObject(StrPas(LogFont.lfFaceName), TObject(FontType));
   Result := 1;
end;

constructor TsFontComboBox.Create(AOwner: TComponent);
begin
   if TrueTypeBMP = nil then
      TrueTypeBMP := CreateBitmap('TTF_FONT');
   if DeviceBMP = nil then
      DeviceBMP := CreateBitmap('PRN_FONT');
   inherited Create(AOwner);
   Style := csDropDownList;
   FDevice := fdScreen;
   FLastBitmap := nil;
   FUpdate := False;
   FEditOffset := Max(TrueTypeBMP.Width, DeviceBMP.Width) + 6;
   Sorted := true;
end;

destructor TsFontComboBox.Destroy;
begin
   inherited Destroy;
end;

procedure TsFontComboBox.CreateWnd;
var
   OldFont: TFontName;
begin
   OldFont := FontName;
   inherited CreateWnd;
   FUpdate := True;
   try
      BuildList;
      inherited Text := '';
      SetFontName(OldFont);
   finally
      FUpdate := False;
   end;
   if AnsiCompareText(FontName, OldFont) <> 0 then
      DoChange;
   SendMessage( Handle, CB_SETITEMHEIGHT, 0,
   	MAKELONG( MaxInteger([TrueTypeBMP.Height + 2, DeviceBMP.Height + 2, GetFontHeight(font) + 2]), 0));
end;

procedure TsFontComboBox.BuildList;
var
   DC: HDC;
   Proc: TFarProc;
begin
   if not HandleAllocated then
      Exit;
   Clear;
   DC := GetDC(0);
   try
      Proc := MakeProcInstance(@EnumFontsProc, HInstance);
      try
         if (FDevice = fdScreen) or (FDevice = fdBoth) then
            EnumFonts(DC, nil, Proc, Pointer(Self));
         if (FDevice = fdPrinter) or (FDevice = fdBoth) then try
            EnumFonts(Printer.Handle, nil, Proc, Pointer(Self));
         except
         end;
      finally
         FreeProcInstance(Proc);
      end;
   finally
      ReleaseDC(0, DC);
   end;
end;

procedure TsFontComboBox.SetFontName(const NewFontName: string);
var
   Item: Integer;
begin
   if FontName <> NewFontName then begin
      if not (csLoading in ComponentState) then begin
         HandleNeeded;
         item := GetFontIndex(NewFontName);
         if item > -1 then begin
            ItemIndex := item;
            inherited Text := NewFontName;
            DoChange;
            Exit;
         end;
         if Style in [csSimple, csDropDown] then
            inherited Text := NewFontName
         else
            ItemIndex := -1;
      end else
         inherited Text := NewFontName;
      DoChange;
   end;
end;

function TsFontComboBox.GetFontName: string;
begin
   Result := Text;
end;

function TsFontComboBox.GetFontIndex(aFontName: String): Integer;
var
   ii: Integer;
begin
   Result := -1;
   for ii := 0 to Items.Count - 1 do
      if (AnsiUpperCase(Items[ii]) = AnsiUpperCase(aFontName)) then begin
         Result := ii;
         Exit;
      end;
end;

function TsFontComboBox.GetTrueTypeOnly: Boolean;
begin
   Result := foTrueTypeOnly in FOptions;
end;

procedure TsFontComboBox.SetOptions(Value: TFontListOptions);
begin
   if Value <> FOptions then begin
      if foNoFixedPitchOnly in Value then
         Exclude(Value, foFixedPitchOnly);
      FOptions := Value;
      Reset;
   end;
end;

procedure TsFontComboBox.SetTrueTypeOnly(Value: Boolean);
begin
   if Value <> TrueTypeOnly then begin
      if Value then
         FOptions := FOptions + [foTrueTypeOnly]
      else
         FOptions := FOptions - [foTrueTypeOnly];
      Reset;
   end;
end;

procedure TsFontComboBox.SetPreviewFont(Value: Boolean);
begin
   if Value <> FPreviewFont then begin
      FPreviewFont := Value;
      Invalidate;
   end;
end;

procedure TsFontComboBox.SetDevice(Value: TFontDevice);
begin
   if Value <> FDevice then begin
      FDevice := Value;
      Reset;
   end;
end;

procedure TsFontComboBox.DrawImage( Index: Integer; Rect: TRect);
var
   Bitmap: TBitmap;
begin
   inherited;
   with Canvas do begin
      Bitmap := GetBmp(index);
      if Bitmap <> nil then begin
         BrushCopy( Bounds( Rect.Left + 2, 1 + (Rect.Top + Rect.Bottom - Bitmap.Height) div 2,
            Bitmap.Width, Bitmap.Height), Bitmap, Bounds(0, 0, Bitmap.Width,
            Bitmap.Height), Bitmap.TransparentColor);
      end;
   end;
end;

procedure TsFontComboBox.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
   if FPreviewFont and (Index  > -1) then
      Canvas.Font.Name := Items[Index];
   inherited DrawItem(Index, Rect, State);
end;

procedure TsFontComboBox.WMFontChange(var Message: TMessage);
begin
   inherited;
   Reset;
end;

procedure TsFontComboBox.WMPaint(var Message: TMessage);
begin
	FLastBitmap := nil;
   inherited;
end;

function TsFontComboBox.GetBmp(index: Integer): TBitMap;
begin
   Result := nil;
   if index = -1 then
      index := GetFontIndex(Text);
   if index > -1 then
      if (Integer(Items.Objects[Index]) and TRUETYPE_FONTTYPE) <> 0 then
         Result := TrueTypeBMP
      else if (Integer(Items.Objects[Index]) and DEVICE_FONTTYPE) <> 0 then
         Result := DeviceBMP;
end;

function TsFontComboBox.InvalidateEditImage: TRect;
var
   bmp: TBitMap;
begin
   if Style in [csSimple, csDropDown] then begin
      bmp := GetBmp(-1);
      if (bmp <> FLastBitmap) or (bmp = nil) then begin
         Result := inherited InvalidateEditImage;
         FLastBitmap := bmp;
      end;
   end else
      Result := inherited InvalidateEditImage;
end;

procedure TsFontComboBox.DoChange;
begin
   if not FUpdate then
      inherited;
end;

procedure TsFontComboBox.Reset;
var
   SaveName: string;
begin
   if HandleAllocated then begin
      Inc(FUpdate);
      try
         SaveName := FontName;
         BuildList;
         FontName := SaveName;
      finally
         Dec(FUpdate);
         if FontName <> SaveName then
            DoChange;
      end;
   end;
end;

function TsFontComboBox.GetDefaultItemHeight: Integer;
begin
   Result := MaxInteger([GetFontHeight(Font)+2, TrueTypeBMP.Height+2, 9]);
end;

function TsFontComboBox.ItemImageAssigned(index: Integer): Boolean;
begin
   Result :=((Integer(Items.Objects[Index]) and TRUETYPE_FONTTYPE) <> 0) or
            ((Integer(Items.Objects[Index]) and DEVICE_FONTTYPE) <> 0);
end;

{ TsPathComboBox }
const
   defaultMask = '*.*';

constructor TsPathComboBox.Create(AOwner: TComponent);
begin
   inherited;
   Style := csDropDownList;
   FFileMask := defaultMask;
   FOptions := [flReadWrite];
   FEditOffset := 0;
   FLastImageIndex := -1;
end;

destructor TsPathComboBox.Destroy;
begin
   inherited;
end;

procedure TsPathComboBox.CreateWnd;
begin
   inherited;
   FillList;
end;

procedure TsPathComboBox.WMPaint(var Message: TMessage);
begin
   FLastImageIndex := -1;
   inherited;
end;

procedure TsPathComboBox.WMKillFocus(var Message: TMessage);
begin
   FLastImageIndex := -1;
   inherited;
end;

procedure TsPathComboBox.WMSetFocus(var Message: TMessage);
begin
   FLastImageIndex := -1;
   inherited;
end;

procedure TsPathComboBox.FillList;
var
   S: String;
   Ptr: array[0..124] of Char;
   flag: Integer;
   ii, w: Integer;
   ShellInfo: TSHFILEINFO;
   saveCursor: TCursor;
begin
   saveCursor := Screen.Cursor;
   Screen.Cursor := crHourGlass;
   Items.Clear;
   S := TerminateDir(LFN2Alternate(UnTerminateDir(FPath))) + FFileMask;
   StrPCopy(Ptr, S);
   flag := 0;
   if flArchive in FOptions then
      flag := DDL_ARCHIVE;
   if flDirectory in Options then
      flag := flag or DDL_DIRECTORY;
   if flDrives in Options then
      flag := flag or DDL_DRIVES;
   if flExclusive in Options then
      flag := flag or DDL_EXCLUSIVE;
   if flHidden in Options then
      flag := flag or DDL_HIDDEN;
   if flReadonly in Options then
      flag := flag or DDL_READONLY;
   if flReadWrite in Options then
      flag := flag or DDL_READWRITE;
   if flSystem in Options then
      flag := flag or DDL_SYSTEM;
   Perform( CB_DIR, flag, Integer(@Ptr));
   FIconHeight := 0;
   FImageList := 0;
   for ii := items.Count - 1 downto 0 do begin
      if Items[ii][1] <> '[' then
         StrPCopy(Ptr, TerminateDir(FPath) + Items[ii])
      else begin
         if Items[ii][2] = '-' then
            StrPCopy(Ptr, Items[ii][3] + ':\')
         else
            StrPCopy(Ptr,
               TerminateDir(FPath) + Copy(Items[ii], 2, Length(Items[ii]) - 2));
      end;
      w := SHGetFileInfo( Ptr, 0, ShellInfo, SizeOf(ShellInfo),
         SHGFI_DISPLAYNAME or SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
      if FImageList = 0 then begin
         FImageList := w;
         ImageList_GetIconSize( FImageList, w, FIconHeight);
         FEditOffset := w + 4;
      end;
      if ShellInfo.szDisplayName = '' then
         Items.Delete(ii)
      else begin
         if Items[ii][1] <> '[' then
            Items[ii] := ExtractFileName(ShellInfo.szDisplayName)
         else if Items[ii][2] <> '-' then
            Items[ii] := '[' + String(ShellInfo.szDisplayName) + ']';
         Items.Objects[ii] := TObject(ShellInfo.iIcon);
      end;
   end;
   SendMessage( Handle, CB_SETITEMHEIGHT, 0, MAKELONG( FIconHeight + 2,0));
   Screen.Cursor := saveCursor;
end;

procedure TsPathComboBox.DrawImage(Index: Integer; Rect: TRect);
begin
   inherited;
   if Index > -1 then
      ImageList_Draw(FImageList, Integer(Items.Objects[Index]), Canvas.Handle,
         Rect.Left + 1, (Rect.Top + Rect.Bottom - FIconHeight) div 2+1,
         ILD_TRANSPARENT);
end;

procedure TsPathComboBox.BeginUpdate;
begin
   Include( FCtrlState, msLock);
end;

procedure TsPathComboBox.EndUpdate;
begin
   Exclude( FCtrlState, msLock);
end;

procedure TsPathComboBox.DoChange;
begin
   if not (msLock in FCtrlState) then
      inherited;
end;

procedure TsPathComboBox.InvalidateEditImage;
var
   R: TRect;
begin
	inherited InvalidateEditImage;
   if FBorderStyle <> bsDefault then
	   RedrawBorders
   else begin
   	Canvas.Brush.Color := Color;
	   Canvas.FrameRect(Rect(1, 2, 2+FeditOffset, Height - 2));
      R := ClientRect;
      R.Right := R.Left + FEditOffset + 1;
      DrawEDGE( Canvas.Handle, R, EDGE_SUNKEN, BF_BOTTOMLEFT);
   end;
end;

function TsPathComboBox.ItemImageAssigned(index: Integer): Boolean;
begin
   Result := Integer(Items.Objects[index]) > -1;
end;

procedure TsPathComboBox.SetFileMask(Value: String);
begin
   if FFileMask <> Value then begin
      FFileMask := Value;
      if not ((msLock in FCtrlState) or (csLoading in ComponentState))then
         FillList;
   end;
end;

procedure TsPathComboBox.SetOptions(Value: TsFilesListOptions);
begin
   if FOptions <> Value then begin
      FOptions := Value;
      if not ((msLock in FCtrlState) or (csLoading in ComponentState))then
         FillList;
   end;
end;

procedure TsPathComboBox.SetPath(Value: String);
begin
   if Value <> FPath then begin
      FPath := Value;
      if not ((msLock in FCtrlState) or (csLoading in ComponentState))then
         FillList;
   end;
end;

function TsPathComboBox.FileMaskStored: Boolean;
begin
   Result := FFileMask <> defaultMask;
end;

function TsPathComboBox.GetDefaultItemHeight: Integer;
begin
   Result := MaxInteger([GetFontHeight(Font) + 2, 9, FIconHeight + 4]);
end;

initialization
   RegisterFlatControl(TsCustomComboBox);

finalization
   TrueTypeBMP.Free;
   DeviceBMP.Free;

end.










