{*******************************************************************************
   Unit
      sCtrls.pas
   Description:
      Transparent, BWCC style CheckBox and RadioButton,
      TsLabel - 3d label with many extras
      TsStatusBar
      TsScrollBar
      TsPanel
   Version:
      '99
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   History:
      '99            - August 1999
                     .Final fixups.
                     .Changes to comply with new TsTransControl
      '99 beta 2     - July 1999
                     .Flicker for TsLabel is redused.
                     .Several small fixes.
      2.4 ('99 beta 1) -  April-May 1999
                     .TsPanel - the general purpose "flatilizer"
                     .Fixed problem with TsLabel and TsWBCCControls autosize.
      2.3a        -  Feb-March 1999
                     .TsSpeedButton moved to sButtons.pas
                     .New methods on TsLabel for monitoring of focus/ mouse movement of
                        more then one control. (since it is a rare case I did not implememt it as a property)
                     .New component TsScrollBar - flat scroll bar***
                     .Fixed bug on hilights when FloatingOnFocusOnly = TRUE
      2.2f			- 	22/01/1999
      					.Fixed up bug with frame when allign is taRightJustify.
                     .Fixed bug with autosize when allign is taRightJustify.
                     .New box style (bwsFlat).
                     .OnEnter and OnExit events are added.
                     .Added standard delphi properties forgotten before (ReadOnly, ParentXXX)
      2.2e        -  14/12/1998
                     .New property ButtonLike for TBWCC controls - it is not always
                     suitable to have button like behavior, so now you can switch it off.
                     .New property AutoSize for TBWSS controls.
                     .New property FocusFrame for TBWCC controls and labels - to be able
                     to switch off the focus frame.
                     .Fixed focus frame painting in TsLabel
                     .Fixed improper beh. of TsLabel when ShowFocus = FALSE and
                        FocusControl <> nil
                     .Fixed size calculation when hilightStyle = hsFlushing
      2.2d        -  3/12/1998
                     .check invalidate problem fixed (I was sick to make such kind of bug!)
                     .Flicker redused for sLabel
      2.2c        -  30/11/1998
                     .Changes suggested by Thomas von Stetten.
                        They marked with TVS in the comment.
      2.2b        -  27/11/1998
                     ShortCut handling for TsCheckBox and TsRadioButton.
                     TsLabel's connection to LinkedControl is improved:
                        -MouseEnter, MouseLeave messages are resend to linked
                        control (looks better)
                        -If it is linked to sXXXEdit or sXXXComboBox then
                        notifications for MouseEnter, MouseLeave messages of this
                        control are handled too.
                     It gives the fill that the label and linked control behave
                     as one logical "unit".
      2.2a        -  25/11/1998
                     handling of "enabled changed" message from the linked control
                     for TsLabel.
      2.2         -  21/11/1998
                     -TsSpeedButton with support for TsGlyphList. Initial realease.
                     -TsStstusBar. Initial realease.
      2.1         -  08/11/1998
                     -TsLabel. Initial release.
                     -Improved painting of disabled control
      2.0*        -  End of Sep. 1998
      			      Initial release
   Comments:
      This components are part of sEditTools.
      Why BWCC ? Whell, I believe that it's style is more "flat" than Microsoft's check box and
      radio button. If you are not agree - use transparent controls from Dream Company.

*     I did not track the versions before, so let's consider it as 2.0
**    Enrico Lodolo - ldlc18k1@bo.nettuno.it -
         Idea and some code in the box drawing part was taken from his
         TBorRadio and TBorCheck. Thanks!
***   This implementation based on TFlatScrollBar with the author's permition
*******************************************************************************}
unit sCtrls;

{$B-,S-,W-,R-}
{$C PRELOAD}
{$I S.Inc}

interface

uses
   SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
   Forms, Dialogs, StdCtrls, ComCtrls, Buttons, sTrans, sGraphics, sGlyphsList,
   sConsts, sCommonTypes, sDefaults;

type
   TsCtrlState = (scsMouseInControl, scsFocused, scsInvalidateCaption, scsTracking);
   TsCtrlStates = set of TsCtrlState;

   TsBWCCControl = class(TsTransControl)
   private
      FAlignment: TLeftRight;
      FAutoSize: Boolean;
      FBoxColor: TColor;
      FButtonLike: Boolean;
      FCheckColor: TColor;
      FCtrlState: TsCtrlStates;
      FDown: Boolean;
      FPaint3D: TFont3d;
      FFont3d: TFont3d;
      FHilightStyle: TTextHilightStyle;
      FHilight3d: THilight3d;
      FFocusFrame: Boolean;
      FReadOnly: Boolean;
      FSpacing: Integer;
      FTextAlignment: TAlignment;
      FTransparentMode: TTransparentMode;
      FVerticalAlignment: TVerticalAlignment;
      FWordWrap: Boolean;
      FOnMouseEnter: TNotifyEvent;
      FOnMouseLeave: TNotifyEvent;
      procedure Font3dChanged(Sender: Tobject);
      procedure AdjustBounds;
      function GetCaption: TCaption;
      procedure SetAlignment(Value: TLeftRight);
      procedure SetAutoSize(Value: Boolean);
      procedure SetBoxColor(Value: TColor);
      procedure SetButtonLike(Value: Boolean);
      procedure SetCaption(const Value: TCaption);
      procedure SetCheckColor(Value: TColor);
      procedure SetFont3d(Value: TFont3d);
      procedure SetHilight3d(Value: THilight3d);
      procedure SetHilightStyle(Value: TTextHilightStyle);
      procedure SetFocusFrame(Value: Boolean);
      procedure SetSpacing(Value: Integer);
      procedure SetTextAlignment(Value: TAlignment);
      procedure SetTransparentMode(Value: TTransparentMode);
      procedure SetVerticalAlignment(Value: TVerticalAlignment);
      procedure SetWordWrap(Value: Boolean);
      procedure PaintCaption(aCanvas: TCanvas);
      procedure InvalidateCaption;
      procedure InvalidateBox;
      procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
      procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
      procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
      procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
		procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
   protected
      FState: TCheckBoxState;
      procedure Paint; override;
      procedure PaintBox(aCanvas: TCanvas; R: TRect); virtual; abstract;
      procedure SetState(Value: TCheckBoxState); virtual;
      procedure Toggle; virtual; abstract;
      procedure SetChecked(const Value: Boolean); virtual;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure KeyDown(var Key: Word; Shift: TShiftSTate); override;
      procedure KeyUp(var Key: Word; Shift: TShiftSTate); override;
      procedure DoEnter; override;
      procedure DoExit; override;
      procedure DoMouseEnter; virtual;
      procedure DoMouseLeave; virtual;
      function EditCanModify: Boolean; virtual;
      function GetChecked: Boolean;
      property Alignment: TLeftRight read FAlignment write SetAlignment default DEF_BWCC_Alignment;
      property AutoSize: Boolean read FAutoSize write SetAutoSize default DEF_BWCC_AutoSize;
      property BoxColor: TColor read FBoxColor write SetBoxColor default DEF_BWCC_BoxColor;
      property ButtonLike: Boolean read FButtonLike write SetButtonLike default DEF_BWCC_ButtonLike;
      property CheckColor: TColor read FCheckColor write SetCheckColor default DEF_BWCC_CheckColor;
      property FocusFrame: Boolean read FFocusFrame write SetFocusFrame default DEF_BWCC_FocusFrame;
      property Font3d: TFont3d read FFont3d write SetFont3d;
      property HilightStyle: TTextHilightStyle read FHilightStyle write SetHilightStyle default DEF_BWCC_HilightSttyle;
      property Hilight3d: THilight3d read FHilight3d write SetHilight3d;
      property ReadOnly: Boolean read FReadOnly write FReadOnly default FALSE;
      property Spacing: Integer read FSpacing write SetSpacing default DEF_BWCC_Spacing;
      property State: TCheckBoxState read FState write SetState default cbUnchecked;
      property TextAlignment: TAlignment read FTextAlignment write SetTextAlignment default DEF_BWCC_TextAlign;
      property TransparentMode: TTransparentMode read FTransparentMode write SetTransparentMode default DEF_BWCC_TransparentMode;
      property VerticalAlignment: TVerticalAlignment read FVerticalAlignment write SetVerticalAlignment default DEF_BWCC_VertAlign;
      property WordWrap: Boolean read FWordWrap write SetWordWrap default FALSE;
      property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
      property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

      property Caption: TCaption read GetCaption write SetCaption;
      property Checked: Boolean read GetChecked write SetChecked default FALSE;
      property Enabled;
   end;

   TsCustomCheckBox = class(TsBWCCControl)
   private
      FAllowGrayed: Boolean;
      FStyle: TsBWCCBoxStyle;
      procedure SetStyle( Value: TsBWCCBoxStyle);
   protected
      procedure PaintBox(aCanvas: TCanvas; R: TRect); override;
      property AllowGrayed: Boolean read FAllowGrayed write FAllowGrayed default FALSE;
      property Style: TsBWCCBoxStyle read FStyle write SetStyle default DEF_CheckBox_Style;
   public
	   constructor Create(Owner: TComponent); override;
      procedure Toggle; override;
   end;

   TsCheckBox = class(TsCustomCheckBox)
   published
      property Alignment;
      property AllowGrayed;
      property AutoSize;
      property BoxColor;
      property ButtonLike;
      property Caption;
      property CheckColor;
      property Checked;
      property Color;
      property Ctl3D;
      property DragCursor;
      property DragMode;
      property Enabled;
      property FocusFrame;
      property Font;
      property Font3d;
      property Hilight3d;
      property HilightStyle;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ReadOnly;
      property ShowHint;
      property Spacing;
      property State;
      property Style;
      property TabOrder;
      property TabStop;
      property TextAlignment;
      property TransparentMode;
      property VerticalAlignment;
      property Visible;
      property WordWrap;
      property OnClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnMouseEnter;
      property OnMouseLeave;
      {$IFDEF VER120} //TVS: Standard Delphi 4 Properties
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

   TsCustomRadioButton = class(TsBWCCControl)
   private
      FGroupIndex: Byte;
      FStyle: TsBWCCRadioStyle;
      procedure SetStyle(Value: TsBWCCRadioStyle);
   protected
      procedure PaintBox(aCanvas: TCanvas; R: TRect); override;
      procedure SetChecked(const Value: Boolean); override;
      property GroupIndex: Byte read FGroupIndex write FGroupIndex default 0;
      property Style: TsBWCCRadioStyle read FStyle write SetStyle default DEF_RadioButtonStyle;
   public
      constructor Create(Owner: TComponent); override;
      procedure Toggle; override;
   end;

   TsRadioButton = class(TsCustomRadioButton)
   published
      property Alignment;
      property AutoSize;
      property BoxColor;
      property ButtonLike;
      property Caption;
      property CheckColor;
      property Checked;
      property Color;
      property Ctl3D;
      property DragCursor;
      property DragMode;
      property Enabled;
      property FocusFrame;
      property Font;
      property Font3d;
      property GroupIndex;
      property Hilight3d;
      property HilightStyle;
      property ParentFont;
      property PopupMenu;
      property ShowHint;
      property Spacing;
      property Style;
      property TabOrder;
      property TabStop;
      property TextAlignment;
      property TransparentMode;
      property VerticalAlignment;
      property Visible;
      property WordWrap;
      property OnClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnMouseEnter;
      property OnMouseLeave;
      {$IFDEF VER120} //TVS: Standard Delphi 4 Properties
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

   TsCustomLabel = class(TGraphicControl)
   private
      FAlignment: TAlignment;
      FAutoSize: Boolean;
      FButtonLike: Boolean;
      FCtrlState: TsCtrlStates;
      FDown: Boolean;
      FFocusControl: TWinControl;
      FFocusFrame: Boolean;
      FFont3D: TFont3D;
      FHilightStyle: TTextHilightStyle;
      FHilight3d: THilight3d;
      FLayout: TsTextLayout;
      FMonitorControls: TList;
      FPaint3D: TFont3d;
      FShowAccelChar: Boolean;
      FShowFocus: Boolean;
      FWordWrap: Boolean;
      FOnMouseEnter: TNotifyEvent;
      FOnMouseLeave: TNotifyEvent;
      FOnMouseClick: TNotifyEvent;
      procedure Font3dChanged(Sender: Tobject);
      procedure AdjustBounds;
      procedure DoDrawText(aCanvas: TCanvas; var Rect: TRect; Flags: Word);
      function GetTransparent: Boolean;
      procedure SetAlignment(Value: TAlignment);
      procedure SetFocusControl(Value: TWinControl);
      procedure SetLayout(Value: TsTextLayout);
      procedure SetButtonLike(Value: Boolean);
      procedure SetFont3D(Value: TFont3D);
      procedure SetHilight3d(Value: THilight3d);
      procedure SetHilightStyle(Value: TTextHilightStyle);
      procedure SetFocusFrame(Value: Boolean);
      procedure SetShowAccelChar(Value: Boolean);
      procedure SetTransparent(Value: Boolean);
      procedure SetWordWrap(Value: Boolean);
      procedure SetShowFocus(Value: Boolean);
      procedure CheckInvalidate(hilight, mouseEvent: Boolean);
      procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
      procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
      procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure STMEnabledChangeNotify(var Message: TMessage); message STM_ENABLEDCHANGENOTIFY;
      procedure STMMouseEnterNotify(var Message: TMessage); message STM_MOUSEENTERNOTIFY;
      procedure STMMouseLeaveNotify(var Message: TMessage); message STM_MOUSELEAVENOTIFY;
   protected
      procedure DoMouseEnter; virtual;
      procedure DoMouseLeave; virtual;
      procedure DoMouseClick; virtual;
      procedure SetAutoSize(Value: Boolean); virtual;
      function GetLabelText: String; virtual;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure Paint; override;
      property Alignment: TAlignment read FAlignment write SetAlignment default DEF_Label_Alignment;
      property AutoSize: Boolean read FAutoSize write SetAutoSize default DEF_Label_AutoSize;
      property ButtonLike: Boolean read FButtonLike write SetButtonLike default DEF_Label_ButtonLike;
      property FocusControl: TWinControl read FFocusControl write SetFocusControl;
      property FocusFrame: Boolean read FFocusFrame write SetFocusFrame default DEF_Label_FocusFrame;
      property Font;
		property Font3D: TFont3D read FFont3D write SetFont3D;
      property HilightStyle: TTextHilightStyle read FHilightStyle write SetHilightStyle default DEF_Label_HilightStyle;
      property Hilight3d: THilight3d read FHilight3d write SetHilight3d;
      property Layout: TsTextLayout read FLayout write SetLayout default DEF_Label_Layout;
      property ShowAccelChar: Boolean read FShowAccelChar write SetShowAccelChar default True;
      property ShowFocus: Boolean read FShowFocus write SetShowFocus default TRUE;
      property Transparent: Boolean read GetTransparent write SetTransparent default DEF_Label_Transparent;
      property WordWrap: Boolean read FWordWrap write SetWordWrap default False;
      property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
      property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
      property OnMouseClick: TNotifyEvent read FOnMouseClick write FOnMouseClick;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure AddMonitorControl(ctrl: TWinControl);
      procedure RemoveMonitorControl(ctrl: TWinControl);
      procedure AddMonitorControls(ctrls: array of TWinControl);
      procedure RemoveMonitorControls(ctrls: array of TWinControl);
      property Canvas;
	end;

   TsLabel = class(TsCustomLabel)
   published
      property AutoSize;
      property Align;
      property Alignment;
      property ButtonLike;
      property Caption;
      property Color;
      property DragCursor;
      property DragMode;
      property Enabled;
      property Font;
      property Font3D;
      property FocusControl;
      property Hilight3d;
      property HilightStyle;
      property FocusFrame;
      property Layout;
      property ShowAccelChar;
      property ShowFocus;
      property Transparent;
      property WordWrap;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnMouseEnter;
      property OnMouseLeave;
      property OnMouseClick;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDrag;
      {$IFDEF VER120}
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;

	TsStatusBar=class(TStatusBar)
   protected
      procedure CreateParams(var Params: TCreateParams); override;
	public
	   constructor Create(AOwner: TComponent); override;
   end;

   {$IFDEF FlatSB}
   TsCustomScrollBar = class(TCustomControl)
   private
      FKind: TScrollBarKind;
      FPosition: Integer;
      FMin: Integer;
      FMax: Integer;
      FPageSize: Integer;
      FRTLFactor: Integer;
      FSmallChange: TScrollBarInc;
      FLargeChange: TScrollBarInc;
      FOnChange: TNotifyEvent;
      FOnScroll: TScrollEvent;
      FMode: TScrollbarMode;
      procedure DoScroll(var Message: TWMScroll);
      function NotRightToLeft: Boolean;
      procedure SetKind(Value: TScrollBarKind);
      procedure SetMax(Value: Integer);
      procedure SetMin(Value: Integer);
      procedure SetPosition(Value: Integer);
      procedure SetPageSize(Value: Integer);
      procedure CNCtlColorScrollBar(var Message: TMessage); message CN_CTLCOLORSCROLLBAR;
      procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
      procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;
      procedure WMSize(var Msg: TWMSize); message WM_SIZE;
      procedure SetMode(Value: TScrollbarMode);
      procedure WMPaint(var Msg: TWMPaint); message WM_PAINT;
   protected
      procedure CreateParams(var Params: TCreateParams); override;
      procedure CreateWnd; override;
      procedure Change; dynamic;
      procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer); dynamic;
      property ScrollbarMode: TScrollbarMode read FMode write SetMode default DEF_ScrollBar_Mode;
      property Kind: TScrollBarKind read FKind write SetKind default sbHorizontal;
      property LargeChange: TScrollBarInc read FLargeChange write FLargeChange default 1;
      property Max: Integer read FMax write SetMax default 100;
      property Min: Integer read FMin write SetMin default 0;
      property PageSize: Integer read FPageSize write SetPageSize;
      property Position: Integer read FPosition write SetPosition default 0;
      property SmallChange: TScrollBarInc read FSmallChange write FSmallChange default 1;
      property TabStop default True;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
      property OnScroll: TScrollEvent read FOnScroll write FOnScroll;
   public
      constructor Create(AOwner: TComponent); override;
      procedure SetParams(APosition, AMin, AMax: Integer);
   end;

   TsScrollBar = class(TsCustomScrollBar)
   published
      property Align;
      property Ctl3D;
      property DragCursor;
      property DragMode;
      property Enabled;
      property Kind;
      property LargeChange default 1;
      property Max default 100;
      property Min default 0;
      property PageSize;
      property ParentCtl3D;
      property ParentShowHint;
      property PopupMenu;
      property Position default 0;
      property ScrollbarMode default sbmFlat;
      property ShowHint;
      property SmallChange default 1;
      property TabOrder;
      property TabStop default True;
      property Visible;
      property OnChange;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnScroll;
      property OnStartDrag;
      {$IFDEF VER120} //Standard Delphi 4 Properties
      property Anchors;
      property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;
   {$ENDIF}

   TsCustomPanel = class(TCustomControl)
   private
      FMouseInControl: Boolean;
      FSolidBorder: Boolean;

      procedure RedrawBorders;
      procedure SetSolidBorder;
      function GetSolidBorder: Boolean;
      function ChildControlFocused: Boolean;

      procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
      procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
      procedure WMFocusChanged(var Message: TMessage); message CM_FOCUSCHANGED;
      procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
      procedure SMMLinkedChanged(var Message: TMessage); message STM_LINKEDCHANGENOTIFY;
    protected
      procedure CreateParams(var Params: TCreateParams); override;
      procedure CreateWnd; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure Paint; override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure SetBounds(left, top, width, height: Integer); override;
   end;

   TsPanel = class(TsCustomPanel)
   published
      property Align;
      property Ctl3D;
      property DragCursor;
      property DragMode;
      property Enabled;
      property ParentCtl3D;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property TabOrder;
      property TabStop default True;
      property Visible;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnStartDrag;
      {$IFDEF VER120} //Standard Delphi 4 Properties
      property Anchors;
      property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF VER120}
   end;


   TsTrackBar = class(TTrackBar)
   published
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
   end;

implementation
{.$R *.DCR}

uses Consts, sFlat, stdUtils, Commctrl;

const
   BoxWidth = 12;

constructor TsBWCCControl.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   Color := clBtnFace;
   FAlignment := DEF_BWCC_Alignment;
   FAutoSize := DEF_BWCC_AutoSize;
   FBoxColor := DEF_BWCC_BoxColor;
   FButtonLike := DEF_BWCC_ButtonLike;
   FCheckColor := DEF_BWCC_CheckColor;
   FFont3d := TFont3d.Create(Font3dChanged);
   FFocusFrame := DEF_BWCC_FocusFrame;
   FHilight3d := THilight3d.Create(Font3dChanged);
   FHilightStyle := DEF_BWCC_HilightSttyle;
   FReadOnly := False;
   FSpacing := DEF_BWCC_Spacing;
   FTextAlignment := DEF_BWCC_TextAlign;
   TransparentMode := DEF_BWCC_TransparentMode;
   FVerticalAlignment := DEF_BWCC_VertAlign;

   Width := 98;
   Height := 20;
   ParentColor := FALSE;
   FPaint3D := TFont3d.Create(nil);
end;

destructor TsBWCCControl.Destroy;
begin
   FHilight3d.Free;
   FFont3d.Free;
   FPaint3D.Free;
   inherited;
end;

procedure TsBWCCControl.CMMouseEnter(var Message: TMessage);
begin
   Include(FCtrlState, scsMouseInControl);
   DoMouseEnter;
   if Enabled and IsFlatApplied and not (scsFocused in FCtrlState) and
      ((FHilightStyle <> hsNone) or (FHilight3D.Active and not FHilight3D.Equal(FFont3D))) then
      InvalidateCaption;
end;

procedure TsBWCCControl.CMMouseLeave(var Message: TMessage);
begin
   Exclude(FCtrlState, scsMouseInControl);
   DoMouseLeave;
   if Enabled and IsFlatApplied and not (scsFocused in FCtrlState) and
      ((FHilightStyle <> hsNone) or (FHilight3D.Active and not FHilight3D.Equal(FFont3D))) then
      InvalidateCaption;
end;

procedure TsBWCCControl.CMEnabledChanged(var Message: TMessage);
begin
   inherited;
   Invalidate;
end;

procedure TsBWCCControl.CMTextChanged(var Message: TMessage);
begin
   Invalidate;
   AdjustBounds;
end;

procedure TsBWCCControl.CMFontChanged(var Message: TMessage);
begin
   inherited;
   AdjustBounds;
end;

procedure TsBWCCControl.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
      if Enabled and CanFocus and IsAccel(CharCode, Caption) then begin
         Toggle;
         Result := 1;
      end else
         inherited;
end;

procedure TsBWCCControl.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
	inherited;
   Message.Result := Message.Result or DLGC_WANTARROWS;
end;

procedure TsBWCCControl.WMSetFocus(var Message: TMessage);
begin
   inherited;
   Include( FCtrlState, scsFocused);
   InvalidateCaption;
end;

procedure TsBWCCControl.WMKillFocus(var Message: TMessage);
begin
   inherited;
   Exclude( FCtrlState, scsFocused);
   //Invalidate;
   InvalidateCaption;
end;

procedure TsBWCCControl.InvalidateCaption;
var
   iR: TRect;
begin
   iR := Bounds(0, 0, Width - BoxWidth - 2 - FSpacing, Height - 2);
   if FAlignment = taLeftJustify then
      OffsetRect( iR, BoxWidth + 2 + FSpacing, 0);
   InvalidateRect( Handle, @iR, FALSE);
end;


procedure TsBWCCControl.InvalidateBox;
var
   R: TRect;
begin
   if FAlignment = taLeftJustify then
      R := Bounds(0, 0, BoxWidth + 2 + FSpacing, Height - 2)
   else
      R := Bounds(Width - BoxWidth - 2 - FSpacing, 0, Width, Height - 2);
   InvalidateRect( Handle, @R, FALSE);
end;


procedure TsBWCCControl.Paint;
var
   R: TRect;
   bmp: TBitmap;

  tmpDC: HDC;
  tmpBmp, oldBmp: THandle;

begin
   inherited;
   case FVerticalAlignment of
      vaTop: begin
         R.Top := 1;
         R.Bottom := BoxWidth + 1;
      end;
      vaBottom: begin
         R.Top := Height - BoxWidth - 1;
         R.Bottom := Height - 1;
      end else begin // vaCenter
         R.Top := (Height div 2) - (BoxWidth div 2);
         R.Bottom := R.Top + BoxWidth;
      end;
   end;

   if FAlignment = taRightJustify then begin
      R.Left := Width - 1 - BoxWidth;
      R.Right := Width - 1;
   end else begin
      R.Left := 1;
      R.Right := BoxWidth + 1;
   end;


   bmp := TBitMap.Create;
   try
      bmp.height := Height;
      bmp.Width := Width;

      if Transparent then begin
      	tmpDC := CreateCompatibleDC(Canvas.Handle);
         tmpBmp := CreateCompatibleBitmap(Canvas.Handle, width, height);
         OldBmp := SelectObject(tmpDC, tmpBmp);
         try
            sTrans.PaintBackground(self, tmpDC);
            BitBlt( bmp.Canvas.handle, 0, 0, width, height,
               tmpDC, 0, 0, srccopy);
         finally
            SelectObject(tmpDC, oldBmp);
         	DeleteObject(tmpBmp);
      	   DeleteDC(tmpDC);
         end;
      end else with Bmp.Canvas do begin
         Brush.Style := bsSolid;
         Brush.Color := Color;
         FillRect(R);
      end;

      Bmp.Canvas.Font := Font;
      PaintBox(Bmp.Canvas, R);
      PaintCaption(Bmp.Canvas);

      BitBlt( Canvas.handle, 0, 0, width, height,
            bmp.canvas.handle, 0, 0, srccopy);
   finally
      bmp.Free;
   end;

{
      if Transparent then
         PaintBackground
      else with Canvas do begin
         Brush.Style := bsSolid;
         Brush.Color := Color;
         FillRect(R);
      end;

      Canvas.Font := Font;
      PaintBox(Canvas, R);
      PaintCaption(Canvas);
}
end;

procedure TsBWCCControl.PaintCaption( aCanvas: TCanvas);
const
   Alignments: array [TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
   WordWraps: array [Boolean] of Word = (0, DT_WORDBREAK);
var
   DrawStyle: Integer;
   R: TRect;
   hilighted: TsHilighted;
   X: Integer;
begin
   DrawStyle := DT_EXPANDTABS or WordWraps [FWordWrap] or Alignments [FTextAlignment];
   with aCanvas do begin
      Font := Font;

      hilighted := hNone;
      if Enabled and not (csDesigning in ComponentState) then begin
         if scsFocused in FCtrlState then
            hilighted := hHilight
         else if (scsMouseInControl in FCtrlState) and IsFlatApplied then
            hilighted := hMouseOver;
      end;

      FPaint3D.Assign(FFont3d);
      if not Enabled then begin
         if NewStyleControls then begin
            FPaint3D.Position := spRightBottom;
            FPaint3D.Spacing := 1;
            FPaint3D.Active := TRUE;
         end;
         Font.Color := clGrayText;
         FPaint3D.Color := clBtnHighlight;
      end;

      if FAlignment = taLeftJustify then
         X := BoxWidth + 2 + FSpacing
         //X := BoxWidth + 7;
      else
         X := 1;
      //R := Bounds(X, 1, Width - BoxWidth - 9, Height - 2);
      R := Bounds(X, 1, Width - BoxWidth - 2 - FSpacing, Height - 2);

      PaintText( aCanvas, PChar(Caption), Length(Caption), R,
         DrawStyle or DT_CALCRECT, FPaint3D, FHilight3d, FHilightStyle, hilighted);


      R.left := X;
      R.Right := X + Width - BoxWidth - FSpacing - 2;

      if FVerticalAlignment = vaBottom then
         OffsetRect(R, 0, Height - R.Bottom)
      else if FVerticalAlignment = vaCenter then
         OffsetRect(R, 0, (Height - R.Bottom) div 2);

      Brush.Style := bsClear;
      if FDown and FButtonLike then
         OffsetRect(R, 1, 1);
      PaintText( aCanvas, PChar(Caption), Length(Caption), R, DrawStyle,
         FPaint3D, FHilight3d, FHilightStyle, hilighted);

      if (scsFocused in FCtrlState) and FFocusFrame then begin
         PaintText( aCanvas, PChar(Caption), Length(Caption), R, DrawStyle or DT_CALCRECT,
            FPaint3D, FHilight3d, FHilightStyle, hilighted);
         case FTextAlignment of
            taCenter:
               OffsetRect(R, (Width - BoxWidth - 9 - R.Right + R.Left) div 2, 0);
            taRightJustify:
               OffsetRect(R, (Width - BoxWidth - 9 - R.Right + R.Left), 0);
         end;
         InflateRect(R, 1, 0);
         Brush.Color := clBtnFace;
         FrameRect(R);
         DrawFocusRect(R);
      end;
   end;
end;

procedure TsBWCCControl.AdjustBounds;
const
   WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
var
   DC: HDC;
   dX, X: Integer;
   Rect: TRect;
begin
   if FAutoSize and ([csLoading, csReading] * ComponentState = []) then begin
      Rect := ClientRect;

      dX := 1;
      if hilightStyle = hsFlushing then
         dX := FHilight3D.Spacing;

      InflateRect(Rect, -dX, 0);
      DC := GetDC(0);
      Canvas.Handle := DC;
      Canvas.Font := Font;

      PaintText( Canvas, PChar(Caption), Length(Caption), Rect, (DT_EXPANDTABS or DT_CALCRECT) or WordWraps[FWordWrap],
         FFont3d, FHilight3D, hsNone);

      Canvas.Handle := 0;
      ReleaseDC(0, DC);

      Inc(Rect.Right, BoxWidth + 6);
      Inc( Rect.Bottom, 2);

      X := Left;
      if FAlignment = taRightJustify then
         Inc(X, Width - Rect.Right - dx);
      InflateRect(Rect, dX, 0);
      SetBounds(X, Top, Rect.Right, Rect.Bottom);
   end;
end;

procedure TsBWCCControl.SetTransparentMode(Value: TTransparentMode);
begin
   if FTransparentMode <> Value then begin
      FTransparentMode := Value;
      Transparent := Value <> tmOpaque;
      Invalidate;
   end;
end;

procedure TsBWCCControl.SetAlignment(Value: TLeftRight);
begin
   if Value <> FAlignment then begin
      FAlignment := Value;
      Invalidate;
   end;
end;

procedure TsBWCCControl.SetAutoSize(Value: Boolean);
begin
   if FAutoSize <> Value then begin
      FAutoSize := Value;
      AdjustBounds;
   end;
end;

procedure TsBWCCControl.SetBoxColor(Value: TColor);
begin
	if Value <> FBoxColor then begin
   	FBoxColor := Value;
      InvalidateBox;
   end;
end;

procedure TsBWCCControl.SetButtonLike(Value: Boolean);
begin
   if FButtonLike <> Value then begin
      FButtonLike := Value;
      {if FAutoSize then
         AdjustBounds;}
   end;
end;

procedure TsBWCCControl.SetSpacing(Value: Integer);
begin
   if FSpacing <> Value then begin
      FSpacing := Value;
      Invalidate;
   end;
end;

procedure TsBWCCControl.SetTextAlignment(Value: TAlignment);
begin
   if FTextAlignment <> Value then begin
      FTextAlignment := Value;
      Invalidate;
   end;
end;

procedure TsBWCCControl.SetVerticalAlignment(Value: TVerticalAlignment);
begin
   if FVerticalAlignment <> Value then begin
      FVerticalAlignment := Value;
      Invalidate;
   end;
end;

procedure TsBWCCControl.SetWordWrap(Value: Boolean);
begin
   if FWordWrap <> Value then begin
      FWordWrap := Value;
      AdjustBounds;
      InvalidateCaption;
   end;
end;

function TsBWCCControl.GetCaption: TCaption;
var
   Buf: array [0..256] of Char;
begin
   GetTextBuf(Buf, 256);
   Result := StrPas(Buf);
end;

procedure TsBWCCControl.SetCaption(const Value: TCaption);
var
   Buffer: array [0..255] of Char;
begin
   if GetCaption <> Value then
      SetTextBuf(StrPCopy(Buffer, Value));
   InvalidateCaption;
end;

procedure TsBWCCControl.SetFont3d(Value: TFont3d);
begin
   if FFont3d <> Value then
      FFont3d.Assign(Value);
end;

procedure TsBWCCControl.Font3dChanged(Sender: Tobject);
begin
   AdjustBounds;
   Invalidate;
end;

procedure TsBWCCControl.SetHilight3d(Value: THilight3d);
begin
   if FHilight3d <> Value then
      FHilight3d.Assign( Value);
end;

procedure TsBWCCControl.SetHilightStyle(Value: TTextHilightStyle);
begin
   if Value <> FHilightStyle then begin
      FHilightStyle := Value;
      if FHilightStyle = hsFlushing then
         FFocusFrame := FALSE;
      if [scsMouseInControl, scsFocused] * FCtrlState <> [] then
         Invalidate;
   end;
end;

procedure TsBWCCControl.SetFocusFrame(Value: Boolean);
begin
   if Value <> FFocusFrame then begin
      FFocusFrame := Value;
      if [scsMouseInControl, scsFocused] * FCtrlState <> [] then
         Invalidate;
   end;
end;

function TsBWCCControl.GetChecked: Boolean;
begin
   Result := FState = cbChecked;
end;

function TsBWCCControl.EditCanModify: Boolean;
begin
   Result := not FReadOnly;
end;

procedure TsBWCCControl.SetState(Value: TCheckBoxState);
begin
   if {EditCanModify and }(FState <> Value) then begin
      FState := Value;
      if FButtonLike then
         DoInvalidateRect(handle, clientrect, FALSE)
      else
         InvalidateBox;
   end;
end;

procedure TsBWCCControl.SetChecked(const Value: Boolean);
const
   checkStates: array [Boolean] of TCheckBoxState = (cbUnchecked, cbChecked);
begin
   if EditCanModify and (Value <> Checked) then begin
      State := checkStates [Value];
      {if FButtonLike then
         DoInvalidateRect(handle, clientrect, FALSE)
      else
         InvalidateBox;}
   end;
end;

procedure TsBWCCControl.SetCheckColor(Value: TColor);
begin
   FCheckColor := Value;
   Invalidate;
end;

procedure TsBWCCControl.DoEnter;
begin
   inherited DoEnter;
end;

procedure TsBWCCControl.DoExit;
begin
   inherited DoExit;
//   Exclude( FCtrlState, scsFocused);
//   Invalidate;
end;

procedure TsBWCCControl.DoMouseEnter;
begin
	if Assigned( FOnMouseEnter) then
      FOnMouseEnter(self);
end;

procedure TsBWCCControl.DoMouseLeave;
begin
	if Assigned( FOnMouseLeave) then
      FOnMouseLeave(self);
end;

procedure TsBWCCControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if Button = mbLeft then begin
      SetFocus;
      inherited;
      MouseCapture := True;
      Include( FCtrlState, scsTracking);
      FDown := TRUE;
      if FButtonLike then
         DoInvalidateRect(handle, clientrect, FALSE)
      else
         InvalidateBox;
   end else
      inherited;
end;

procedure TsBWCCControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if Button = mbLeft then begin
      MouseCapture := False;
      Exclude( FCtrlState, scsTracking);
      FDown := FALSE;
      if PtInRect(ClientRect, Point(X, Y)) then
         Toggle;
   end;
   inherited;
end;

procedure TsBWCCControl.MouseMove(Shift: TShiftState; X, Y: Integer);
var
   oldDown: Boolean;
begin
   if scsTracking in FCtrlState {MouseCapture }then begin
      oldDown := FDown;
      FDown := (X >= 0) and (X <= Width) and (Y >= 0) and (Y <= Height);
      if oldDown <> FDown then begin
         if FButtonLike then
            DoInvalidateRect(handle, clientrect, FALSE)
         else
            InvalidateBox;
      end;
   end;
   inherited;
end;

procedure TsBWCCControl.KeyDown(var Key: Word; Shift: TShiftSTate);
begin
   if (Key = vk_Space) and not FDown then begin
      FDown := TRUE;
      if FButtonLike then
         DoInvalidateRect(handle, clientrect, FALSE)
      else
         InvalidateBox;
   end;
   inherited KeyDown(Key, Shift);
end;

procedure TsBWCCControl.KeyUp(var Key: Word; Shift: TShiftSTate);
begin
   if Key = vk_Space then begin
      FDown := False;
      Toggle;
   end;
end;

{*******************************************************************************
   TsCheckBox
*******************************************************************************}

constructor TsCustomCheckBox.Create(Owner: TComponent);
begin
   inherited;
	FStyle := DEF_CheckBox_Style;
end;

procedure TsCustomCheckBox.SetStyle( Value: TsBWCCBoxStyle);
begin
   if FStyle <> Value then begin
      FStyle := Value;
      Invalidate;
   end;
end;

procedure TsCustomCheckBox.Toggle;
begin
   if EditCanModify then begin
      if FAllowGrayed and (FState = cbUnchecked) then
         State := cbGrayed
      else
         Checked := not Checked;
      Click;
   end else begin
      if FButtonLike then
         DoInvalidateRect(handle, clientrect, FALSE)
      else
         InvalidateBox;
   end;
end;

const
   ColorMap: array[1..2, Boolean] of TColor =
      ((clBtnHighLight, clBtnShadow), (clBtnShadow, clBtnHighLight));

procedure TsCustomCheckBox.PaintBox(aCanvas: TCanvas; R: TRect);
begin
   with aCanvas do begin
      Brush.Color := FBoxColor;
      if transparentMode <> tmTransparent then
         FillRect(R);
      if FDown then begin
         Pen.Color := clBlack;
         Pen.Width := 2;
         Rectangle( R.Left + 1, R.Top + 1, R.Right + 1, R.Bottom + 1);
         Pen.Width := 1;
      end else begin
         if FStyle = bwsFlat then
				Brush.Color := Font.Color;
         FrameRect(R);
         if FStyle <> bwsFlat then begin
         	Pen.Color := ColorMap[1, Boolean(Ord(FStyle))];
         	PolyLine([Point(R.Left, R.Bottom), R.TopLeft, Point(R.Right, R.Top)]);
         	Pen.Color := ColorMap[2, Boolean(Ord(FStyle))];
         	PolyLine([Point(R.Right, R.Top), R.BottomRight, Point(R.Left, R.Bottom)]);
        	end;
      end;

      case State of
         cbChecked: begin
               if FStyle = bwsSunken then
                  OffsetRect(R, 0, -1);

               Pen.Color := FCheckColor;
               Pen.Width := 1;
               Dec(R.Top);
               Dec(R.Bottom);
               Polyline([Point(R.Left + 2, R.Top + BoxWidth div 2 + 1), Point(R.Left + 2, R.Bottom - 1)]);
               Polyline([Point(R.Left + 3, R.Top + BoxWidth div 2), Point(R.Left + 3, R.Bottom - 2)]);
               Polyline([Point(R.Left + 2, R.Bottom - 1), Point(R.Right - 2, R.Top + 3)]);
               Polyline([Point(R.Left + 3, R.Bottom - 1), Point(R.Right - 1, R.Top + 3)]);
            end;
         cbGrayed: begin
               OffsetRect( R, 2, 2);
               Dec(R.Right, 3);
               Dec(R.Bottom, 3);
               if FDown then begin
                  Pen.Color := clBtnFace;
                  FrameRect(R);
               end;
               Pen.Color := ColorMap[1, Boolean(FStyle)];
               Brush.Color := FCheckColor;
               Rectangle(R.Left, R.Top, R.Right, R.Bottom);
         end;
      end;
   end;
end;

{*******************************************************************************
   TsRadioButton
*******************************************************************************}
constructor TsCustomRadioButton.Create(Owner: TComponent);
begin
   inherited;
   FStyle := DEF_RadioButtonStyle;
end;

procedure TsCustomRadioButton.SetStyle(Value: TsBWCCRadioStyle);
begin
   if FStyle <> Value then begin
      FStyle := Value;
      Invalidate;
   end;
end;

procedure TsCustomRadioButton.Toggle;
begin
   if EditCanModify then begin
      if not Checked then begin
         Checked := TRUE;
         Click;
      end else begin
         if FButtonLike then
            DoInvalidateRect(handle, clientrect, FALSE)
         else
            InvalidateBox;
      end;
   end;
end;

procedure TsCustomRadioButton.SetChecked(const Value: Boolean);
var
   ii: Integer;
   ctrl: TsCustomRadioButton;
begin
   inherited;
   if Value and (Parent <> nil) then for ii := 0 to Parent.ControlCount - 1 do
      if Parent.Controls[ii] is TsCustomRadioButton then begin
         ctrl := TsCustomRadioButton(Parent.Controls[ii]);
         if( ctrl <> Self) and ( ctrl.FGroupIndex = FGroupIndex) then
            Ctrl.SetChecked(FALSE);
      end;
end;

procedure TsCustomRadioButton.PaintBox(aCanvas: TCanvas; R: TRect);
var
   HalfBoxWidth: Integer;
begin
   HalfBoxWidth := BoxWidth div 2;
   with aCanvas do begin
      if FDown then begin
         Pen.Color := clBlack;
         PolyLine([Point(R.Left + HalfBoxWidth, R.Top), Point(R.Left, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Bottom), Point(R.Right, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Top)]);
         PolyLine([Point(R.Left + HalfBoxWidth, R.Top + 1), Point(R.Left + 1, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Bottom - 1), Point(R.Right - 1, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Top + 1)]);
      end else begin
			if FStyle = brsFlat then
				Pen.Color := Font.Color
         else
         	Pen.Color := ColorMap[1, Checked];
         PolyLine([Point(R.Left + HalfBoxWidth, R.Top), Point(R.Left, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Bottom)]);
         if FStyle <> brsFlat then
         	Pen.Color := ColorMap[2, Checked];
         PolyLine([Point(R.Left + HalfBoxWidth, R.Bottom), Point(R.Right, R.Top + HalfBoxWidth),
            Point(R.Left + HalfBoxWidth, R.Top)]);
         if FTransparentMode = tmSemiTransparent then begin
            Pen.Color := FBoxColor;
            Brush.Color := FBoxColor;
            InflateRect(R, -1, -1);
            dec(HalfBoxWidth);
            Polygon([Point(R.Left + HalfBoxWidth, R.Top), Point(R.Left, R.Top + HalfBoxWidth),
                     Point(R.Left + HalfBoxWidth, R.Bottom), Point(R.Right, R.Top + HalfBoxWidth),
                     Point(R.Left + HalfBoxWidth, R.Top)]);
            InflateRect(R, 1, 1);
            Inc(HalfBoxWidth);
         end;
      end;
      if Checked then begin
         Pen.Color := FCheckColor;
         Inc(R.Left, HalfBoxWidth);
         Inc(R.Top, HalfBoxWidth);

         PolyLine([Point(R.Left - 1, R.Top - 1), Point(R.Left + 2, R.Top - 1)]);
         PolyLine([Point(R.Left - 2, R.Top), Point(R.Left + 3, R.Top)]);
         PolyLine([Point(R.Left - 1, R.Top + 1), Point(R.Left + 2, R.Top + 1)]);
         PolyLine([Point(R.Left, R.Top - 2), Point(R.Left, R.Top + 3)]);
      end;
   end;
end;

{*******************************************************************************
   TsLabel
*******************************************************************************}
constructor TsCustomLabel.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FAlignment := DEF_Label_Alignment;
   FAutoSize := DEF_Label_AutoSize;
	FButtonLike := DEF_Label_ButtonLike;
   FFocusFrame := DEF_Label_FocusFrame;
   FFont3d := TFont3d.Create(Font3dChanged);
   FHilight3d := THilight3d.Create(Font3dChanged);
   FHilightStyle := DEF_Label_HilightStyle;
   FLayout := DEF_Label_Layout;
   FPaint3D := TFont3d.Create(nil);
   if not DEF_Label_Transparent then
      ControlStyle := ControlStyle + [csOpaque];
   Width := 65;
   Height := 17;

   FShowAccelChar := TRUE;
   FShowFocus := TRUE;
end;

destructor TsCustomLabel.Destroy;
begin
   FFont3d.Free;
   FHilight3d.Free;
   FMonitorControls.Free;
   FPaint3D.Free;
   inherited;
end;


procedure TsCustomLabel.AddMonitorControl(ctrl: TWinControl);
begin
   if FMonitorControls = nil then
      FMonitorControls := TList.Create;
   FMonitorControls.Add(ctrl);
end;

procedure TsCustomLabel.RemoveMonitorControl(ctrl: TWinControl);
var
   index: Integer;
begin
   if FMonitorControls = nil then
      Exit;
   index := FMonitorControls.IndexOf( ctrl);
   if index > -1 then
      FMonitorControls.Delete(index);
   if FMonitorControls.Count = 0 then begin
      FMonitorControls.Free;
      FMonitorControls := nil;
   end;
end;

procedure TsCustomLabel.AddMonitorControls(ctrls: array of TWinControl);
var
   ii: Integer;
begin
   if FMonitorControls = nil then
      FMonitorControls := TList.Create;
   for ii := Low(ctrls) to High(ctrls) do
      FMonitorControls.Add(ctrls[ii]);
end;

procedure TsCustomLabel.RemoveMonitorControls(ctrls: array of TWinControl);
var
   ii: Integer;
   index: Integer;
begin
   if FMonitorControls = nil then
      Exit;
   for ii := Low(ctrls) to High(ctrls) do begin
      index := FMonitorControls.IndexOf( ctrls[ii]);
      if index > -1 then
         FMonitorControls.Delete(index);
   end;
   if FMonitorControls.Count = 0 then begin
      FMonitorControls.Free;
      FMonitorControls := nil;
   end;
end;

function TsCustomLabel.GetLabelText: String;
begin
   Result := Caption;
end;

procedure TsCustomLabel.DoDrawText(aCanvas: TCanvas; var Rect: TRect; Flags: Word);
var
   Text: string;
   hilighted: TsHilighted;
begin
   Text := GetLabelText;
   if (Flags and DT_CALCRECT <> 0) and ((Text = '') or FShowAccelChar and
      (Text[1] = '&') and (Text[2] = #0)) then
      Text := Text + ' ';
   if not FShowAccelChar then
      Flags := Flags or DT_NOPREFIX;
   aCanvas.Font := Font;


   hilighted := hNone;
   if Enabled and not (csDesigning in ComponentState) then begin
      if scsFocused in FCtrlState then
         hilighted := hHilight
      else if (scsMouseInControl in FCtrlState) and IsFlatApplied then
         hilighted := hMouseOver;
   end;

   FPaint3D.Assign(FFont3D);
   if not Enabled then begin
      if NewStyleControls then begin
         FPaint3D.Position := spRightBottom;
         FPaint3D.Spacing := 1;
         FPaint3D.Active := TRUE;
      end;
      aCanvas.Font.Color := clGrayText;
      FPaint3D.Color := clBtnHighlight;
   end;
   PaintText( aCanvas, PChar(Text), Length(Text), Rect, Flags,
      FPaint3d, FHilight3d, FHilightStyle, hilighted);
end;

procedure TsCustomLabel.Paint;
const
   Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
   WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
var
   Rect: TRect;
   DrawStyle: Integer;
   Bmp: TBitMap;
   H, W: Integer;
begin
   Bmp := TBitmap.Create;
   W := Width;
   H := Height;
   if FDown then begin
      Inc(W, 2);
      Inc(H, 2);
   end;
   BMP.Width := W;
   BMP.Height := H;

   with BMP.Canvas do try
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(ClientRect);
      Brush.Style := bsClear;
      Rect := ClientRect;
      InflateRect(Rect, -2, -1);
      DrawStyle := DT_EXPANDTABS or WordWraps[FWordWrap] or Alignments[FAlignment];
      { Calculate vertical layout }
      if FLayout <> tlTop then begin
         DoDrawText(BMP.Canvas, Rect, DrawStyle or DT_CALCRECT);
         Rect.Left := ClientRect.Left;
         Rect.Right := ClientRect.Right;
         if FLayout = tlBottom then
            OffsetRect(Rect, 0, Height - Rect.Bottom)
         else
            OffsetRect(Rect, 0, (Height - Rect.Bottom) div 2);
      end;
      if FDown then
         OffsetRect(Rect, 1, 1);

      DoDrawText( BMP.Canvas, Rect, DrawStyle);
      if FShowFocus and Assigned(FFocusControl) and FFocusControl.Focused and
         FFocusFrame and not (csDesigning in ComponentState) then begin
         InflateRect(Rect, 1, 0);
         Brush.Color := clBtnFace;
         FrameRect(Rect);
         DrawFocusRect(Rect);
      end;
      if Transparent then
         DrawBitmapTransparent( Canvas, 0, 0, Bmp, Self.Color)
      else
         BitBlt( Canvas.handle, 0, 0, width, height, bmp.canvas.handle, 0, 0, srccopy);
   finally
      BMP.Free;
   end;
end;

procedure TsCustomLabel.AdjustBounds;
const
   WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
var
   DC: HDC;
   dX, X: Integer;
   Rect: TRect;
begin
   if FAutoSize then begin
      Rect := ClientRect;

      dX := 1;
      if hilightStyle = hsFlushing then
         dX := FHilight3D.Spacing;

      InflateRect(Rect, -dX, 0);
      DC := GetDC(0);
      Canvas.Handle := DC;
      DoDrawText( Canvas, Rect, (DT_EXPANDTABS or DT_CALCRECT) or WordWraps[FWordWrap]);
      Canvas.Handle := 0;
      ReleaseDC(0, DC);
      X := Left;

      if FAlignment = taRightJustify then
         Inc(X, Width - Rect.Right - 2 * dx);
      InflateRect(Rect, dX, 0);

      Inc( Rect.Right);
      if HilightStyle = hsUnderline then
         Inc( Rect.Bottom, 1 + Ord(FButtonLike));

      SetBounds(X, Top, Rect.Right, Rect.Bottom);
   end;
end;

procedure TsCustomLabel.SetAlignment(Value: TAlignment);
begin
   if FAlignment <> Value then begin
      FAlignment := Value;
      Invalidate;
   end;
end;

procedure TsCustomLabel.SetAutoSize(Value: Boolean);
begin
   if FAutoSize <> Value then begin
      FAutoSize := Value;
      AdjustBounds;
   end;
end;

procedure TsCustomLabel.SetLayout(Value: TsTextLayout);
begin
   if FLayout <> Value then begin
      FLayout := Value;
      Invalidate;
   end;
end;

procedure TsCustomLabel.SetButtonLike(Value: Boolean);
begin
   if FButtonLike <> Value then begin
      FButtonLike := Value;
      if FHilightStyle = hsUnderline then
         AdjustBounds;
      {if FAutoSize then
         AdjustBounds;}
   end;
end;

procedure TsCustomLabel.Font3dChanged(Sender: Tobject);
begin
   AdjustBounds;
   Invalidate;
end;

procedure TsCustomLabel.SetFont3D(Value: TFont3D);
begin
   if FFont3d <> Value then
      FFont3d.Assign(Value);
end;

procedure TsCustomLabel.SetHilight3d(Value: THilight3d);
begin
   if FHilight3d <> Value then
      FHilight3d.Assign( Value);
end;

procedure TsCustomLabel.SetHilightStyle(Value: TTextHilightStyle);
begin
   if Value <> FHilightStyle then begin
      FHilightStyle := Value;
      if FHilightStyle = hsFlushing then
         FFocusFrame := FALSE
      else if FHilightStyle = hsUnderline then
         AdjustBounds;
      if [scsMouseInControl, scsFocused] * FCtrlState <> [] then
         Invalidate;
   end;
end;

procedure TsCustomLabel.SetFocusFrame(Value: Boolean);
begin
   if Value <> FFocusFrame then begin
      FFocusFrame := Value;
      if [scsMouseInControl, scsFocused] * FCtrlState <> [] then
         Invalidate;
   end;
end;

function TsCustomLabel.GetTransparent: Boolean;
begin
   Result := not (csOpaque in ControlStyle);
end;

procedure TsCustomLabel.SetFocusControl(Value: TWinControl);
begin
   if Value <> FFocusControl then begin
      if FFocusControl <> nil then
         FFocusControl.Perform( STM_LINKEDCHANGENOTIFY, LongInt(self), 0);
      FFocusControl := Value;
      if Value <> nil then begin
         Value.FreeNotification(Self);
         Value.Perform( STM_LINKEDCHANGENOTIFY, LongInt(self), 1);
         Enabled := FFocusControl.Enabled;
      end;
      if FShowFocus then
         Invalidate;
   end;
end;

procedure TsCustomLabel.SetShowAccelChar(Value: Boolean);
begin
   if FShowAccelChar <> Value then begin
      FShowAccelChar := Value;
      Invalidate;
   end;
end;

procedure TsCustomLabel.SetTransparent(Value: Boolean);
begin
   if Transparent <> Value then begin
      if Value then
         ControlStyle := ControlStyle - [csOpaque]
      else
         ControlStyle := ControlStyle + [csOpaque];
      Invalidate;
   end;
end;

procedure TsCustomLabel.SetShowFocus(Value: Boolean);
begin
   if FShowFocus <> Value then begin
      FShowFocus := Value;
      Invalidate;
   end;
end;

procedure TsCustomLabel.SetWordWrap(Value: Boolean);
begin
   if FWordWrap <> Value then begin
      FWordWrap := Value;
      AdjustBounds;
   end;
end;

procedure TsCustomLabel.Notification(AComponent: TComponent; Operation: TOperation);
var
   index: Integer;
begin
   inherited Notification(AComponent, Operation);
   if (Operation = opRemove) then begin
      if AComponent = FFocusControl then
         FocusControl := nil;
      if FMonitorControls <> nil then begin
         index := FMonitorControls.IndexOf( AComponent);
         if index > -1 then begin
            FMonitorControls.Delete(index);
            if FMonitorControls.Count = 0 then begin
               FMonitorControls.Free;
               FMonitorControls := nil;
            end;
         end;
      end;
   end;
end;

procedure TsCustomLabel.DoMouseEnter;
begin
	if Assigned( FOnMouseEnter) then
      FOnMouseEnter(self);
end;

procedure TsCustomLabel.DoMouseLeave;
begin
	if Assigned( FOnMouseLeave) then
      FOnMouseLeave(self);
end;

procedure TsCustomLabel.DoMouseClick;
begin
	if Assigned( FOnMouseClick) then
      FOnMouseClick(self);
   if FFocusControl <> nil then
      with FFocusControl do if CanFocus then
         SetFocus;
end;

procedure TsCustomLabel.CMFocusChanged(var Message: TCMFocusChanged);
var
   Active: Boolean;
begin
   Active := Assigned(FFocusControl) and (Message.Sender = FFocusControl);
   if not Active and (FMonitorControls <> nil) then
      Active := (FMonitorControls.IndexOf(Message.Sender) > -1);
   if (scsFocused in FCtrlState) <> Active then begin
      if FShowFocus then
         CheckInvalidate(Active, FALSE);
   end;
   inherited;
end;

procedure TsCustomLabel.CMTextChanged(var Message: TMessage);
begin
   Invalidate;
   AdjustBounds;
end;

procedure TsCustomLabel.CMFontChanged(var Message: TMessage);
begin
   inherited;
   AdjustBounds;
end;

procedure TsCustomLabel.CMDialogChar(var Message: TCMDialogChar);
begin
   if (FFocusControl <> nil) and Enabled and ShowAccelChar and
      IsAccel(Message.CharCode, Caption) then with FFocusControl do
      if CanFocus then begin
         SetFocus;
         Message.Result := 1;
      end;
end;

procedure TsCustomLabel.CMMouseEnter(var Message: TMessage);
var
   ii: Integer;
begin
   CheckInvalidate(TRUE, TRUE);
   if (FFocusControl <> nil) and Enabled then
      FFocusControl.Perform(CM_MouseEnter, 1, Longint(Self));
   if FMonitorControls <> nil then
      for ii := 0 to FMonitorControls.Count - 1 do
         TControl(FMonitorControls[ii]).Perform(CM_MouseEnter, 1, Longint(Self));
   DoMouseEnter;
end;

procedure TsCustomLabel.CMMouseLeave(var Message: TMessage);
var
   ii: Integer;
begin
   CheckInvalidate(FALSE, TRUE);
   if (FFocusControl <> nil) and Enabled then
      FFocusControl.Perform(CM_MouseLeave, 1, Longint(Self));
   if FMonitorControls <> nil then
      for ii := 0 to FMonitorControls.Count - 1 do
         TControl(FMonitorControls[ii]).Perform(CM_MouseLeave, 1, Longint(Self));
   DoMouseLeave;
end;

procedure TsCustomLabel.STMEnabledChangeNotify(var Message: TMessage);
var
   isEnabled: Boolean;
   ii: Integer;
begin
   if (Message.Result = 0) and
      ((Message.WParam = LongInt(FFocusControl)) or
      ((FMonitorControls <> nil) and (FMonitorControls.IndexOf(Pointer(Message.WParam)) > -1))) then begin
      isEnabled := FALSE;
      if FFocusControl <> nil then
         isEnabled := FFocusControl.Enabled;
      if FMonitorControls <> nil then for ii := 0 to FMonitorControls.Count - 1 do
         isEnabled := TWinControl(FMOnitorControls[ii]).Enabled or IsEnabled;
      Enabled := isEnabled;
      Message.Result := 1;
   end;
end;

procedure TsCustomLabel.STMMouseEnterNotify(var Message: TMessage);
begin
   if (Message.Result = 0) and
      ((Message.WParam = LongInt(FFocusControl)) or
      ((FMonitorControls <> nil) and (FMonitorControls.IndexOf(Pointer(Message.WParam)) > -1))) then begin
      CheckInvalidate(TRUE, TRUE);
      Message.Result := 1;
   end;
end;

procedure TsCustomLabel.STMMouseLeaveNotify(var Message: TMessage);
begin
   if (Message.Result = 0) and
      ((Message.WParam = LongInt(FFocusControl)) or
      ((FMonitorControls <> nil) and (FMonitorControls.IndexOf(Pointer(Message.WParam)) > -1))) then begin
      CheckInvalidate(FALSE, TRUE);
      Message.Result := 1;
   end;
end;

procedure TsCustomLabel.CheckInvalidate(hilight, mouseEvent: Boolean);
var
   needRefresh: Boolean;
begin
//   if not FHilight3D.Active and (HilightStyle = hsNone) then
//      Exit;

   needRefresh := Enabled and not (csDesigning in ComponentState) and
                  ((not (scsMouseInControl in FCtrlState) and not MouseEvent) or
                  (not (scsFocused in FCtrlState) and MouseEvent) or
                  (FFocusFrame and  not (scsFocused in FCtrlState) and not mouseEvent)) and
                  ((FHilightStyle <> hsNone) or (FHilight3D.Active and not FHilight3D.Equal(FFont3D, MouseEvent)));



   if {not needRefresh and (Hilight3d.MouseOverColor <> Hilight3d.color) and
      (MouseEvent and not (scsFocused in FCtrlState)) or}
      (not MouseEvent and (scsMouseInControl in FCtrlState)) and Enabled then
      needRefresh := TRUE;


   if mouseEvent then begin
      needRefresh := needRefresh and IsFlatApplied;
      if hilight then
         Include(FCtrlState, scsMouseInControl)
      else
         Exclude(FCtrlState, scsMouseInControl);
   end else begin
      if FloatingOnFocusOnly then
         needRefresh := TRUE;
      if hilight then
         Include(FCtrlState, scsFocused)
      else
         Exclude(FCtrlState, scsFocused);
   end;
   if needRefresh then
      Refresh;
end;

procedure TsCustomLabel.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if Enabled and not (scsFocused in FCtrlState) and FButtonLike and (Button = mbLeft) and
      (FFocusControl <> nil) and (FFocusControl.CanFocus) then begin
      inherited;
      MouseCapture := True;
      Include( FCtrlState, scsTracking);
      FDown := TRUE;
      Invalidate;
   end else
      inherited;
end;

procedure TsCustomLabel.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if Enabled and not (scsFocused in FCtrlState) and FButtonLike and (Button = mbLeft) and
      (FFocusControl <> nil) and (FFocusControl.CanFocus) then begin
      MouseCapture := False;
      Exclude( FCtrlState, scsTracking);
      if PtInRect(ClientRect, Point(X, Y)) then
         DoMouseClick;
      FDown := FALSE;
      Invalidate;
   end;
   inherited;
end;

procedure TsCustomLabel.MouseMove(Shift: TShiftState; X, Y: Integer);
var
   oldDown: Boolean;
begin
   if scsTracking in FCtrlState{MouseCapture} then begin
      oldDown := FDown;
      FDown := (X >= 0) and (X <= Width) and (Y >= 0) and (Y <= Height);
      if oldDown <> FDown then
         Invalidate;
   end;
   inherited;
end;

{*******************************************************************************
   TsStatusBar
*******************************************************************************}

constructor TsStatusBar.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
   ControlStyle := ControlStyle + [csAcceptsControls];
   //ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
   //   csSetCaption, csOpaque, csDoubleClicks, csReplicatable];
end;

procedure TsStatusBar.CreateParams(var Params: TCreateParams);
const
  GripStyles: array[Boolean] of DWORD = (CCS_TOP, SBARS_SIZEGRIP);
begin
  inherited CreateParams(Params);
  with Params do begin
      Style := Style and not CCS_TOP;
      Style := Style and not SBARS_SIZEGRIP;
      Style := Style or GripStyles[SizeGrip];
      //WindowClass.style := WindowClass.style and not CS_HREDRAW;
  end;
end;

{*******************************************************************************
   TsScrollBar
*******************************************************************************}
{$IFDEF FlatSB}
constructor TsCustomScrollBar.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   Width := 121;
   Height := GetSystemMetrics(SM_CYHSCROLL);
   TabStop := True;
   ControlStyle := [csDoubleClicks, csOpaque];
   FKind := sbHorizontal;
   FPosition := 0;
   FMin := 0;
   FMax := 100;
   FSmallChange := 1;
   FLargeChange := 1;
   FMode:= DEF_ScrollBar_Mode;
   ImeMode := imDisable;
end;

procedure TsCustomScrollBar.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
   if NotRightToLeft then
      FRTLFactor := 1
   else
      FRTLFactor := -1;
end;

procedure TsCustomScrollBar.CreateWnd;
begin
   inherited CreateWnd;
   InitializeFlatSB(Handle);
   FlatSB_ShowScrollBar(Handle, Integer(FKind), True);
   FlatSB_SetScrollRange(Handle, Integer(FKind), FMin, FMax, False);
   if NotRightToLeft then
      FlatSB_SetScrollPos(Handle, Integer(FKind), FPosition, True)
   else
      FlatSB_SetScrollPos(Handle, Integer(FKind), FMax - FPosition, True);

   if FKind = sbVertical then
      FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Integer(FMode), True)
   else
      FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Integer(FMode), True);

   if FKind = sbHorizontal then
      FlatSB_SetScrollProp(Handle, WSB_PROP_CYHSCROLL, Height-1, True)
   else
      FlatSB_SetScrollProp(Handle, WSB_PROP_CXVSCROLL, Width, True);
   ControlStyle := ControlStyle-[csOpaque];
end;

function TsCustomScrollBar.NotRightToLeft: Boolean;
begin
   Result := (not IsRightToLeft) or (FKind = sbVertical);
end;

procedure TsCustomScrollBar.SetKind(Value: TScrollBarKind);
begin
   if FKind <> Value then begin
      FKind := Value;
      if not (csLoading in ComponentState) then
         SetBounds(Left, Top, Height, Width);
      RecreateWnd;
   end;
end;

procedure TsCustomScrollBar.SetParams(APosition, AMin, AMax: Integer);
begin
   if AMax < AMin then
      raise EInvalidOperation.Create(SScrollBarRange);
   if APosition < AMin then
      APosition := AMin;
   if APosition > AMax then
      APosition := AMax;
   if (FMin <> AMin) or (FMax <> AMax) then begin
      FMin := AMin;
      FMax := AMax;
      if HandleAllocated then
         FlatSB_SetScrollRange(Handle, Integer(FKind), AMin, AMax, FPosition = APosition);
   end;
   if FPosition <> APosition then begin
      FPosition := APosition;
      if HandleAllocated then
         if NotRightToLeft then
            FlatSB_SetScrollPos(Handle, Integer(FKind), FPosition, True)
         else
            FlatSB_SetScrollPos(Handle, Integer(FKind), FMax - FPosition, True);
      Change;
   end;
end;

procedure TsCustomScrollBar.SetPosition(Value: Integer);
begin
   SetParams(Value, FMin, FMax);
end;

procedure TsCustomScrollBar.SetPageSize(Value: Integer);
var
   ScrollInfo: TScrollInfo;
begin
   if (FPageSize = Value) or (FPageSize > FMax) then
      Exit;
   FPageSize := Value;
   ScrollInfo.nPage := Value;
   ScrollInfo.fMask := SIF_PAGE;
   if HandleAllocated then
      FlatSB_SetScrollInfo(Handle, Integer(FKind), ScrollInfo, True);
end;

procedure TsCustomScrollBar.SetMin(Value: Integer);
begin
   SetParams(FPosition, Value, FMax);
end;

procedure TsCustomScrollBar.SetMax(Value: Integer);
begin
   SetParams(FPosition, FMin, Value);
end;

procedure TsCustomScrollBar.Change;
begin
   inherited Changed;
   if Assigned(FOnChange) then
      FOnChange(Self);
end;

procedure TsCustomScrollBar.Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer);
begin
   if Assigned(FOnScroll) then
      FOnScroll(Self, ScrollCode, ScrollPos);
end;

procedure TsCustomScrollBar.DoScroll(var Message: TWMScroll);
var
   ScrollPos: Integer;
   NewPos: Longint;
   ScrollInfo: TScrollInfo;
begin
   with Message do begin
      NewPos := FPosition;
      case TScrollCode(ScrollCode) of
         scLineUp:
            Dec(NewPos, FSmallChange * FRTLFactor);
         scLineDown:
            Inc(NewPos, FSmallChange * FRTLFactor);
         scPageUp:
            Dec(NewPos, FLargeChange * FRTLFactor);
         scPageDown:
            Inc(NewPos, FLargeChange * FRTLFactor);
         scPosition, scTrack: with ScrollInfo do begin
            cbSize := SizeOf(ScrollInfo);
            fMask := SIF_ALL;
            FlatSB_GetScrollInfo(Handle, Integer(FKind), ScrollInfo);
            NewPos := nTrackPos;
            if not NotRightToLeft then NewPos := FMax - NewPos;
         end;
         scTop:
            NewPos := FMin;
         scBottom:
            NewPos := FMax;
      end;
      if NewPos < FMin then
         NewPos := FMin;
      if NewPos > FMax then
         NewPos := FMax;
      ScrollPos := NewPos;
      Scroll(TScrollCode(ScrollCode), ScrollPos);
      SetPosition(ScrollPos);
   end;
end;

procedure TsCustomScrollBar.CNCtlColorScrollBar(var Message: TMessage);
begin
   with Message do
      CallWindowProc(DefWndProc, Handle, Msg, WParam, LParam);
end;

procedure TsCustomScrollBar.WMVScroll(var Msg: TWMVScroll);
begin
   DoScroll(Msg);
end;

procedure TsCustomScrollBar.WMHScroll(var Msg: TWMHScroll);
begin
   DoScroll(Msg);
end;

procedure TsCustomScrollBar.SetMode(Value: TScrollbarMode);
begin
   FMode:= Value;
   if FKind = sbVertical then
      FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Integer(FMode), True)
   else
      FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Integer(FMode), True);
end;

procedure TsCustomScrollBar.WMSize(var Msg: TWMSize);
begin
   inherited;
   if FKind=sbHorizontal then
      FlatSB_SetScrollProp(Handle, WSB_PROP_CYHSCROLL, Height-1, True)
   else
      FlatSB_SetScrollProp(Handle, WSB_PROP_CXVSCROLL, Width, True);
end;

procedure TsCustomScrollBar.WMPaint(var Msg: TWMPaint);
begin
   inherited;
   Canvas.Brush.Color:= clBtnFace;
   Canvas.FillRect(BoundsRect);
end;
{$ENDIF}

constructor TsCustomPanel.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
      csOpaque, csDoubleClicks, csReplicatable];
   Color := clBtnFace;
   BorderWidth := 3;

   Width := 185;
   Height := 41;
end;

destructor TsCustomPanel.Destroy;
begin
   inherited;
end;

procedure TsCustomPanel.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
end;

procedure TsCustomPanel.CreateWnd;
begin
   inherited;
   SetSolidBorder;
end;

procedure TsCustomPanel.WMNCPaint(var Message: TMessage);
begin
   inherited;
   RedrawBorders;
end;

procedure TsCustomPanel.WMSetFocus(var Message: TMessage);
var
   ii: Integer;
   ctrl: TWinControl;
begin
   inherited;
   SetSolidBorder;
   ctrl := nil;
   for ii := 0 to ControlCount -1 do
      if (Controls[ii] is TWinControl) and TWinControl(Controls[ii]).TabStop then
         if (ctrl = nil) or (ctrl.TabOrder > TWinControl(Controls[ii]).TabOrder) then
            ctrl := TWinControl(Controls[ii]);
   if ctrl <> nil then
      ctrl.SetFocus;
end;

procedure TsCustomPanel.WMKillFocus(var Message: TMessage);
begin
   SetSolidBorder;
   inherited;
end;

procedure TsCustomPanel.WMFocusChanged(var Message: TMessage);
begin
   inherited;
   SetSolidBorder;
end;

procedure TsCustomPanel.CMMouseEnter(var Message: TMessage);
var
  P: TPoint;
begin
   inherited;
   if not FMouseInControl and Enabled then begin
      GetCursorPos(P);
      if (Message. wParam = 1) or PointInRect(ScreenToClient(P), GetClientRect) then begin
         FMouseInControl := TRUE;
         SetSolidBorder;
         Notify( self, STM_MOUSEENTERNOTIFY, 0);
      end;
   end;
end;

procedure TsCustomPanel.CMMouseLeave(var Message: TMessage);
var
  P: TPoint;
begin
   inherited;
   if FMouseInControl and Enabled then begin
      GetCursorPos(P);
      if (Message. wParam = 1) or not PointInRect(ScreenToClient(P), GetClientRect) then begin
         FMouseInControl := FALSE;
         SetSolidBorder;
         Notify( self, STM_MOUSELEAVENOTIFY, 0);
      end;
   end;
end;

procedure TsCustomPanel.SetBounds(left, top, width, height: Integer);
begin
   inherited;
   Refresh;
end;

function TsCustomPanel.ChildControlFocused: Boolean;
var
   ii: Integer;
   h: HWND;
begin
   h := GetFocus;
   Result := h = Handle;
   if not Result then for ii := 0 to ControlCount -1 do
      if Controls[ii] is TWinControl and
         (TWinControl(controls[ii]).Handle = h) then begin
         Result := TRUE;
         Exit;
      end;
end;

procedure TsCustomPanel.CMEnabledChanged(var Message: TMessage);
var
   ii: Integer;
begin
   inherited;
   for ii := 0 to ControlCount -1 do
      Controls[ii].Enabled := Enabled;
   Notify( self, STM_ENABLEDCHANGENOTIFY, Integer(Enabled));
end;

procedure TsCustomPanel.SMMLinkedChanged(var Message: TMessage);
var
   ii: Integer;
   lb: TsCustomLabel;
begin
   lb := TsCustomLabel(Message.wParam);
   for ii := 0 to ControlCount - 1 do
      if Controls[ii] is TWinControl then begin
         if Message.lParam = 0 then
            lb.RemoveMonitorControl( TWinControl(Controls[ii]))
         else
            lb.AddMonitorControl(TWinControl(Controls[ii]));
      end;
end;

procedure TsCustomPanel.Paint;
var
   Rect: TRect;
begin
   inherited;
   Rect := GetClientRect;
   with Canvas do begin
      Brush.Color := Color;
      FillRect(Rect);
   end;
   RedrawBorders;
end;

procedure TsCustomPanel.Notification(AComponent: TComponent; Operation: TOperation);
begin
	inherited;
{
         if Message.lParam = 0 then
            RemoveMonitorControl( TWinControl(Controls[ii]))
         else
            AddMonitorControl(TWinControl(Controls[ii]));
 }
   if Operation = opInsert then begin
   end else begin
   end;
end;

procedure TsCustomPanel.SetSolidBorder;
var
   sb: Boolean;
begin
   sb := GetSolidBorder;
   if sb <> FSolidBorder then begin
      FSolidBorder := sb;
      if ([csLoading, csReading] * ComponentState = []) then
         RedrawBorders;
   end;
end;

function TsCustomPanel.GetSolidBorder: Boolean;
begin
   Result := ((csDesigning in ComponentState) and Enabled) or
      (not(csDesigning in ComponentState) and
      ((FMouseInControl and IsFlatApplied) or ChildControlFocused));
end;

procedure TsCustomPanel.RedrawBorders;
begin
   InternalRedrawBorder( Handle, FSolidBorder);
end;


initialization
   RegisterFlatControl(TsBWCCControl);
   RegisterFlatControl(TsCustomPanel);

finalization

end.
