{*******************************************************************************
   Unit
      sDBCombos.pas
   Description:
      Flat style DB combo box controls.
   Versions:
      '99
	sHistory:
      '99         August 1999
               .Final fixups
      '99 beta 2  July 1999
               .Published new properties, introdused by TsCustomComboBox:
                  BorderStyle and
                  DosKey
               .New property EnableAutoAddValues. When combo's style = csDropDownList
                  and the value in the field does not exists in the popup list, then,
                  if EnableAutoAddValues = TRUE, this value is added to the list.
                  Sometime it's usefull.
      '99 beta 1 	March of 1999
      			.Initial release
   Autor(s):
      Delphi VCL Extensions (RX).**
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      These components are part of sEditTools.

      **In fact this controls were taken from Delphi VCL Extensions (RX) with minimum
         modifications.
*******************************************************************************}
unit sDBCombos;

interface

uses SysUtils, Windows, Messages, Classes, Controls, Forms,
  Graphics, Menus, StdCtrls, ExtCtrls, DB, DBTables, DBCtrls, sCombos;

type
   TIdxDisplayMode = (dmFieldLabels, dmFieldNames, dmIndexName);

   { TDBIndexCombo }
   TsDBIndexComboBox = class(TsCustomComboBox)
   private
      FDataLink: TDataLink;
      FUpdate: Boolean;
      FNoIndexItem: PString;
      FEnableNoIndex: Boolean;
      FChanging: Boolean;
      FDisplayMode: TIdxDisplayMode;
      FExcludeIndexFieldNames: TStringList;
      FExcludeIndexNAmes: TStringList;
      function GetDataSource: TDataSource;
      procedure SetDataSource(Value: TDataSource);
      function GetIndexFieldName(var AName: string): Boolean;
      procedure SetNoIndexItem(const Value: string);
      function GetNoIndexItem: string;
      procedure SetEnableNoIndex(Value: Boolean);
      procedure SetDisplayMode(Value: TIdxDisplayMode);
      procedure SetExcludeIndexFieldNames( Value: TStringList);
      procedure SetExcludeIndexNames( Value: TStringList);
      procedure ActiveChanged;
      procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
   protected
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure FillIndexList(List: TStrings);
      function IncludedIndex( iDef: TIndexDef): Boolean;
      procedure Change; override;
      procedure UpdateList; virtual;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
   published
      property DataSource: TDataSource read GetDataSource write SetDataSource;
      property NoIndexItem: string read GetNoIndexItem write SetNoIndexItem;
      property EnableNoIndex: Boolean read FEnableNoIndex write SetEnableNoIndex default False;
      property BorderStyle;
      property Color;
      property Ctl3D;
      property DefaultHeight;
      property DisabledStyle;
      property DisabledFont;
      property DisplayMode: TIdxDisplayMode read FDisplayMode write SetDisplayMode default dmFieldLabels;
      property DosKey;
      property DragCursor;
      property DragMode;
      property DropDownCount;
      property Enabled;
      property ExcludeIndexFieldNames: TStringList read FExcludeIndexFieldNames write SetExcludeIndexFieldNames;
      property ExcludeIndexNAmes: TStringList read FExcludeIndexNAmes write SetExcludeIndexNAmes;
      property Flat;
      property Font;
      property ImeMode;
      property ImeName;
      property ItemHeight;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      {$IFDEF FlatSB}
      property ScrollBarMode;
      {$ENDIF FlatSB}
      property ShowHint;
      property Sorted;
      property TabOrder;
      property TabStop;
      property Visible;
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnStartDrag;
      {$IFDEF VER120}
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
      {$ENDIF}
   end;

   TsCustomDBComboBox = class(TsCustomComboBox)
   private
      FDataLink: TFieldDataLink;
      FPaintControl: TPaintControl;
      procedure DataChange(Sender: TObject);
      procedure EditingChange(Sender: TObject);
      function GetDataField: string;
      function GetDataSource: TDataSource;
      function GetField: TField;
      function GetReadOnly: Boolean;
      procedure SetDataField(const Value: string);
      procedure SetDataSource(Value: TDataSource);
      procedure SetEditReadOnly;
      procedure SetItems(Value: TStrings);
      procedure SetReadOnly(Value: Boolean);
      procedure UpdateData(Sender: TObject);
      function GetComboText: string; virtual;
      procedure SetComboText(const Value: string); virtual;
      procedure CMExit(var Message: TCMExit); message CM_EXIT;
      procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
      procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
   protected
      procedure Change; override;
      procedure Click; override;
      procedure ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer); override;
      procedure CreateWnd; override;
      procedure DropDown; override;
      function GetPaintText: string; virtual;
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
      procedure KeyPress(var Key: Char); override;
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure SetStyle(Value: TComboBoxStyle); override;
      procedure WndProc(var Message: TMessage); override;
      property ComboText: string read GetComboText write SetComboText;
      property DataField: string read GetDataField write SetDataField;
      property DataSource: TDataSource read GetDataSource write SetDataSource;
      property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      {$IFDEF VER120}
      function ExecuteAction(Action: TBasicAction): Boolean; override;
      function UpdateAction(Action: TBasicAction): Boolean; override;
      function UseRightToLeftAlignment: Boolean; override;
      {$ENDIF}
      property Field: TField read GetField;
      property Items write SetItems;
      property Text;
   end;

   TsDBComboBox = class(TsCustomDBComboBox)
   private
      FValues: TStrings;
      FEnableValues: Boolean;
      FEnableAutoAddValues: Boolean;
      procedure SetEnableValues(Value: Boolean);
      procedure SetEnableAutoAddValues(Value: Boolean);
      procedure SetValues(Value: TStrings);
      procedure ValuesChanged(Sender: TObject);
   protected
      procedure SetStyle(Value: TComboBoxStyle); override;
      function GetComboText: string; override;
      function GetPaintText: string; override;
      procedure SetComboText(const Value: string); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
   published
      property Style; { must be published before Items }
      property BorderStyle;
      property Color;
      property Ctl3D;
      property DataField;
      property DataSource;
      property DisabledStyle;
      property DisabledFont;
      property DefaultHeight;
      property DosKey;
      property DragMode;
      property DragCursor;
      property DropDownCount;
      property Enabled;
      property EnableValues: Boolean read FEnableValues write SetEnableValues;
      property EnableAutoAddValues: Boolean read FEnableAutoAddValues write SetEnableAutoAddValues;
      property Font;
      property Flat;
      {$IFDEF VER120}
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      {$ENDIF}
      property ImeMode;
      property ImeName;
      property ItemHeight;
      property Items;
      property ListBorderStyle;
      property ListColor;
      property ListCursor;
      property ListHilight;
      property ListPopupAnimation;
      property ListScrollAnimation;
      property ListWidth;
      property LookupStyle;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ReadOnly;
      {$IFDEF FlatSB}
      property ScrollBarMode;
      {$ENDIF FlatSB}
      property ShowHint;
      property Sorted;
      property TabOrder;
      property TabStop;
      property Values: TStrings read FValues write SetValues;
      property Visible;
      property OnChange;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnDrawItem;
      property OnDropDown;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMeasureItem;
      property OnStartDrag;
      {$IFDEF VER120}
      property OnEndDock;
      property OnStartDock;
      {$ENDIF}
  end;

implementation
{.$R *.DCR}

uses Bde ,DBConsts;

{ TKeyDataLink }
type
   TKeyDataLink = class(TDataLink)
   private
      FCombo: TsDBIndexComboBox;
   protected
      procedure ActiveChanged; override;
      procedure DataSetChanged; override;
      procedure DataSetScrolled(Distance: Integer); override;
   public
      constructor Create(ACombo: TsDBIndexComboBox);
      destructor Destroy; override;
   end;

constructor TKeyDataLink.Create(ACombo: TsDBIndexComboBox);
begin
   inherited Create;
   FCombo := ACombo;
end;

destructor TKeyDataLink.Destroy;
begin
   FCombo := nil;
   inherited Destroy;
end;

procedure TKeyDataLink.ActiveChanged;
begin
   if FCombo <> nil then
      FCombo.ActiveChanged;
end;

procedure TKeyDataLink.DataSetChanged;
begin
   if FCombo <> nil then
      FCombo.ActiveChanged;
end;

procedure TKeyDataLink.DataSetScrolled(Distance: Integer);
begin
  { ignore this data event }
end;

{ TsDBIndexComboBox }

constructor TsDBIndexComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FDataLink := TKeyDataLink.Create(Self);
   Style := csDropDownList;
   FUpdate := False;
   FNoIndexItem := NullStr;
   FEnableNoIndex := False;
   FExcludeIndexFieldNames := TStringList.Create;
   FExcludeIndexNames := TStringList.Create
end;

destructor TsDBIndexComboBox.Destroy;
begin
   FDataLink.Free;
   FDataLink := nil;
   DisposeStr(FNoIndexItem);
   FNoIndexItem := NullStr;
   FExcludeIndexFieldNames.Free;
   FExcludeIndexNames.Free;
   inherited Destroy;
end;

procedure TsDBIndexComboBox.SetNoIndexItem(const Value: string);
begin
   if Value <> FNoIndexItem^ then begin
      AssignStr(FNoIndexItem, Value);
      if not (csLoading in ComponentState) then
         ActiveChanged;
   end;
end;

procedure TsDBIndexComboBox.SetEnableNoIndex(Value: Boolean);
begin
   if FEnableNoIndex <> Value then begin
      FEnableNoIndex := Value;
      if not (csLoading in ComponentState) then
         ActiveChanged;
   end;
end;

procedure TsDBIndexComboBox.SetDisplayMode(Value: TIdxDisplayMode);
begin
   if (Value <> FDisplayMode) then begin
      FDisplayMode := Value;
      if not (csLoading in ComponentState) then
         UpdateList;
   end;
end;

procedure TsDBIndexComboBox.SetExcludeIndexFieldNames( Value: TStringList);
begin
   if not (csLoading in ComponentState) then
      FExcludeIndexNames.Clear;
   FExcludeIndexFieldNames.Assign(Value);
   if not (csLoading in ComponentState) then
      UpdateList;
end;

procedure TsDBIndexComboBox.SetExcludeIndexNames( Value: TStringList);
begin
   if not (csLoading in ComponentState) then
      FExcludeIndexFieldNames.Clear;
   FExcludeIndexNames.Assign(Value);
   if not (csLoading in ComponentState) then
      UpdateList;
end;

function TsDBIndexComboBox.GetNoIndexItem: string;
begin
   Result := FNoIndexItem^;
end;

function TsDBIndexComboBox.GetDataSource: TDataSource;
begin
   if FDataLink <> nil then
      Result := FDataLink.DataSource
   else
      Result := nil;
end;

procedure TsDBIndexComboBox.SetDataSource(Value: TDataSource);
begin
   FDataLink.DataSource := Value;
   if Value <> nil then
      Value.FreeNotification(Self);
   if not (csLoading in ComponentState) then
      ActiveChanged;
end;

procedure TsDBIndexComboBox.ActiveChanged;
begin
   if not (Enabled and FDataLink.Active and FDataLink.DataSet.InheritsFrom(TTable)) then begin
      Clear;
      ItemIndex := -1;
   end else
      UpdateList;
end;

procedure TsDBIndexComboBox.Loaded;
begin
   inherited Loaded;
   ActiveChanged;
end;

procedure TsDBIndexComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);
   if (Operation = opRemove) and (FDataLink <> nil) and (AComponent = DataSource) then
      DataSource := nil;
end;

procedure TsDBIndexComboBox.CMEnabledChanged(var Message: TMessage);
begin
   inherited;
   if not (csLoading in ComponentState) then
      ActiveChanged;
end;

function TsDBIndexComboBox.GetIndexFieldName(var AName: string): Boolean;
begin
   Result := True;
   if ItemIndex >= 0 then begin
      if EnableNoIndex and (Items[ItemIndex] = NoIndexItem) then
         AName := ''
      else begin
         AName := TIndexDef(Items.Objects[ItemIndex]).Fields;
         if AName = '' then begin
            AName := TIndexDef(Items.Objects[ItemIndex]).Name;
            Result := False;
         end;
      end;
   end else
      AName := '';
end;

function TsDBIndexComboBox.IncludedIndex( iDef: TIndexDef): Boolean;
var
   ii: Integer;
begin
   for ii := 0 to FExcludeIndexFieldNames.Count - 1 do begin
      Result := CompareText( iDef.Fields, FExcludeIndexFieldNames[ii]) <> 0;
      if not Result then
         Exit;
   end;
   for ii := 0 to FExcludeIndexNames.Count - 1 do begin
      Result := CompareText( iDef.Name, FExcludeIndexNames[ii]) <> 0;
      if not Result then
         Exit;
   end;
   Result := TRUE;
end;

procedure TsDBIndexComboBox.FillIndexList(List: TStrings);
var
   AFld: string;
   Pos: Integer;
   I: Integer;
begin
   List.Clear;
   if not FDataLink.Active then
      Exit;
   with FDataLink.DataSet as TTable do begin
      for I := 0 to IndexDefs.Count - 1 do with IndexDefs[I] do
         if IncludedIndex(IndexDefs[I]) and not (ixExpression in Options) then begin
            if FDisplayMode = dmIndexName then
               AFld := Name
            else begin
               AFld := '';
               Pos := 1;
               while Pos <= Length(Fields) do begin
                  if AFld <> '' then
                     AFld := AFld + '; ';
                  case FDisplayMode of
                     dmFieldLabels:
                        AFld := AFld + FieldByName(ExtractFieldName(Fields, Pos)).DisplayLabel;
                     dmFieldNames:
                        AFld := AFld + FieldByName(ExtractFieldName(Fields, Pos)).FieldName;
                  end;
               end;
            end;
            if List.IndexOf(AFld) < 0 then
               List.AddObject(AFld, IndexDefs[I]);
         end;
   end;
   if EnableNoIndex then
      if List.IndexOf(NoIndexItem) < 0 then
         List.AddObject(NoIndexItem, nil);
end;

procedure TsDBIndexComboBox.Change;
var
   ABookmark: TBookmark;
   AName: string;
begin
   if Enabled and FDataLink.Active and not FChanging and
      FDataLink.DataSet.InheritsFrom(TTable) and
      not (csLoading in ComponentState) then begin
      ABookmark := nil;
      with FDataLink.DataSet as TTable do begin
         if Database.IsSQLBased then
            ABookmark := GetBookmark;
         try
            if GetIndexFieldName(AName) then begin
               IndexFieldNames := AName;
               if (AName = '') and (IndexDefs.Count > 0) then
                  IndexName := '';
            end else begin
               if AName = '' then
                  IndexFieldNames := '';
               IndexName := AName;
            end;
            if (ABookmark <> nil) then begin
               with TTable(Self.FDataLink.DataSet) do
                  if Active and not (Bof and Eof) and BookmarkValid(ABookmark) then try
                     GotoBookmark(ABookmark);
                  except
                  end;
            end;
         finally
            if ABookmark <> nil then
               FreeBookmark(ABookmark);
         end;
      end;
   end;
   inherited Change;
end;

procedure TsDBIndexComboBox.UpdateList;
   function FindIndex(Table: TTable): Integer;
   var
      I: Integer;
      IdxFields: string;
   begin
      Result := -1;
      IdxFields := '';
      if Table.IndexFieldNames <> '' then
         for I := 0 to Table.IndexFieldCount - 1 do begin
            if IdxFields <> '' then
               IdxFields := IdxFields + ';';
            IdxFields := IdxFields + Table.IndexFields[I].FieldName;
         end;
      for I := 0 to Items.Count - 1 do begin
         if (Items.Objects[I] <> nil) and (((IdxFields <> '') and
            (AnsiCompareText(TIndexDef(Items.Objects[I]).Fields, IdxFields) = 0)) or
            ((Table.IndexName <> '') and
            (AnsiCompareText(TIndexDef(Items.Objects[I]).Name, Table.IndexName) = 0))) then begin
            Result := I;
            Exit;
         end;
      end;
      if EnableNoIndex and FDataLink.Active then
         if (Table.IndexFieldNames = '') and (Table.IndexName = '') then
            Result := Items.IndexOf(NoIndexItem);
   end;

begin
   if Enabled and FDataLink.Active then try
      Items.BeginUpdate;
      try
         if FDataLink.DataSet.InheritsFrom(TTable) then begin
            TTable(FDataLink.DataSet).IndexDefs.Update;
            FillIndexList(Items);
            ItemIndex := FindIndex(TTable(FDataLink.DataSet));
            FChanging := True;
         end else
            Items.Clear;
      finally
         Items.EndUpdate;
      end;
   finally
      FChanging := False;
   end;
end;

{TsDBComboBox}
constructor TsCustomDBComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   ControlStyle := ControlStyle + [csReplicatable];
   FDataLink := TFieldDataLink.Create;
   FDataLink.Control := Self;
   FDataLink.OnDataChange := DataChange;
   FDataLink.OnUpdateData := UpdateData;
   FDataLink.OnEditingChange := EditingChange;
   FPaintControl := TPaintControl.Create(Self, 'COMBOBOX');
end;

destructor TsCustomDBComboBox.Destroy;
begin
   FPaintControl.Free;
   FDataLink.OnDataChange := nil;
   FDataLink.OnUpdateData := nil;
   FDataLink.Free;
   FDataLink := nil;
   inherited Destroy;
end;

procedure TsCustomDBComboBox.Loaded;
begin
   inherited Loaded;
   if (csDesigning in ComponentState) then
      DataChange(Self);
end;

procedure TsCustomDBComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);
   if (Operation = opRemove) and (FDataLink <> nil) and
      (AComponent = DataSource) then
      DataSource := nil;
end;

procedure TsCustomDBComboBox.CreateWnd;
begin
   inherited CreateWnd;
   SetEditReadOnly;
end;

procedure TsCustomDBComboBox.DataChange(Sender: TObject);
begin
   if DroppedDown then
      Exit;
   if FDataLink.Field <> nil then
      ComboText := FDataLink.Field.Text
   else if csDesigning in ComponentState then
      ComboText := Name
   else
      ComboText := '';
end;

procedure TsCustomDBComboBox.UpdateData(Sender: TObject);
begin
   FDataLink.Field.Text := ComboText;
end;

procedure TsCustomDBComboBox.SetComboText(const Value: string);
var
   I: Integer;
   Redraw: Boolean;
begin
   if Value <> ComboText then begin
      if Style <> csDropDown then begin
         Redraw := (Style <> csSimple) and HandleAllocated;
         if Redraw then
            SendMessage(Handle, WM_SETREDRAW, 0, 0);
         try
            if Value = '' then
               I := -1
            else
               I := Items.IndexOf(Value);
            ItemIndex := I;
         finally
            if Redraw then begin
               SendMessage(Handle, WM_SETREDRAW, 1, 0);
               Invalidate;
            end;
         end;
         if I >= 0 then
            Exit;
      end;
      if Style in [csDropDown, csSimple] then
         Text := Value;
   end;
end;

function TsCustomDBComboBox.GetComboText: string;
var
   I: Integer;
begin
   if Style in [csDropDown, csSimple] then
      Result := Text
   else begin
      I := ItemIndex;
      if I < 0 then
         Result := ''
      else
         Result := Items[I];
   end;
end;

procedure TsCustomDBComboBox.Change;
begin
   FDataLink.Edit;
   inherited Change;
   FDataLink.Modified;
end;

procedure TsCustomDBComboBox.Click;
begin
   FDataLink.Edit;
   inherited Click;
   FDataLink.Modified;
end;

procedure TsCustomDBComboBox.DropDown;
begin
   inherited DropDown;
end;

function TsCustomDBComboBox.GetDataSource: TDataSource;
begin
   Result := FDataLink.DataSource;
end;

procedure TsCustomDBComboBox.SetDataSource(Value: TDataSource);
begin
   {$IFDEF VER123}
   if not (FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
   {$ENDIF}
      FDataLink.DataSource := Value;
   if Value <> nil then
      Value.FreeNotification(Self);
end;

function TsCustomDBComboBox.GetDataField: string;
begin
   Result := FDataLink.FieldName;
end;

procedure TsCustomDBComboBox.SetDataField(const Value: string);
begin
   FDataLink.FieldName := Value;
end;

function TsCustomDBComboBox.GetReadOnly: Boolean;
begin
   Result := FDataLink.ReadOnly;
end;

procedure TsCustomDBComboBox.SetReadOnly(Value: Boolean);
begin
   FDataLink.ReadOnly := Value;
end;

function TsCustomDBComboBox.GetField: TField;
begin
   Result := FDataLink.Field;
end;

procedure TsCustomDBComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
   inherited KeyDown(Key, Shift);
   if Key in [VK_BACK, VK_DELETE, VK_UP, VK_DOWN, 32..255] then begin
      if not FDataLink.Edit and (Key in [VK_UP, VK_DOWN]) then
         Key := 0;
   end;
end;

procedure TsCustomDBComboBox.KeyPress(var Key: Char);
begin
   inherited KeyPress(Key);
   if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
      not FDataLink.Field.IsValidChar(Key) then begin
      MessageBeep(0);
      Key := #0;
   end;
   case Key of
      ^H, ^V, ^X, #32..#255:
         FDataLink.Edit;
      #27: begin
         FDataLink.Reset;
         SelectAll;
      end;
   end;
end;

procedure TsCustomDBComboBox.EditingChange(Sender: TObject);
begin
   SetEditReadOnly;
end;

procedure TsCustomDBComboBox.SetEditReadOnly;
begin
   if (Style in [csDropDown, csSimple]) and HandleAllocated then
      SendMessage( EditHandle, EM_SETREADONLY, Ord(not FDataLink.Editing), 0);
end;

procedure TsCustomDBComboBox.WndProc(var Message: TMessage);
begin
   if not (csDesigning in ComponentState) then case Message.Msg of
      WM_COMMAND: if TWMCommand(Message).NotifyCode = CBN_SELCHANGE then
         if not FDataLink.Edit then begin
            if Style <> csSimple then
               PostMessage(Handle, CB_SHOWDROPDOWN, 0, 0);
            Exit;
         end;
      CB_SHOWDROPDOWN:
         if Message.WParam <> 0 then
            FDataLink.Edit
         else if not FDataLink.Editing then
            DataChange(Self); {Restore text}
      WM_CREATE,
      WM_WINDOWPOSCHANGED,
      CM_FONTCHANGED:
         FPaintControl.DestroyHandle;
   end;
   inherited WndProc(Message);
end;

procedure TsCustomDBComboBox.ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer);
begin
   if not (csDesigning in ComponentState) then case Message.Msg of
      WM_LBUTTONDOWN:
        if (Style = csSimple) and (ComboWnd <> EditHandle) then
            if not FDataLink.Edit then
               Exit;
   end;
   inherited ComboWndProc(Message, ComboWnd, ComboProc);
end;

procedure TsCustomDBComboBox.CMExit(var Message: TCMExit);
begin
   try
      FDataLink.UpdateRecord;
   except
      SelectAll;
      if CanFocus then
         SetFocus;
      raise;
   end;
   inherited;
end;

procedure TsCustomDBComboBox.CMGetDatalink(var Message: TMessage);
begin
   Message.Result := Longint(FDataLink);
end;

procedure TsCustomDBComboBox.WMPaint(var Message: TWMPaint);
var
   S: string;
   R: TRect;
   P: TPoint;
   Child: HWND;
begin
   if csPaintCopy in ControlState then begin
      S := GetPaintText;
      if Style = csDropDown then begin
         SendMessage(FPaintControl.Handle, WM_SETTEXT, 0, Longint(PChar(S)));
         SendMessage(FPaintControl.Handle, WM_PAINT, Message.DC, 0);
         Child := GetWindow(FPaintControl.Handle, GW_CHILD);
         if Child <> 0 then begin
            Windows.GetClientRect(Child, R);
            Windows.MapWindowPoints(Child, FPaintControl.Handle, R.TopLeft, 2);
            GetWindowOrgEx(Message.DC, P);
            SetWindowOrgEx(Message.DC, P.X - R.Left, P.Y - R.Top, nil);
            IntersectClipRect(Message.DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
            SendMessage(Child, WM_PAINT, Message.DC, 0);
         end;
      end else begin
         SendMessage(FPaintControl.Handle, CB_RESETCONTENT, 0, 0);
         if Items.IndexOf(S) <> -1 then begin
            SendMessage(FPaintControl.Handle, CB_ADDSTRING, 0, Longint(PChar(S)));
            SendMessage(FPaintControl.Handle, CB_SETCURSEL, 0, 0);
         end;
         SendMessage(FPaintControl.Handle, WM_PAINT, Message.DC, 0);
      end;
   end else
      inherited;
end;

function TsCustomDBComboBox.GetPaintText: string;
begin
   if FDataLink.Field <> nil then
      Result := FDataLink.Field.Text
   else
      Result := '';
end;

procedure TsCustomDBComboBox.SetItems(Value: TStrings);
begin
   Items.Assign(Value);
   DataChange(Self);
end;

procedure TsCustomDBComboBox.SetStyle(Value: TComboBoxStyle);
begin
   if (Value = csSimple) and Assigned(FDatalink) and FDatalink.DatasourceFixed then
      DatabaseError(SNotReplicatable);
   inherited SetStyle(Value);
end;

{$IFDEF VER120}
function TsCustomDBComboBox.UseRightToLeftAlignment: Boolean;
begin
   Result := DBUseRightToLeftAlignment(Self, Field);
end;

function TsCustomDBComboBox.ExecuteAction(Action: TBasicAction): Boolean;
begin
   Result := inherited ExecuteAction(Action) or (FDataLink <> nil) and
      FDataLink.ExecuteAction(Action);
end;

function TsCustomDBComboBox.UpdateAction(Action: TBasicAction): Boolean;
begin
   Result := inherited UpdateAction(Action) or (FDataLink <> nil) and
      FDataLink.UpdateAction(Action);
end;
{$ENDIF}

{ TsDBComboBox }
constructor TsDBComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FValues := TStringList.Create;
   TStringList(FValues).OnChange := ValuesChanged;
   EnableValues := False;
end;

destructor TsDBComboBox.Destroy;
begin
   TStringList(FValues).OnChange := nil;
   FValues.Free;
   inherited Destroy;
end;

procedure TsDBComboBox.ValuesChanged(Sender: TObject);
begin
   if FEnableValues then
      DataChange(Self);
end;

function TsDBComboBox.GetPaintText: string;
var
   I: Integer;
begin
   Result := '';
   if FDataLink.Field <> nil then begin
      if FEnableValues then begin
         I := Values.IndexOf(FDataLink.Field.Text);
         if I >= 0 then
            Result := Items.Strings[I]
      end else
         Result := FDataLink.Field.Text;
   end;
end;

function TsDBComboBox.GetComboText: string;
var
   I: Integer;
begin
   if (Style in [csDropDown, csSimple]) and (not FEnableValues) then
      Result := Text
   else begin
      I := ItemIndex;
      if (I < 0) or (FEnableValues and (FValues.Count < I + 1)) then
         Result := ''
      else if FEnableValues then
         Result := FValues[I]
      else
         Result := Items[I];
   end;
end;

procedure TsDBComboBox.SetComboText(const Value: string);
var
   I: Integer;
   Redraw: Boolean;
begin
   if Value <> ComboText then begin
      if Style <> csDropDown then begin
         Redraw := (Style <> csSimple) and HandleAllocated;
         if Redraw then
            SendMessage(Handle, WM_SETREDRAW, 0, 0);
         try
            if Value = '' then
               I := -1
            else if FEnableValues then
               I := Values.IndexOf(Value)
            else begin
               I := Items.IndexOf(Value);
               if FEnableAutoAddValues and (I = -1) then begin
                  Items.Add(Value);
                  I := Items.IndexOf(Value);
               end;
            end;
            if I >= Items.Count then
               I := -1;
            ItemIndex := I;
         finally
            if Redraw then begin
               SendMessage(Handle, WM_SETREDRAW, 1, 0);
               Invalidate;
            end;
         end;
         if I >= 0 then
            Exit;
      end;
      if Style in [csDropDown, csSimple] then
         Text := Value;
   end;
end;

procedure TsDBComboBox.SetEnableAutoAddValues(Value: Boolean);
begin
   if FEnableAutoAddValues <> Value then begin
      if Value then
         FEnableValues := FALSE;
      FEnableAutoAddValues := Value;
      DataChange(Self);
   end;
end;

procedure TsDBComboBox.SetEnableValues(Value: Boolean);
begin
   if FEnableValues <> Value then begin
      if Value and (Style in [csDropDown, csSimple]) then
         Style := csDropDownList;
      if Value then
         FEnableAutoAddValues := FALSE;
      FEnableValues := Value;
      DataChange(Self);
   end;
end;

procedure TsDBComboBox.SetValues(Value: TStrings);
begin
   FValues.Assign(Value);
end;

procedure TsDBComboBox.SetStyle(Value: TComboboxStyle);
begin
   if (Value in [csSimple, csDropDown]) and FEnableValues then
      Value := csDropDownList;
   inherited SetStyle(Value);
end;

end.
