{*******************************************************************************
   Unit
      sDBLookup.pas
   Description:
      Lookup combo box
   Versions:
   	'99
	History:
      '99      - Aug. 1999
                  .painting fix of combo. Had a problem with several combos with a same
                     list's dataSource, list's data/key fields and dataSource but different data fields.
      '99 beta 2  - June-July 1999
                  .Small fix on repaint when changing the style of lookupcombo
                  .Changed disabled style drawing policy
                  .When Style = csDropDownList click on edit area drops down the
                     list (like in normal combo box)
                  .Fixed bugs:
                     +Handling of the animation scroll (sometime it shows duplicate items
                        at the nutton of the list)
                     +Fix scroll bar handling
                     +Scroll bar size (see Thomas letter)
                     +CMDialogChar conflict with default or cancel buttons
                  .Options property. Right now has only 1 value in the set:
                     -soDisplayInvalidKeyValues. This mean that if Combo could not find value
                     in the List's DataSet then it displays Key field value.
                     Idea came from Thomas van Stetten and I agree with him - sometime it
                     is usefull.
                  .OnChange event
      '99 beta 1 - April-May 1999
                  .Rewrite of original combo (see note below)
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      This control is a part of sEditTools.

      Long time ago I rewrote the Borland's lookup controls to make lookup combo
      like the combo box with csDropDown style (with an edit box where user can
      type the text he is looking for). This version is a partial rewrite (mainly
      concerning the look and feel of the controls) to meet the new standarts I
      developed for myself in sEditTools package.
*******************************************************************************}
unit sDBLookup;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, sCommonTypes, Dialogs,
  DB, DBTables, DBCtrls, ImgList, StdCtrls, sGraphics, sDefaults, Forms;

type
	TsDBSearchComboBox = class;

   TComboInPlaceEdit = class(TCustomEdit)
  	private
   	FCombo: TsDBSearchComboBox;
      FKey: Word;
      FTextLog: String;
      procedure SetText( s: String);
  	protected
   	procedure DblClick; override;
    	procedure Click; override;
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    	procedure KeyPress(var Key: Char); override;
    	procedure KeyUp(var Key: Word; Shift: TShiftState); override;
      procedure RestoreValue;
    	procedure UpdateContents; virtual;
    	procedure WndProc(var Message: TMessage); override;
  	public
    	constructor Create(AOwner: TComponent); override;
    	procedure Deselect;
    	procedure FocusEdit;
    	procedure UnfocusEdit;
    	procedure DoInvalidate;
    	property  Visible;
  end;

	{ TsDBLookupControl }
	TsDBLookupControl = class;

  	TsDataSourceLink = class(TDataLink)
  	private
    	FDBLookupControl: TsDBLookupControl;
      FLock: Boolean;
  	protected
    	procedure ActiveChanged; override;
    	procedure RecordChanged(Field: TField); override;
  	end;

  	TsListSourceLink = class(TDataLink)
  	private
    	FDBLookupControl: TsDBLookupControl;
      Fsteps: Integer;
  	protected
    	procedure ActiveChanged; override;
    	procedure DataSetChanged; override;
      procedure DataSetScrolled(Distance: Integer); override;
      property Steps: Integer read FSteps;
  	end;

   TsLookupState = (lsLookupMode, lsListActive, lsFocused, lsPopup, lsTracking,
      lsTimerActive, lsLockPosition, lsEditMode);
   TsLookupStates = set of TsLookupState;

  	TsDBLookupControl = class(TCustomControl)
  	private
      FCaseSensitive: Boolean;
      FDataField: TField;
      FDataFieldName: string;
    	FDataLink: TsDataSourceLink;
      FHilightIndex: Integer;
    	FKeyField: TField;
    	FKeyFieldName: string;
      FKeyValue: Variant;
    	FListField: TField;
      FListFieldIndex: Integer;
    	FListFieldName: string;
      FImageField: TField;
      FImageFieldName: String;
      FListFields: TList;
    	FListLink: TsListSourceLink;
    	FLookupSource: TDataSource;
    	FMasterField: TField;
    	FSearchText: string;
      FState: TsLookupStates;

    	function CanModify: Boolean;
      procedure CheckNotCircular;
    	procedure CheckNotLookup;
    	procedure DataLinkActiveChanged;
    	procedure DataLinkRecordChanged(Field: TField); dynamic;
    	function GetBorderSize: Integer; dynamic;
    	function GetDataSource: TDataSource;
    	function GetKeyFieldName: string;
    	function GetListSource: TDataSource;
    	function GetReadOnly: Boolean;
    	procedure KeyValueChanged; virtual; abstract;
    	procedure ListLinkActiveChanged; virtual;
    	procedure ListLinkDataChanged; virtual;
    	function LocateKey: Boolean;
    	procedure ProcessSearchKey(Key: Char);
      function  ProcessSearch(Str: String): Boolean;
    	procedure SetDataFieldName(const Value: string);
    	procedure SetDataSource(Value: TDataSource);
    	procedure SetKeyFieldName(const Value: string);
    	procedure SetListFieldName(const Value: string);
    	procedure SetListSource(Value: TDataSource);
      procedure SetImageFieldName(const Value: String);
    	procedure SetLookupMode(Value: Boolean);
    	procedure SetReadOnly(Value: Boolean);
    	procedure WMGetDlgCode(var Message: TMessage); message WM_GETDLGCODE;
    	procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
    	procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
  protected
      procedure SetKeyValue(const Value: Variant); virtual;
      procedure SelectKeyValue(Value: Variant); virtual;
      function DoLocate(S: string; Options: TLocateOptions): Boolean; virtual;
    	procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    	property DataField: string read FDataFieldName write SetDataFieldName;
    	property DataSource: TDataSource read GetDataSource write SetDataSource;
    	property KeyField: string read GetKeyFieldName write SetKeyFieldName;
    	property KeyValue: Variant read FKeyValue write SetKeyValue;
    	property ListField: string read FListFieldName write SetListFieldName;
    	property ListFieldIndex: Integer read FListFieldIndex write FListFieldIndex default 0;
      property ListSource: TDataSource read GetListSource write SetListSource;
      property ImageField: String read FImageFieldName write SetImageFieldName;
    	property ParentColor default False;
    	property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    	property TabStop default True;
      property SearchText: String read FSearchText write FSearchText;
      property CaseSensitive: Boolean read FCaseSensitive write FCaseSensitive;
   public
    	constructor Create(AOwner: TComponent); override;
    	destructor Destroy; override;
   end;

   TGetImageIdProc = procedure(Sender: TObject; DataSet: TDataSet; var imageIndex: Integer) of object;

   TsDBLookupListBox = class(TsDBLookupControl)
  	private
      FBorderStyle: TBorderStyle;
      FHilight: TsHilight;
    	FKeySelected: Boolean;
    	FMousePos: Integer;
    	FRecordIndex: Integer;
    	FRecordCount: Integer;
    	FRowCount: Integer;
      FImageChangeLink: TChangeLink;
      FImageOffset: Integer;
      FImages: TCustomImageList;
      FItemHeight: Integer;
      FListScrollAnimation: TsScrollAnimationStyle;
      FTextWidth: Integer;
      FScrolling: Boolean;
      FOnGetImageId: TGetImageIdProc;
      FOnImageChange: TNotifyEvent;

      procedure SetHilight(Value: TsHilight);
      procedure MoveBy(delta: Integer; animated: Boolean = TRUE);
      procedure ScrollBy(delta: Integer; animated: Boolean = TRUE);
      procedure ScrollFirst(animated: Boolean = TRUE);
      procedure ScrollLast(animated: Boolean = TRUE);

    	function GetKeyIndex: Integer;
    	procedure KeyValueChanged; override;
    	procedure ListLinkActiveChanged; override;
    	procedure ListLinkDataChanged; override;
    	procedure SelectCurrent; dynamic;
    	procedure SelectItemAt(X, Y: Integer);
    	procedure SetBorderStyle(Value: TBorderStyle);
    	procedure SetRowCount(Value: Integer);

      function GetImageId: Integer;
      procedure ImageListChange(Sender: TObject);
      procedure SetImages(const Value: TCustomImageList);

      procedure StopTimer;
    	procedure StopTracking;
    	procedure TimerScroll;
    	procedure UpdateScrollBar(forceUpdate: Boolean = FALSE);
      procedure PaintItem(Rect: TRect; Index: Integer; Selected: Boolean);
      procedure PaintItemTo(DC: HDC; Rect: TRect; Index: Integer; Selected: Boolean);
    	procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    	procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    	procedure WMCancelMode(var Message: TMessage); message WM_CANCELMODE;
    	procedure WMTimer(var Message: TMessage); message WM_TIMER;
    	procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
  protected
    	procedure CreateParams(var Params: TCreateParams); override;
    	procedure CreateWnd; override;
      procedure Loaded; override;
      function DoLocate(S: string; Options: TLocateOptions): Boolean; override;
    	procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    	procedure KeyPress(var Key: Char); override;
    	procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    	procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    	procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    	procedure Paint; override;
      property OnImageChange: TNotifyEvent read FOnImageChange write FOnImageChange;
  public
    	constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
    	procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    	property KeyValue;
  published
    	property Align;
    	property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
      property CaseSensitive;
    	property Color;
    	property Ctl3D;
    	property DataField;
    	property DataSource;
    	property DragCursor;
    	property DragMode;
    	property Enabled;
    	property Font;
      property Hilight: TsHilight read FHilight write SetHilight;
    	property KeyField;
    	property ListField;
    	property ListFieldIndex;
    	property ListSource;
    	property ParentColor;
    	property ParentCtl3D;
    	property ParentFont;
    	property ParentShowHint;
    	property PopupMenu;
    	property ReadOnly;
    	property RowCount: Integer read FRowCount write SetRowCount stored False;
    	property ShowHint;
    	property TabOrder;
    	property TabStop;
    	property Visible;
      property Images: TCustomImageList read FImages write SetImages;
      property ImageField;
    	property OnClick;
    	property OnDblClick;
    	property OnDragDrop;
    	property OnDragOver;
    	property OnEndDrag;
    	property OnEnter;
    	property OnExit;
      property OnGetImageId: TGetImageIdProc read FOnGetImageId write FOnGetImageId;
    	property OnKeyDown;
    	property OnKeyPress;
    	property OnKeyUp;
    	property OnMouseDown;
    	property OnMouseMove;
    	property OnMouseUp;
    	property OnStartDrag;
   end;

 	TsPopupDataList = class(TSDBLookupListBox)
  	private
   	FCombo: TsDBSearchComboBox;
      FBorderStyle: TsListBorderStyle;
      procedure SetBorderStyle(Value: TsListBorderStyle);
      procedure AniNCPaint(tDC: HDC; var R: TRect);
    	procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
      procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
      procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
  	protected
      procedure SelectCurrent; override;
    	procedure CreateParams(var Params: TCreateParams); override;
      function GetBorderSize: Integer; override;
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  	public
    	constructor Create(AOwner: TComponent); override;
      property BorderStyle: TsListBorderStyle read FBorderStyle write SetBorderStyle;
  	end;

   TdbComboBoxStyle = csDropDown..csDropDownList;
   TsDBSearchOptions = set of (soDisplayInvalidKeyValues);
   TsDBDataState = set of (ssInvalidKey);

  	TsDBSearchComboBox = class(TsDBLookupControl)
  	private
      FAlignment: TAlignment;
      FBorderStyle: TsBorderStyle;
      FButtonWidth: Integer;
      FChanging: Boolean;
      FDataList: TSPopupDataList;
      FDisabledFont: TFont;
      FDisabledStyle: TsDisabledStyle;
      FDosKey: Boolean;
    	FDropDownRows: Integer;
    	FDropDownWidth: Integer;
    	FDropDownAlign: TDropDownAlign;
      FEdit: TComboInPlaceEdit;
      FEmptyString: String;
      FFlat: Boolean;
      FListVisible: Boolean;
      FListPopupAnimation: TsListAnimationType;
      FLookupStyle: TsLookupStyle;

      FMouseInControl: Boolean;
      FOptions: TsDBSearchOptions;
      FPressed: Boolean;
      FDataState: TsDBDataState;
      FStyle: TdbComboBoxStyle;
    	FText: string;

    	FOnDropDown: TNotifyEvent;
    	FOnCloseUp: TNotifyEvent;
      FOnGetImageId: TGetImageIdProc;
      FOnChange: TNotifyEvent;

      function GetSolidBorder: boolean;
      function GetImages: TCustomImageList;
      function GetHilight: TsHilight;

      procedure SetImages(Value: TCustomImageList);
      procedure SetFlat(const Value: Boolean);
      procedure SetHilight(Value: TsHilight);
      procedure SetButtonWidth;
      procedure SetDisabledFont(Value: TFont);
      procedure SetBorderStyle(const Value: TsBorderStyle);

      procedure SetEmptyString(Value: String);
      function GetListColor: TColor;
      procedure SetListColor(Value: TColor);
      function GetListBorderStyle: TsListBorderStyle;
      procedure SetListBorderStyle(Value: TsListBorderStyle);
      function GetListCursor: TCursor;
      procedure SetListCursor(Value: TCursor);
      procedure SetStyle(Value: TdbComboBoxStyle);
      procedure SetDisabledStyle(const Value: TsDisabledStyle);
      procedure SetListScrollAnimation(const Value: TsScrollAnimationStyle);
      function GetListScrollAnimation: TsScrollAnimationStyle;
      procedure SetCtl3D(const Value: Boolean);
      function GetCtl3D: Boolean;


    	procedure ListMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      procedure OnImageChange(Sender: TObject);
      procedure EditChange(Sender: TObject);
    	procedure StopTracking;
    	procedure TrackButton(X, Y: Integer);
      function GetEditRect: TRect;
      function GetImageRect: TRect;
      procedure SetEditRect;
      function GetImageId: Integer;
      function GetButtonRect: TRect;
      function GetEmpty: Boolean;
      procedure AniShow( Handle: HWND);
    	procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
      procedure CMEnabledChanged (var Msg: TMessage); message CM_ENABLEDCHANGED;
    	procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    	procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
    	procedure WMCancelMode(var Message: TMessage); message WM_CANCELMODE;
      procedure CNKeyDown(var Message: TWMKeyDown); message CN_KEYDOWN;
    	procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
      procedure WMKillFocus(var Message: TMessage);   message WM_KILLFOCUS; 
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
      procedure WMNCPaint(var Message: TWMPaint); message WM_NCPAINT;
  protected
    	procedure CreateParams(var Params: TCreateParams); override;
      procedure CreateWnd; override;
      procedure SelectKeyValue(Value: Variant); override;
      procedure SetKeyValue(const Value: Variant); override;
      function GetBorderSize: Integer; override;
      procedure Loaded; override;
    	procedure KeyValueChanged; override;
    	procedure ListLinkActiveChanged; override;

      procedure RedrawBorders;
      procedure DrawDisabled(DC: HDC);
      procedure PaintButton;
      procedure PaintButtonBorder;

      {$IFDEF V1_COMP}      {Back compatibility}
      procedure ReadDefaultDraw(Reader: TReader);
      procedure WriteNothing(Writer: TWriter);
      procedure DefineProperties(Filer: TFiler); override;
      {$ENDIF}

      procedure DoExit; override;
      procedure Click; override;
      procedure DataLinkRecordChanged(Field: TField); override;
    	procedure Paint; override;
    	procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    	procedure KeyPress(var Key: Char); override;
    	procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    	procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    	procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      property SolidBorder: Boolean read getSolidBorder;
  public
    	constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
    	procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    	procedure CloseUp(Accept: Boolean);
    	procedure DropDown;
    	property KeyValue;
    	property ListVisible: Boolean read FListVisible;
    	property Text: string read FText;
      property Empty: Boolean read GetEmpty;
      property DataState: TsDBDataState read FDataState;
  published
      property Flat: Boolean read FFlat write SetFlat stored FALSE;
      property BorderStyle: TsBorderStyle read FBorderStyle write SetBorderStyle default DEF_Combo_BorderStyle;
    	property Color;
    	property Ctl3D: Boolean read GetCtl3D write SetCtl3D stored FALSE;
      property CaseSensitive;
    	property DataField;
    	property DataSource;
      property DisabledFont: TFont read FDisabledFont write SetDisabledFont;
      property DisabledStyle: TsDisabledStyle read FdisabledStyle write SetDisabledStyle default dsDefault;
      property DosKey: Boolean read FDosKey write FDosKey default DEF_Combo_DosKey;
    	property DragCursor;
    	property DragMode;
    	property DropDownAlign: TDropDownAlign read FDropDownAlign write FDropDownAlign default daLeft;
    	property DropDownRows: Integer read FDropDownRows write FDropDownRows default 7;
      property EmptyString: String read FEmptyString write SetEmptyString;
      property Font;
      property KeyField;
      property ListHilight: TsHilight read GetHilight write SetHilight;
      property ListColor: TColor read GetListColor write SetListColor default DEF_Combo_ListColor;
      property ListCursor: TCursor read GetListCursor write SetListCursor default DEF_Combo_ListCursor;
      property ListBorderStyle: TsListBorderStyle read GetListBorderStyle write SetListBorderStyle;// default DEF_Combo_ListBorderStyle;
    	property ListField;
    	property ListFieldIndex;
      property ListPopupAnimation: TsListAnimationType read FListPopupAnimation write FListPopupAnimation default DEF_Combo_ListPopupAnimation;
      property ListScrollAnimation: TsScrollAnimationStyle read GetListScrollAnimation write SetListScrollAnimation default DEF_Combo_ListScrollAnimation;
      property ListWidth: Integer read FDropDownWidth write FDropDownWidth default 0;
    	property ListSource;
      property LookupStyle: TsLookupStyle read FLookupStyle write FLookupStyle default DEF_DBSearchCombo_LookupStyle;
      property ImageField;
      property Options: TsDBSearchOptions read FOptions write FOptions;
    	property ParentColor;
    	property ParentCtl3D;
    	property ParentFont;
    	property ParentShowHint;
    	property PopupMenu;
    	property ReadOnly;
    	property ShowHint;
      property Style: TdbComboBoxStyle read FStyle write SetStyle default csDropDown;
    	property TabOrder;
    	property TabStop;
    	property Visible;
      property Images: TCustomImageList read GetImages write SetImages;
      property Enabled;
    	property OnClick;
    	property OnCloseUp: TNotifyEvent read FOnCloseUp write FOnCloseUp;
    	property OnDragDrop;
    	property OnDragOver;
    	property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;
    	property OnEndDrag;
    	property OnEnter;
    	property OnExit;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
      property OnGetImageId: TGetImageIdProc read FOnGetImageId write FOnGetImageId;
    	property OnKeyDown;
    	property OnKeyPress;
    	property OnKeyUp;
    	property OnMouseDown;
    	property OnMouseMove;
    	property OnMouseUp;
    	property OnStartDrag;
   end;

implementation
{.$R *.DCR}

uses DBConsts, sFlat, StdUtils, sConsts, Commctrl;

constructor TComboInPlaceEdit.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FCombo := AOwner as TsDBSearchComboBox;
  	ParentCtl3D := False;
  	Ctl3D := False;
  	TabStop := False;
  	BorderStyle := bsNone;
end;

procedure TComboInPlaceEdit.DblClick;
begin
   FCombo.DblClick;
end;

procedure TComboInPlaceEdit.Click;
begin
  	FCombo.Click;
end;

procedure TComboInPlaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
  	procedure SendToParent;
  	begin
    	FCombo.KeyDown(Key, Shift);
    	Key := 0;
  	end;
  	procedure ParentEvent;
  	var
    	ComboKeyDown: TKeyEvent;
  	begin
    	ComboKeyDown := FCombo.OnKeyDown;
    	if Assigned(ComboKeyDown) then
         ComboKeyDown(FCombo, Key, Shift);
  	end;

begin
   FKey := Key;
  	case Key of
   	VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT:
         SendToParent;
		VK_ESCAPE:
         Text := '';
  	end;
  	if Key <> 0 then begin
   	FTextLog := Text;
      Include( FCombo.FState, lsEditMode);
    	ParentEvent;
    	inherited KeyDown(Key, Shift);
  	end;
end;

procedure TComboInPlaceEdit.KeyPress(var Key: Char);
begin
  	FCombo.KeyPress(Key);
   if Key = #13 then key := #0;
  	if Key <> #0 then begin
      SelLength := Length(Text) - SelStart;
      inherited KeyPress(Key);
   end;
end;

procedure TComboInPlaceEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
   FCombo.KeyUp(Key, Shift);
end;

procedure TComboInPlaceEdit.WndProc(var Message: TMessage);
begin
  	if Message.Msg = WM_SETFOCUS then begin
   	if GetParentForm(Self).SetFocusedControl(FCombo) then
         Dispatch(Message);
   end else if Message.Msg = WM_KILLFOCUS then begin
      FCombo.WMKillFocus(message);
      inherited WndProc(Message);
   end else
      inherited WndProc(Message);
end;

procedure TComboInPlaceEdit.Deselect;
begin
  	SendMessage(Handle, EM_SETSEL, $7FFFFFFF, Longint($FFFFFFFF));
end;

procedure TComboInPlaceEdit.DoInvalidate;
begin
  	ValidateRect(Handle, nil);
  	InvalidateRect(Handle, nil, True);
end;

procedure TComboInPlaceEdit.UnFocusEdit;
begin
   if HandleAllocated and IsWindowVisible(Handle) then
    	if Focused then
         Windows.SetFocus(FCombo.Handle);
end;

procedure TComboInPlaceEdit.FocusEdit;
begin
   if FCombo.Focused then begin
   	Windows.SetFocus(Handle);
      SelectAll;
   end;
end;

procedure TComboInPlaceEdit.SetText( s: String);
var
	pos: Integer;
   lastOnChange: TNotifyEvent;
begin
	pos := SelStart;
	lastOnChange := OnChange;
   OnChange := nil;
   Text := S;
   SelStart := pos;
   OnChange := lastOnChange;
end;

procedure TComboInPlaceEdit.UpdateContents;
begin
	if FCombo.Text <> Text then
      SetText(FCombo.Text);
   SelectAll;
end;

procedure TComboInPlaceEdit.RestoreValue;
begin
   SetText( FTextLog);
   Beep;
end;

{ TSDataSourceLink }
procedure TsDataSourceLink.ActiveChanged;
begin
   if FDBLookupControl <> nil then
      FDBLookupControl.DataLinkActiveChanged;
end;
procedure TsDataSourceLink.RecordChanged(Field: TField);
begin
   if (FDBLookupControl <> nil) {and not FLock }then
      FDBLookupControl.DataLinkRecordChanged(Field);
end;

{ TSListSourceLink }
procedure TsListSourceLink.ActiveChanged;
begin
   if FDBLookupControl <> nil then
      FDBLookupControl.ListLinkActiveChanged;
end;

procedure TsListSourceLink.DataSetChanged;
begin
   if FDBLookupControl <> nil then
      FDBLookupControl.ListLinkDataChanged;
end;

procedure TsListSourceLink.DataSetScrolled(Distance: Integer);
begin
   FSteps := Distance;
   inherited;
end;

{ TsDBLookupControl }
function VarEquals(const V1, V2: Variant): Boolean;
begin
   Result := False;
   try
      Result := V1 = V2;
   except
   end;
end;

var
   SearchTickCount: Integer = 0;

constructor TsDBLookupControl.Create(AOwner: TComponent);
begin
  	inherited Create(AOwner);
  	if NewStyleControls then
    	ControlStyle := [csOpaque]
  	else
    	ControlStyle := [csOpaque, csFramed];
  	ParentColor := False;
  	TabStop := True;
  	FLookupSource := TDataSource.Create(Self);
  	FDataLink := TSDataSourceLink.Create;
  	FDataLink.FDBLookupControl := Self;
  	FListLink := TSListSourceLink.Create;
  	FListLink.FDBLookupControl := Self;
  	FListFields := TList.Create;
  	FKeyValue := Null;
end;

destructor TsDBLookupControl.Destroy;
begin
  	FListFields.Free;
  	FListLink.FDBLookupControl := nil;
  	FListLink.Free;
  	FDataLink.FDBLookupControl := nil;
  	FDataLink.Free;
  	inherited Destroy;
end;

function TsDBLookupControl.CanModify: Boolean;
begin
  	Result := (lsListActive in FState) and not ReadOnly and ((FDataLink.DataSource = nil) or
    	(FMasterField <> nil) and FMasterField.CanModify);
end;

procedure TsDBLookupControl.CheckNotCircular;
begin
  	if FDataLink.Active and FDataLink.DataSet.IsLinkedTo(ListSource) then
      Raise Exception.Create('Circular Data Link');
end;

procedure TsDBLookupControl.CheckNotLookup;
begin
  	if lsLookupMode in FState then
      Raise Exception.Create('Property Defined By Lookup');
  	if FDataLink.DataSourceFixed then
      Raise Exception.Create('Data Source Fixed');
end;

procedure TsDBLookupControl.DataLinkActiveChanged;
begin
  	FDataField := nil;
  	FMasterField := nil;
  	if FDataLink.Active and (FDataFieldName <> '') then begin
      CheckNotCircular;
    	FDataField := FDataLink.DataSet.FieldByName(FDataFieldName);
    	FMasterField := FDataField;
  	end;
  	SetLookupMode((FDataField <> nil) and FDataField.Lookup);
  	DataLinkRecordChanged(nil);
end;

procedure TsDBLookupControl.DataLinkRecordChanged(Field: TField);
begin
  	if (Field = nil) or (Field = FMasterField) then
    	if FMasterField <> nil then
      	SetKeyValue(FMasterField.Value)
      else
      	SetKeyValue(Null);
end;

function TsDBLookupControl.GetBorderSize: Integer;
begin
   Result := 4
end;

function TsDBLookupControl.GetDataSource: TDataSource;
begin
  	Result := FDataLink.DataSource;
end;

function TsDBLookupControl.GetKeyFieldName: string;
begin
   if lsLookupMode in FState then
      Result := ''
   else
      Result := FKeyFieldName;
end;

function TsDBLookupControl.GetListSource: TDataSource;
begin
   if lsLookupMode in FState then
      Result := nil
   else
      Result := FListLink.DataSource;
end;

function TsDBLookupControl.GetReadOnly: Boolean;
begin
   Result := FDataLink.ReadOnly;
end;

procedure TsDBLookupControl.ListLinkActiveChanged;
var
   DataSet: TDataSet;
   ResultField: TField;
begin
   Exclude(FState, lsListActive);
   FKeyField := nil;
   FListField := nil;
   FImageField := nil;

   FListFields.Clear;
   if FListLink.Active and (FKeyFieldName <> '') then begin
      CheckNotCircular;
      DataSet := FListLink.DataSet;
      FKeyField := DataSet.FieldByName(FKeyFieldName);
      DataSet.GetFieldList(FListFields, FListFieldName);
      if lsLookupMode in FState then begin
         ResultField := DataSet.FieldByName(FDataField.LookupResultField);
         if FListFields.IndexOf(ResultField) < 0 then
            FListFields.Insert(0, ResultField);
         FListField := ResultField;
      end else begin
         if FListFields.Count = 0 then
            FListFields.Add(FKeyField);
         if (FListFieldIndex >= 0) and (FListFieldIndex < FListFields.Count) then
            FListField := FListFields[FListFieldIndex]
         else
            FListField := FListFields[0];

      end;
      if FImageFieldName <> '' then
         FImageField := DataSet.FieldByName(FImageFieldName);
      Include(FState, lsListActive);
   end;
end;

procedure TsDBLookupControl.ListLinkDataChanged;
begin
end;

function TsDBLookupControl.LocateKey: Boolean;
var
   Options: TLocateOptions;
begin
   Result := False;
   if csLoading in ComponentState then
      Exit;
   try
      Options := [];
      if (FMasterField <> nil) and (FListField <> nil) and (FMasterField.Size < FListField.Size) then
         Options := [loPartialKey];
      if not VarIsNull(FKeyValue) and FListLink.DataSet.Active and
         FListLink.DataSet.Locate(FKeyFieldName, FKeyValue, Options) then
         Result := True;
   except
   end;
end;

procedure TsDBLookupControl.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);
   if Operation = opRemove then begin
      if (FDataLink <> nil) and (AComponent = DataSource) then
         DataSource := nil;
      if (FListLink <> nil) and (AComponent = ListSource) then
         ListSource := nil;
   end;
end;

function TsDBLookupControl.DoLocate(S: string; Options: TLocateOptions): Boolean;
begin
   Result := FListLink.DataSet.Locate( FListField.FieldName, S, Options);
end;

const
   _SearchTimeOut = 200;

procedure TsDBLookupControl.ProcessSearchKey(Key: Char);
var
   TickCount: Integer;
   S: string;
   Options: TLocateOptions;
begin
   if (FListField <> nil) and (FListField.FieldKind = fkData) and (FListField.DataType = ftString) then
      case Key of
         #8, #27:
            FSearchText := '';
         #32..#255: if CanModify then begin
            TickCount := GetTickCount;
            if TickCount - SearchTickCount > _SearchTimeOut then
               FSearchText := '';
            SearchTickCount := TickCount;
            if Length(FSearchText) < 32 then begin
               S := FSearchText + Key;
               Options := [loPartialKey];
               if not FCaseSensitive then
                  Include(Options, loCaseInsensitive);
               if DoLocate( S, Options) then begin
                  SelectKeyValue(FKeyField.Value);
                  FSearchText := S;
               end;
            end;
         end;
      end;
end;

function TsDBLookupControl.ProcessSearch(Str: String): Boolean;
var
   Options: TLocateOptions;
begin
   Result := FALSE;
  	if (FListField <> nil) and (FListField.FieldKind = fkData) and (FListField.DataType = ftString) then
    	if CanModify then begin
      	if Length(Str) < 32 then with FListLink.DataSet do
         	if Str <> '' then begin
               Options := [loPartialKey];
               if not FCaseSensitive then
                  Include(Options, loCaseInsensitive);
         		Result := DoLocate( Str, Options)
            end else begin
               First;
					Result := TRUE;
            end;
         if Result then begin
            //SelectCurrent;
         	SelectKeyValue(FKeyField.Value);
            FSearchText := Str;
         end;
		end;
end;

procedure TsDBLookupControl.SelectKeyValue(Value: Variant);
var
   b: Boolean;
begin
   if FMasterField <> nil then begin
      if not VarEquals( FMasterField.Value, Value) then begin
         FDataLink.FLock := TRUE;
         b := FDataLink.Edit;
         FDataLink.FLock := FALSE;
         if b then
         	FMasterField.Value := Value;
      end;
  	end else
    	SetKeyValue(Value);
end;

procedure TsDBLookupControl.SetDataFieldName(const Value: string);
begin
   if FDataFieldName <> Value then begin
      FDataFieldName := Value;
      DataLinkActiveChanged;
   end;
end;

procedure TsDBLookupControl.SetDataSource(Value: TDataSource);
begin
   FDataLink.DataSource := Value;
   if Value <> nil then
      Value.FreeNotification(Self);
end;

procedure TsDBLookupControl.SetKeyFieldName(const Value: string);
begin
   CheckNotLookup;
   if FKeyFieldName <> Value then begin
      FKeyFieldName := Value;
      ListLinkActiveChanged;
   end;
end;

procedure TsDBLookupControl.SetKeyValue(const Value: Variant);
begin
   if not VarEquals(FKeyValue, Value) then //begin
      FKeyValue := Value;

      KeyValueChanged;
   //end;

end;

procedure TsDBLookupControl.SetListFieldName(const Value: string);
begin
   if FListFieldName <> Value then begin
      FListFieldName := Value;
      ListLinkActiveChanged;
   end;
end;

procedure TsDBLookupControl.SetImageFieldName(const Value: String);
begin
   if FListFieldName <> Value then begin
      FListFieldName := Value;
      ListLinkActiveChanged;
   end;
end;

procedure TsDBLookupControl.SetListSource(Value: TDataSource);
begin
   CheckNotLookup;
   FListLink.DataSource := Value;
   if Value <> nil then
      Value.FreeNotification(Self);
end;

procedure TsDBLookupControl.SetLookupMode(Value: Boolean);
begin
   if (lsLookupMode in FState) <> Value then
      if Value then begin
         FMasterField := FDataField.DataSet.FieldByName(FDataField.KeyFields);
         FLookupSource.DataSet := FDataField.LookupDataSet;
         FKeyFieldName := FDataField.LookupKeyFields;
         Include( FState, lsLookupMode);
         FListLink.DataSource := FLookupSource;
      end else begin
         FListLink.DataSource := nil;
         Exclude( FState, lsLookupMode);
         FKeyFieldName := '';
         FLookupSource.DataSet := nil;
         FMasterField := FDataField;
      end;
end;

procedure TsDBLookupControl.SetReadOnly(Value: Boolean);
begin
   FDataLink.ReadOnly := Value;
end;

procedure TsDBLookupControl.WMGetDlgCode(var Message: TMessage);
begin
   Message.Result := DLGC_WANTARROWS or DLGC_WANTCHARS;
end;

procedure TsDBLookupControl.WMKillFocus(var Message: TMessage);
begin
   Exclude(FState, lsFocused);
   Invalidate;
   inherited;
end;

procedure TsDBLookupControl.WMSetFocus(var Message: TMessage);
begin
   Include(FState, lsFocused);
   Invalidate;
   inherited;
end;

{ TSLookupListBox}
constructor TSDBLookupListBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FItemHeight := 13;
   FTextWidth := 5;

   ControlStyle := ControlStyle + [csDoubleClicks];
   Width := 121;
   FBorderStyle := bsSingle;
   FScrolling := FALSE;
   RowCount := 7;

   FImageChangeLink := TChangeLink.Create;
   FImageChangeLink.OnChange := ImageListChange;

   FHilight := TsHilight.Create;
   FHilight.Font.AssignToFont(Font);
end;

destructor TsDBLookupListBox.Destroy;
begin
   FHilight.Free;
   inherited;
end;

procedure TSDBLookupListBox.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
   with Params do if FBorderStyle = bsSingle then
      if NewStyleControls and Ctl3D then
         ExStyle := ExStyle or WS_EX_CLIENTEDGE
      else
         Style := Style or WS_BORDER;
end;

procedure TsDBLookupListBox.CreateWnd;
begin
   inherited CreateWnd;
   UpdateScrollBar;
end;

procedure TsDBLookupListBox.Loaded;
begin
   inherited;
end;

procedure TSDBLookupListBox.SetHilight(Value: TsHilight);
begin
   FHilight.Assign(Value);
end;

function TSDBLookupListBox.GetKeyIndex: Integer;
var
   FieldValue: Variant;
begin
   if not VarIsNull(FKeyValue) then
      for Result := 0 to FRecordCount - 1 do begin
         FListLink.ActiveRecord := Result;
         FieldValue := FKeyField.Value;
         FListLink.ActiveRecord := FRecordIndex;
         if VarEquals(FieldValue, FKeyValue) then Exit;
      end;
   Result := -1;
end;

type
   TDataSetCrack = class(TDataSet)
      property CurrentRecord;
   end;

function TSDBLookupListBox.DoLocate(S: string; Options: TLocateOptions): Boolean;
var
   cr: Integer;
   no: Integer;
   delta: Integer;
begin
   cr := TDataSetCrack(FListLink.dataSet).CurrentRecord;
   if FListLink.dataSet.IsSequenced then
   	no := FListLink.dataSet.RecNo
   else
      no := 0;
   Result := inherited DoLocate(S, Options);
   if Result and FListLink.dataSet.IsSequenced then begin
      FScrolling := True;
      SelectCurrent;
      delta := FListLink.dataSet.RecNo - no;
	   if not ((FHilightIndex + Delta) in [0..FRowCount-1]) then
		   ScrollBy( delta, saLookup in FListScrollAnimation)
      else begin
         delta := TDataSetCrack(FListLink.dataSet).CurrentRecord - cr;
		   if not ((FHilightIndex + Delta) in [0..FRowCount-1]) then
			   ScrollBy( delta, saLookup in FListScrollAnimation)
      end;
      FScrolling := FALSE;
   end;
   //Refresh;
end;

procedure TSDBLookupListBox.ScrollBy(delta: Integer; animated: Boolean = TRUE);
var
   scroller: TsListScroller;
   Top, step, Index, ii: Integer;
begin
   if Animated then begin
   	Canvas.Font := Font;
      scroller := TsListScroller.Create(Handle, FItemHeight, FRowCount);
      try
      	scroller.PaintItemProc := PaintItemTo;
         if delta > 0 then
         	step := -1
         else
      	   step := 1;

         Top := 1;

         Index := FRecordIndex - Delta - step;

         if Delta < 0 then
      	   delta := -delta;

         scroller.WaitStep := Delta * ListScrollWaitSteps;
         if lsTimerActive in FState then
         	scroller.WaitStep := 4 * ListScrollWaitSteps;
         scroller.waittime := ListScrollWaitTime;

         for ii := 1 to delta do begin
         	scroller.ScrollEx(top, Index, step, -1, TRUE);
            UpdateWindow(handle);
            Dec( Index, step);
         end;
      finally
   	   scroller.Free;
      end;
   end else
      Invalidate;
end;

procedure TSDBLookupListBox.MoveBy(delta: Integer; animated: Boolean = TRUE);
var
   KeyIndex: Integer;
   R: TRect;
begin
   if (FListLink.DataSet.EOF and (Delta > 0)) or
      (FListLink.DataSet.BOF and (Delta < 0)) then
      Exit;

   // "unselect" item
   R := ClientRect;
   R.Top := FHilightIndex * FItemHeight;
   R.Bottom := R.Top + FItemHeight;
   PaintItem( R, FHilightIndex, FALSE);

   FScrolling := TRUE;
   KeyIndex := GetKeyIndex;

   if KeyIndex >= 0 then begin
      if KeyIndex <> FRecordIndex then
         FListLink.DataSet.MoveBy(KeyIndex - FRecordIndex)
   end else begin
      KeyValueChanged;
      Delta := 0;
   end;

   FListLink.DataSet.MoveBy(Delta);
   SelectCurrent;

   if not ((FHilightIndex + Delta) in [0..FRowCount-1]) and
      not ((FListLink.Steps = 0) or ((FListLink.Steps = 1) and (FListLink.DataSet.EOF))) then begin
      if FListLink.Steps >= 0 then begin
         if Delta > FListLink.Steps then
         	Delta := FListLink.Steps;
      end else begin
         if Delta < FListLink.Steps then
         	Delta := FListLink.Steps;
      end;
      ScrollBy(delta, animated);
   end else begin
      R.Top := FRecordIndex * FItemHeight;
      R.Bottom := R.Top + FItemHeight;
      PaintItem( R, FRecordIndex, TRUE);
   end;
   FScrolling := FALSE;
   UpdateScrollBar;
end;

procedure TSDBLookupListBox.ScrollFirst(animated: Boolean = TRUE);
begin
   if not FListLink.DataSet.BOF then begin
      if animated then
         ScrollBy(-FRowCount);
      FListLink.DataSet.First;
   end;
end;

procedure TSDBLookupListBox.ScrollLast(animated: Boolean = TRUE);
begin
   if not FListLink.DataSet.EOF then begin
      if animated then
         ScrollBy(FRowCount);
      FListLink.DataSet.Last;
   end;
end;

procedure TSDBLookupListBox.KeyDown(var Key: Word; Shift: TShiftState);
var
   Delta: Integer;
begin
   inherited KeyDown(Key, Shift);
   if CanModify then begin
      Delta := 0;
      case Key of
         VK_UP, VK_LEFT:
            Delta := -1;
         VK_DOWN, VK_RIGHT:
            Delta := 1;
         VK_PRIOR:
            Delta := 1 - FRowCount;
         VK_NEXT:
            Delta := FRowCount - 1;
         VK_HOME: if not FListLink.DataSet.BOF then
            Delta := -Maxint;
         VK_END: if not FListLink.DataSet.EOF then
            Delta := Maxint;
      end;
      if Delta <> 0 then begin
         FSearchText := '';
         if Delta = -Maxint then
            ScrollFirst(saKey in FListScrollAnimation)
         else if Delta = Maxint then
            ScrollLast(saKey in FListScrollAnimation)
         else
            MoveBy(delta, saKey in FListScrollAnimation);
      end else
         SelectCurrent;
   end;
end;

procedure TSDBLookupListBox.KeyPress(var Key: Char);
begin
   inherited KeyPress(Key);
   if not (lsPopUp in FState) then
      ProcessSearchKey(Key);
end;

procedure TSDBLookupListBox.KeyValueChanged;
begin
   if (lsListActive in FState) and not (lsLockPosition in FState) and not LocateKey then
      FListLink.DataSet.First;
end;

procedure TSDBLookupListBox.ListLinkActiveChanged;
begin
   try
      inherited;
   finally
      if lsListActive in FState then
         KeyValueChanged
      else
         ListLinkDataChanged;
   end;
end;

procedure TsDBLookupListBox.ListLinkDataChanged;
begin
   if lsListActive in FState then begin
      FRecordCount := FListLink.RecordCount;
      FRecordIndex := FListLink.ActiveRecord;
      FKeySelected := not VarIsNull(FKeyValue) or not FListLink.DataSet.BOF;
   end else begin
      FRecordIndex := 0;
      FRecordCount := 0;
      FKeySelected := False;
   end;

   if not FScrolling and HandleAllocated then
      UpdateScrollBar;
end;

procedure TSDBLookupListBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if Button = mbLeft then begin
      FSearchText := '';
      if not (lsPopup in FState) then begin
         SetFocus;
         if not (lsFocused in FState) then
            Exit;
      end;
      if CanModify then
         if ssDouble in Shift then begin
            if FRecordIndex = Y div FItemHeight then
               DblClick;
         end else begin
            MouseCapture := True;
            Include( FState, lsTracking);
            SelectItemAt(X, Y);
         end;
   end;
   inherited MouseDown(Button, Shift, X, Y);
end;

procedure TSDBLookupListBox.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
   if lsTracking in FState then begin
      SelectItemAt(X, Y);
      FMousePos := Y;
      TimerScroll;
   end;
   inherited MouseMove(Shift, X, Y);
end;

procedure TSDBLookupListBox.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if lsTracking in FState then begin
      StopTracking;
      SelectItemAt(X, Y);
   end;
   inherited MouseUp(Button, Shift, X, Y);
end;

procedure TSDBLookupListBox.PaintItemTo(DC: HDC; Rect: TRect; Index: Integer; Selected: Boolean);
var
   keptDC: HDC;
begin
   KeptDC := Canvas.Handle;
   Canvas.Handle := DC;
   try
      PaintItem(Rect, Index, Selected);
   finally
      Canvas.Handle := KeptDC;
   end;
end;

procedure DoPaintImage( Canvas: TCanvas; R: TRect; ImageIndex: Integer; ImageList: TCustomImageList);
begin
   Canvas.FillRect( R);
   if ImageIndex > -1 then
      ImageList_Draw( ImageList.Handle, imageIndex, Canvas.Handle,
         R.Left + 1, (R.Top + R.Bottom - ImageList.Height) div 2+1,
         ILD_TRANSPARENT);
end;

procedure DoDrawItem(Canvas: TCanvas; Rect: TRect; ListFields: TList; TextWidth, ImageWidth, ImageIndex: Integer;
   ImageList: TCustomImageList; selected: Boolean; hilight: TsHilight);
const
   EdgeS: array[Boolean] of Integer = (EDGE_SUNKEN, EDGE_RAISED);
var
   LastFieldIndex, ii, W, X: Integer;
   Field: TField;
   S: String;
   R: TRect;
begin
	LastFieldIndex := ListFields.Count - 1;

   if Selected then begin
      Canvas.Font.Assign(Hilight.Font);
      if (Hilight <> nil) then case hilight.Style of
         lhsUnderline:
            Canvas.Font.Style := Canvas.Font.Style + [fsUnderline];
         else
            Canvas.Brush.Color := Hilight.Color;
      end;
   end;

   R := Rect;
	R.Right := ImageWidth;
   DoPaintImage( Canvas, R, ImageIndex, ImageList);

   if Selected and (Hilight <> nil) and (Hilight.Style = lhsOffice97) then begin
      if ImageIndex > -1 then
         DrawFlatEDGE(Canvas, R, FALSE);
      Canvas.Brush.Color := Hilight.Color;
   end;

   for ii := 0 to LastFieldIndex do begin
   	Field := ListFields[ii];
      if ii < LastFieldIndex then
      	W := Field.DisplayWidth * TextWidth + 4
      else
      	W := WidthOf(Rect) - R.Right;

      S := Field.DisplayText;
      X := 2;
      case Field.Alignment of
      	taRightJustify:
         	X := W - Canvas.TextWidth(S) - 3;
         taCenter:
         	X := (W - Canvas.TextWidth(S)) div 2;
      end;

      R.Left := R.Right;
      R.Right := R.Right + W;

      Canvas.TextRect(R, R.Left + X, R.Top, S);

      if ii < LastFieldIndex then begin
      	Canvas.MoveTo(R.Right, R.Top);
         Canvas.LineTo(R.Right, R.Bottom);
         Inc(R.Right);
         if R.Right >= WidthOf(Rect) then
         	Break;
      end;
   end;
   if Selected then case Hilight.Style of
      lhsFrame, lhsSunkedFrame:
         DrawFlatEDGE(Canvas, Rect, Hilight.Style = lhsSunkedFrame);
      lhsEDGE, lhsSunkedEDGE:
         DrawEDGE(Canvas.Handle, R,
            EdgeS[(Hilight.Style = lhsEdge)], BF_RECT);
   end;
end;

procedure TSDBLookupListBox.PaintItem(Rect: TRect; Index: Integer; Selected: Boolean);
var
   R: TRect;
   itemSelected: Boolean;
begin
   //LastFieldIndex := FListFields.Count - 1;
   Canvas.Font := Font;

   Canvas.Font.Color := Font.Color;
   Canvas.Brush.Color := Color;
   R := Rect;
   itemSelected := FALSE;

   if FScrolling and (Index >= FRecordCount) then
      Index := FRecordCount - 1;

   if index < FRecordCount then begin
      if Selected then
         itemSelected := not FKeySelected and (index = 0);

      FListLink.ActiveRecord := Index;

      if Selected and not VarIsNull(FKeyValue) and VarEquals(FKeyField.Value, FKeyValue) then begin
         itemSelected := TRUE;
         FHilightIndex := index;
      end;
		DoDrawItem(Canvas, Rect, FListFields, FTextWidth, FImageOffset, GetImageID, FImages,
         itemSelected, FHilight);
      Canvas.Font.Assign(Font);
      Canvas.Brush.Color := Color;
   end;
   if index >= FRecordCount then
      Canvas.FillRect(Rect);
   if itemSelected and ([lsFocused, lsPopup] * FState <> []) and
      (FHilight.Style in [lhsDefault, lhsUnderline]) then
      Canvas.DrawFocusRect(Rect);
end;

procedure TSDBLookupListBox.Paint;
var
   ii: Integer;
   R: TRect;
begin
   if FScrolling then
      Exit;
   Canvas.Font := Font;
   if ColorToRGB(Color) <> ColorToRGB(clBtnFace) then
      Canvas.Pen.Color := clBtnFace
   else
      Canvas.Pen.Color := clBtnShadow;

   for ii := 0 to FRowCount - 1 do begin
      Canvas.Font.Color := Font.Color;
      Canvas.Brush.Color := Color;
      R := ClientRect;
      R.Top := ii * FItemHeight;
      R.Bottom := R.Top + FItemHeight;
      if ii < FRecordCount then
         PaintItem( R, ii, TRUE)
      else
         Canvas.FillRect( R);
   end;
   if FRecordCount <> 0 then
      FListLink.ActiveRecord := FRecordIndex;
end;

procedure TSDBLookupListBox.SelectCurrent;
begin
   Include( FState, lsLockPosition);
   try
      SelectKeyValue(FKeyField.Value);
   finally
      Exclude( FState, lsLockPosition);
   end;
end;

procedure TSDBLookupListBox.SelectItemAt(X, Y: Integer);
var
   Delta: Integer;
begin
   if Y < 0 then
      Y := 0;
   if Y >= ClientHeight then
      Y := ClientHeight - 1;
   Delta := Y div FItemHeight - FRecordIndex;

   if Delta <> 0 then
      MoveBy(delta);
end;

procedure TSDBLookupListBox.SetBorderStyle(Value: TBorderStyle);
begin
   if FBorderStyle <> Value then begin
      FBorderStyle := Value;
      RecreateWnd;
      RowCount := RowCount;
   end;
end;

procedure TSDBLookupListBox.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
   BorderSize, Rows: Integer;
begin
   BorderSize := GetBorderSize;
   Rows := (AHeight - BorderSize) div FItemHeight;
   if Rows < 1 then
      Rows := 1;

   if (FListLink.DataSet <> nil) and FListLink.DataSet.IsSequenced and
      (FListLink.DataSet.RecordCount < Rows) then
      Rows := FListLink.DataSet.RecordCount;

   FRowCount := Rows;
   if FListLink.BufferCount <> Rows then begin
      FListLink.BufferCount := Rows;
      ListLinkDataChanged;
   end;
   inherited SetBounds(ALeft, ATop, AWidth, Rows * FItemHeight + BorderSize);
end;

procedure TSDBLookupListBox.SetRowCount(Value: Integer);
begin
   if Value < 1 then
      Value := 1;
   if Value > 100 then
      Value := 100;
   if (FListLink.DataSet <> nil) and FListLink.DataSet.IsSequenced and
      (FListLink.DataSet.RecordCount < Value) then
      Value := FListLink.DataSet.RecordCount;
   Height := Value * FItemHeight + GetBorderSize;
end;

function TSDBLookupListBox.GetImageId: Integer;
begin
   Result := -1;
   if FImageField <> nil then try
      Result := FImageField.AsInteger;
   except
   end;
   if Assigned(FOnGetImageId) and (FListLink <> nil) and (FListLink.DataSet <> nil) then
      FOnGetImageId( self, FListLink.DataSet, Result);
end;

procedure TSDBLookupListBox.ImageListChange(Sender: TObject);
begin
   FImageOffset := 0;
   if Sender = Images then begin
      FImageOffset := Images.Width + 2;
      if FItemHeight < Images.Height + 2 then
         FItemHeight := Images.Height + 2;
      Refresh;
   end;
   if Assigned(FOnImageChange) then
      FOnImageChange(self);
end;

procedure TSDBLookupListBox.SetImages(const Value: TCustomImageList);
begin
   FImageOffset := 0;
   if FImages <> nil then
      FImages.UnRegisterChanges(FImageChangeLink);
   FImages := Value;
   if FImages <> nil then begin
      FImages.RegisterChanges(FImageChangeLink);
      FImages.FreeNotification(Self);
   	FImageOffset := Images.Width + 2;
      if FItemHeight < Images.Height + 2 then
         FItemHeight := Images.Height + 2;
   end;
   if not (csLoading in ComponentState) then
      Invalidate;
   if Assigned(OnImageChange) then
      OnImageChange(self);
end;

procedure TSDBLookupListBox.StopTimer;
begin
   if lsTimerActive in FState then begin
      KillTimer(Handle, 1);
      Exclude(Fstate, lsTimerActive);
   end;
end;

procedure TSDBLookupListBox.StopTracking;
begin
   if lsTracking in FState then begin
      StopTimer;
      Exclude( FState, lsTracking);
      MouseCapture := False;
   end;
end;

procedure TSDBLookupListBox.TimerScroll;
var
   Delta, Distance, Interval: Integer;
begin
   Delta := 0;
   Distance := 0;
   if FMousePos < 0 then begin
      Delta := -1;
      Distance := -FMousePos;
   end;
   if FMousePos >= ClientHeight then begin
      Delta := 1;
      Distance := FMousePos - ClientHeight + 1;
   end;

   if Delta = 0 then
      StopTimer
   else begin
      MoveBy(Delta, saAuto in FListScrollAnimation);
      SelectCurrent;
      Interval := 200 - Distance * 15;
      if Interval < 0 then
         Interval := 0;
      SetTimer(Handle, 1, Interval, nil);
      Include(Fstate, lsTimerActive);
   end;
end;

procedure TSDBLookupListBox.UpdateScrollBar(forceUpdate: Boolean = FALSE);
var
   SIOld, SINew: TScrollInfo;
begin
   if FListLink.Active and HandleAllocated then with FListLink.DataSet do begin
      SIOld.cbSize := sizeof(SIOld);
      SIOld.fMask := SIF_ALL;
      GetScrollInfo(Self.Handle, SB_VERT, SIOld);
      SINew := SIOld;

      if IsSequenced then begin
         SINew.nMin := 1;
         SINew.nPage := FRowCount;
         SINew.nMax := Integer(DWORD(RecordCount) + SINew.nPage - 1);
         if State in [dsInactive, dsBrowse, dsEdit] then
            SINew.nPos := RecNo;
      end else begin
         if FRecordCount = FRowCount then begin
            SINew.nMin := 0;
            SINew.nMax := 4;
            if not FListLink.DataSet.BOF then
               if not FListLink.DataSet.EOF then
                  SINew.nPos := 2
               else
                  SINew.nPos := 4;
         end;
      end;
      if ForceUpdate or (SINew.nMin <> SIOld.nMin) or (SINew.nMax <> SIOld.nMax) or
        (SINew.nPage <> SIOld.nPage) or (SINew.nPos <> SIOld.nPos) then
        SetScrollInfo(Self.Handle, SB_VERT, SINew, True);
   end;
end;

procedure TSDBLookupListBox.CMCtl3DChanged(var Message: TMessage);
begin
   if NewStyleControls and (FBorderStyle = bsSingle) then begin
      RecreateWnd;
      RowCount := RowCount;
   end;
   inherited;
end;

procedure TSDBLookupListBox.CMFontChanged(var Message: TMessage);
begin
   inherited;
   Canvas.Font := Font;
   FItemHeight := Canvas.TextHeight('O');
   if Images <> nil then
      FItemHeight := Max(FItemHeight, Images.Height + 2);
   FTextWidth := Canvas.TextWidth('0');
   Height := Height;

	if EqualFonts(Font, FHilight.Font) then
		FHilight.Font.AssignToFont( Font);
end;

procedure TSDBLookupListBox.WMCancelMode(var Message: TMessage);
begin
   StopTracking;
   inherited;
end;

procedure TSDBLookupListBox.WMTimer(var Message: TMessage);
begin
   TimerScroll;
end;

procedure TSDBLookupListBox.WMVScroll(var Message: TWMVScroll);
begin
   FSearchText := '';
   with Message do
      case ScrollCode of
         SB_LINEUP:
            MoveBy(-FRecordIndex - 1, saScrollBar in FListScrollAnimation);
         SB_LINEDOWN:
            MoveBy(FRecordCount - FRecordIndex, saScrollBar in FListScrollAnimation);
         SB_PAGEUP:
            MoveBy(-FRecordIndex - FRecordCount + 2, saScrollBar in FListScrollAnimation);
         SB_PAGEDOWN:
            MoveBy(FRecordCount - FRecordIndex + FRecordCount - 2, saScrollBar in FListScrollAnimation);
         SB_THUMBPOSITION: if FListLink.Active and HandleAllocated then begin
            if FListLink.DataSet.IsSequenced then begin
               if Pos <= 1 then
                  ScrollFirst(saKey in FListScrollAnimation)
               else if Pos >= FListLink.DataSet.RecordCount then
                  ScrollLast(saKey in FListScrollAnimation)
               else
                  MoveBy( Pos - FListLink.DataSet.RecNo + 1, saScrollBar in FListScrollAnimation);
               Refresh;
            end else case Pos of
               0: ScrollFirst(saKey in FListScrollAnimation);
               1: MoveBy(-FRecordIndex - FRecordCount + 1, saScrollBar in FListScrollAnimation);
               2: Exit;
               3: MoveBy(FRecordCount - FRecordIndex + FRecordCount - 2, saScrollBar in FListScrollAnimation);
               4: ScrollLast(saKey in FListScrollAnimation);
            end;
         end;
         SB_BOTTOM: begin
            if saKey in FListScrollAnimation then
               ScrollBy(FRowCount);
            FListLink.DataSet.Last;
            Refresh;
         end;
         SB_TOP: begin
            if saKey in FListScrollAnimation then
               ScrollBy(-FRowCount);
            FListLink.DataSet.First;
            Refresh;
         end;
      end;
end;

{ TSPopupDataList }
constructor TSPopupDataList.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FItemHeight := Canvas.TextHeight('O');
   FTextWidth := Canvas.TextWidth('0');
  	FCombo := AOwner as TsDBSearchComboBox;
  	ControlStyle := ControlStyle + [csNoDesignVisible, csReplicatable];
   FBorderStyle := lbsEdge;
   FState := [lsPopup];
   Cursor := crHandPoint;
end;

procedure TSPopupDataList.CreateParams(var Params: TCreateParams);
begin
  	inherited CreateParams(Params);
  	with Params do begin
      if FBorderStyle = lbsEdge then
         Style := WS_POPUP
      else
         Style := WS_POPUP or WS_BORDER;

      if FBorderStyle = lbsEdge then
         ExStyle := WS_EX_TOOLWINDOW or WS_EX_CLIENTEDGE
      else
         ExStyle := WS_EX_TOOLWINDOW;
    	WindowClass.Style := CS_SAVEBITS;
  	end;
end;

function TSPopupDataList.GetBorderSize: Integer;
begin
   Result := inherited GetBorderSize;
   if FBorderStyle <> lbsEdge then
      Dec(Result, 2);
end;

procedure TSPopupDataList.SelectCurrent;
var
	field: TField;
begin
   inherited;
   Exclude( FCombo.FState, lsEditMode);
   field := FListFields[FListFieldIndex];
	if FCombo.FText <> field.DisplayText then
   	FCombo.FText := field.DisplayText;
end;

procedure TsPopupDataList.WMMouseActivate(var Message: TMessage);
begin
  	Message.Result := MA_NOACTIVATE;
end;

procedure TsPopupDataList.WMNCPaint(var Message: TMessage);
var
   DC: HDC;
   R: TRect;
begin
   inherited;
   GetWindowRect(Handle, R);
   OffsetRect (R, -R.Left, -R.Top);
   DC := GetWindowDC(Handle);
   DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_ADJUST);
   ReleaseDC(Handle, DC);
end;

procedure TsPopupDataList.WMMouseMove(var Message: TWMMouseMove);
begin
   inherited;
   if not (lsTracking in FState) then
      SelectItemAt(Message.XPos, Message.YPos);
end;

procedure TsPopupDataList.KeyDown(var Key: Word; Shift: TShiftState);
begin
   inherited KeyDown(Key, Shift);
   if CanModify and (Key in [VK_UP, VK_LEFT, VK_DOWN, VK_RIGHT, VK_PRIOR, VK_NEXT, VK_HOME, VK_END]) then
      FCombo.Paint;
end;

procedure TsPopupDataList.SetBorderStyle(Value: TsListBorderStyle);
var
   dlStyle: Longint;
begin
   if FBorderStyle <> Value then begin
      FBorderStyle := Value;
      if HandleAllocated then begin
         dlStyle := GetWindowLong(Handle, GWL_STYLE);
         if FBorderStyle = lbsEdge then
            dlStyle := dlStyle and not WS_BORDER
         else
            dlStyle := dlStyle or WS_BORDER;
         SetWindowLong(Handle, GWL_STYLE, dlStyle);

         dlStyle := GetWindowLong(Handle, GWL_EXSTYLE);
         if FBorderStyle = lbsEdge then
            dlStyle := dlStyle or WS_EX_CLIENTEDGE
         else
            dlStyle := dlStyle and not WS_EX_CLIENTEDGE;
         SetWindowLong(Handle, GWL_EXSTYLE, dlStyle);
      end;
   end;
end;

procedure TsPopupDataList.AniNCPaint(tDC: HDC; var R: TRect);
var
   Brush: HBrush;
   sR: TRect;
   flag: Integer;

   SI: TScrollInfo;
   cy: Integer;

   sWidth: Integer;
   sArowHeight: Integer;
begin
   Brush := CreateSolidBrush(ColorToRGB(Color));
   FillRect(tDC, R, Brush);
   DeleteObject(Brush);

	if FBorderStyle = lbsSmallEdge then
   	InflateRect(R, 1, 1);
   if FBorderStyle <> lbsDefault then
   	DrawEdge( tDC, R, EDGE_RAISED, BF_RECT or BF_ADJUST)
   else begin
   	Brush := CreateSolidBrush(GetSysColor(COLOR_WINDOWFRAME));
      FrameRect( tDC, R, Brush);
      DeleteObject(Brush);
      InflateRect(R, -1, -1);
   end;

   if FRecordCount = FRowCount then begin
      sWidth := GetSystemMetrics( SM_CXVSCROLL); // Width of arrow bitmap on vertical scroll bar
      sArowHeight := GetSystemMetrics( SM_CYVSCROLL); // Height of arrow bitmap on vertical scroll bar

   	sR := R;
      sR.Left := sR.Right - sWidth;

      Brush := CreateSolidBrush( GetSysColor(COLOR_3DLight));
      FillRect( tDC, sR, Brush);
      DeleteObject(Brush);

      sR.Bottom := sR.Top + sArowHeight;
      flag := 0;

      DrawFrameControl( tDC, sR, DFC_SCROLL, DFCS_SCROLLUP or flag);

      sR.Bottom := R.Bottom;
      sR.Top := sR.Bottom - sArowHeight;
      DrawFrameControl( tDC, sR, DFC_SCROLL, DFCS_SCROLLDOWN or flag);

      SI.cbSize:= SizeOf(SI);
      SI.fMask:= SIF_RANGE or SIF_POS or SIF_PAGE;
      GetScrollInfo( Handle, SB_VERT, SI);

      sR.Top := R.Top + sArowHeight;

      cy := GetSystemMetrics(SM_CYVTHUMB);

      if SI.nPos > 0 then
         sR.Top := sR.Top + Trunc(0.5+(((HeightOf(R) - 2 * sArowHeight) - cy) /
                  ((SI.nMax + 1) - Integer(SI.nPage)) * SI.nPos));
      sR.Bottom := sR.Top + cy;

      DrawEdge( tDC, sR, EDGE_RAISED, BF_RECT or BF_ADJUST );

      Brush := CreateSolidBrush( GetSysColor(COLOR_SCROLLBAR));
      FillRect( tDC, sR, Brush);
      DeleteObject(Brush);

      R.Right := R.Right - sWidth - 2;
      R.Bottom := R.Bottom - 2
   end else begin
   	R.Right := R.Right - 2;
      R.Bottom := R.Bottom - 2
   end;
end;

{ TDBLookupComboBox }
constructor TsDBSearchComboBox.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
  	ControlStyle := ControlStyle + [csReplicatable];
  	Width := 145;
  	Height := 0;
  	FDataList := TSPopupDataList.Create(Self);
  	FDataList.Visible := False;
  	FDataList.Parent := Self;
  	FDataList.OnMouseUp := ListMouseUp;
   FDataList.OnImageChange := OnImageChange;
   ListColor := DEF_Combo_ListColor;
   ListCursor := DEF_Combo_ListCursor;
   ListBorderStyle := DEF_Combo_ListBorderStyle;
   ListScrollAnimation := DEF_Combo_ListScrollAnimation;
   FListPopupAnimation := DEF_Combo_ListPopupAnimation;

  	FEdit := TComboInPlaceEdit.Create(self);
 	FEdit.Parent := self;
  	FEdit.FCombo := self;
   FEdit.OnChange := EditChange;

   BorderStyle := DEF_Edit_BorderStyle;
   FDosKey := DEF_Combo_DosKey;
  	FDropDownRows := 7;
   FDisabledFont := TFont.Create;
   FDisabledFont.Assign(Font);
   SetButtonWidth;
   FLookupStyle := DEF_DBSearchCombo_LookupStyle;

   FDisabledStyle := dsDefault;
   FStyle := csDropDown;
end;

destructor TsDBSearchComboBox.Destroy;
begin
   FDataList.Free;
   FEdit.Free;
   inherited;
end;

procedure TsDBSearchComboBox.CreateParams(var Params: TCreateParams);
begin
  	inherited CreateParams(Params);
   with Params do case FBorderStyle of
      _bsNone:
         Style := Style and not WS_BORDER;
      _bsSingle:
         Style := Style or WS_BORDER;
      bsDefault, bsFlat: begin
         Style := Style and not WS_BORDER;
         ExStyle := ExStyle or WS_EX_CLIENTEDGE;
      end;
   end;
end;

procedure TsDBSearchComboBox.CreateWnd;
begin
   SetButtonWidth;
   inherited;
   if FEdit <> nil then
      FEdit.HandleNeeded;
   Height := 0;
end;

procedure TsDBSearchComboBox.Loaded;
begin
   inherited;
   if FFlat then begin
      if FBorderStyle = _bsNone then
         FBorderStyle := bsFlat;
   end;
end;

procedure TsDBSearchComboBox.SetButtonWidth;
begin
   if Style = csSimple then
      FButtonWidth := 0
   else begin
      if FFlat then
      	FButtonWidth := 11
      else
      	FButtonWidth := GetSystemMetrics(SM_CXVSCROLL);
   end;
end;

function TsDBSearchComboBox.GetCtl3D: Boolean;
begin
   Result := inherited Ctl3D;
end;

procedure TsDBSearchComboBox.SetCtl3D(const Value: Boolean);
begin
   if csLoading in ComponentState then begin
      {since new version of controls dows not store Ctl3D, we can come to this point ONLY with
       previous verions}
      inherited Ctl3D := FBorderStyle in [bsFlat, bsDefault];
   end else begin
      if Value then
         FBorderStyle := bsDefault
      else
         FBorderStyle := _bsSingle;
      inherited Ctl3D := Value;
      if HandleAllocated then
         RecreateWnd;
   end;
end;

procedure TsDBSearchComboBox.SetFlat(const Value: Boolean);
begin
   if csLoading in ComponentState then begin
      {since new version of controls dows not store Ctl3D, we can come to this point ONLY with
       previous verions}
      FFlat := FBorderStyle = bsFlat;
   end else if Value <> FFlat then begin
      FFlat := Value;
      if Value then
         FBorderStyle := bsFlat
      else
         FBorderStyle := bsDefault;
      SetButtonWidth;
      if not Ctl3D then
         Ctl3D := TRUE
      else if HandleAllocated then
         RecreateWnd;
   end;
end;

procedure TsDBSearchComboBox.SetBorderStyle(const Value: TsBorderStyle);
begin
  if FBorderStyle <> Value then begin
      FBorderStyle := Value;
      case FBorderStyle of
         _bsNone: begin
            FFlat := FALSE;
            inherited Ctl3D := FALSE;
         end;
         _bsSingle: begin
            FFlat := FALSE;
            inherited Ctl3D := FALSE;
         end;
         bsFlat: begin
            FFlat := TRUE;
            inherited Ctl3D := TRUE;
         end;
         bsDefault: begin
            FFlat := FALSE;
            inherited Ctl3D := TRUE;
         end;
      end;
      if HandleAllocated then
         RecreateWnd;
  end;
end;

procedure TsDBSearchComboBox.SetHilight(Value: TsHilight);
begin
   FDataList.Hilight := Value;
end;

function TsDBSearchComboBox.GetHilight: TsHilight;
begin
   Result := FDataList.Hilight;
end;

procedure TsDBSearchComboBox.SetEmptyString(Value: String);
begin
   if Value <> FEmptyString then begin
      FEmptyString := Value;
      KeyValueChanged;
   end;
end;

procedure TsDBSearchComboBox.SetDisabledFont(Value: TFont);
begin
   FDisabledFont.Assign(Value);
   if not Enabled then
      Invalidate;
end;

function TsDBSearchComboBox.GetListColor: TColor;
begin
   Result := FDataList.Color;
end;

procedure TsDBSearchComboBox.SetListColor(Value: TColor);
begin
   FDataList.Color := Value;
end;

function TsDBSearchComboBox.GetListBorderStyle: TsListBorderStyle;
begin
   Result := FDataList.BorderStyle;
end;

procedure TsDBSearchComboBox.SetListBorderStyle(Value: TsListBorderStyle);
begin
   FDataList.BorderStyle := Value;
end;

function TsDBSearchComboBox.GetListCursor: TCursor;
begin
   Result := FDataList.Cursor;
end;

procedure TsDBSearchComboBox.SetListCursor(Value: TCursor);
begin
   FDataList.Cursor := Value;
end;

function TsDBSearchComboBox.GetImages: TCustomImageList;
begin
   Result := FDataList.Images;
end;

procedure TsDBSearchComboBox.SetImages(Value: TCustomImageList);
begin
   FDataList.Images := Value;
   SetEditRect;
   Refresh;
end;

{$IFDEF V1_COMP}
procedure TsDBSearchComboBox.ReadDefaultDraw(Reader: TReader);
begin
   if Reader.ReadBoolean then
      DisabledStyle := dsDefault
   else
      DisabledStyle := dsOffice97;
end;

procedure TsDBSearchComboBox.WriteNothing(Writer: TWriter);
begin
end;

procedure TsDBSearchComboBox.DefineProperties(Filer: TFiler);
begin
   inherited;
   Filer.DefineProperty( 'DefaultDraw', ReadDefaultDraw, WriteNothing, FALSE);
end;
{$ENDIF V1_COMP}

procedure TsDBSearchComboBox.SetDisabledStyle(const Value: TsDisabledStyle);
begin
   if FDisabledStyle <> Value then begin
      FDisabledStyle := Value;
      SetEditRect;
      Invalidate;
   end;
end;

procedure TsDBSearchComboBox.SetListScrollAnimation(const Value: TsScrollAnimationStyle);
begin
   FDataList.FListScrollAnimation := Value;
end;

function TsDBSearchComboBox.GetListScrollAnimation: TsScrollAnimationStyle;
begin
   Result := FDataList.FListScrollAnimation;
end;

function TsDBSearchComboBox.GetEmpty: Boolean;
begin
   Result := FText = FEmptyString;
end;

const
   isVisible: array[Boolean] of Integer = (SWP_HIDEWINDOW, SWP_SHOWWINDOW);

procedure TsDBSearchComboBox.SetStyle(Value: TdbComboBoxStyle);
var
   R: TRect;
begin
   if FStyle <> Value then begin
      FStyle := Value;
      case FStyle of
         csDropDown, csSimple: begin
            if FEdit = nil then begin
              	FEdit := TComboInPlaceEdit.Create(self);
             	FEdit.Parent := self;
              	FEdit.FCombo := self;
               FEdit.OnChange := EditChange;
            end;
            SetButtonWidth;
            SetEditRect;
         end;
         csDropDownList: begin
            FEdit.Free;
            FEdit := nil;
            SetButtonWidth;
         end;
      end;
   end;
   R := Rect(0, 0, Width, Height);
   InvalidateRect(Handle, @R, TRUE);
end;

procedure TsDBSearchComboBox.RedrawBorders;
begin
   case FBorderStyle of
      bsFlat:
         InternalRedrawBorder( Handle, GetSolidBorder);
{      _bsSingle:
         InternalRedrawSingleBorder(Handle, Color);
      _bsNone:
         InternalEriseBorder(Handle, Color);}
   end;
   if Enabled or (FDisabledStyle <> dsEditTools) then
      PaintButtonBorder;
end;

function TsDBSearchComboBox.GetImageRect: TRect;
begin
   if FdataList.FImageOffset > 0 then begin
      Result := ClientRect;
      Result.Right := Result.Left + FDataList.FImageOffset + 1;
   end else
      Result := Rect(0, 0, 0, 0);
end;

function TsDBSearchComboBox.GetEditRect: TRect;
begin
   SetButtonWidth;
   Result := ClientRect;
   if (Style <> csSimple) and (Enabled or (FDisabledStyle <> dsEditTools)) then
      Dec(Result.Right, FButtonWidth);
   if (FdataList.FImageOffset > 0) and (Style <> csDropDownList) then
      Inc(Result.Left, FdataList.FImageOffset);
   InflateRect(Result, -1, -1);
end;

procedure TsDBSearchComboBox.SetEditRect;
var
   R: TRect;
begin
   if (FEdit <> nil) and FEdit.HandleAllocated then begin
      R := GetEditRect;
      SetWindowPos(FEdit.Handle, HWND_TOP, R.Left, R.Top, WidthOf(R), HeightOf(R), SWP_NOREDRAW);
      SetWindowPos(FEdit.Handle, 0, 0, 0, 0, 0, SWP_NOMOVE + SWP_NOSIZE + SWP_NOACTIVATE + isVisible[Enabled]);
      FEdit.Visible := Enabled;
   end;
end;

procedure TsDBSearchComboBox.OnImageChange(Sender: TObject);
begin
   SetEditRect;
end;

function TsDBSearchComboBox.GetBorderSize: Integer;
begin
   if FBorderStyle = bsFlat then
      Result := inherited GetBorderSize
   else if FBorderStyle in [_bsNone, _bsSingle] then
      Result := 2
   else
      Result := 6;
end;

function TsDBSearchComboBox.GetSolidBorder: Boolean;
begin
   Result := (csDesigning in ComponentState) or
      (not(csDesigning in ComponentState) and
      (ListVisible or (GetFocus = Handle) or ((FEdit <> nil) and (GetFocus = FEdit.Handle)) or
      (FMouseInControl and IsFlatApplied)));
end;

function TsDBSearchComboBox.GetButtonRect: TRect;
begin
   if Style = csSimple then
      Result := Rect(0, 0, 0, 0)
   else begin
      Result := ClientRect;
	   Inc(Result.Left, ClientWidth - FButtonWidth);
   end;
end;

procedure TsDBSearchComboBox.PaintButton;
var
   R: TRect;
   state: Integer;
begin
   if Style = csSimple then
      Exit;
   R := GetButtonRect;
   state := DFCS_SCROLLDOWN;
   if FBorderStyle = bsFlat then begin
      state := state or DFCS_FLAT;
      InflateRect(R, 1, 0); // to have a normal size of down arrow
   end;
   if not Enabled then
      state := state or DFCS_INACTIVE;
   if FPressed or ListVisible then
      state := state or DFCS_PUSHED;
   DrawFrameControl(Canvas.Handle, R, DFC_SCROLL, state);
   if FFlat then
      PaintButtonBorder;
end;

type
   TWinControlCrack = class(TWinControl)
      property Color;
   end;

procedure TsDBSearchComboBox.PaintButtonBorder;
const
   Colors: array[Boolean] of TColor = (clBtnShadow, clWindow);
var
   R: TRect;
begin
   if Style = csSimple then
      Exit;
   R := GetButtonRect;
   if FFlat then with Canvas do begin
      if SolidBorder then begin
         Brush.Color := Colors[FPressed or ListVisible];
         FrameRect(R);
         Pen.Color := Colors[not FPressed and not ListVisible];
         Polyline([Point(R.Right-2, R.Top), Point(R.Left, R.Top), Point(R.Left, R.Bottom-1)]);
         Pen.Color := clWindow;
         PolyLine([Point(R.Left - 2, R.Top), Point(R.Left - 2, R.Bottom)]);
         Pen.Color := clBtnFace;
         Polyline([Point(R.Left-1, R.Top), Point(R.Left-1, R.Bottom)]);
      end else begin
         Brush.Color := clWindow;
         FrameRect(R);
         InflateRect(R, 0, -1);
         Pen.Color := Color;
         if not Enabled and (DisabledStyle = dsOffice97) then
            Pen.Color := TWinControlCrack(Parent).Color;
         Polyline([Point(R.Left-2, R.Top), Point(R.Left-2, R.Bottom)]);
         Polyline([Point(R.Left-1, R.Top), Point(R.Left-1, R.Bottom)]);
      end;
   end;
end;

procedure TsDBSearchComboBox.WMPaint(var Message: TWMPaint);
var
   DC: HDC;
   PS: TPaintStruct;
begin
   DC := Message.DC;
   if DC = 0 then
      DC := BeginPaint(Handle, PS);

   try
      Canvas.Handle := DC;

      if not Enabled and (FDisabledStyle = dsOffice97) then
         Canvas.Brush.Color := TWinControlCrack(Parent).Color
      else
         Canvas.Brush.Color := Color;

      if Enabled or (FDisabledStyle <> dsEditTools) then begin
         if ColorToRGB(Canvas.Brush.Color) <> ColorToRGB(clBtnFace) then
            Canvas.Pen.Color := clBtnFace
         else
            Canvas.Pen.Color := clBtnShadow;
         PaintWindow(DC);
         PaintButton;
      end else
         DrawDisabled(DC);
   finally
      if Message.DC = 0 then
         EndPaint(Handle, PS);
   end;
   RedrawBorders;
end;

procedure TsDBSearchComboBox.WMNCPaint(var Message: TWMPaint);
begin
   if BorderStyle <> bsFlat then
      inherited;
   if Enabled or (FDisabledStyle <> dsEditTools) then
      PaintButton;
end;

const
   alignments: array[TAlignment] of Word = (0, DT_RIGHT, DT_CENTER);

procedure TsDBSearchComboBox.DrawDisabled(DC: HDC);
var
   R: TRect;
   H: THandle;
begin
   Canvas.Handle := DC;
   Canvas.Brush.Color := inherited Color;
   Canvas.FillRect(ClientRect);

   H := SelectObject( Canvas.Handle, FDisabledFont.Handle);
   SetTextColor( Canvas.Handle, FDisabledFont.Color);

   R := GetEditRect;
   DrawText( Canvas.Handle, PChar(FText), Length(FText), R, DT_VCENTER or DT_NOCLIP or alignments[FAlignment]);

   SelectObject( Canvas.Handle, H);
   SetTextColor( Canvas.Handle, Font.Color);

   DoPaintImage( Canvas, GetImageRect, GetImageID, FDataList.Images);
   if FFlat then
      InternalRedrawBorder( Handle, FALSE);
end;

function TsDBSearchComboBox.GetImageId: Integer;
begin
   if FListVisible then
      Result := FDataList.GetImageID
   else begin
      Result := -1;
      if not Empty then
         if FImageField <> nil then try
            Result := FImageField.AsInteger;
         except
         end;
      if Assigned(FOnGetImageId) and (FListLink <> nil) and (FListLink.DataSet <> nil) then
         FOnGetImageId( self, FListLink.DataSet, Result);
   end;
end;

procedure TsDBSearchComboBox.Paint;
var
   R: TRect;
begin
   if Enabled and Assigned(FEdit) then begin
      if not (lsEditMode in FState) then
         FEdit.UpdateContents;
      DoPaintImage( Canvas, GetImageRect, GetImageID, FDataList.Images);
   end else begin
      R := GetEditRect;
      InflateRect(R, 1, 0);
      Canvas.FillRect(R);
      InflateRect(R, -1, 0);

      Canvas.Font.Color := Font.Color;

      if FListVisible then
			FDataList.PaintItemTo(Canvas.Handle, R, FDataList.FRecordIndex, FALSE)
      else begin
         DoPaintImage( Canvas, GetImageRect, GetImageID, FDataList.Images);
         DrawText( Canvas.Handle, PChar(FText), Length(FText), R, DT_VCENTER or DT_NOCLIP or alignments[FAlignment]);
         if Focused then begin
            R := ClientRect;
            Dec(R.Right, FButtonWidth+2); // at this point FButtonWidth is always > 0
            InflateRect(R, -1, -1);
            Canvas.DrawFocusRect(R);
         end;
      end;
   end;
end;

procedure TsDBSearchComboBox.WMSetFocus(var Message: TMessage);
begin
	if Assigned(FEdit) then
      FEdit.FocusEdit;
   inherited;
   if not (csDesigning in ComponentState) then
      RedrawBorders;
end;

procedure TsDBSearchComboBox.WMKillFocus(var Message: TMessage);
begin
	inherited;
{   if HWND(Message.wParam) = handle then begin
      if FListVisible then
         CloseUp(TRUE);//FALSE);
      if not (csDesigning in ComponentState) then
         RedrawBorders;
   end;}
end;

procedure TsDBSearchComboBox.CMMouseEnter(var Message: TMessage);
begin
   inherited;
   if not FMouseInControl and Enabled then begin
      FMouseInControl := TRUE;
      RedrawBorders;
      Notify( self, STM_MOUSEENTERNOTIFY, 0);
   end;
end;

procedure TsDBSearchComboBox.CMMouseLeave(var Message: TMessage);
begin
   inherited;
   if FMouseInControl and Enabled then begin
      FMouseInControl := FALSE;
      RedrawBorders;
      Notify( self, STM_MOUSELEAVENOTIFY, 0);
   end;
end;

procedure TsDBSearchComboBox.CNKeyDown(var Message: TWMKeyDown);
begin
   with Message do begin
      Result := 1;
      if not (csDesigning in ComponentState) and Enabled and Visible and
         FListVisible and (Parent <> nil) and Parent.Showing then begin
         if Perform(CM_CHILDKEY, CharCode, Integer(Self)) <> 0 then
            Exit;
         case CharCode of
            VK_RETURN, VK_EXECUTE, VK_ESCAPE, VK_CANCEL: begin
               CloseUp( CharCode in [VK_RETURN, VK_EXECUTE]);
               Exit;
            end;
         end;
      end;
      Result := 0;
      inherited;
   end;
end;

procedure TsDBSearchComboBox.DataLinkRecordChanged(Field: TField);
begin
   inherited;
   if not FChanging and Assigned(FEdit) and FEdit.HandleAllocated then
      FEdit.SetText(FText);
end;

procedure TsDBSearchComboBox.AniShow( Handle: HWND);
begin
   SetWindowPos( Handle, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_SHOWWINDOW);
end;

procedure TsDBSearchComboBox.DropDown;
var
	P: TPoint;
  	ii, Y: Integer;
  	S: string;
   AnimatingFlag: Boolean;
   downUp: Boolean;
begin
   if Style = csSimple then
      Exit;
  	if not FListVisible and (lsListActive in FState) then begin
      SetFocus;
    	if Assigned(FOnDropDown) then
         FOnDropDown(Self);

    	FDataList.Font := Font;
    	if FDropDownWidth > 0 then
      	FDataList.Width := FDropDownWidth
      else
      	FDataList.Width := Width;
      FDataList.FOnGetImageId := FOnGetImageId;
    	FDataList.ReadOnly := not CanModify;
    	FDataList.RowCount := FDropDownRows;
    	FDataList.KeyField := FKeyFieldName;
    	for ii := 0 to FListFields.Count - 1 do
      	S := S + TField(FListFields[ii]).FieldName + ';';
    	FDataList.ListField := S;
    	FDataList.ListFieldIndex := FListFields.IndexOf(FListField);
    	FDataList.ListSource := FListLink.DataSource;
    	FDataList.KeyValue := KeyValue;
    	P := Parent.ClientToScreen(Point(Left, Top));
    	Y := P.Y + Height;
      downUp := FALSE;
    	if Y + FDataList.Height > Screen.Height then begin
         Y := P.Y - FDataList.Height;
         downUp := TRUE;
      end;
    	case FDropDownAlign of
      	daRight: Dec(P.X, FDataList.Width - Width);
      	daCenter: Dec(P.X, (FDataList.Width - Width) div 2);
    	end;

      FDataList.Left := P.X;
      FDataList.Top := Y;
      AnimateDropDown(  FDataList.Handle, @AnimatingFlag,
         Rect(0, 0, FDataList.Width, FDataList.Height), FDataList.AniNCPaint,
         AniShow, FListPopupAnimation, TRUE, downUp);
    	FListVisible := True;
      FDataList.UpdateScrollBar(TRUE);
  	end;
end;

procedure TsDBSearchComboBox.CloseUp(Accept: Boolean);
var
  	ListValue: Variant;
begin
  	if FListVisible then begin
    	if GetCapture <> 0 then
         SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    	ListValue := FKeyField.Value;
    	SetWindowPos(FDataList.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or
      	SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_HIDEWINDOW);
      FListVisible := False;
    	Invalidate;
    	if Accept and CanModify then
         SelectKeyValue(ListValue);
         //SetKeyValue(ListValue);
    	if Assigned(FOnCloseUp) then
         FOnCloseUp(Self);
  	end;
end;

procedure TsDBSearchComboBox.EditChange(Sender: TObject);
var
	sRes: Boolean;
   p1: Integer;
   S: string;
begin
   if FChanging then
      Exit;
   FChanging := True;
   p1 := FEdit.SelStart;
   s := FEdit.Text;

	if FListVisible then
   	sRes := FDataList.ProcessSearch( FEdit.Text)
  	else
    	sRes := ProcessSearch( FEdit.Text);
  	if not sRes then
      FEdit.RestoreValue
   else begin
      FKeyValue := FDataList.FKeyValue;
   	//KeyValueChanged;
   end;
   case FLookupStyle of
      lsMatch: begin
         if AnsiCompareText( FListField.AsString, FEdit.Text) = 0 then begin
            FEdit.Text := FListField.AsString;
            FEdit.SelectAll;
         end;
     	end;
      lsAutoComplete: if sRes then begin
         if not (FEdit.FKey in [8, VK_DELETE]) then begin
	         FEdit.Text := FListField.AsString;
   	      FEdit.SelStart := p1;
      	   FEdit.SelLength := Length(FEdit.Text) - p1;
         end else
         	FEdit.Text := S;
         FEdit.Perform(EM_SCROLLCARET, 0, 0);
      end;
   end;
   FChanging := FALSE;
end;

procedure TsDBSearchComboBox.KeyDown(var Key: Word; Shift: TShiftState);
var
  	Delta: Integer;
begin
   Exclude( FState, lsEditMode);
  	inherited KeyDown(Key, Shift);
   if Key = VK_ESCAPE then begin
      if FListVisible then
         CloseUp(FALSE);
  	end else if (lsListActive in FState) and ((Key = VK_UP) or (Key = VK_DOWN)) then
    	if ssAlt in Shift then begin
      	if FListVisible then
            CloseUp(True)
         else
            DropDown;
      	Key := 0;
    	end else if not FListVisible then begin
        	if not LocateKey then
          	FListLink.DataSet.First
        	else begin
          	if Key = VK_UP then
               Delta := -1
            else
               Delta := 1;
          	FListLink.DataSet.MoveBy(Delta);
        	end;
        	SelectKeyValue(FKeyField.Value);
        	Key := 0;
      end;
  	if (Key <> 0) and FListVisible then
      FDataList.KeyDown(Key, Shift);
end;

procedure TsDBSearchComboBox.KeyPress(var Key: Char);
var
   aForm: TCustomForm;
begin
   if (Key = #13) and FDosKey then begin
      aForm := GetParentForm(Self);
      if aForm <> nil then
         SendMessage( aForm.Handle, WM_NEXTDLGCTL, 0, 0);
      Key := #0;
      Exit;
   end;

	if (FLookupStyle = lsDropDown) and (Key in [#32..#255]) then
	  	 DropDown;
	inherited KeyPress(Key);
   if FListVisible and (Key in [#13, #27]) then
   	CloseUp(Key = #13);
   if FEdit = nil then begin
      if FListVisible then begin
         FDataList.ProcessSearchKey(Key);
         FKeyValue := FDataList.FKeyValue;
   	   KeyValueChanged;
      end else
      	ProcessSearchKey(Key);
   end;
end;

procedure TsDBSearchComboBox.KeyValueChanged;
   procedure Clear;
   begin
    	FText := FEmptyString;
    	FAlignment := taLeftJustify;
   end;
begin
   Exclude( FDataState, ssInvalidKey);
  	if lsLookupMode in FState then begin
    	FText := FDataField.DisplayText;
    	FAlignment := FDataField.Alignment;
  	end else if not VarIsNull(FKeyValue) then begin
      if LocateKey then begin
 			FText := FListField.DisplayText;
         FAlignment := FListField.Alignment;
 		end else if (soDisplayInvalidKeyValues in FOptions) and (FKeyField <> nil) then begin
 			FText := FKeyValue;
         FAlignment := FKeyField.Alignment;
         Include( FDataState, ssInvalidKey);
 		end else
         Clear;
  	end else
      Clear;
   Invalidate;
end;

procedure TsDBSearchComboBox.ListLinkActiveChanged;
begin
  	inherited;
  	KeyValueChanged;
end;

procedure TsDBSearchComboBox.Click;
begin
	if Assigned(FEdit) and not (lsEditMode in FState) then
      FEdit.UpdateContents;
   inherited;
end;

procedure TsDBSearchComboBox.ListMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  	if Button = mbLeft then
    	CloseUp(PtInRect(FDataList.ClientRect, Point(X, Y)));
end;

procedure TsDBSearchComboBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   SetFocus;
   if (Button = mbLeft) and
      (PtInRect(Rect(ClientWidth - FButtonWidth, 0, ClientWidth, ClientHeight), Point(X, Y)) or
      ((Style = csDropDownList) and PtInRect(ClientRect, Point(X, Y))))then begin
    	if FListVisible then
      	CloseUp(False)
      else begin
        	MouseCapture := True;
         Include( FState, lsTracking);
        	TrackButton(X, Y);
        	DropDown;
      end;
	end;
   inherited MouseDown(Button, Shift, X, Y);
end;

procedure TsDBSearchComboBox.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  	ListPos: TPoint;
  	MousePos: TSmallPoint;
begin
  	if lsTracking in FState then begin
    	TrackButton(X, Y);
    	if FListVisible then begin
      	ListPos := FDataList.ScreenToClient(ClientToScreen(Point(X, Y)));
      	if PtInRect(FDataList.ClientRect, ListPos) then begin
        		StopTracking;
        		MousePos := PointToSmallPoint(ListPos);
        		SendMessage(FDataList.Handle, WM_LBUTTONDOWN, 0, Integer(MousePos));
        		Exit;
      	end;
    	end;
  	end;
  	inherited MouseMove(Shift, X, Y);
end;

procedure TsDBSearchComboBox.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  	StopTracking;
  	inherited MouseUp(Button, Shift, X, Y);
end;

procedure TsDBSearchComboBox.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  	inherited SetBounds(ALeft, ATop, AWidth, GetFontHeight(Font) + GetBorderSize + 2);
   SetEditRect;
   if FDataList <> nil then
      ListWidth := ListWidth;
end;

procedure TsDBSearchComboBox.StopTracking;
begin
  	if lsTracking in FState then begin
    	TrackButton(-1, -1);
      Exclude( FState, lsTracking);
    	MouseCapture := False;
  	end;
end;

procedure TsDBSearchComboBox.TrackButton(X, Y: Integer);
var
  	NewState: Boolean;
begin
	NewState := PtInRect(Rect(ClientWidth - FButtonWidth, 0, ClientWidth,
    	ClientHeight), Point(X, Y));
  	if FPressed <> NewState then begin
    	FPressed := NewState;
      PaintButton;
  	end;
end;

procedure TsDBSearchComboBox.CMCancelMode(var Message: TCMCancelMode);
begin
  	if (Message.Sender <> Self) and (Message.Sender <> FDataList) then
    	CloseUp(False);
end;

procedure TsDBSearchComboBox.CMEnabledChanged(var Msg: TMessage);
begin
   inherited;
   SetEditRect;
   Notify( self, STM_ENABLEDCHANGENOTIFY, Ord(ENABLED));
   Invalidate;
end;

procedure TsDBSearchComboBox.CMFontChanged(var Message: TMessage);
begin
  	inherited;
  	Height := 0;
end;

procedure TsDBSearchComboBox.CMColorChanged(var Message: TMessage);
begin
   inherited;
   if Assigned(FEdit) then
      FEdit.Color := Color;
   Invalidate;
end;

procedure TsDBSearchComboBox.CMGetDataLink(var Message: TMessage);
begin
   Message.Result := Integer(FDataLink);
end;

procedure TsDBSearchComboBox.WMCancelMode(var Message: TMessage);
begin
  	StopTracking;
  	inherited;
end;

procedure TsDBSearchComboBox.DoExit;
begin
   inherited;
   if Assigned(FEdit) then
      FEdit.UnFocusEdit;
   if FListVisible then
      CloseUp(TRUE)
   else
      SelectKeyValue(FDataList.KeyValue);
end;

procedure TsDBSearchComboBox.SetKeyValue(const Value: Variant);
begin
   inherited;
   FDataList.KeyValue := Value;
end;

procedure TsDBSearchComboBox.SelectKeyValue(Value: Variant);
var
   b: Boolean;
begin
   b := VarEquals(Value, FKeyValue);
   if (FEdit <> nil) and ((FEdit.Text = '') or (FEdit.Text = FEmptyString)) then
      Value := NULL;
   if not (FLookupStyle in [lsMatch]{, lsAutoComplete]}) or not FChanging or
   	((FEdit <> nil) and (FEdit.Text = FListField.AsString)) then
   	inherited;
   if not b and Assigned(FOnChange) then
      FOnChange(self);
end;

initialization
   RegisterFlatControl( TsDBSearchComboBox);

end.
