{*******************************************************************************
   Utit:
      sGraphics.pas
   Description:
      Commonly used graphic and text-drawing routines
   Versions:
      '99
   Author(s):
      Dimitry Statilko - support@sedittools.com**
      Partial Copyright (c) 1995,97 Borland International***
   URL:
      http://www.sedittools.com
   History:
      '99    - July 1999
               .Small adjustment in AnimateDropDown for smoother animation of
                  small lists
               .Fixed bug with drop down animation when AnimationType = anNone
                  (thanks to Thomas vas Stetten) 
      '99 beta 1 -  April-may 1999
               .General purpose animation procedure to animate drop-down/close-up
      2.0e  -  Feb-march 1999
               .New method Equal in TsFont3D
               .New class TsHilight3D inherited from TsFont3d with one more property
                  MouseOverColor. Used by TWinControl child to distinct visualy between
                  hilight (focused) and mouse over states.
               .Added new parameter hilighted in PaintText to handle hilight
               properly when HilightStyle = hsNone, but hilight3d differs from
               font3d
      2.0d  -  16/12/1998
               .Added TGlyphList and TGlyphCache from Delphi's Buttons.pas
      2.0c  -  14/12/1998
               .Small fix in PaintText to hand handle Font3D with Active = FALSE.
      2.0b  -  25/11/1998
               1. Two more selection styles - underline and color
               2. PaintText's first parameter changed from DC to TCanvas.
               3. Some fixes.
      2.0a  -  18/11/1998
               Small change for the TsGlyphList
		2.0*  - 	End of Sep. 1998
      			Initial release
*     I did not track the versions before, so let's consider it as 2.0
**    Some functions here were copied (some of them modified) from RX's VCLUtils
      unit.
***   TGlyphList and TGlyphCache are taken from Borland Delphi's Buttons.pas
*******************************************************************************}

unit sGraphics;

interface

uses Windows, Classes, Graphics, Controls, SysUtils, sCommonTypes, sDefaults;

type
   TFont3d = class(TPersistent)
   private
      FActive: Boolean;
      FColor: TColor;
      FPosition: TShadowPosition;
      FSpacing: ShortInt;
      FOnChange: TNotifyEvent;
      procedure SetActive(Value: Boolean);
      procedure SetPosition(Value: TShadowPosition);
      procedure SetColor(Value: TColor);
      procedure SetSpacing(Value: ShortInt);
      procedure Changed;
   public
      constructor Create( aOnChange: TNotifyEvent);
      procedure Assign(Source: TPersistent); override;
      function Equal(Value: TFont3D): Boolean; overload; virtual;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
   published
      property Active: Boolean read FActive write SetActive default DEF_Font3D_Active;
      property Color: TColor read FColor write SetColor default DEF_Font3D_Color;
      property Position: TShadowPosition read FPosition write SetPosition default DEF_Font3D_Position;
      property Spacing: ShortInt read FSpacing write SetSpacing default DEF_Font3D_Spacing;
   end;

   THilight3d = class(TFont3d)
   private
      FMouseOverColor: TColor;
      procedure SetMouseColor(Value: TColor);
   public
      constructor Create( aOnChange: TNotifyEvent);
      procedure Assign(Source: TPersistent); override;
      function Equal(Value: TFont3D; mouseCheck: Boolean = TRUE): Boolean; reintroduce; overload;
   published
      property Color default DEF_Hilight3d_Color;
      property MouseOverColor: TColor read FMouseOverColor write SetMouseColor default DEF_Hilight3d_MouseOverColor;
   end;

   TsHilightFont = class(TFont)
   public
      constructor Create;
      procedure AssignToFont(Source: TFont);
   published
      property Color default DEF_Hilight_FontColor;
      property Style default DEF_Hilight_FontStyle;
   end;

   TsHilight = class(TPersistent)
   private
   	FColor: TColor;
      FFont: TsHilightFont;
      FStyle: TListHilightStyle;
      procedure SetFont(Value: TsHilightFont);
   public
   	constructor Create;
      destructor Destroy; override;
      procedure Assign(Source: TPersistent); override;
   published
   	property Color: TColor read FColor write FColor default DEF_Hilight_Color;
      property Font: TsHilightFont read FFont write SetFont;
      property Style: TListHilightStyle read FStyle write FStyle default DEF_Hilight_Style;
   end;

   TPaintItemProc = procedure( DC: HDC; R: TRect; Index: Integer; Selected: Boolean) of Object;

   TsListScroller = class
   private
      sDC, tDC: HDC;
      Bmp, oldBmp: HBitmap;
      R: TRect;

      FListHandle: HWND;
      FItemHeight: Integer;
      FItemsCount: Integer;
      FPaintItemProc: TPaintItemProc;
   public
      WaitTime: Integer;
      WaitStep: Integer;
      ScrollCounter: Integer;
      constructor Create(listHandle: HWND; ItemHeight, ItemsCount: Integer);
      destructor Destroy; override;
      procedure Scroll( var topItem: Integer; step, ItemIndex: Integer; selected: Boolean);
      procedure ScrollEx( topItem: Integer; Index, step, ItemIndex: Integer; selected: Boolean);
      property PaintItemProc: TPaintItemProc read FPaintItemProc write FPaintItemProc;
   end;


function PaintText( Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; Font3d, Hilight3d: TFont3D; hilightStyle: TTextHilightStyle = hsNone;
   hilighted: TsHilighted = hNone): Integer;

function DrawShadowText(Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; ShadowSize: Byte; ShadowColor: TColorRef;
   ShadowPos: TShadowPosition): Integer;

function DrawFlashText( Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; ShadowSize: Byte; ShadowColor: TColorRef): Integer;

procedure DrawFlatEDGE(Canvas: TCanvas; R: TRect; Down: Boolean);

type
   PBoolean = ^Boolean;
   TAnimateNCPaint = procedure(tDC: HDC; var R: TRect) of object;
   TAnimateShow = procedure(Handle: HWND) of object;

procedure AnimateDropDown( CtrlHandle: HWND; AnimatingFlag: PBoolean;
   R: TRect; OnNCPaint: TAnimateNCPaint; OnShow: TAnimateShow; AnimationType: TsAnimationType;
	open: Boolean; DownUp: Boolean = FALSE; RightLeft: Boolean = FALSE; anchorBmp: Boolean = FALSE);

{
   from RX
}
procedure StretchBitmapRectTransparent(Dest: TCanvas; DstX, DstY, DstW, DstH: Integer;
   SrcRect: TRect; Bitmap: TBitmap; TransparentColor: TColor);

procedure StretchBltTransparent(DstDC: HDC; DstX, DstY, DstW, DstH: Integer;
  SrcDC: HDC; SrcX, SrcY, SrcW, SrcH: Integer; Palette: HPalette;
  TransparentColor: TColorRef);

procedure DrawTransparentBitmapRect(DC: HDC; Bitmap: HBitmap; DstX, DstY,
   DstW, DstH: Integer; SrcRect: TRect; TransparentColor: TColorRef);

procedure DrawBitmapTransparent(Dest: TCanvas; DstX, DstY: Integer;
  Bitmap: TBitmap; TransparentColor: TColor);

const
   PaletteMask = $02000000;
   TransparentMask = $02000000;


type
   TGlyphList = class(TImageList)
   private
      Used: TBits;
      FCount: Integer;
      function AllocateIndex: Integer;
   public
      constructor CreateSize(AWidth, AHeight: Integer);
      destructor Destroy; override;
      function AddMasked(Image: TBitmap; MaskColor: TColor): Integer;
      procedure Delete(Index: Integer);
      property Count: Integer read FCount;
   end;

   TGlyphCache = class
   private
      GlyphLists: TList;
   public
      constructor Create;
      destructor Destroy; override;
      function GetList(AWidth, AHeight: Integer): TGlyphList;
      procedure ReturnList(List: TGlyphList);
      function Empty: Boolean;
  end;

function EqualFonts(const f1, f2: TFont): Boolean;
procedure DoInvalidateRect(h: THandle; R: TRect; eraseBack: Boolean);


implementation

uses Consts, stdUtils, Messages;

function EqualFonts(const f1, f2: TFont): Boolean;
begin
   Result :=(f1.Charset = f2.Charset) and
            (f1.Height = f2.Height) and
    			(f1.Name = f2.Name) and
            (f1.Pitch = f2.Pitch) and
    			(f1.Size = f2.Size) and
    			(f1.Style = f2.Style);
end;

procedure DoInvalidateRect(h: THandle; R: TRect; eraseBack: Boolean);
begin
   InvalidateRect(h, @R, eraseBack);
end;


constructor TFont3d.Create( aOnChange: TNotifyEvent);
begin
   FOnChange := aOnChange;
   FActive := DEF_Font3D_Active;
   FPosition := DEF_Font3D_Position;
   FColor := DEF_Font3D_Color;
   FSpacing := DEF_Font3D_Spacing;
end;

procedure TFont3d.SetActive(Value: Boolean);
begin
   if FActive <> Value then begin
      FActive := Value;
      Changed;
   end;
end;

procedure TFont3d.SetPosition(Value: TShadowPosition);
begin
   if FPosition <> Value then begin
      FPosition := Value;
      Changed;
   end;
end;

procedure TFont3d.SetColor(Value: TColor);
begin
   if FColor <> Value then begin
      FColor := Value;
      Changed;
   end;
end;

procedure TFont3d.SetSpacing(Value: ShortInt);
begin
   if FSpacing <> Value then begin
      FSpacing := Value;
      Changed;
   end;
end;

procedure TFont3d.Changed;
begin
   if Assigned(FOnChange) then
      FOnChange(self);
end;

procedure TFont3d.Assign(Source: TPersistent);
begin
   if (Source = nil) or not (Source is TFont3d) then
      Error_MissmatchSource(Source, className);
   with TFont3d(Source) do begin
      self.FActive := FActive;
      self.FColor := FColor;
      self.FPosition := FPosition;
      self.FSpacing := FSpacing;
   end;
   Changed;
end;

function TFont3d.Equal(Value: TFont3D): Boolean;
begin
   Result := (Value.FColor = FColor) and (Value.FPosition = FPosition) and
      (Value.FSpacing = FSpacing);
end;

constructor THilight3d.Create( aOnChange: TNotifyEvent);
begin
   inherited;
   FColor := DEF_Hilight3d_Color;
   FMouseOverColor := DEF_Hilight3d_MouseOverColor;
end;

procedure THilight3d.SetMouseColor(Value: TColor);
begin
   if FMouseOverColor <> Value then begin
      FMouseOverColor := Value;
      Changed;
   end;
end;

function THilight3d.Equal(Value: TFont3D; mouseCheck: Boolean = TRUE): Boolean;
begin
   Result := (Value.FPosition = FPosition) and (Value.FSpacing = FSpacing);
   if Result then begin
      if mouseCheck then
         Result := Value.FColor = FMouseOverColor
      else
         Result := Value.FColor = FColor;
   end;
end;

procedure THilight3d.Assign(Source: TPersistent);
begin
   if source is THilight3D then
      self.FMouseOverColor := FMouseOverColor;
   inherited;
end;

constructor TsHilightFont.Create;
begin
   inherited Create;
   Color := DEF_Hilight_FontColor;
   Style := DEF_Hilight_FontStyle;
end;

{procedure TsHilightFont.Assign(Source: TPersistent);
var
   aColor: TColor;
   aStyle: TFontStyles;
begin
   aColor := Color;
   aStyle := Style;
   inherited;
   if not (Source is TsHilightFont) then begin
      Color := aColor;
      Style := aStyle;
   end;
end;}

procedure TsHilightFont.AssignToFont(Source: TFont);
var
   aColor: TColor;
   aStyle: TFontStyles;
begin
   aColor := Color;
   aStyle := Style;
   Assign(Source);
   if not (Source is TsHilightFont) then begin
      Color := aColor;
      Style := aStyle;
   end;
end;


constructor TsHilight.Create;
begin
	FFont := TsHilightFont.Create;
   FColor := DEF_Hilight_Color;
   FStyle := DEF_Hilight_Style;
end;

destructor TsHilight.Destroy;
begin
	FFont.Free;
   inherited;
end;

procedure TsHilight.SetFont(Value: TsHilightFont);
begin
	FFont.Assign(Value);
end;

procedure TsHilight.Assign(Source: TPersistent);
begin
   if (Source = nil) or not (Source is TsHilight) then
      Error_MissmatchSource(Source, className);
   FColor := TsHilight(Source).FColor;
   FStyle := TsHilight(Source).FStyle;
   FFont.Assign(TsHilight(Source).FFont);
end;

{TsListScroller}
constructor TsListScroller.Create(listHandle: HWND; itemHeight, itemsCount: Integer);
begin
   FListHandle := listHandle;
   FItemHeight := itemHeight;
   FItemsCount := itemsCount;

   Windows.GetClientRect(FListHandle, R);
   sDC := GetDC(FListHandle);
   tDC := CreateCompatibleDC(sDC);
   bmp := CreateCompatibleBitmap(sDC, WidthOf(R), FItemHeight);
   oldBmp := SelectObject(tDC, Bmp);

   WaitTime := 2;
   WaitStep := 1;
   ScrollCounter := 0;
end;

destructor TsListScroller.Destroy;
begin
   SelectObject(tDC, oldBmp);
   DeleteObject(bmp);
   DeleteDC(tDC);

   ReleaseDC(FListHandle, sDC);
   inherited;
end;

procedure TsListScroller.Scroll( var topItem: Integer; step, ItemIndex: Integer; selected: Boolean);
var
   Index: Integer;
begin
   if step < 0 then begin
      index := topItem + FItemsCount;
      Inc(topItem);
//      y := HeightOf(R) + step;
   end else begin
      index := topItem - 1;
      Dec(topItem);
//      y := step;
   end;
   if not selected then
      selected := Index = ItemIndex;

   ScrollEx( topItem, Index, step, ItemIndex, selected);
end;

procedure TsListScroller.ScrollEx( topItem: Integer; Index, step, ItemIndex: Integer; selected: Boolean);
var
   y, ii: Integer;
   w, h: Integer;
   ticks, t: Integer;
   cR: TRect;
begin
   if Index < 0 then begin
      Index := 0;
      selected := False;
   end;
   if Assigned(FPaintItemProc) then
   	FPaintItemProc(tDC, Rect( 0, 0, WidthOf(R), FItemHeight), Index, Selected);

   cR := R;
   if step < 0 then begin
      y := HeightOf(cR) + step;
      Dec(cR.Bottom);
   end else begin
      y := step;
      Inc(cR.Top);
   end;

   w := WidthOf(cR);
   h := FItemHeight;
   ticks := GetTickCount;

   for ii := 1 to FItemHeight do begin
   	ScrollWindow(FListHandle, 0, step, nil, @cR);
      ValidateRect(FListHandle, @cR);

      if step < 0 then begin
         BitBlt(sDC, 0, y, w, h, tDC, 0, 0, SRCCOPY);
	      Dec(y);
         Dec(cR.Bottom);
      end else begin
	      BitBlt(sDC, 0, 0, w, ii, tDC, 0, h - ii, SRCCOPY);
         Inc(cR.Top);
      end;

      Inc(ScrollCounter);
      if (WaitStep = 1) or (ScrollCounter mod WaitStep = 0) then begin
         if WaitTime in [0..100] then
            repeat
               t := GetTickCount;
               Dec( t, ticks);
            until t > WaitTime;
      end;
      ticks := GetTickCount;
   end;
end;

{}
function PaintText( Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; Font3d, Hilight3d: TFont3D; hilightStyle: TTextHilightStyle = hsNone;
   hilighted: TsHilighted = hNone): Integer;
var
   saveStyle: TFontStyles;
   saveColor: TColor;
   aColor: TColor;
   font3Dcreated: Boolean;
begin
   Result := 0;
   font3Dcreated := FALSE;
   if Font3D = nil then begin
      Font3D := TFont3D.Create(nil);
      Font3D.Spacing := 0;
      font3Dcreated := TRUE;
   end;
   if (hilighted <> hNone) and (hilightStyle in [hsMoving, hsRaising]) and ((Format and (0 xor DT_CALCRECT)) = 0) then begin
      if (hilight3d <> nil) and hilight3d.Active then begin
         if not Font3d.Active then
            Font3d.Spacing := 0;
         case Hilight3d.Position of
            spLeftTop:
               if hilightStyle = hsMoving then
                  OffsetRect(Rect, -Hilight3d.Spacing, -Hilight3d.Spacing);
            spRightBottom:
               if hilightStyle = hsMoving then
                  OffsetRect(Rect, Hilight3d.Spacing, Hilight3d.Spacing)
               else
                  OffsetRect( Rect, - Hilight3d.Spacing + Font3d.Spacing,
                                    - Hilight3d.Spacing + Font3d.Spacing);
            spLeftBottom: begin
               if hilightStyle = hsMoving then
                  OffsetRect(Rect, -Hilight3d.Spacing, Hilight3d.Spacing)
               else
                  OffsetRect(Rect, 0, -Hilight3d.Spacing + Font3d.Spacing);
            end;
            spRightTop: begin
               if hilightStyle = hsMoving then
                  OffsetRect(Rect, Hilight3d.Spacing, -Hilight3d.Spacing)
               else
                  OffsetRect(Rect, -Hilight3d.Spacing + Font3d.Spacing, 0);
            end;
         end;
      end;
   end;
   if (hilighted <> hNone) and (Hilight3d <> nil) and (Hilight3d.Active) then begin
      if (hilighted = hMouseOver) and (Hilight3D is THilight3D) then
         aColor := THilight3D(Hilight3d).MouseOverColor
      else
         aColor := Hilight3d.Color;
      case hilightStyle of
         hsNone:
            Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
               Hilight3d.Spacing, ColorToRGB(aColor), Hilight3d.Position);
         hsRaising:
            Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
               Hilight3d.Spacing, ColorToRGB(aColor), Hilight3d.Position);
         hsFlushing:
            Result := DrawFlashText( Canvas, Str, Count, Rect, Format,
               Hilight3d.Spacing, ColorToRGB(aColor));
         hsMoving:
            Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
               Font3D.Spacing, ColorToRGB(aColor), Font3D.Position);
         hsUnderline: begin
            saveStyle := Canvas.Font.Style;
            Canvas.Font.Style := Canvas.Font.Style + [fsUnderline];
            Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
               Font3d.Spacing, ColorToRGB(aColor), Font3d.Position);
            Canvas.Font.Style := saveStyle;
         end;
         hsColor: begin
            saveColor := Canvas.Font.Color;
            Canvas.Font.Color := aColor;//Hilight3d.Color;
            Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
               Font3d.Spacing, ColorToRGB(Font3d.Color), Font3d.Position);
            Canvas.Font.Color := saveColor;
         end;
      end;
   end else if (Font3d <> nil) and Font3d.Active then
      Result := DrawShadowText( Canvas, Str, Count, Rect, Format,
         Font3d.Spacing, ColorToRGB(Font3d.Color), Font3d.Position)
   else
      Result := DrawText( Canvas.Handle, Str, Count, Rect, Format);
   if font3Dcreated then
      font3D.Free;

{   if hilightStyle = hsRaising then
      Hilight3d.Spacing := Hilight3d.Spacing - font3D.Spacing;}

end;

function DrawShadowText(Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; ShadowSize: Byte; ShadowColor: TColorRef;
   ShadowPos: TShadowPosition): Integer;
var
   RText, RShadow: TRect;
   Color: TColorRef;
begin
   RText := Rect;
   if ShadowSize > 0 then begin
      RShadow := Rect;
      case ShadowPos of
         spLeftTop:
            OffsetRect(RText, ShadowSize, ShadowSize);
         spRightBottom:
            OffsetRect(RShadow, ShadowSize, ShadowSize);
         spLeftBottom: begin
            OffsetRect(RText, ShadowSize, 0);
            OffsetRect(RShadow, 0, ShadowSize);
         end;
         spRightTop: begin
            OffsetRect(RText, 0, ShadowSize);
            OffsetRect(RShadow, ShadowSize, 0);
         end;
      end; { case }
      Color := SetTextColor(Canvas.Handle, ShadowColor);
      DrawText(Canvas.Handle, Str, Count, RShadow, Format);
      SetTextColor(Canvas.Handle, Color);
   end;
   Result := DrawText(Canvas.Handle, Str, Count, RText, Format);
   if Result > 0 then
      Inc(Result, ShadowSize);
   if (Format and (0 xor DT_CALCRECT)) > 0 then
      UnionRect(Rect, RText, RShadow);
end;

function DrawFlashText( Canvas: TCanvas; Str: PChar; Count: Integer; var Rect: TRect;
   Format: Word; ShadowSize: Byte; ShadowColor: TColorRef): Integer;
var
   Color: TColorRef;
   RShadow: TRect;
   x, y: Integer;
begin
   Result := 0;
   Color := SetTextColor(Canvas.Handle, ShadowColor);
   for x := 1 to ShadowSize do begin
      for y := 1 to ShadowSize do begin
         RShadow := Rect;
         OffsetRect( RShadow, x, y);
         Result := DrawText( Canvas.Handle, Str, Count, RShadow, Format);
         if Result > 0 then begin
            RShadow := Rect;
            OffsetRect( RShadow, x, -y);
            Result := DrawText( Canvas.Handle, Str, Count, RShadow, Format);
         end;
         if Result > 0 then begin
            RShadow := Rect;
            OffsetRect( RShadow, -x, y);
            Result := DrawText( Canvas.Handle, Str, Count, RShadow, Format);
         end;
         if Result > 0 then begin
            RShadow := Rect;
            OffsetRect( RShadow, -x, -y);
            Result := DrawText( Canvas.Handle, Str, Count, RShadow, Format);
         end;
      end;
   end;

   if Result > 0 then
      Inc(Result, 2 * ShadowSize);

   SetTextColor(Canvas.Handle, Color);
   DrawText(Canvas.Handle, Str, Count, Rect, Format);

{   if (Format and (0 xor DT_CALCRECT)) > 0 then
      InflateRect(Rect, ShadowSize, ShadowSize);}
end;

procedure DrawFlatEDGE(Canvas: TCanvas; R: TRect; Down: Boolean);
begin
	Canvas.Brush.Color := cl3DDkShadow;
	Canvas.FrameRect(R);
   Canvas.Pen.Color := cl3DLight;
   if Down then
		Canvas.PolyLine( [Point(R.Left, R.Bottom-1), Point(R.Right-1, R.Bottom-1), Point(R.Right-1, R.Top-1)])
	else
		Canvas.PolyLine( [Point(R.Left, R.Bottom-1), Point(R.Left, R.Top), Point(R.Right-1, R.Top)])
end;

procedure AnimateDropDown( CtrlHandle: HWND; AnimatingFlag: PBoolean;
   R: TRect; OnNCPaint: TAnimateNCPaint; OnShow: TAnimateShow; AnimationType: TsAnimationType;
	open: Boolean; DownUp: Boolean = FALSE; RightLeft: Boolean = FALSE; anchorBmp: Boolean = FALSE);
const
   ShowFlag: array[Boolean] of Integer = (SW_HIDE, SW_SHOW);
var
   w, h: Integer;
   x, y: Integer;

   rx, ry: Integer;
   rw, rh: Integer;

   dx, dy: Integer;
   cx, cy: Integer;
   px, py: Integer;

   pR: TRect;


   steps: Integer;

   tmpDC, sDC, tDC: HDC;
   bmp, oldBmp: THandle;
   tmpBmp, tmpOldBmp: THandle;
   Brush: HBrush;
   Rgn: HRgn;

begin
   if AnimationType = anNone then try
      if AnimatingFlag <> nil then
      	AnimatingFlag^ := TRUE;
      if Assigned( OnShow) then
         OnShow(CtrlHandle)
      else
         ShowWindow( CtrlHandle, ShowFlag[open]);
      Exit;
   finally
      if AnimatingFlag <> nil then
   	   AnimatingFlag^ := FALSE;
   end;


  if AnimationType = anRandom then
      AnimationType := TsAnimationType(Random(3) + 1);

   if AnimatingFlag <> nil then
   	AnimatingFlag^ := TRUE;

   h := HeightOf(R);
   w := WidthOf(R);

{   flag := 0;
   if open then
   	flag := SWP_NOREDRAW;}

   sDC := GetWindowDC(CtrlHandle);
   tDC := CreateCompatibleDC(sDC);
   bmp := CreateCompatibleBitmap(sDC, w, h);
   oldBmp := SelectObject(tDC, bmp);

   if Assigned(OnNCPaint) then
      OnNCPaint( tDC, R);

   tmpDC := CreateCompatibleDC(sDC);
   tmpBmp := CreateCompatibleBitmap(sDC, w, h);
   tmpOldBmp := SelectObject(tmpDC, tmpBmp);

   if open then begin
	   Rgn := CreateRectRgn(0, 0, 1, 1);
   	SetWindowRgn(CtrlHandle, Rgn, FALSE);
      if Assigned( OnShow) then
         OnShow(CtrlHandle)
      else
   	   ShowWindow( CtrlHandle, SW_SHOW);
   end;

   SendMessage( CtrlHandle, WM_PAINT, tmpDC, 0);

   BitBlt( tDC, R.Left, R.Top, R.Right, R.Bottom, tmpDC, 0, 0, SRCCOPY);

   SelectObject(tmpDC, tmpOldBmp);
   DeleteObject( tmpBmp);
   DeleteDC( tmpDC);


   {calculating offsets}
   steps := 15;
   x := 0;
   y := 0;

   if AnimationType in [anHSlide, anUnfold] then begin
      dx := w div steps;
      if w mod steps > 10 then begin// 2/3 of steps
         Inc(dx);
         steps := w div dx;
      end;
      cx := dx;
      if not Open or RightLeft then
         cx := w;
      if RightLeft and Open then
         x := w - dx;
   end else begin
      dx := 0;
      cx := w;
   end;

   if AnimationType in [anVSlide, anUnfold] then begin
      dy := h div steps;
      if h mod steps > 10 then begin// 2/3 of steps
         Inc(dy);
         steps := min( h div dy, steps);
      end;
      cy := dy;
      if not Open or DownUp then
         cy := h;
      if DownUp and Open then
         y := h - dy;
   end else begin
      dy := 0;
      cy := h;
   end;

   if not Open then begin
      dx := -dx;
      dy := -dy;
   end;

   AnchorBmp := AnchorBmp or DownUp or RightLeft;
   pR := Rect( x, y, cx, cy);

   while steps > 0 do begin
  	   Rgn := CreateRectRgn(pR.Left, pR.Top, pR.Right, pR.Bottom);
      SetWindowRgn(CtrlHandle, Rgn, not open);

      if AnchorBmp then
	      BitBlt( sDC, pR.Left, pR.Top, pR.Right, pR.Bottom, tDC, 0, 0, SRCCOPY)
      else
	      BitBlt( sDC, pR.Left, pR.Top, pR.Right, pR.Bottom, tDC, w-WidthOf(pR), h-HeightOf(pR), SRCCOPY);

      if RightLeft then
         Dec( pR.Left, dx)
      else
         Inc( pR.Right, dx);

      if DownUp then
         Dec(pR.Top, dy)
      else
         Inc(pR.Bottom, dy);

      Sleep(5);
      Dec(steps);
   end;
   if AnimatingFlag <> nil then
      AnimatingFlag^ := FALSE;
   SetWindowRgn(CtrlHandle, 0, not open);

   if open then
	   BitBlt( sDC, 0, 0, w, h, tDC, 0, 0, SRCCOPY);

   SelectObject(tDC, oldBmp);
   DeleteObject( bmp);
   DeleteDC( tDC);
   ReleaseDC( CtrlHandle, sDC);

end;

{
   From RX.
}

procedure StretchBitmapTransparent(Dest: TCanvas; Bitmap: TBitmap;
  TransparentColor: TColor; DstX, DstY, DstW, DstH, SrcX, SrcY,
  SrcW, SrcH: Integer);
var
   CanvasChanging: TNotifyEvent;
   Temp: TBitmap;
begin
   if DstW <= 0 then
      DstW := Bitmap.Width;
   if DstH <= 0 then
      DstH := Bitmap.Height;
   if (SrcW <= 0) or (SrcH <= 0) then begin
      SrcX := 0; SrcY := 0;
      SrcW := Bitmap.Width;
      SrcH := Bitmap.Height;
   end;
   if not Bitmap.Monochrome then
      SetStretchBltMode(Dest.Handle, STRETCH_DELETESCANS);
   CanvasChanging := Bitmap.Canvas.OnChanging;
   try
      Bitmap.Canvas.OnChanging := nil;
{$IFDEF RX_D3}
    {if Bitmap.HandleType = bmDIB then begin
      Temp := TBitmap.Create;
      Temp.Assign(Bitmap);
      Temp.HandleType := bmDDB;
    end
    else} Temp := Bitmap;
{$ELSE}
      Temp := Bitmap;
{$ENDIF}
      try
         if TransparentColor = clNone then begin
            StretchBlt(Dest.Handle, DstX, DstY, DstW, DstH, Temp.Canvas.Handle,
               SrcX, SrcY, SrcW, SrcH, Dest.CopyMode);
         end else begin
{$IFDEF RX_D3}
            if TransparentColor = clDefault then
               TransparentColor := Temp.Canvas.Pixels[0, Temp.Height - 1];
{$ENDIF}
            if Temp.Monochrome then
               TransparentColor := clWhite
            else
               TransparentColor := ColorToRGB(TransparentColor);
            StretchBltTransparent(Dest.Handle, DstX, DstY, DstW, DstH,
               Temp.Canvas.Handle, SrcX, SrcY, SrcW, SrcH, Temp.Palette,
               TransparentColor);
         end;
      finally
{$IFDEF RX_D3}
      {if Bitmap.HandleType = bmDIB then Temp.Free;}
{$ENDIF}
      end;
   finally
      Bitmap.Canvas.OnChanging := CanvasChanging;
   end;
end;

procedure StretchBltTransparent(DstDC: HDC; DstX, DstY, DstW, DstH: Integer;
   SrcDC: HDC; SrcX, SrcY, SrcW, SrcH: Integer; Palette: HPalette;
   TransparentColor: TColorRef);
var
   Color: TColorRef;
   bmAndBack, bmAndObject, bmAndMem, bmSave: HBitmap;
   bmBackOld, bmObjectOld, bmMemOld, bmSaveOld: HBitmap;
   MemDC, BackDC, ObjectDC, SaveDC: HDC;
   palDst, palMem, palSave, palObj: HPalette;
begin
  { Create some DCs to hold temporary data }
   BackDC := CreateCompatibleDC(DstDC);
   ObjectDC := CreateCompatibleDC(DstDC);
   MemDC := CreateCompatibleDC(DstDC);
   SaveDC := CreateCompatibleDC(DstDC);
  { Create a bitmap for each DC }
   bmAndObject := CreateBitmap(SrcW, SrcH, 1, 1, nil);
   bmAndBack := CreateBitmap(SrcW, SrcH, 1, 1, nil);
   bmAndMem := CreateCompatibleBitmap(DstDC, DstW, DstH);
   bmSave := CreateCompatibleBitmap(DstDC, SrcW, SrcH);
   { Each DC must select a bitmap object to store pixel data }
   bmBackOld := SelectObject(BackDC, bmAndBack);
   bmObjectOld := SelectObject(ObjectDC, bmAndObject);
   bmMemOld := SelectObject(MemDC, bmAndMem);
   bmSaveOld := SelectObject(SaveDC, bmSave);
  { Select palette }
   palDst := 0; palMem := 0; palSave := 0; palObj := 0;
   if Palette <> 0 then begin
      palDst := SelectPalette(DstDC, Palette, True);
      RealizePalette(DstDC);
      palSave := SelectPalette(SaveDC, Palette, False);
      RealizePalette(SaveDC);
      palObj := SelectPalette(ObjectDC, Palette, False);
      RealizePalette(ObjectDC);
      palMem := SelectPalette(MemDC, Palette, True);
      RealizePalette(MemDC);
   end;
   { Set proper mapping mode }
   SetMapMode(SrcDC, GetMapMode(DstDC));
   SetMapMode(SaveDC, GetMapMode(DstDC));
   { Save the bitmap sent here }
   BitBlt(SaveDC, 0, 0, SrcW, SrcH, SrcDC, SrcX, SrcY, SRCCOPY);
   { Set the background color of the source DC to the color,         }
   { contained in the parts of the bitmap that should be transparent }
   Color := SetBkColor(SaveDC, ColorToRGB(TransparentColor) or PaletteMask);
   { Create the object mask for the bitmap by performing a BitBlt()  }
   { from the source bitmap to a monochrome bitmap                   }
   BitBlt(ObjectDC, 0, 0, SrcW, SrcH, SaveDC, 0, 0, SRCCOPY);
   { Set the background color of the source DC back to the original  }
   SetBkColor(SaveDC, Color);
   { Create the inverse of the object mask }
   BitBlt(BackDC, 0, 0, SrcW, SrcH, ObjectDC, 0, 0, NOTSRCCOPY);
   { Copy the background of the main DC to the destination }
   BitBlt(MemDC, 0, 0, DstW, DstH, DstDC, DstX, DstY, SRCCOPY);
   { Mask out the places where the bitmap will be placed }
   StretchBlt(MemDC, 0, 0, DstW, DstH, ObjectDC, 0, 0, SrcW, SrcH, SRCAND);
   { Mask out the transparent colored pixels on the bitmap }
   BitBlt(SaveDC, 0, 0, SrcW, SrcH, BackDC, 0, 0, SRCAND);
   { XOR the bitmap with the background on the destination DC }
   StretchBlt(MemDC, 0, 0, DstW, DstH, SaveDC, 0, 0, SrcW, SrcH, SRCPAINT);
   { Copy the destination to the screen }
   BitBlt(DstDC, DstX, DstY, DstW, DstH, MemDC, 0, 0, SRCCOPY);
   { Restore palette }
   if Palette <> 0 then begin
      SelectPalette(MemDC, palMem, False);
      SelectPalette(ObjectDC, palObj, False);
      SelectPalette(SaveDC, palSave, False);
      SelectPalette(DstDC, palDst, True);
   end;
   { Delete the memory bitmaps }
   DeleteObject(SelectObject(BackDC, bmBackOld));
   DeleteObject(SelectObject(ObjectDC, bmObjectOld));
   DeleteObject(SelectObject(MemDC, bmMemOld));
   DeleteObject(SelectObject(SaveDC, bmSaveOld));
   { Delete the memory DCs }
   DeleteDC(MemDC);
   DeleteDC(BackDC);
   DeleteDC(ObjectDC);
   DeleteDC(SaveDC);
end;


procedure StretchBitmapRectTransparent(Dest: TCanvas; DstX, DstY,
  DstW, DstH: Integer; SrcRect: TRect; Bitmap: TBitmap;
  TransparentColor: TColor);
begin
  with SrcRect do
    StretchBitmapTransparent(Dest, Bitmap, TransparentColor,
    DstX, DstY, DstW, DstH, Left, Top, Right - Left, Bottom - Top);
end;

procedure DrawTransparentBitmapRect(DC: HDC; Bitmap: HBitmap; DstX, DstY,
   DstW, DstH: Integer; SrcRect: TRect; TransparentColor: TColorRef);
var
   hdcTemp: HDC;
begin
   hdcTemp := CreateCompatibleDC(DC);
   try
      SelectObject(hdcTemp, Bitmap);
      with SrcRect do
         StretchBltTransparent(DC, DstX, DstY, DstW, DstH, hdcTemp,
            Left, Top, Right - Left, Bottom - Top, 0, TransparentColor);
   finally
      DeleteDC(hdcTemp);
   end;
end;

procedure DrawBitmapTransparent(Dest: TCanvas; DstX, DstY: Integer;
  Bitmap: TBitmap; TransparentColor: TColor);
begin
   StretchBitmapTransparent(Dest, Bitmap, TransparentColor, DstX, DstY,
      Bitmap.Width, Bitmap.Height, 0, 0, Bitmap.Width, Bitmap.Height);
end;


{ TGlyphList }

constructor TGlyphList.CreateSize(AWidth, AHeight: Integer);
begin
   inherited CreateSize(AWidth, AHeight);
   Used := TBits.Create;
end;

destructor TGlyphList.Destroy;
begin
   Used.Free;
   inherited Destroy;
end;

function TGlyphList.AllocateIndex: Integer;
begin
   Result := Used.OpenBit;
   if Result >= Used.Size then begin
      Result := inherited Add(nil, nil);
      Used.Size := Result + 1;
   end;
   Used[Result] := True;
end;

function TGlyphList.AddMasked(Image: TBitmap; MaskColor: TColor): Integer;
begin
   Result := AllocateIndex;
   ReplaceMasked(Result, Image, MaskColor);
   Inc(FCount);
end;

procedure TGlyphList.Delete(Index: Integer);
begin
   if Used[Index] then begin
      Dec(FCount);
      Used[Index] := False;
   end;
end;

{ TGlyphCache }

constructor TGlyphCache.Create;
begin
   inherited Create;
   GlyphLists := TList.Create;
end;

destructor TGlyphCache.Destroy;
begin
   GlyphLists.Free;
   inherited Destroy;
end;

function TGlyphCache.GetList(AWidth, AHeight: Integer): TGlyphList;
var
  ii: Integer;
begin
   for ii := GlyphLists.Count - 1 downto 0 do begin
      Result := GlyphLists[ii];
      with Result do
         if (AWidth = Width) and (AHeight = Height) then
            Exit;
   end;
   Result := TGlyphList.CreateSize(AWidth, AHeight);
   GlyphLists.Add(Result);
end;

procedure TGlyphCache.ReturnList(List: TGlyphList);
begin
   if List = nil then
      Exit;
   if List.Count = 0 then begin
      GlyphLists.Remove(List);
      List.Free;
   end;
end;

function TGlyphCache.Empty: Boolean;
begin
   Result := GlyphLists.Count = 0;
end;

end.
