{*******************************************************************************
   Unit
      sItemGlyphs.pas
   Description:
      property editor for "ImageIndexes" property
   Versions:
      '99
   History:
      1.0('99)  -   Feb-March 1999
               .Initial implementation
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      This property editor is a part of sEditTools
*******************************************************************************}
unit sItemGlyphs;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ComCtrls, ToolWin, DsgnIntf, ImgList, sCtrls, StdUtils;

type
  TItemGlyphsForm = class(TForm)
      ToolBar1: TToolBar;
      OkBtn: TToolButton;
      CancelBtn: TToolButton;
      ToolButton1: TToolButton;
      RemoveBtn: TToolButton;
      ImageList1: TImageList;
      StatusBar1: TStatusBar;
      ListBox: TListBox;
      Splitter1: TSplitter;
      ScrollBox1: TScrollBox;
      GlyphsList: TPaintBox;
      procedure ListBoxDrawItem(Control: TWinControl; Index: Integer;
         Rect: TRect; State: TOwnerDrawState);
      procedure FormCreate(Sender: TObject);
      procedure FormDestroy(Sender: TObject);
      procedure GlyphsListPaint(Sender: TObject);
      procedure OkBtnClick(Sender: TObject);
      procedure CancelBtnClick(Sender: TObject);
      procedure RemoveBtnClick(Sender: TObject);
      procedure ListBoxClick(Sender: TObject);
      procedure Splitter1Moved(Sender: TObject);
      procedure GlyphsListMouseDown(Sender: TObject; Button: TMouseButton;
         Shift: TShiftState; X, Y: Integer);
      procedure ListBoxDragOver(Sender, Source: TObject; X, Y: Integer;
         State: TDragState; var Accept: Boolean);
      procedure ListBoxDragDrop(Sender, Source: TObject; X, Y: Integer);
      procedure GlyphsListMouseMove(Sender: TObject; Shift: TShiftState; X,
         Y: Integer);
	    procedure ListBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
   	   Y: Integer);
    procedure ListBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
   private
      FImages: TImageList;
      FSelectedImage: Integer;
      FSelectedItem: Integer;
      function GetImagePoint(index: Integer): TPoint;
      procedure InvalidateGlyphItem(index: Integer);
   public
      function Call( anImageList: TCustomImageList; anIndexes: TsImageIdList; anItems: TStrings): Boolean;
   end;

   TImageIndexesProperty = class(TPropertyEditor)
   public
      procedure Edit; override;
      function GetAttributes: TPropertyAttributes; override;
   end;

implementation
{$R *.DFM}

uses Commctrl, sCombos, sConsts;

function TItemGlyphsForm.Call( anImageList: TCustomImageList; anIndexes: TsImageIdList; anItems: TStrings): Boolean;
var
   ii: Integer;
   imageIndex: Integer;
begin
   FSelectedImage := -1;

   FImages.Assign(anImageList);
   ListBox.Items.Assign(anItems);

   if anIndexes <> nil then for ii := 0 to ListBox.Items.Count - 1 do
      ListBox.Items.Objects[ii] := TObject(anIndexes[ii]);

   GlyphsList.Height := (FImages.Count div (GlyphsList.Width div FImages.Width)) *
      FImages.Height + 4;

   Result := ShowModal = mrOk;
   if Result and (anIndexes <> nil) then begin
      for ii := 0 to ListBox.Items.Count - 1 do begin
      	try
         	imageIndex := Integer(ListBox.Items.Objects[ii]);
      	except
				imageIndex := -1;
      	end;
         anIndexes[ii] := imageIndex;
      end;
   end;
end;

procedure TItemGlyphsForm.FormCreate(Sender: TObject);
begin
   FImages := TImageList.Create(self);
end;

procedure TItemGlyphsForm.FormDestroy(Sender: TObject);
begin
   FImages.Free;
end;

procedure TItemGlyphsForm.ListBoxDrawItem(Control: TWinControl; Index: Integer;
   Rect: TRect; State: TOwnerDrawState);
var
   Text: array[0..255] of Char;
   i: Integer;
begin
   with ListBox do begin
      Canvas.FillRect(Rect);
      try
         i := Integer(Items.Objects[Index]);
      except
         i := -1;
      end;
      if i in [0..FImages.Count- 1] then
         FImages.Draw(Canvas, Rect.Left + 2, (Rect.Top + Rect.Bottom - FImages.Height) div 2, i);
      StrPCopy(Text, Items[Index]);
      Rect.Left := Rect.Left + 26;
      DrawText(Canvas.Handle, Text, StrLen(Text), Rect, DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
   end;
end;

procedure TItemGlyphsForm.OkBtnClick(Sender: TObject);
begin
   ModalResult := mrOk;
end;

procedure TItemGlyphsForm.CancelBtnClick(Sender: TObject);
begin
   ModalResult := mrCancel;
end;

procedure TItemGlyphsForm.RemoveBtnClick(Sender: TObject);
begin
   with ListBox do begin
      Items.Objects[ItemIndex] := TObject(-1);
      invalidate;
   end;
end;

procedure TItemGlyphsForm.ListBoxClick(Sender: TObject);
begin
   RemoveBtn.Enabled := ListBox.ItemIndex > -1;
end;

procedure TItemGlyphsForm.Splitter1Moved(Sender: TObject);
begin
   GlyphsList.Height := (FImages.Count div (GlyphsList.Width div FImages.Width)) *
      FImages.Height + 4;
end;

function TItemGlyphsForm.GetImagePoint(index: Integer): TPoint;
var
   cols: Integer;
begin
   cols := (GlyphsList.Width+4) div (FImages.Width + 2);
   if cols = 0 then
      Exit;

   Result.X := 2 + (index mod cols) * (FImages.Width + 2);
   Result.Y := 2 + (index div cols) * (FImages.Height + 2);
end;

procedure TItemGlyphsForm.InvalidateGlyphItem(index: Integer);
var
   P: TPoint;
begin
   with GlyphsList do begin
      P := GetImagePoint(index);
      if index = FSelectedImage then
         Canvas.Brush.Color := clHighlight
      else
         Canvas.Brush.Color := clInfoBk;
      Canvas.FillRect(Rect(P.X, P.Y, P.X + FImages.Width, P.Y + FImages.Height));
      FImages.Draw(Canvas, P.X, P.Y, index);
   end;
end;

procedure TItemGlyphsForm.GlyphsListPaint(Sender: TObject);
var
   ii: Integer;
begin
   GlyphsList.Canvas.Brush.Color := clInfoBk;
   GlyphsList.Canvas.FillRect(ClientRect);
   for ii := 0 to FImages.Count - 1 do
      InvalidateGlyphItem(ii);
end;

procedure TItemGlyphsForm.GlyphsListMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
   cols: Integer;
   index, oldIndex: Integer;
begin
   cols := (GlyphsList.Width+4) div (FImages.Width + 2);
   if cols = 0 then
      Exit;

   index := (Y div (FImages.Height + 2)) * cols + (X div (FImages.Width + 2));

   if (index <> FSelectedImage) and (index in [0..FImages.Count - 1]) then begin
      oldIndex := FSelectedImage;
      FSelectedImage := index;
      if FSelectedImage > - 1 then
         InvalidateGlyphItem(oldIndex);
      InvalidateGlyphItem(FSelectedImage);
   end;
end;

procedure TItemGlyphsForm.GlyphsListMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
   FSelectedItem := -1;
   if ssLeft in Shift then
      GlyphsList.BeginDrag(True);
end;

procedure TItemGlyphsForm.ListBoxDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
   Accept := (Source = GlyphsList) or (Source = ListBox);
   if Accept then
      with ListBox do ItemIndex := ItemAtPos(Point(X, Y), FALSE);
end;

procedure TItemGlyphsForm.ListBoxDragDrop(Sender, Source: TObject; X, Y: Integer);
begin
   with ListBox do begin
      Items.Objects[ItemAtPos(Point(X, Y), FALSE)] := TObject(FSelectedImage);
      if FSelectedItem > -1 then
         Items.Objects[FSelectedItem] := TObject(-1);
      invalidate;
   end;
end;

procedure TItemGlyphsForm.ListBoxMouseDown(Sender: TObject;
	  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   if ([ssLeft, ssRight] * Shift <> []) then
	   ListBox.ItemIndex := ListBox.ItemAtPos(Point(X, Y), False);
end;

procedure TItemGlyphsForm.ListBoxMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
   FSelectedItem := ListBox.ItemIndex;
   if (FSelectedItem > -1) and (ListBox.Items.Objects[FSelectedItem] <> TObject(-1)) and
      ([ssLeft, ssRight] * Shift <> []) then begin

      FSelectedImage := Integer(ListBox.Items.Objects[FSelectedItem]);

      if (ssRight in Shift) then
         FSelectedItem := -1;
      ListBox.BeginDrag(True);
   end;
end;

type
   TsComboBoxCrack = class(TsComboBox)
      property ImageIndexes;
      property Images;
   end;

procedure TImageIndexesProperty.Edit;
var
	cmp: TPersistent;
   images: TCustomImageList;
   imageIndexes: TsImageIdList;
   items: TStrings;
begin
	cmp := GetComponent(0);
   if cmp.InheritsFrom(TsComboBox) then begin
      images := TsComboBoxCrack(cmp).Images;
      imageIndexes := TsComboBoxCrack(cmp).ImageIndexes;
      items := TsCustomComboBox(cmp).Items;
   end else
      Exit;

   with TItemGlyphsForm.Create(nil) do try
      Call( Images, ImageIndexes, Items);
   finally
      Free;
   end;
end;

function TImageIndexesProperty.GetAttributes: TPropertyAttributes;
begin
	Result := [paDialog, paRevertable];
end;



end.
