{*******************************************************************************
   Unit
      smemo.pas
   Description:
      sMemo control - flat TMemo child with TsLabel support
   Versions:
      '99
   History:
      '99    - July 1999
               .Border style implementation
      '99 beta 1 -  June 1999
               .Several small changes
      1.1   -  20/02/1999
               .Added flat scroll bars
      1.0*	- 	10/12/1998
               .Initial release
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      Whell, instead of flat borders this is nothing new. I made it to keep the
      same feel of interface for my applications
*******************************************************************************}
unit sMemo;

interface

uses Windows, StdCtrls, Controls, Messages, Classes,
     Graphics, ExtCtrls, SysUtils, sEdits, sFlat, sCtrls, sCommonTypes, sDefaults;

{$I S.inc}

type
   TsMemo = class(TMemo)
   private
      FFlat: Boolean;
      FSolidBorder: Boolean;
      FBorderStyle: TsBorderStyle;
      {$IFDEF FlatSB}
      FScrollBarMode: TScrollbarMode;
      FFirst, FCount, FOffset, FWidth: Integer;
      function GetOffset: Integer;
      function GetTextWidth: Integer;
      procedure SetScrollBarMode(const Value: TScrollbarMode);
      procedure UpdateFlatSB;
      procedure CheckUpdateFlatSB;
      {$ENDIF}
      procedure SetSolidBorder;
      procedure SetFlat(const Value: Boolean);
      procedure SetBorderStyle(const Value: TsBorderStyle);
      procedure AdjustBorderStyleValue;
      procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
      procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
      procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
      procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure CMEnabledChanged(var Msg: TMessage); message CM_ENABLEDCHANGED;
      {$IFDEF FlatSB}
      procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
      procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;
      procedure WMSize(var Msg: TWMSize); message WM_SIZE;
      procedure CMFontChanged(var Msg: TMessage); message CM_FONTCHANGED;
      {$ENDIF}
   protected
      FEditState: TsEditState;
      procedure CreateWnd; override;
      {$IFDEF FlatSB}
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
      procedure KeyUp(var Key: Word; Shift: TShiftState); override;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure Change; override;
      {$ENDIF}
      procedure RedrawBorders; virtual;
      function GetSolidBorder: Boolean; virtual;
      property SolidBorder: Boolean read FSolidBorder;
      procedure SetEditRect; virtual;
   public
      constructor Create(AOwner: TComponent); override;
   published
      property BorderStyle: TsBorderStyle read FBorderStyle write SetBorderStyle default DEF_Memo_BorderStyle;
      property Ctl3D stored FALSE;
      property Flat: Boolean read FFlat write SetFlat stored FALSE;
      {$IFDEF FlatSB}
      property ScrollbarMode: TScrollbarMode read FScrollBarMode write SetScrollBarMode default DEF_Memo_ScrollBarMode;
      {$ENDIF}
   end;


procedure Register;

implementation

uses sTdUtils, sConsts, Commctrl, Forms;

constructor TsMemo.Create(AOwner: TComponent);
begin
   inherited;
   ControlStyle := ControlStyle - [csSetCaption, csFramed];
   BorderStyle := DEF_Memo_BorderStyle;
   {$IFDEF FlatSB}
   FCount := 0;
   FFirst := -1;
   FOffset := 1;
   FWidth := 0;
   FScrollBarMode := DEF_Memo_ScrollBarMode;//sbmRegular;
   {$ENDIF}
end;

procedure TsMemo.CreateWnd;
begin
   AdjustBorderStyleValue;
   inherited;
   SetEditRect;
   {$IFDEF FlatSB}
   if ScrollBars<>ssNone then
      InitializeFlatSB(Handle);
   UpdateFlatSB;
   FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Integer(FScrollBarMode), True);
   FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Integer(FScrollBarMode), True);
   {$ENDIF}
   SetSolidBorder;
end;

procedure TsMemo.WMPaint(var Message: TWMPaint);
var
   DC: HDC;
   PS: TPaintStruct;
begin
   try
      DC := BeginPaint(Handle, PS);
      PaintWindow(DC);
      RedrawBorders;
   finally
      EndPaint(Handle, PS);
   end;
end;

procedure TsMemo.WMNCPaint(var Message: TMessage);
begin
   inherited;
   RedrawBorders;
end;

procedure TsMemo.WMSetFocus(var Message: TMessage);
begin
   inherited;
   SetSolidBorder;
end;

procedure TsMemo.WMKillFocus(var Message: TMessage);
begin
   SetSolidBorder;
   inherited;
end;

procedure TsMemo.CMMouseEnter(var Message: TMessage);
begin
   inherited;
   if not (msMouseInControl in FEditState) and Enabled then begin
      Include(FEditState, msMouseInControl);
      SetSolidBorder;
      Notify( self, STM_MOUSEENTERNOTIFY, 0);
   end;
end;

procedure TsMemo.CMMouseLeave(var Message: TMessage);
begin
   inherited;
   if (msMouseInControl in FEditState) and Enabled then begin
      Exclude(FEditState, msMouseInControl);
      SetSolidBorder;
      Notify( self, STM_MOUSELEAVENOTIFY, 0);
   end;
end;

procedure TsMemo.CMEnabledChanged(var Msg: TMessage);
begin
   inherited;
   Notify( self, STM_ENABLEDCHANGENOTIFY, Ord(ENABLED));
end;

{$IFDEF FlatSB}
function TsMemo.GetOffset: Integer;
var
   P: TPoint;
   sP: TSize;
   S: string;
   DC: HDC;
begin
   P := CaretPos;
   S := Lines[p.y];
   SetLength(S, p.x);
   DC := GetDC(Handle);
   Windows.GetTextExtentPoint( DC, PChar(S), Length(S), sP);
   ReleaseDC(Handle, DC);
   Result := sP.cX;
end;

function TsMemo.GetTextWidth: Integer;
var
   ii: Integer;
   P: TSize;
   DC: HDC;
begin
   Result := 0;
   DC := GetDC(Handle);
   try
      for ii := 0 to Lines.Count - 1 do begin
         Windows.GetTextExtentPoint( DC, PChar(Lines[ii]), Length(Lines[ii]), P);
         if Result < P.cX then
            Result := P.cX;
      end;
   finally
      releaseDC(Handle, DC);
   end;
   //Dec(Result, GetOffset);
end;
{$ENDIF FlatSB}

function TsMemo.GetSolidBorder: Boolean;
begin
   Result := (csDesigning in ComponentState) or
      (Focused or ((msMouseInControl in FEditState) and IsFlatApplied));
end;

procedure TsMemo.SetSolidBorder;
var
   sb: Boolean;
begin
   sb := GetSolidBorder;
   if sb <> FSolidBorder then begin
      FSolidBorder := sb;
      RedrawBorders;
   end;
end;

procedure TsMemo.RedrawBorders;
begin
   if FFlat then
      InternalRedrawBorderEx( Handle, SolidBorder, 1);
end;

procedure TsMemo.SetFlat(const Value: Boolean);
begin
   if Value <> FFlat then begin
      FFlat := Value;
      RecreateWnd;
   end;
end;

procedure TsMemo.AdjustBorderStyleValue;
begin
   if FFlat and Ctl3D then
      FBorderStyle := bsFlat
   else if inherited BorderStyle = bsSingle then begin
      FFlat := FALSE;
      if Ctl3D then
         FBorderStyle := bsDefault
      else
         FBorderStyle := _bsSingle;
   end else
      FBorderStyle := _bsNone;
end;

procedure TsMemo.SetBorderStyle(const Value: TsBorderStyle);
var
   oldCtl3D: Boolean;
begin
  if FBorderStyle <> Value then begin
      oldCtl3D := Ctl3D;
      FBorderStyle := Value;
      case FBorderStyle of
         _bsNone: begin
            FFlat := FALSE;
            inherited BorderStyle := bsNone;
         end;
         _bsSingle: begin
            FFlat := FALSE;
            inherited BorderStyle := bsSingle;
            Ctl3D := FALSE;
         end;
         bsFlat: begin
            FFlat := TRUE;
            inherited BorderStyle := bsSingle;
            Ctl3D := TRUE;
         end;
         bsDefault: begin
            FFlat := FALSE;
            inherited BorderStyle := bsSingle;
            Ctl3D := TRUE;
         end;
      end;
      if (oldCtl3D = Ctl3D) and HandleAllocated then
         RecreateWnd;
  end;
end;

procedure TsMemo.SetEditRect;
var
  	R: TRect;
begin
   if HandleAllocated and FFlat then begin
      R := GetClientRect;
      InflateRect(R, -1, -1);
      SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
   end;
end;

{$IFDEF FlatSB}
procedure TsMemo.WMVScroll(var Msg: TWMVScroll);
begin
   inherited;
   FlatSB_SetScrollPos(Handle, SB_VERT, Perform(EM_GETFIRSTVISIBLELINE, 0, 0), True);
   FFirst := Perform(EM_GETFIRSTVISIBLELINE, 0, 0);
end;

procedure TsMemo.WMHScroll(var Msg: TWMHScroll);
begin
   inherited;
   FlatSB_SetScrollPos(Handle, SB_HORZ, GetOffset, True);
   FOffset := GetOffset;
end;

procedure TsMemo.WMSize(var Msg: TWMSize);
begin
  inherited;
  UpdateFlatSB;
end;

procedure TsMemo.CMFontChanged(var Msg: TMessage);
begin
  inherited;
  UpdateFlatSB;
end;

procedure TsMemo.Change;
begin
   inherited;
   CheckUpdateFlatSB;
end;

procedure TsMemo.KeyDown(var Key: Word; Shift: TShiftState);
begin
   inherited;
   CheckUpdateFlatSB;
end;

procedure TsMemo.KeyUp(var Key: Word; Shift: TShiftState);
begin
   inherited;
   CheckUpdateFlatSB;
end;

procedure TsMemo.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   inherited;
   CheckUpdateFlatSB;
end;

procedure TsMemo.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   inherited;
   CheckUpdateFlatSB;
end;

procedure TsMemo.SetScrollBarMode(const Value: TScrollbarMode);
begin
   FScrollBarMode := Value;
   FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Integer(FScrollBarMode), True);
   FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Integer(FScrollBarMode), True);
end;

procedure TsMemo.CheckUpdateFlatSB;
begin
   if (FCount <> Lines.Count) or
      (FFirst <> Perform(EM_GETFIRSTVISIBLELINE, 0, 0)) or
      (FOffset <> GetOffset) or (FWidth <> GetTextWidth) then
      UpdateFlatSB;
end;

procedure TsMemo.UpdateFlatSB;
var
   N: Integer;
   SI: TScrollInfo;
   W: Integer;
begin
   if ScrollBars in [ssVertical, ssBoth] then begin
      SI.cbSize := SizeOf(SI);
      SI.fMask := SIF_RANGE or SIF_POS or SIF_PAGE or SIF_DISABLENOSCROLL;
      SI.nMin := 0;
      SI.nMax := Lines.Count-1;
      SI.nPos := Perform(EM_GETFIRSTVISIBLELINE, 0, 0);
      N := Perform(EM_LINEINDEX, SI.nPos+1, 0);
      N := Perform(EM_POSFROMCHAR, N, 0) shr 16 -1;
      SI.nPage := ClientHeight div N;
      FlatSB_SetScrollInfo(Handle, SB_VERT, SI, True);
      FCount := SI.nMax+1;
      FFirst := SI.nPos;
   end;
   if ScrollBars in [ssHorizontal, ssBoth] then begin
      W := GetTextWidth;
      SI.cbSize := SizeOf(SI);
      SI.fMask := SIF_RANGE or SIF_POS or SIF_PAGE or SIF_DISABLENOSCROLL;
      SI.nMin := 0;

      {if W < ClientWidth then
        SI.nMax := 0
      else}
      SI.nMax := W;

      SI.nPos := {ClientWidth}GetOffset;
      SI.nPage := ClientWidth;

      FlatSB_SetScrollInfo(Handle, SB_HORZ, SI, True);
      FOffset := GetOffset;
      FWidth := W;
   end;
end;
{$ENDIF}


procedure Register;
begin
   RegisterComponents('sEditTools', [TsMemo]);
end;

end.
