{*******************************************************************************
   Unit
      sTrans.pas
   Description:
      TCustomControl which provide the transparency mechanizme for its child controls
   Versions:
   	'99
	History:
      '99   - August 1999
               Fully rewriten (hope - the last time)
   Autor(s):
      Dimitry Statilko - support@sedittools.com**
   URL:
      http://www.sedittools.com
   Comments:
      When TMDIWallpeper define is on full transparency is not supported
*******************************************************************************}

unit sTrans;

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   StdCtrls, ExtCtrls;

type
   TsTransControl = class(TCustomControl)
   private
      FTransparent: Boolean;
      procedure SetTransparent(Value: Boolean);
      procedure WMEraseBkgnd(var Msg: TWMEraseBkgnd); message WM_ERASEBKGND;
      procedure WMPaint(var Msg: TWMPaint); message WM_PAINT;
      procedure WMMove(var Msg: TMessage); message WM_MOVE;
      procedure WMSize(var Msg: TMessage); message WM_SIZE;
      procedure CMParentColorChanged(var Msg: TMessage); message CM_PARENTCOLORCHANGED;
   protected
      property Transparent: Boolean read FTransparent write SetTransparent default FALSE;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure PaintBackground;
      procedure PaintBackgroundRect(R: TRect);
      procedure FrameRect(R: TRect; penWidth: Integer);
   end;

procedure DoInvalidateRect(h: THandle; R: TRect; eraseBackground: Boolean);

procedure PaintBackground(Ctrl: TWinControl; DC: HDC);
procedure PaintBackgroundRect(Ctrl: TWinControl; DC: HDC; R: TRect);


const
   TrueTransparency: Boolean = FALSE;

implementation

{$I S.Inc}
uses sGraphics, sTdUtils
{$IFDEF MDIWALLPAPER_SUPPORT}
   ,MDIWallp
{$ENDIF}
;

procedure DoInvalidateRect(h: THandle; R: TRect; eraseBackground: Boolean);
begin
   InvalidateRect(h, @R, eraseBackground);
end;

type
  TWinCtl = class(TWinControl);

{$IFDEF MDIWALLPAPER_SUPPORT}
type
   TComponentClass = class of TComponent;

function FindComponentClass(aOwner: TComponent; CompClass: TComponentClass): TComponent;
var
   ii: Integer;
begin
   Result := nil;
   for ii := 0 to aOwner.ComponentCount - 1 do
      if aOwner.Components[ii] is CompClass then begin
         Result := aOwner.Components[ii];
         Exit;
      end;
end;
{$ENDIF}

procedure PaintBackgroundTo(Ctrl: TWinControl; DC: HDC; P: TPoint);
var
   SaveIndex: Integer;
	{$IFDEF MDIWALLPAPER_SUPPORT}
   cmp: TComponent;
   tmpDC: HDC;
	{$ENDIF}
begin
   if (Ctrl = nil) and (Ctrl.Parent = nil) then
      Exit;
	{$IFDEF MDIWALLPAPER_SUPPORT}
   cmp := FindComponentClass(Ctrl.Parent, TMDIWallpaper);
	{$ENDIF}
   if TrueTransparency then begin
   	{$IFDEF MDIWALLPAPER_SUPPORT}
      if (cmp <> nil) {and (TMDIWallpaper(cmp).Client.FormStyle = fsMDIForm)} then
      	SendMessage( TMDIWallpaper(cmp).ClientHandle, WM_ERASEBKGND, DC, 0);
      {$ENDIF}
   	Ctrl.Parent.PaintTo(DC, -P.X-1, -P.Y-1);
   end else begin
		SaveIndex := SaveDC(DC);
   	try
   		MoveWindowOrg(DC, -P.X, -P.Y);
			{$IFDEF MDIWALLPAPER_SUPPORT}
   	   cmp := FindComponentClass(Ctrl.Parent, TMDIWallpaper);
   		if (cmp <> nil) then
	      	SendMessage( TMDIWallpaper(cmp).ClientHandle, WM_ERASEBKGND, DC, 0);
	   	else
			{$ENDIF}
	      SendMessage( Ctrl.Parent.Handle, WM_ERASEBKGND, DC, 0);
   	   TWinCtl(Ctrl.Parent).PaintControls( DC, nil);
	   finally
   	  	RestoreDC(DC, SaveIndex);
   	end;
   end;
end;

procedure PaintBackground(Ctrl: TWinControl; DC: HDC);
var
  P: TPoint;
begin
   if (Ctrl <> nil) and (Ctrl.Parent <> nil) then begin
   	P := Ctrl.ClientOrigin;
      ScreenToClient(Ctrl.Parent.Handle, P);
      PaintBackgroundTo(Ctrl, DC, P);
   end;
end;

procedure PaintBackgroundRect(Ctrl: TWinControl; DC: HDC; R: TRect);
var
  P: TPoint;
  tmpDC: HDC;
  tmpBmp, oldBmp: THandle;
begin
   if (Ctrl <> nil) and (Ctrl.Parent <> nil) then begin
   	tmpDC := CreateCompatibleDC(DC);
      tmpBmp := CreateCompatibleBitmap(DC, widthOf(R), heightOf(R));
      OldBmp := SelectObject(tmpDC, tmpBmp);
      try
         P := Ctrl.ClientOrigin;
         Inc(P.X, R.Left);
         Inc(P.Y, R.Top);
         ScreenToClient(Ctrl.Parent.Handle, P);
      	PaintBackgroundTo(Ctrl, tmpDC, P);
         BitBlt(DC, R.Left, R.Top, widthOf(R), heightOf(R), tmpDC, 0, 0, SRCCOPY);
      finally
         SelectObject(tmpDC, oldBmp);
      	DeleteObject(tmpBmp);
      	DeleteDC(tmpDC);
      end;
  	end;
end;

{
}
constructor TsTransControl.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   ControlStyle := ControlStyle - [csOpaque];
   FTransparent := FALSE;
end;

destructor TsTransControl.Destroy;
begin
   inherited;
end;

procedure TsTransControl.SetTransparent(Value: Boolean);
begin
   if Value <> FTransparent then begin
      FTransparent := Value;
      Invalidate;
   end;
end;

procedure TsTransControl.PaintBackground;
begin
   if FTransparent then
      sTrans.PaintBackground(self, Canvas.Handle)
   else
      PaintBackgroundRect( ClientRect);
end;

procedure TsTransControl.PaintBackgroundRect( R: TRect);
begin
   if FTransparent then begin
      sTrans.PaintBackgroundRect( Self, Canvas.Handle, R);
   end else with Canvas do begin
      Brush.Style := bsSolid;
      Brush.Color := Color;
      FillRect(R);
   end;
end;

procedure TsTransControl.CMParentColorChanged(var Msg: TMessage);
begin
   inherited;
   invalidate;
end;

procedure TsTransControl.WMEraseBkgnd(var Msg: TWMEraseBkgnd);
begin
   if FTransparent then begin
      if csPaintCopy in ControlState then
         Exit;
      sTrans.PaintBackground(self, Msg.DC)
   end else
      //PaintBackground;
      inherited;
   Msg.Result := 1;
end;

procedure TsTransControl.WMPaint(var Msg: TWMPaint);
begin
   if FTransparent and (csPaintCopy in ControlState) then
      Exit;
   inherited;
end;

procedure TsTransControl.WMMove(var Msg: TMessage);
begin
   inherited;
   Invalidate;
end;

procedure TsTransControl.WMSize(var Msg: TMessage);
begin
   inherited;
   Invalidate;
end;

procedure TsTransControl.FrameRect(R: TRect; penWidth: Integer);
var
   MemDC: HDC;
   aBitmap, oldObject: HBITMAP;
   ii: Integer;
begin
   if Transparent then begin
      InflateRect(R, -penWidth, -PenWidth);
      MemDC := CreateCompatibleDC( Canvas.Handle);
      aBitmap := CreateCompatibleBitmap( Canvas.Handle, R.Right - R.Left, R.Bottom - R.Top);
      oldObject := SelectObject( MemDC, aBitmap);
      BitBlt( MemDC, 0, 0, R.Right - R.Left, R.Bottom - R.Top, Canvas.Handle,
         R.Left, R.Top, SRCCOPY);
      InflateRect(R, penWidth, PenWidth);
      PaintBackgroundRect(R);
      InflateRect(R, -penWidth, -PenWidth);
      BitBlt( Canvas.Handle, R.left, R.Top, R.Right - R.Left, R.Bottom - R.Top, MemDC, 0, 0, SRCCOPY);
      SelectObject( MemDC, oldObject);
      DeleteObject( aBitmap);
      DeleteDC( MemDC);
   end else for ii := 0 to penWidth do begin
      Canvas.Brush.Color := Color;
      Canvas.FrameRect(R);
      InflateRect(R, -1, -1);
   end;
end;

end.

