unit SEDReg;

interface

procedure Register;

implementation


uses
   SysUtils, Classes, DsgnIntf, TypInfo, db, Dialogs, Forms, FiltEdit, MaskProp,
   TiledBmp, sFileUtils, sGraphics, sTrans, sCtrls, sPickDate, sEdits, sCombos, sButtons,
   sBrowseFolder, sGlyphsList, sDBCtrls, sConsts, sMask, sMemo, sDBCombos, sItemGlyphs,
   stdUtils, sDBLookup;

{$I S.INC}
{$R BUTTONS.RES}

type
   TFilenameProperty = class(TStringProperty)
   protected
      function GetFilter: string; virtual;
   public
      procedure Edit; override;
      function GetAttributes: TPropertyAttributes; override;
   end;

   TDirnameProperty = class(TStringProperty)
   public
      procedure Edit; override;
      function GetAttributes: TPropertyAttributes; override;
   end;

function TFilenameProperty.GetFilter: string;
begin
   Result := SDefaultFilesFilter;
end;

procedure TFilenameProperty.Edit;
var
   FileOpen: TOpenDialog;
begin
   FileOpen := TOpenDialog.Create(Application);
   try
      FileOpen.Filename := GetValue;
      FileOpen.InitialDir := ExtractFilePath(FileOpen.Filename);
      if (ExtractFileName(FileOpen.Filename) = '') or not
         ValidFileName(ExtractFileName(FileOpen.Filename)) then
         FileOpen.Filename := '';
      FileOpen.Filter := GetFilter;
      FileOpen.Options := FileOpen.Options + [ofHideReadOnly];
      if FileOpen.Execute then
         SetValue(FileOpen.Filename);
   finally
      FileOpen.Free;
   end;
end;

function TFilenameProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog, paRevertable];
end;

{ TDirnameProperty }

procedure TDirnameProperty.Edit;
var
   FolderName: string;
begin
   FolderName := GetValue;
   if BrowseDirectory(FolderName, SBrowseCaption) then
      SetValue(FolderName);
end;

function TDirnameProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog, paRevertable];
end;

type
   TsStringProperty = class(TStringProperty)
   public
      function GetAttributes: TPropertyAttributes; override;
      procedure GetValueList(List: TStrings); virtual;
      procedure GetValues(Proc: TGetStrProc); override;
   end;

function TsStringProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TsStringProperty.GetValues(Proc: TGetStrProc);
var
   ii: Integer;
   Values: TStringList;
begin
   Values := TStringList.Create;
   try
      GetValueList(Values);
      for ii := 0 to Values.Count - 1 do Proc(Values[ii]);
   finally
      Values.Free;
   end;
end;

procedure TsStringProperty.GetValueList(List: TStrings); 
begin
end;

{ TDataFieldProperty}
type
  TDataFieldProperty = class(TsStringProperty)
  public
    function GetDataSourcePropName: string; virtual;
    procedure GetValueList(List: TStrings); override;
  end;

function TDataFieldProperty.GetDataSourcePropName: string;
begin
  Result := 'DataSource';
end;

procedure TDataFieldProperty.GetValueList(List: TStrings);
var
   Instance: TComponent;
   PropInfo: PPropInfo;
   DataSource: TDataSource;
begin
   Instance := TComponent(GetComponent(0));
   PropInfo := TypInfo.GetPropInfo(Instance.ClassInfo, GetDataSourcePropName);
   if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkClass) then begin
      DataSource := TObject(GetOrdProp(Instance, PropInfo)) as TDataSource;
      if (DataSource <> nil) and (DataSource.DataSet <> nil) then
         DataSource.DataSet.GetFieldNames(List);
  end;
end;

{ TListFieldProperty }
type
  TListFieldProperty = class(TDataFieldProperty)
  public
    function GetDataSourcePropName: string; override;
  end;

function TListFieldProperty.GetDataSourcePropName: string;
begin
  Result := 'ListSource';
end;

procedure Register;
begin
   RegisterComponents(SEditToolsPageCaption, [TsEdit, TsLinkEdit, TsMaskLinkEdit, TsDateEdit,
         TsNumberEdit, TsFilenameEdit, TsDirectoryEdit, TsMemo]);
   RegisterComponents(SEditToolsPageCaption, [TsComboBox, TsColorComboBox,
         TsFontComboBox, TsPathComboBox]);
   RegisterComponents(SEditToolsPageCaption, [TsCheckBox, TsRadioButton, TsLabel,
         TsSpeedButton, TsBitBtn, TsStatusBar, {$IFDEF FlatSB}TsScrollBar, {$ENDIF}
         TsPanel, TsTrackBar]);
   RegisterComponents(SEditToolsPageCaption, [TsCalendar]);
   RegisterComponents(SEditToolsPageCaption, [TTiledBmp, TsGlyphList]);

   {Database components}
   RegisterComponents(sDBEditToolsPageCaption, [TsDBEdit, TsDBLinkEdit, TsDBDateEdit,
      TsDBNumberEdit, TsDBMaskLinkEdit,
   	TsDBCheckBox, TsDBText, TsDBMemo,
      TsDBIndexComboBox, TsDBComboBox,
      TsDBSearchComboBox]);

   {Property editors}
   RegisterClasses([TsImageIdList]);
   RegisterPropertyEditor( TypeInfo(string), TsFileNameEdit, 'Filter', TFilterProperty);
   RegisterPropertyEditor( TypeInfo(string), TsFileNameEdit, 'Path', TFilenameProperty);
   RegisterPropertyEditor( TypeInfo(string), TsDirectoryEdit, 'Path', TDirnameProperty);
   RegisterPropertyEditor( TypeInfo(string), TsFileDirEdit, 'InitialDir', TDirnameProperty);
   RegisterPropertyEditor( TypeInfo(string), TsPathComboBox, 'Path', TDirnameProperty);
   RegisterPropertyEditor( TypeInfo(string), TsCustomMaskEdit, 'EditMask', TMaskProperty);
   //RegisterPropertyEditor( TypeInfo(TsGlyphCollection), TsGlyphList, 'List', TGlyphListProperty);
	RegisterPropertyEditor( TypeInfo(Integer), TPersistent, 'GlyphListId', TGlyphIdProperty);
   RegisterPropertyEditor( TypeInfo(TsImageIdList), TsComboBox, 'ImageIndexes', TImageIndexesProperty);
   RegisterPropertyEditor(TypeInfo(string), TsDBLookupControl, 'KeyField', TListFieldProperty);
   RegisterPropertyEditor(TypeInfo(string), TsDBLookupControl, 'ListField', TListFieldProperty);
   RegisterPropertyEditor(TypeInfo(string), TsDBLookupControl, 'ImageField', TListFieldProperty);

   {Component editors}
   RegisterComponentEditor(TsGlyphList, TGlyphsListEditor);
end;

end.
